package algo.alignment;

import lmu.utils.*;
import java.io.*;
import java.util.*;



/**
 * This class provides the modified Smith-Waterman alignment algorithm from O.
 * Gotoh (for further information refer to O. Gotoh. An improved algorithm for
 * matching biological sequences. J. Mol. Biol., 162:705-708, 1982.) to align
 * two sequence instances. The method is able to calculate global, local and
 * freeShift alignments. As a consequence of the algoritm's utilization of
 * affine gap costs (weight w_k to a gap of length k will be calculated as u*k +
 * v), two gap penalties, open gap costs (v) and extension gap costs (u) have to
 * be specifified by the user (for details see method documentation).
 * 
 * @author cgx
 * 
 */

public class Gotoh
{
    Logger log = LogConfig.getLogger();
    private static final long INF = Long.MIN_VALUE/2;

    public interface SimCalculator
    {
        /** in target there is memory allocated, but if the
         * implementator decides to care about memory handling
         * another long[] can be returned
         * 
         * otherwise default implementator returns target 
         */
        public long calcSim(int i, int j);
        public long[] calcSim(int i, long[] target);
        public long getFactor();
    }

    
    public enum AlignMode
    {

        GLOBAL("global"), //1
        GLOBFREE("globfree"), //start with local but ends with freeshift
        FREEGLOB("freeglob"), //start with freeshift but ends with global

        LOCAL("local"),//2
        FREESHIFT("freeshift") //3
        ;

        String name;

        AlignMode(String n)
        {
            name=n;
        }
        public static AlignMode get(String n)
        {
            for (AlignMode am : AlignMode.values())
            {
                if (am.name.equals(n))
                {
                    return am;
                }
            }
            return null;
        }




    };

    // alignment parameters affine gap consts ( weight w_k to a gap of length k will be calculated
    // as gap_extend*k + gap_open )
    double gap_open=-12;
    double gap_extend=-1;

    
    

    int startx, starty;
    private AlignMode alignMode=AlignMode.FREESHIFT;
    private float score;

    

    // dynamic programming & pointer matrix


    public Gotoh()
    {

    }
    /**
    
     */
    public Gotoh(double openGapCost, double extendGapCost, AlignMode alignMode)
    {
        this.gap_open = openGapCost;
        this.gap_extend = extendGapCost;
        this.alignMode = alignMode;
        
        
        
    }


    /**
     * Method returns the openGapCost parameter
     * 
     * @return
     */
    public double getGapOpen()
    {
        return gap_open;
    }

    /**
     * Method sets the openGapCost parameter
     * 
     * @param
     */
    public void setGapOpen(double openGapCost)
    {
        this.gap_open = openGapCost;
        
    }

    /**
     * Method returns the extendGapCost parameter
     * 
     * @return
     */
    public double getGapExtend()
    {
        return gap_extend;
    }

    /**
     * Method sets the extendGapCost parameters
     * 
     * @param
     */
    public void setGapExtend(double extendGapCost)
    {
        this.gap_extend = extendGapCost;
    }

    /**
     * Method returns the alignMode parameter
     * 
     * @return
     */
    public AlignMode getAlignMode()
    {
        return alignMode;
    }

    public void setAlignMode(AlignMode mode)
    {
        this.alignMode=mode;
    }
    public float getScore()
    {
        return score;
    }
    
    //matrices

    long[][] M;
    long[][] noAlignM;
    
    long[] Ivec;
    long[] simVec;


    public long[][] getM()
    {
        return M;
    }


    public void reallocate(int xlen, int ylen, boolean noalign)
    {
        if (noalign && noAlignM!=null && Ivec.length>=ylen)
        {
            return;
        }

        if (!noalign && M!=null && M.length>=xlen && M[0].length>=ylen)
        {
            return;
        }

        int i=(int)(xlen/100.0f)+1;
        int j=(int)(ylen/100.0f)+1;
        int nlx = 100*i;
        int nly = 100*j;

        log.trace("reallocate "+nlx+","+nly +" xlen, ylen: "+xlen+","+ylen+" i,j"+ i+","+j);
        if (noalign)
        {
            noAlignM=new long[2][nly];
        }
        else
        {
            M=new long[nlx][nly];
        }
        
        Ivec = new long[nly];
        simVec = new long[nly];

    }

    int current_n, current_m;

    

    public static class FullMatrixSimCalculator implements SimCalculator
    {
        long[][] m;
        long factor;
        FullMatrixSimCalculator(long[][] m, long factor)
        {
            this.m=m;
            this.factor=factor;
        }

        public long calcSim(int i, int j)
        {
            return m[i][j];
        }
        public long[] calcSim(int i, long[] target)
        {
            return m[i];
        }
        public long getFactor()
        {
            return factor;
        }
    }

    public double score(int n, int m, long[][] simmatrix, long factor)
    {
        return align(n,m,simmatrix, factor, true);
    }
    
    public Alignment align(int n, int m, long[][] simmatrix, long factor)
    {
        Alignment a = new Alignment(n, m);
        a.score=align(n,m,simmatrix, factor, false);
        a.alignment=getAlignedPositions();
        return a;
    }

    public double score(int n, int m, SimCalculator sc)
    {
        return align(n,m,sc, true);
    }
    public Alignment align(int n, int m, SimCalculator sc)
    {
        Alignment a = new Alignment(n, m);
        a.score=align(n,m,sc, false);
        a.alignment=getAlignedPositions();
        return a;
        
    }
    private double align(int n, int m, long[][] simmatrix, long factor, boolean noalign) 
    {
        return align_unnormed(n,m,new FullMatrixSimCalculator(simmatrix, factor), noalign)/(double)factor;
    }

    private double align(int n, int m, SimCalculator sc, boolean noalign) 
    {
        return align_unnormed(n,m,sc, noalign)/(double)sc.getFactor();
    }

    SimCalculator current_simcalculator = null;

    private long align_unnormed(int n, int m, SimCalculator sc, boolean noalign) 
    {
        reallocate(n + 1, m+1, noalign);

        current_n=n;
        current_m=m;
        current_simcalculator = sc;
        
        startx=starty=-1;
        long gapFirst = (long)(sc.getFactor()*(gap_open+gap_extend));
        long gapIncrement = (long)(sc.getFactor()*gap_extend);

        //System.out.printf("GOTOHCGX calc ali for n,m=%d,%d  mode: %s gapfirst: %d inc: %d\n", n, m, alignMode, gapFirst, gapIncrement);

        long max = 0;
        int max_i = 0;
        int max_j = 0;
        long Dval;
        long v1, v2;

        long[] Ivec = this.Ivec;
        long[] simVec = this.simVec;
        int i, j, lasti, lastj;

        long[][] M = (noalign)?this.noAlignM:this.M;

        M[0][0] = 0;
        
        long initGapIncrement = gapIncrement;

        switch (alignMode)
        {
            case GLOBAL:
            case GLOBFREE:
                initGapIncrement=gapIncrement;
                break;
            default:
                initGapIncrement=0;
                break;
        }
        
        switch (alignMode)
        {
            case GLOBAL:
            case GLOBFREE:
                for (j=1; j<=m; j++)
                {
                    M[0][j] = M[0][j-1]+gapIncrement;
                }
                break;
            default:
                for (j=1; j<=m; j++)
                {
                    M[0][j] = 0;
                }
                break;

        }


        for (j=1; j<=m; Ivec[j++]=INF);

        


        long worstscore=(alignMode==AlignMode.LOCAL)?0:INF;

        

        int lastMIndex=0;
        int currentMIndex=0;

        Pair<Integer,Integer> maxScorePos =new Pair<Integer,Integer>(0,0);
        
        long bestRowScore=Long.MIN_VALUE;
        int bestRowpos = -1;

        for (i=1, lasti=0; lasti<n; i++, lasti++)
        {
            if (noalign)
            {
                lastMIndex=currentMIndex;
                currentMIndex=1-lastMIndex;
            }
            else
            {
                lastMIndex=lasti;
                currentMIndex=i;
            }
            
            
            M[currentMIndex][0]=M[lastMIndex][0]+initGapIncrement;

            //System.out.printf("M[%d,%d]=%d (initgap: %d)\n",currentMIndex, 0, M[currentMIndex][0], initGapIncrement);
            bestRowScore=Long.MIN_VALUE;
            bestRowpos = -1;
            simVec=sc.calcSim(lasti, simVec);
            Dval = INF;
            //System.out.printf("\t");
            for (j=1, lastj=0; lastj<m; j++, lastj++)
            {
                Ivec[j]=Math.max(Ivec[j]+gapIncrement,M[lastMIndex][j]+gapFirst);
                Dval = Math.max(Dval+gapIncrement,M[currentMIndex][lastj]+gapFirst);
                if ( bestRowScore < (M[currentMIndex][j] = Math.max(worstscore, Math.max(M[lastMIndex][lastj]+simVec[lastj], Math.max(Ivec[j], Dval)))))
                {
                    bestRowScore=M[currentMIndex][j];
                    bestRowpos=j;
                }

                //System.out.printf("\t%d,%d -> %d I: %d D: %d diag: %d sim: %d\n", currentMIndex,j, M[currentMIndex][j], Ivec[i], Dval,
                //                    M[lastMIndex][lastj], simVec[j]);
                /*System.out.printf("%d,%d: %d sim %d M[%d][%d]=%d ivec: %d\t", i,j,M[currentMIndex][j], simVec[lastj], lastMIndex, lastj, M[lastMIndex][lastj], Ivec[j]);
                if (j%10==0)
                {
                    System.out.printf("\n\t");
                }
                */

            }

            //System.out.printf("\n\ni: %d/%d bestscore: %d at: %d max: %d last [%d][%d]: %d gap Ivec: %d\n", i,n, bestRowScore, bestRowpos, max, currentMIndex, m, M[currentMIndex][m], Ivec[j]);
            //System.out.printf("gap open: %d inc: %d\n", gapFirst, gapIncrement);
            if (alignMode==AlignMode.GLOBAL)
            {
                max = M[currentMIndex][m];
                continue;
            }
                
            
            if (alignMode == AlignMode.FREESHIFT || alignMode == AlignMode.GLOBFREE)
            {
                if ( max>M[currentMIndex][m] && i<n )
                    continue;
                
                if (max<M[currentMIndex][m])
                {
                    max=M[currentMIndex][m];
                    startx=i;
                    starty=m;
                    //System.out.printf("update max: %d,%d -> %d\n", startx, starty, max);
                }
                if (i<n)
                    continue;
                
                
                if (max<bestRowScore)
                {
                    max=bestRowScore;
                    startx=i;
                    starty=bestRowpos;
                    //System.out.printf("update max: %d,%d -> %d\n", startx, starty, max);
                }
                continue;
            }

            //local

            if (max>=bestRowScore)
                continue;
            
            max=bestRowScore;
            startx=i;
            starty=bestRowpos;
        }   
        if (alignMode==AlignMode.GLOBAL) 
        {
            startx=n;
            starty=m;
        }
        return max;


    }

    public Vector<Pair<Integer,Integer> > getAlignedPositions()
    {
        
        return getAlignedPositions(startx, starty);
    }

    public Pair<Integer,Integer> getAliStart()
    {
        return new Pair<Integer,Integer>(startx,starty);

    }
    public Vector<Pair<Integer,Integer> > getAlignedPositions(int i, int j)
    {

        if (current_simcalculator==null)
        {
            throw new RuntimeException("no simcalculator yet set!");
        }
        
        Vector<Pair<Integer,Integer> >aligned_positions = new Vector<Pair<Integer,Integer> >();

        if (i<=0 || j<=0)
        {
            return aligned_positions;
        }
    
                         
        SimCalculator sc = this.current_simcalculator;

        long gapFirst = (long)(sc.getFactor()*(gap_open+gap_extend));
        long gapIncrement = (long)(sc.getFactor()*gap_extend);

        long s;

        long gaptest;
        int testindex;
        int step;
        

        
        do
        {
            s = M[i][j];
            
            if (s == M[i-1][j-1]+sc.calcSim(i-1,j-1))
            {
                aligned_positions.add(new Pair<Integer,Integer>(i-1, j-1));
                i--; j--;
                continue;
            }

            for (step=1, gaptest=gapFirst; (i>=step || j>=step); step++, gaptest+=gapIncrement)
            {

                if (i>=step &&  M[i][j]==M[i-step][j]+gaptest)
                {
                    i-=step;
                    break;
                }          
                
                if (j>=step && M[i][j]==M[i][j-step]+gaptest)
                {
                    j-=step;
                    break;
                }
            }
            
        } 
        while (i!=0 && j!=0 && (alignMode!=AlignMode.LOCAL || M[i][j]!=0));


        //System.out.println("aligned positions: "+aligned_positions);
        return aligned_positions;
    }
    
    
    static public double rescore(Vector<Pair<Integer,Integer> > ali, int qsize, int tsize, SimCalculator sc, float gap_open, float gap_extend, 
                                AlignMode alignMode)
    {
        return rescore(ali, qsize,tsize,sc,gap_open,gap_extend,alignMode,false);
    }


    static public long rescoreUnormed(Vector<Pair<Integer,Integer> > ali, int qsize, int tsize, SimCalculator sc, long lgo, long lge, 
                                AlignMode alignMode, boolean verbose)
    {
        int xgaps=0;
        int ygaps=0;
        long checkscore=0;

        
        if (verbose)
        {
            System.out.println("rescore in mode "+alignMode+ " ali: "+ ali);
        }
        Pair<Integer,Integer> lastmatch = null;

        int open_xgap = 0;
        int open_ygap = 0;
        int end_ygap = 0;
        int end_xgap = 0;
        int open_gaps=0;
        int end_gaps=0;

        if (ali.size()>0)
        {
            open_xgap=ali.get(0).getFirst();
            open_ygap=ali.get(0).getSecond();

            open_gaps = Math.min(open_xgap, open_ygap);

            end_xgap = qsize-ali.get(ali.size()-1).getFirst()-1;
            end_ygap = tsize-ali.get(ali.size()-1).getSecond()-1;

            end_gaps = Math.min(end_xgap, end_ygap);


        }

        long initgapscore=0;
        if (alignMode==AlignMode.FREESHIFT && ali.size()>0 && end_gaps>0)
        {
            initgapscore+=lgo;
            initgapscore+=(lge*end_gaps);
            
        }
        long total_matchscores=0;
        for (int ai=ali.size()-1; ai>=0; ai--)
        {
            Pair<Integer,Integer> p = ali.get(ai);
            if (lastmatch!=null)
            {
                int gap1=lastmatch.getFirst()-p.getFirst()-1;
                int gap2=lastmatch.getSecond()-p.getSecond()-1;

                if (gap1>0)
                {
                    checkscore+=lgo;
                    checkscore+=(gap1*lge);
                    xgaps+=gap1;

                }
                if (gap2>0)
                {
                    checkscore+=lgo;
                    checkscore+=(gap2*lge);
                    ygaps+=gap2;

                }

            }

            long matchscore=sc.calcSim(p.getFirst(),p.getSecond());
            
            checkscore+=matchscore;
            total_matchscores+=matchscore;
            if (verbose)
            {
                System.out.printf(" (%d,%d) = %d , total match: %d check: %d prefgep check:%d \n", p.getFirst(), p.getSecond(), matchscore, total_matchscores, checkscore, checkscore+initgapscore);
            }

            lastmatch=p;
        }

        //freeshift add end costs
        if (ali.size()>0)
        {
            if (verbose)
            {
                System.out.printf(" open gaps: (%d,%d): min %d, end gaps (%d,%d) min :: %d lgo: %d lge: %d\n", 
                                  open_xgap, open_ygap, open_gaps, 
                                  end_xgap, end_ygap, end_gaps, lgo, lge);
            }

            switch (alignMode)
            {
                case FREEGLOB:
                case FREESHIFT:
                    if (open_gaps>0)
                    {
                        checkscore+=lgo;
                        checkscore+=(open_gaps*lge);

                    }

                    break;
                case GLOBFREE:
                case GLOBAL:
                    if (open_xgap>0)
                    {
                        checkscore+=lgo;
                        checkscore+=(open_xgap*lge);
                    }
                    if (open_ygap>0)
                    {
                        checkscore+=lgo;
                        checkscore+=(open_ygap*lge);
                    }
                    break;
                default:
                    break;

            }

            switch (alignMode)
            {
            
                case FREESHIFT:
                case GLOBFREE:
                    if (end_gaps>0)
                    {
                        checkscore+=lgo;
                        checkscore+=(end_gaps*lge);

                    }

                    break;
                case GLOBAL:
                case FREEGLOB:
                    if (end_xgap>0)
                    {
                        checkscore+=lgo;
                        checkscore+=(end_xgap*lge);
                    }
                    if (end_ygap>0)
                    {
                        checkscore+=lgo;
                        checkscore+=(end_ygap*lge);
                    }
    
                    break;
                case LOCAL:

                    break;
            }


        }
        else
        {
            if (alignMode==AlignMode.GLOBAL )
            {
                checkscore+=(2*lgo);
                checkscore+=(lge*(qsize+tsize));
            }
        }

        return checkscore;
    }


    static public double rescore(Vector<Pair<Integer,Integer> > ali, int qsize, int tsize, SimCalculator sc, float gap_open, float gap_extend, 
                                AlignMode alignMode, boolean verbose)
    {

        long lgo=(long)(sc.getFactor()*gap_open);
        long lge=(long)(sc.getFactor()*gap_extend);
        long checkscore = rescoreUnormed(ali,qsize,tsize,sc,lgo,lge,alignMode,verbose);
        return checkscore/sc.getFactor();
    }

    
}

