package isoformtransfer.algo;

import genome.Gene;
import genome.GenomicUtils;
import genome.Protein;
import isoformtransfer.InferredExon;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.List;
import java.util.Vector;

import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.Pair;

public abstract class ExternalAlgo implements TransferAlgo {
	
    protected Logger log = LogConfig.getLogger();
    protected boolean useProtein = false;
    
	protected File tmp_gene_dna;
	protected boolean needFastaHeader = false;
	/**
	 * @param line -- output of external program
	 * @return transformation into InferredExon
	 */
	public abstract List<InferredExon> parseExternalProgramOutputLine(String line);
	
	/**
	 * Returns the command parameters 
	 * @param cdna -- file with input CDNA data
	 * @param DNA -- file with input DNA data
	 * @return parameters
	 */
	public abstract List<String> getCommand(File cdna, File DNA);
	
	public boolean useProtein(){
		return useProtein;
	}
	
	@Override
	public Object clone()
	{
	    try
	    {
	        return super.clone();
	    }catch(Exception e)
	    {
	        throw new RuntimeException(e);
	    }
	}
	@Override
	public void setToGene(Gene toGene,String geneSequence) {
		if (tmp_gene_dna != null && tmp_gene_dna.exists()){
			tmp_gene_dna.delete();
		}
		if( geneSequence != null){

			try
			{
				tmp_gene_dna = File.createTempFile("trans_gene", ".tmp");
				tmp_gene_dna.deleteOnExit();
			}
			catch (IOException e)
			{
				throw new RuntimeException("i/o error: "+e.getMessage()+" while creating tmpfiles for sim4cc transfers", e);
			}    

			try
			{
				PrintWriter pw = new PrintWriter(tmp_gene_dna);
				if ( needFastaHeader )pw.write(">DNA\n");
				fasta.FastaLikeReader.writeFasta(geneSequence, pw);
				pw.close();

			}
			catch (IOException e)
			{
				throw new RuntimeException("i/o error: "+e.getMessage()+" while writing tmp genome file for sim4cc", e);
			}
		}
	}
	public void postProess(){
		
	}
	@Override
	public Vector<InferredExon> transfer(Protein toProtein,String cDNA, String protSeq) {

	    
	    if ( tmp_gene_dna == null)
	    {
	        throw new RuntimeException("target gene not set!");
	    }
		File tmp_query = null;
		//create tmp file
		try
		{
			tmp_query = File.createTempFile("trans_query", ".tmp");
			tmp_query.deleteOnExit();
		}
		catch (IOException e)
		{
			throw new RuntimeException("i/o error: "+e.getMessage()+" while creating tmpfiles for transfers", e);
		}    
		//write cDNA to tmp file
	     try
         {
	 		PrintWriter pw = new PrintWriter(tmp_query);
	        if( needFastaHeader) pw.write(">Query\n");
	        if ( useProtein){
	        	fasta.FastaLikeReader.writeFasta(protSeq, pw);
	        }else{
	        	fasta.FastaLikeReader.writeFasta(cDNA, pw);
	        }
	        
	        pw.close();
         }
         catch (IOException e)
         {
             throw new RuntimeException("i/o error: "+e.getMessage()+" while writing tmp cdna file", e);
         }

		List<String> command = this.getCommand(tmp_query,tmp_gene_dna);
		log.trace("will invoke: "+command);

		ProcessBuilder builder = new ProcessBuilder(command  );

		builder.redirectErrorStream(true);

		Vector<InferredExon> rv = new Vector<InferredExon>();

		try
		{
			this.setEnviroment(builder);
			Process process = builder.start();
			
			BufferedReader reader = new BufferedReader (new InputStreamReader(process.getInputStream()));

			String line = null;

			while ((line = reader.readLine ()) != null)
			{

				List<InferredExon> exon = this.parseExternalProgramOutputLine(line); 

				if (exon==null || exon.size() == 0)
				{
					log.trace("skipped: "+line);
					continue;
				}

				/*
				// to close to be a real intron, merge the region... 
				if (rv.size()>0 && rv.get(rv.size()-1).target.getSecond()+MINIMUM_INTRON_SIZE >= exon.target.getFirst())
				{
					//merge
					InferredExon pre = rv.get(rv.size()-1);
					InferredExon merged = new InferredExon();
					merged.infoline=pre.infoline+" :: "+exon.infoline;
					merged.source = new Pair<Integer,Integer>(pre.source.getFirst(), exon.source.getSecond());
					merged.target = new Pair<Integer,Integer>(pre.target.getFirst(), exon.target.getSecond());

					merged.dna_seqid = (pre.dna_seqid*pre.getTargetLength() + exon.dna_seqid * exon.getTargetLength())/(pre.getTargetLength()+exon.getTargetLength());

					rv.set(rv.size()-1, merged);
					continue;
				}
				*/
				for(InferredExon inferredExon : exon){
			        final int src_start=inferredExon.source.getFirst();
	                final int src_end=inferredExon.source.getSecond();
					if ( src_end-src_start < 3 || inferredExon.target.getSecond()-inferredExon.target.getFirst() < 3 ) continue; //to short, or matches in reverse strand
					rv.add(inferredExon);
					
				}
				
			}

			reader.close();

			process.waitFor();


			int exitValue = process.exitValue();

			if ( exitValue != 0)
			{
				//throw new RuntimeException("");
				log.error("exit with return value != 0");
				return null;

			}

		}
		
		catch (IOException e)
		{
			throw new RuntimeException("i/o error: "+e.getMessage()+" while running sim4cc", e);
		}
		catch (InterruptedException iex)
		{
			throw new RuntimeException("sim4cc run interrupted: "+iex.getMessage(), iex);
		}
		if (tmp_query.exists()){
			tmp_query.delete();
		}
		if ( tmp_gene_dna.exists()){
		    tmp_gene_dna.delete();
		    tmp_gene_dna = null;
		}

		return rv;
	}

	public void setEnviroment(ProcessBuilder builder) {
		// by default: do nothing
	}
	public Vector<InferredExon> removeUTR(Vector<InferredExon> exons,String targetGeneSequence, int tssStart, int tssEnd){
		int startExon = -1;
		int endExon = -1;
		for(int i = 0 ; i< exons.size(); i++){
			
			if (startExon == -1 && exons.get(i).source.first > tssStart) startExon = i;
			if (endExon == -1 && exons.get(i).source.first > tssEnd) endExon = i;
		}
		
		int bestStart = 0;
		int bestStartX = Integer.MAX_VALUE;
		System.out.println(startExon + "\t" + endExon);
		for(int i= exons.get(startExon).target.first ; i<= exons.get(startExon).target.second-3; i++){
			String sequence = targetGeneSequence.substring(i,exons.get(startExon).target.second);
			String translated = GenomicUtils.translate(sequence);
			int x=0;
			for(int k = 0 ; k< translated.length();k++){
				if ( translated.charAt(k) == 'X')x++;
			}
			if ( x < bestStartX){
				bestStartX = x;
				bestStart = i;
			}
		}
		
		int bestEnd = 0;
		int bestEndtX = Integer.MAX_VALUE;
		
		for(int i= exons.get(endExon).target.second +3; i>= exons.get(endExon).target.first; i--){
			String sequence = targetGeneSequence.substring(exons.get(endExon).target.first,i);
			String translated = GenomicUtils.translate(sequence);
			int x=0;
			for(int k = 0 ; k< translated.length();k++){
				if ( translated.charAt(k) == 'X')x++;
			}
			if ( x < bestEndtX){
				bestEndtX = x;
				bestEnd = i;
			}
		}
		
	
		Vector<InferredExon> inferred= new Vector<InferredExon>();
		InferredExon newStart = new InferredExon();
		newStart.target = new Pair(exons.get(bestStart).target.first+bestStart,exons.get(bestStart).target.second);
		newStart.source = exons.get(bestStart).source;
		inferred.add(newStart);
		
		for(int i = bestStart+1 ; i<= bestEnd-1;i++){
			inferred.add(exons.get(i));
		}

		InferredExon newEnd = new InferredExon();
		newEnd.target = new Pair(exons.get(bestEnd).target.first,exons.get(bestEnd).target.first+bestEnd);
		newEnd.source = exons.get(bestStart).source;
		inferred.add(newEnd);
		
		return inferred;
	}
	

}
