package isoformtransfer.isar;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Vector;

import javax.imageio.ImageIO;

import genome.Organism;
import isoformtransfer.algo.Spaln2;
import isoformtransfer.algo.TransferAlgo;
import isoformtransfer.isar.PartialOrderMultipleAlignment.DebugStat;
import isoformtransfer.isar.PartialOrderMultipleAlignment.OracleStat;
import isoformtransfer.isar.PartialOrderMultipleAlignment.UnmatchedStrategy;
import isoformtransfer.isar.oracle.MultipleAlignmentOracle;
import isoformtransfer.isar.oracle.MultiplePairwiseOracle;
import isoformtransfer.isar.oracle.Oracle;
import isoformtransfer.isar.oracle.PairwiseAlignmentOracle;
import isoformtransfer.isar.oracle.SplicedAlignmentOracle;
import lmu.utils.ConfigFileGetter;
import lmu.utils.ConfigFileGetter.ConfigFileNotFound;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.NumUtils;
import lmu.utils.ObjectGetter;
import lmu.utils.OptionParser;
import lmu.utils.Pair;
import lmu.utils.SimpleOptionParser;
import lmu.utils.StringUtils;
import protein.algo.external.MultipleAlignmentAlgo;
import protein.algo.external.PRRN;

public class ISAR {
    
    public static void printISARInfo(PrintStream ps)
    {
        ps.println("\nISAR v.1.0 (24-02-2016).\nCopyright 2014-2016 by Robert Pesch, Gergely Csaba and Ralf Zimmer ({pesch, csaba, zimmer}@bio.ifi.lmu.de).\n");
        ps.println("THIS SOFTWARE IS PROVIDED \"AS IS\", WITH ABSOLUTELY NO WARRANTY EXPRESSED OR IMPLIED. ANY USE IS AT YOUR OWN RISK.\n");
        ps.flush();
        
    }
    static class OrganismNameReader
    {
        static OrganismNameReader instance = null;
        private HashMap<String, String> map;
        
        private OrganismNameReader()
        {
            Logger log = LogConfig.getLogger();
            
            //File f = new File("mapping/organism");
            File f;
            try {
                log.info("Read organism mapping");
                f = ConfigFileGetter.getFile(ISAR.class, "mapping", "organism");
            } catch (ConfigFileNotFound e) {
                log.warn("Cannot read taxID-organism mapping");
                return;
            }
            
            map = FileUtils.readMap(f, "\t", 1, 0);
            
        }
        
        private static void init()
        {
            if ( instance == null) 
                instance = new OrganismNameReader();
        
        }
        public static String getShortName(int taxId)
        {
            init();
          
            if ( instance != null && instance.map != null)
            {
                String name = instance.map.get(taxId +"");
                
                String[] tokens =  name.split("\\s+");
                if ( tokens.length < 2)
                    return name;
                return tokens[0].charAt(0) + ". " + tokens[1];
                
                
            }
            return null;
        }
    }
    
    /*
    @Test
    public void test()
    {
        PartialOrderMultipleAlignment isar = ISAR.getISAR(new File("/home/proj/projekte/expressionlab/isar/Rab/msa/RAB1A.1B.xml.gz"),UnmatchedStrategy.ONLY_EXTENDED,false);
        for(POMAGene g1 : isar.getGenes())
        {
            for(POMAGene g2 : isar.getGenes())
            {
                Vector<UPair<Region1D>> m = isar.getPOA().getRegionMatchInfo().getMatchedRegions(g1, g2);
          //      System.out.println(m);
            }
        }
       
        List<HashSet<Pair<POMAGene, Integer>>> it = isar.getAlignment();
        for(HashSet<Pair<POMAGene,Integer>> i : it)
        {
            System.out.println(i);
        }
        
    }
    */
    public static PartialOrderMultipleAlignment getISAR(File f)
    {
        return new PartialOrderMultipleAlignment(f);
    }
    public static PartialOrderMultipleAlignment getISAR(File f,UnmatchedStrategy  unmatchedStrategy,boolean skipFormSplicedAlignment)
    {
        return new PartialOrderMultipleAlignment(f,unmatchedStrategy,skipFormSplicedAlignment);
    }
    
    public static class ISAROptionParser extends SimpleOptionParser{
        HashMap<String,List<Option>> category = new LinkedHashMap<>();
        HashMap<Option,String> type = new HashMap<>();
        
        
        public ISAROptionParser(String ... keys)
        {
             addSimpleOptions(keys);
        }
        @Override
        protected String getName() {
            return "ISAR";
        }
        public void setOptional(String ... keys)
        {
            for (String k: keys)
            {
                Option o =super.getOption(k);
                
                o.default_value="";
            }
        }
        public void addType(String name, String ... keys )
        {
            for(String key : keys){
                Option option = this.getOption(key);
                if ( option == null)  throw new RuntimeException("Unknown option:" + name);
                
                this.type.put(option,name);
            }
        }
        public void addToCategory(String name, String ... keys )
        {
            if ( !this.category.containsKey(name))
                this.category.put(name, new ArrayList<>());
            
            for(String key : keys){
                Option option = this.getOption(key);
                if ( option == null)  throw new RuntimeException("Unknown option:" + name);
                
                this.category.get(name).add(option);
            }
        }
        
        @Override
        public String getHelp()
        {
            StringBuffer sb = new StringBuffer();
            Vector<Option> params = new Vector<Option>();
            params.addAll(this.getOptions());
            Collections.sort(params);
            
            int maxName = NumUtils.max(params, (Option o) -> o.getOptionName().length());
            int maxDefault = NumUtils.max(params, (Option o) -> o.default_value==null?0:(o.default_value.length()+1));
            int maxType = NumUtils.max(params, (Option o) -> o.type==null?0:o.type.length());
            
            
            Set<Option> processed = new HashSet<Option>();
            
            for(Entry<String,List<Option>> e : category.entrySet())
            {
                sb.append(e.getKey().toUpperCase() + ":\n");
                for (Option o : e.getValue())
                {
                    processed.add(o);
                    sb.append(o.getHelp(maxName,maxDefault,maxType)+"\n");
                }   
                sb.append("\n");
            }
            
            sb.append("GENERAL:\n");
            for (Option o : params)
            {
                if ( processed.contains(o))
                    continue;
                sb.append(o.getHelp(maxName,maxDefault,maxType)+"\n");
            }
            return sb.toString();
        }
    }
    
    
    public static PartialOrderMultipleAlignment getPOMA(SimpleOptionParser cmdLine) throws IOException{
        Logger log = LogConfig.getLogger();
        
        if ( cmdLine.isOptionSet("alignment") && cmdLine.isOptionSet("prrn"))
        {
            log.fatal("Cannot combine the options alignment and prrn. Choose only one of them.");
            return null;
        }
        /*
        if ( cmdLine.isOptionSet("alignment") && !cmdLine.isOptionSet("init_from_msa"))
        {
            log.info("Activate init_from_msa swith" );
            cmdLine.getOption("init_from_msa").is_set =true;
            
        }
        */
        
        UnmatchedStrategy unmatchedStrategy = null;
        
        if ( cmdLine.getValue("update_mode").equals("both"))
            unmatchedStrategy = UnmatchedStrategy.BOTH;
        else if ( cmdLine.getValue("update_mode").equals("basic"))
            unmatchedStrategy = UnmatchedStrategy.ONLY_BASIC;
        else if ( cmdLine.getValue("update_mode").equals("extended")){
            unmatchedStrategy = UnmatchedStrategy.ONLY_EXTENDED;
        }else{
            log.error("Unknown update_mode: %s", cmdLine.getValue("update_mode"));
            return null;
        }

        TransferAlgo transferAlgo = null;
        if ( cmdLine.isOptionSet("spaln2"))
        {
            transferAlgo = new Spaln2(cmdLine.getFile("spaln2"), true,true);
        }else{
            URL spaln =ISAR.class.getClassLoader().getResource("spaln");
            if ( spaln != null)
            {
                File f = new File(spaln.getFile());
                transferAlgo = new Spaln2(f, true,true);
            }
        }
        Integer threads = cmdLine.getInt("threads");
        List<Integer> taxIds = new ArrayList<Integer>();
        boolean useTaxIDs = true;
        List<POMAGene> genes = new ArrayList<POMAGene>();
        
        HashSet<String> g_ids = new HashSet<>();
        HashSet<String> p_ids = new HashSet<>();
        
        boolean cf = cmdLine.isSet("common_fasta");
        
        if(cf)
            log.info("Common fasta file flag is set. Note that this option is experimental. No isoform interrelationsships are considered and no spliced aligment oracle can be used");

        //read genes from fasta
        List<File> pomas = new ArrayList<>();
        for(String i : cmdLine.getValue("in").split(","))
        {
            File f = new File(i);
            if ( !f.exists())
            {
                log.fatal(f + " does not exist");
                return null;
            }
            log.info("Read: %s",f);
            
            if (  isISAR(f))
            {
                pomas.add(f);
                continue;
            }
            
            if ( cf)
            {
                genes.addAll(POMAGene.readFromCommonFasta(f));
                continue;
            }
            
            
            POMAGene pg = new POMAGene(f,transferAlgo);
            if ( pg.getProteins().size()==0)
            {
                log.warn("Skip gene: %s (no protein information given)",pg.getId());
                continue;
            }
            genes.add(pg);
            
        }
        //check genes
        for(POMAGene pg: genes){
            if ( g_ids.contains(pg.getId())){
                log.fatal("Redundant gene id:" + pg.getId()  + " in input files");
                return null;
            }
            g_ids.add(pg.getId());
            
            for(String proteinId : pg.getProteins())
            {
                if ( p_ids.contains(proteinId)){
                    log.fatal("Redundant protein id:" + proteinId + " in input files");
                    return null;
                }
                p_ids.add(proteinId);
                    
            }
            
            if (pg.getTaxId() != null && pg.getTaxId() != -1)
            {
                
                if ( taxIds.contains(pg.getTaxId()) )
                {
                    if (  !taxIds.get(taxIds.size()-1).equals(pg.getTaxId()))
                    {
                        useTaxIDs = false;
                    }
                    continue;
                }
                
                taxIds.add(pg.getTaxId());
            }else{
                useTaxIDs = false;
            }
        }

        File out = cmdLine.getFile("out");
        
        //setup oracles
        
        
        MultipleAlignmentOracle msaOra = null;
        
        if ( cmdLine.isOptionSet("prrn"))
        {
            File mulRes = null;
            if ( cmdLine.isOptionSet("prrn_output"))
            {
                mulRes = cmdLine.getFile("prrn_output");
            }else{
                mulRes = File.createTempFile("prrn.", ".fasta");
            }
            MultipleAlignmentAlgo prrn = new PRRN(cmdLine.getFile("prrn"),false,10);
            msaOra = new MultipleAlignmentOracle(mulRes,prrn,genes,useTaxIDs?taxIds:null);
        }
        if ( cmdLine.isOptionSet("alignment"))
        {
            List<File> files = new ArrayList<>();
            for(String f : cmdLine.getValue("alignment").split(","))
            {
                File file = new File(f);
                if (! file.exists())
                {
                    log.fatal(f + " does not exist");
                    return null;
                }
                files.add(file);
            }
            msaOra = new MultipleAlignmentOracle(files.toArray(new File[files.size()]));
        }
        
        
        SplicedAlignmentOracle pwSplicedAlignedOra = null;
        if ( cmdLine.isOptionSet("spaln2"))
        {
            if (cmdLine.isOptionSet("spaln2") )
            {
                TransferAlgo tranferAlgo =  new Spaln2(cmdLine.getFile("spaln2"), true, true);
                pwSplicedAlignedOra = new SplicedAlignmentOracle(threads,tranferAlgo);
            }        
        }

        PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(genes,unmatchedStrategy,out.getName());
        if ( out != null && out.getParent() != null)
            poma.setDebugConfiguration(new File(out.getParent()), DebugStat.NONE, true);
        //read genes from input isars
        for(File f: pomas)
        {
            //recompute
            PartialOrderMultipleAlignment inPoma = null;
            
            if ( cmdLine.isSet("recompute"))
            {
                inPoma = PartialOrderMultipleAlignment.getEmptyPOMA(f) ;
            }else{
                inPoma = new PartialOrderMultipleAlignment(f);
                    
            }
            
            boolean ret =merge(g_ids,p_ids,poma,inPoma);
            if ( !ret) 
                return null;
        }
        if (poma.getGenes().size()== 0)
        {
            log.warn("Nothing to align");
            return null;
        }
        PairwiseAlignmentOracle pw = cmdLine.isSet("no_refinement")?null:new PairwiseAlignmentOracle(threads);

        MultiplePairwiseOracle mulpw = new MultiplePairwiseOracle(msaOra,pw);
        
        List<Oracle> oracles = new ArrayList<>();
        if ( cmdLine.isSet("init_from_msa"))
        {
            oracles.add(msaOra);
        }
        oracles.add(mulpw);
        if (pwSplicedAlignedOra != null )
        {
            oracles.add(pwSplicedAlignedOra);
        }
        if ( cmdLine.isSet("no_filter"))
        {
            poma.setMIN_SEQ_ID(0f);
            poma.setMIN_COVERAGE(0f);
            poma.setMIN_MATCHES(0);
        }else{
            poma.setMIN_SEQ_ID((double)cmdLine.getDouble("min_region_seq_id"));
            poma.setMIN_COVERAGE((float)cmdLine.getDouble("min_exon_coverage"));;
            poma.setMIN_MATCHES(cmdLine.getInt("min_matches_region"));;
        }
        log.info("Compute ISAR with the following oracles:" + StringUtils.join(",",oracles,(Oracle o)->o.getName()));
        
        poma.createGraph(oracles.toArray(new Oracle[]{}));
        
        if ( !cmdLine.isSet("no_xml"))
            poma.writePOMA(cmdLine.getFile("out"));

        List<HashSet<Pair<POMAGene, Integer>>> mapping = null;
        
        if ( cmdLine.isSet("exon_view") || cmdLine.isSet("fasta"))
        {
            
            log.info("Create complete mapping");
            mapping = poma.getCompleteMapping();
        }
        
        
    //    File out = cmdLine.getFile("out");
        
        if ( cmdLine.isSet("exon_view"))
        {
            try{
                Drawer.SCALE = cmdLine.getDouble("scale_factor");
                BufferedImage bim = Drawer.drawAlignmentComplete(poma, mapping, true, true,true,30, poma.getGenes());
                File outImg = new File(out + ".png");
                
                log.info("Write: %s",outImg);
                
                ImageIO.write(bim, "png", outImg);
            }catch(Exception e)
            {
                log.warn("Error creating exon view");
                e.printStackTrace();
            }

        }
        
        if (cmdLine.isSet("fasta"))
        {
            Formater formater= new Formater(poma,mapping);;
               
            File outFasta = new File(out + ".fasta");
            log.info("Write: %s",outFasta);
            
            PrintWriter pwi = FileUtils.getWriter(outFasta);
            
            formater.printISAR(pwi,true,poma.getGenes(),0);
            pwi.close();
        }
        

        poma.writeTimeLog();
        return poma;
    }
    public static boolean isISAR(File f)
    {
        BufferedReader reader = FileUtils.getReader(f);
        boolean ret = false;
        try{
            String firstLine = reader.readLine();
            if ( firstLine.trim().startsWith("<PartialOrderMulAlignment")) ret = true;
        }catch(Exception e)
        {
            throw new RuntimeException(e);
        }
        return ret;
        //return doc.getRootElement().getName().equals(PartialOrderMultipleAlignment.POMA_XML_STRING);
    }
    public static boolean merge( HashSet<String> g_ids,  HashSet<String> p_ids, PartialOrderMultipleAlignment poma, PartialOrderMultipleAlignment inPoma)
    {
        Logger log = LogConfig.getLogger();
        
        for(POMAGene gene : inPoma.getGenes())
        {
            if ( g_ids.contains(gene.getId())) //same gene in input files
            {
                POMAGene targetGene = poma.getGene(gene.getId());
                if (! targetGene.geneSequence.equals(gene.geneSequence))
                {
                    log.fatal("Cannot extend ISARs, due to inconsistent gene sequence for gene:" + gene.getId() + " in input files");
                    return false;
                }

                for(String pId : gene.getProteins())
                {
                    if (! targetGene.getProteins().contains(pId))
                    {
                        if ( p_ids.contains(pId))
                        {
                            log.fatal("Redundant protein id:" + pId  + " in input files");
                            return false;
                        }
                        gene.addProtein(pId, gene.getSequence(pId, true), gene.getRegionVector(pId));
                        p_ids.add(pId);

                    }else{
                        if (! gene.getSequence(pId, true).equals(targetGene.getSequence(pId, true)))
                        {
                            log.fatal("Cannot extend ISARs, due to inconsistent protein sequence for gene/protein:" + gene.getId() + "/" +  pId +  "in input files");
                            return false;
                        }
                        if (! gene.getRegionVector(pId).equals(targetGene.getRegionVector(pId)))
                        {
                            log.fatal("Cannot extend ISARs, due to inconsistent exon annotationfor gene/protein:" + gene.getId() + "/" +  pId +  "in input files");
                            return false;
                        }
                    }
                }

                targetGene.init();
                continue;
            }

            //check for protein id clash
            for(String pId : gene.getProteins())
            {
                if ( p_ids.contains(pId))
                {
                    log.fatal("Redundant protein id:" + pId  + " in input files");
                    return false;
                }
                p_ids.add(pId);
                
            }
            g_ids.add(gene.getId());
            
            poma.registerPOMAGene(gene);
        }
        
        //copy alignments
        OracleStat stat =  poma.merge(inPoma);
        
        if ( stat == null)
            return false;
        
        log.info("Merge ISAR %s with %s",poma.getPOMAID() , inPoma.getPOMAID());
        
        
        
        return true;
    }
    public static class ISARBatch {
        public static void main(String[] args) throws IOException {
            ISAROptionParser cmd = ISAR.getCmd();
            cmd.setOptional("in","out");
            
            cmd.addSimpleOptions("batch_input","root_input_dir");
            cmd.setFile("batch_input");
            cmd.setDir("root_input_dir");
            
            
            LogConfig lc = LogConfig.configureLog(null, true);
            

            if (!OptionParser.parseParams(args,true, false, true,  cmd,lc)){
                return;
            }
            
            Vector<String> lines = FileUtils.getLines(cmd.getFile("batch_input"));
            
            for(String line : lines)
            {
             
                Vector<String> map = ObjectGetter.map(line.split(","),(String p) -> cmd.getValue("root_input_dir") + "/" +  p);
                
                cmd.setDefault("in", lmu.utils.StringUtils.join(",", map));
                cmd.setDefault("out", "tmp/" + new File(map.get(0)).getName());
                ISAR.getPOMA(cmd);
            }
            
        }
    }
    public static ISAROptionParser getCmd()
    {
        ISAROptionParser cmdLine = new ISAROptionParser("in","out","alignment","exon_view","scale_factor","fasta","spaln2","prrn","no_refinement","prrn_output","min_region_seq_id","min_matches_region","min_exon_coverage","no_filter","threads","update_mode","init_from_msa","common_fasta","no_xml","recompute");
        cmdLine.setFile("prrn","spaln2");
        cmdLine.setDouble("min_region_seq_id","min_exon_coverage");
        cmdLine.setInt("min_matches_region","threads");
        cmdLine.setOutFile("out");
        
        cmdLine.getOption("exon_view").setHelp("output an image representation of the alignment as PNG file");
        cmdLine.getOption("scale_factor").setHelp("scale factor for exon view");
        cmdLine.setDefault("scale_factor", "1.0");
        cmdLine.getOption("fasta").setHelp("output the multiple sequence alignment as FASTA file");
        
        cmdLine.getOption("no_refinement").setHelp("skip pairwise alignment oracle");
        cmdLine.getOption("alignment").setHelp("comma separated list of precomputed alignment(s) in FASTA/CLUSTAL file format");
        cmdLine.getOption("threads").setHelp("number of parallel threads");
        cmdLine.getOption("in").setHelp("comma separated list of FASTA/ISAR files e.g. examples/RAB1A.fasta,examples/RAB1B.fasta");
        cmdLine.getOption("out").setHelp("output ISAR file e.g. RAB");
        cmdLine.getOption("spaln2").setHelp("path to SPALN2 executable");
        cmdLine.getOption("prrn").setHelp("path to PRRN executable");
        cmdLine.getOption("prrn_output").setHelp("temporary PRRN alignment output file");
        cmdLine.getOption("min_region_seq_id").setHelp("minimum region sequence identity");
        cmdLine.getOption("min_exon_coverage").setHelp("minimum exon coverage");
        cmdLine.getOption("min_matches_region").setHelp("minimum matches in region");
        cmdLine.getOption("recompute").setHelp("ignore matches in a given input ISAR");
        
        cmdLine.getOption("common_fasta").setHelp("input files are ordinary fasta files (experimental)");
        cmdLine.getOption("update_mode").setHelp("mode used to query unmatched regions");
        cmdLine.getOption("init_from_msa").setHelp("initialize the ISAR with computed PRRN, or given pre-computed sequence alignment");
        cmdLine.getOption("no_filter").setHelp("set min_region_seq_id=min_exon_coverage=min_matches_region=0");
        cmdLine.getOption("no_xml").setHelp("do not write a POMA XML files");
        cmdLine.setDefault("min_region_seq_id", "0.4");
        cmdLine.setDefault("min_exon_coverage", "0.5");
        cmdLine.setDefault("min_matches_region", "10");
        
        cmdLine.getOption("update_mode").setEnumParam(1, "basic","extended","both");
        cmdLine.addToCategory("Input/Output", "in","out");
        cmdLine.addToCategory("Additional Outputs", "exon_view","scale_factor","fasta","no_xml");
        cmdLine.addToCategory("Oracle", "prrn","prrn_output","spaln2","alignment","no_refinement");
        cmdLine.addToCategory("Block filter", "min_region_seq_id","min_exon_coverage","min_matches_region","no_filter");
        
        cmdLine.getOption("alignment").setType("FILES");
        
        cmdLine.getOption("in").setType("FILES");
        cmdLine.getOption("out").setType("FILE");
        
        cmdLine.getOption("prrn").setType("EXEC");
        cmdLine.getOption("spaln2").setType("EXEC");
        cmdLine.getOption("prrn_output").setType("FILE");
        
        cmdLine.getOption("scale_factor").setType("N");
        cmdLine.getOption("min_region_seq_id").setType("N");
        cmdLine.getOption("min_exon_coverage").setType("N");
        cmdLine.getOption("min_matches_region").setType("N");
        cmdLine.getOption("threads").setType("N");
        cmdLine.getOption("update_mode").setType("STR");
        
        cmdLine.setDefault("update_mode", "both");
        cmdLine.setOptional("spaln2","prrn","alignment","prrn_output");
        cmdLine.setDefault("threads", "1");
        cmdLine.setSwitches("exon_view","fasta","common_fasta","init_from_msa","no_filter","no_refinement","no_xml","recompute");
        
        return cmdLine;
    }
    public static void main(String[] args) throws Exception
    {
        //System.in.read();
        
        ISAROptionParser cmdLine = getCmd();
        
        printISARInfo(System.err);
        
        LogConfig lc = LogConfig.configureLog(null, true);
        
        if (!OptionParser.parseParams(args,true, false, true,  cmdLine,lc)){
            return;
        }
        
        Logger log = LogConfig.getLogger();
        
        PartialOrderMultipleAlignment poma = ISAR.getPOMA(cmdLine);
        if ( poma == null){
            log.fatal("Computation aborted");
            System.exit(1);
        }
        
        System.exit(0);
    }
}
