package isoformtransfer.isar;

import java.io.File;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.stream.XMLStreamReader;

import org.jdom.Element;

import algo.alignment.Alignment;
import fasta.SimpleFastaReader;
import genome.Gene;
import genome.GenomicUtils;
import genome.Organism;
import genome.Protein;
import isoformtransfer.InferredExon;
import isoformtransfer.algo.TransferAlgo;
import lmu.utils.FileUtils;
import lmu.utils.FileUtils.LineIterator;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.MapBuilder;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.StringUtils;
import lmu.utils.XMLUtils;
import lmu.utils.tuple.Tuple3;
import protein.algo.ProteinAligner;
import protein.aminoacid.AminoAcidType;


public class POMAGene {
    public static final String ELEM_NAME="POMAGENE";

    private Logger log = LogConfig.getLogger();

    public RegionVector geneExons;
    public RegionVector origGeneExons;
    public RegionVector alignedRegions;
    
    public Gene gene; 
    private Integer taxId = -1;
    
    private String id;
    private String name ="";
    
    public String geneSequence;
    
    private Set<String> proteinIds = new HashSet<>();
    private HashMap<String,String> proteinIds2Sequence = new HashMap<>();
    private HashMap<String,RegionVector> proteinIds2RegionVector = new HashMap<>();
    
    private static Pattern dna = Pattern.compile("[ACGT]+");
    
    public enum SplicedSite
    {
        CANONICAL,
        NONE_CANONICAL
    }
    
    //genomic coordinates which should be matched
    private HashSet<Integer> geneRelevantPoints = new HashSet<Integer>();
    
    public HashSet<Integer> getGeneRelevantPoints()
    {
        return geneRelevantPoints;
    }
    
    public Gene getOrigGene()
    {
        return gene;
    }
    public Integer getTaxId()
    {
        return taxId;
    }
    public String getProtein(RegionVector rv)
    {
        for(Entry<String, RegionVector> e : proteinIds2RegionVector.entrySet())
        {
            if ( e.getValue().equals(rv))
                return e.getKey();
        }
        return null;
    }
    public boolean isFrameShift()
    {
        return isFrameShift(null);
    }

    public  RegionVector getProteinCodingIntersecting(String proteinId, Region1D target, boolean debug)
    {
      
        RegionVector ret = new RegionVector();
        if ( target.getLength() < 3) return ret;
        int l = 0;
        for (Region1D r : this.getRegionVector(proteinId).getRegions() )
        {
           // if ( target.getX1() >= r.getX2()) break;
            if ( r.getOverlap(target) == 0)
            {
                l+=r.getLength();
                continue;
            }
            
            Region1D intersected = r.getIntersectedRegion(target);
            if (ret.size() == 0 && intersected.getLength() < 3) 
            {
                l+=r.getLength();
                continue;
            }
            
            int dist = intersected.getX1()-r.getX1();
            l+=dist;
            if ( ret.size() == 0)
            {
             //   System.out.println(l);
                if ( l%3 != 0)
                {
                    int off = 3-(l%3);
                    intersected = new Region1D(intersected.getX1()+off,intersected.getX2(),null);
    
                }
            }
            ret.addRegion(intersected);
        }
        
        if ( debug && ret.size() > 0 && ret.getCoveredLength() > 3)
        {
            String sequence = this.getSequence(proteinId, false);
            String subSequence = GenomicUtils.getSplicedSeq(this.geneSequence, ret, false);
            subSequence = subSequence.substring(0, subSequence.length()-1);
            
            if ( !sequence.contains(subSequence))
            {
                log.fatal(new RuntimeException(String.format("%s is not part of %s (Protein:%s NotMapped:%s Created:%s)",subSequence,sequence,this.getRegionVector(proteinId),target,ret)));
                System.exit(1);
            } 
        }
       
        return ret;
    }

    
    public static Pair<String,SplicedSite> getSplicedSite(POMAGene gene,Region1D intron)
    {
        return getSplicedSite(gene,intron.getX1(),intron.getX2());
    }
    public static Pair<String, SplicedSite> getSplicedSite(POMAGene gene,int start, int end)
    {
        String site = getSplicedSiteStr(gene,start,end);

        SplicedSite ss = (site.equals("GTAG")) ?  SplicedSite.CANONICAL : SplicedSite.NONE_CANONICAL;

        return Pair.create(site, ss);
    }
    public static String getSplicedSiteStr(POMAGene gene,int start, int end)
    {
        
        if ( end-2 <0 ) return "";
        String acceptor = gene.geneSequence.substring(start,start+2);
        String donor = gene.geneSequence.substring(end-2,end);

        String site = acceptor+donor;
        site = site.toUpperCase();

        return site;

    }
    public boolean isFrameShift(RegionVector rv1, RegionVector rv2)
    {
        HashMap<Integer,Integer> frames = new HashMap<>();
        int k = 0;
        for(Region1D exon : rv1.getRegions())
        {
            for(int i= exon.getX1();i<exon.getX2();i++)
            {
                int frame = k++%3;
                
                frames.put(i,frame);
            }
        }
        k = 0;
        for(Region1D exon : rv1.getRegions())
        {
            for(int i= exon.getX1();i<exon.getX2();i++)
            {
                int frame = k++%3;
                if ( frames.containsKey(i) && !frames.get(i).equals(frame))
                    return true;
            }
        }
        
        return false;
    }
    
    public boolean isFrameShift(PrintStream pw)
    {
    
        HashMap<Integer,HashSet<Integer>> posToFrame = new HashMap<Integer,HashSet<Integer>>();
        HashMap<Integer,HashSet<String>> posToId = new HashMap<>();
        
        for(String proteinId : getProteins())
        {
            RegionVector exons = getRegionVector(proteinId);
            int frame = 0;
            for(Region1D<Object> exon : exons.getRegions())
            {
                for(int i = exon.getX1(); i< exon.getX2(); i++)
                {
                    MapBuilder.update(posToFrame,i,frame);
                    frame=(frame+1)%3;
                    
                    MapBuilder.update(posToId, i,proteinId);
                }
            }
        }
        for(Entry<Integer, HashSet<Integer>> e : posToFrame.entrySet())
        {
            if ( e.getValue().size() > 1)
            {
                if( pw != null)
                {
                    pw.printf("Position:%d Frames:%s Isoforms:%s\n", e.getKey() , e.getValue().toString(), posToId.get(e.getKey()) );
                    
                    int min = Integer.MAX_VALUE;
                    int max = Integer.MIN_VALUE;
                    
                    List<String> proteinIds = new ArrayList<>(posToId.get(e.getKey()));
                    List<StringBuffer> DNA = new ArrayList<StringBuffer>();
                    List<StringBuffer> frame = new ArrayList<StringBuffer>();
                    List<Integer> cframe = new ArrayList<Integer>();
                    
                    
                    for(int j = 0 ; j< proteinIds.size(); j++ )
                    {
                        String proteinId = proteinIds.get(j);
                        RegionVector rv = getRegionVector(proteinId);
                        
                        min = Math.min(min,rv.getX1());
                        max = Math.max(max,rv.getX2());
                        
                        DNA.add(new StringBuffer());
                        frame.add(new StringBuffer());
                        cframe.add(0);
                    }
                    
                    for(int i = min ; i< max ; i++){
                        boolean consider = false;
                        for(int j = 0 ; j< proteinIds.size(); j++ )
                        {
                            RegionVector rv = getRegionVector(proteinIds.get(j));
                            
                            if ( rv.isIn(i))
                            {
                                consider =true;
                                break;
                            }
                        }
                        
                        if ( !consider)
                            continue;
                        
                        for(int j = 0 ; j< proteinIds.size(); j++ )
                        {
                            RegionVector rv = getRegionVector(proteinIds.get(j));
                            if ( rv.isIn(i))
                            {
                                int f = cframe.get(j);
                                DNA.get(j).append(geneSequence.charAt(i));
                                frame.get(j).append(f);
                                cframe.set(j,(f+=1)%3);
                            }else
                            {
                                frame.get(j).append(" ");
                                DNA.get(j).append("-");
                            }
                        }
                 
                    }
                    for(int j = 0 ; j< proteinIds.size(); j++ )
                    {
                        pw.printf("%s %s\n",proteinIds.get(j) , frame.get(j));
                        pw.printf("%s %s\n",proteinIds.get(j) , DNA.get(j));
                        
                    }
                }
                return true;
            }
        }
        return false;
    }
    public Set<String> getProteins()
    {
        return proteinIds;
    }
    
    public RegionVector getRegionVector(String proteinId)
    {
        return proteinIds2RegionVector.get(proteinId);
    }
    public Element toXML(){
        Element root = new Element(ELEM_NAME);
        //root.setAttribute("geneExons", geneExons.getXML());
        root.setAttribute("taxId",taxId==null?"":taxId+"");
        root.setAttribute("geneId",getId());
        root.setAttribute("name", name==null?"":name);
        
        root.setAttribute("gene_seq", this.geneSequence);
        
        Element isoforms = new Element("Isoforms");
        
        for(String proteinId : proteinIds)
        {
            Element isoform = new Element("Isoform");
            isoform.setAttribute("id", proteinId);
            isoform.setAttribute("sequence", proteinIds2Sequence.get(proteinId));
            isoform.getChildren().add(proteinIds2RegionVector.get(proteinId).getXML());
            
            isoforms.getChildren().add(isoform);
        }
        root.getChildren().add(isoforms);
        
        return root;
    }
    public void init() {
        this.geneExons = generateGeneExons();
        this.origGeneExons = generateGeneExons();
        this.alignedRegions = new RegionVector();
    }
    public POMAGene(){
        
    }
    
    public static List<POMAGene> readFromCommonFasta(File f) {
        List<POMAGene> genes = new ArrayList<>();
        
        Logger log = LogConfig.getLogger();

        HashMap<Character,List<String>> asToCodons = new HashMap<>();
        for( Entry<String, Character> e : AminoAcidType.code.entrySet() ) 
        {
            if (! asToCodons.containsKey(e.getValue()))
            {
                asToCodons.put(e.getValue(), new ArrayList<>());
            }
            asToCodons.get(e.getValue()).add(e.getKey());
        }
        ArrayList<String> t = new ArrayList<String>();
        t.add("TGA");
        asToCodons.put('X', t);
        
        SimpleFastaReader reader = new SimpleFastaReader(f);
        while(reader.hasNext())
        {
            Pair<String, String> line = reader.next();
            Tuple3<String,String,Integer> header = getHeader(line.getFirst());
            
            POMAGene gene = new POMAGene();
            
            gene.id = header.getFirst();
            gene.name = header.getSecond();
            gene.taxId = header.getThird();
            
            gene.proteinIds.add(header.getFirst());
            
            String sequence = line.getSecond().toUpperCase();
            String protSeq = null;
            
            Matcher matcher = dna.matcher(sequence);
            if ( matcher.matches() )
            {
                log.info("Sequence %s: DNA sequence detected",header.getFirst());
                protSeq = GenomicUtils.translate(sequence);
            }else{
                log.info("Sequence %s: protein sequence detected",header.getFirst());
                protSeq = sequence;
                
                //generate one valid translation for protein
                StringBuffer cDNA = new StringBuffer();
                for(int i = 0 ; i < sequence.length(); i++)
                {
                    List<String> possibleCodons = asToCodons.get(sequence.charAt(i));
                    if ( possibleCodons == null )
                    {
                        log.warn("Unknown AS (%s) in sequence %s (replace with X)",sequence.charAt(i),header.getFirst());
                        possibleCodons = asToCodons.get('X');
                    }
                    
                    //take first codon (could use any)
                    cDNA.append(possibleCodons.get(0)); 
                }
                sequence = cDNA.toString();
            }
            gene.proteinIds2Sequence.put(header.getFirst(), sequence);
            
            RegionVector exons = new RegionVector();
            Region1D exon = new Region1D(0,sequence.length());
            exons.addRegion(exon);
            

            gene.proteinIds2RegionVector.put(header.getFirst(), exons);
            gene.geneSequence = sequence;
            gene.init();;
            genes.add(gene);
        }
        
        return genes;
    }


    private static Tuple3<String,String,Integer> getHeader(String line)
    {
        String id = "";
        String name = "";
        Integer taxId = -1;
        
        boolean comment = line.startsWith(">");
        
        String[] tokens = line.substring(comment?1:0).split("\\s+");
        id = tokens[0];
        
        if ( tokens.length>1)
            name = tokens[1];
        if ( tokens.length >2)
        {
            try{
                taxId = Integer.valueOf(tokens[2]);
            }catch(NumberFormatException e)
            {
                //no number, just skip
            }
        }
        return Tuple3.create(id, name, taxId);
    }
    public POMAGene(File f, TransferAlgo algo)
    {
        
        boolean gene = true;
        LineIterator it = FileUtils.getLineIterator(f);
        String currentProteinId = null;
        String currentSequence = "";
        this.geneSequence="";
        RegionVector currentExons = new RegionVector();
        
        while(it.hasNext())
        {
            String line = it.next();
            if (line.startsWith(">"))
            {
                Tuple3<String,String,Integer> header = getHeader(line);
                
                if ( gene && this.id !=null)
                    gene = false;
                
                if ( gene)
                {
                    
                    this.id = header.getFirst();
                    this.name = header.getSecond();
                    this.taxId = header.getThird();
                    
                }else{
                    if ( currentProteinId != null){
                        if ( this.geneSequence.length() == 0)
                            throw new RuntimeException("No gene sequence provided");
                        
                        boolean ret = addProtein(currentProteinId,currentExons,currentSequence,algo);
                        if ( ret)
                        {
                            log.info("Add protein %s for gene %s",currentProteinId,this.id);
                        }
    
                    }
                    currentProteinId = header.getFirst();
                    currentExons = new RegionVector();
                    currentSequence = "";
                }
            }else if ( line.startsWith(";"))
            {
                String[] reg = line.substring(1).split("\\s+");
                
                Integer start = Integer.valueOf(reg[0]);
                Integer end = Integer.valueOf(reg[1]);
                
                currentExons.addRegion(new Region1D(start, end));
            }else{
                String sequence =  line.trim().toUpperCase();
                if ( gene){
                    this.geneSequence += sequence;
                }else{
                    currentSequence +=line;
                }
            }
        }
        boolean ret = addProtein(currentProteinId,currentExons,currentSequence,algo);
        if ( ret)
        {
            log.info("Add protein %s for gene %s",currentProteinId,this.id);
        }
        init();
    }
    
    private boolean addProtein(String pId, RegionVector exons, String sequence, TransferAlgo algo)
    {    
        if ( (exons == null && sequence == null) || (exons != null && exons.getCoveredLength()==0 && sequence !=null && sequence.trim().length()==0))
        {
            log.warn("Neither sequence nor exons information are provided for protein %s",pId);
            return false;
        }
        boolean spAligned = false;
        if ( exons.getCoveredLength() == 0 && sequence != null)
        {
            if ( algo  == null)
            {
                log.fatal("Cannot map sequence to gene sequence. Provide exon information, or spliced-aligner.");
                sequence = null;
                return false;
            }
            algo.setToGene(null, this.geneSequence);

            String protSeq = sequence;
            Matcher matcher = dna.matcher(sequence);
            if ( matcher.matches() )
            {
                log.info("Translate sequence for protein %s to amino acid",pId);
                protSeq = GenomicUtils.translate(sequence);
            }
            log.info("Map sequence for protein %s to gene sequence",pId);
            exons = InferredExon.toRegionVector(algo.transfer(null, null, protSeq));

            ProteinAligner protAligner = new ProteinAligner();
            String predicted =  GenomicUtils.getSplicedSeq(this.geneSequence, exons, false);
            Alignment ali = protAligner.align(protSeq, predicted);
            float seqId = (float)ali.getNumId(protSeq, predicted)/(2f*(protSeq.length()+predicted.length()));
            
            log.info("Alignment:\n\n%s", ali.getFormattedBy(80,pId, protSeq, "Spliced-aligned", predicted));
            log.info("Predicted exons: %s", StringUtils.join(",", exons.getRegions(), (Region1D r) ->r.getX1() + "-" + r.getX2()));
            spAligned=true;
        }
        String dnaSeq = GenomicUtils.getSplicedSeq(this.geneSequence, exons, true);
        String asSeq = GenomicUtils.translate(dnaSeq);

   
        //String sequence = 
        
        if ( !spAligned && ( (!sequence.equals(dnaSeq) && !sequence.equals(asSeq)) ))
        {
            log.fatal("Given sequence for protein %s does not match to exon annotations.\nGiven:    %s\nExpected: %s\nor:       %s\nCheck input!",pId,sequence,dnaSeq,asSeq);

            return false;
        }
        
        int nonCoding = getNonCodingAS(asSeq);
        
        if ( nonCoding>0)
        {
            log.warn("Protein sequence for protein %s contains %d non-coding codons",pId,nonCoding);
            log.debug("Sequence: %s (%s)",asSeq,dnaSeq);
        }
        this.proteinIds.add(pId);
        this.proteinIds2RegionVector.put(pId, exons);
        this.proteinIds2Sequence.put(pId, dnaSeq);
        return true;
        
    }
    public POMAGene(String id, Integer taxId){
        this.id = id;
        this.taxId = taxId;
    }
    
    public boolean isIn(boolean onlyAnnotated, int pos)
    {
        RegionVector rv = onlyAnnotated?this.origGeneExons:this.geneExons;
        
        for(Region1D r1 :rv.getRegions()){
            if ( pos >= r1.getX1() && pos <= r1.getX2()) return true;
        }
        
        return false;
        
    }
    
    public boolean isAnnotated(int pos){
       /* for(Region1D r1 : this.origGeneExons.getRegions()){
            if ( pos >= r1.getX1() && pos <= r1.getX2()) return true;
        }
        
        return false;
        */
        return isIn(true,pos);
    }
    public boolean isAnnotated(Region1D r){
        for(Region1D r1 : this.origGeneExons.getRegions()){
            if ( r1.contains(r)) return true;
        }
        
        return false;
    }
    
    public void addProtein(String proteinId, String sequence, RegionVector rv) {
        if ( this.getProteins().contains(proteinId))
        {
            throw new RuntimeException("Protein already contained");
        }
        this.proteinIds.add(proteinId);
        this.proteinIds2Sequence.put(proteinId, sequence);
        this.proteinIds2RegionVector.put(proteinId, rv);
        
        
    }
    
    public String getSequence(RegionVector rv, boolean dna){
        return GenomicUtils.getSplicedSeq(geneSequence, rv, dna);
    }
  
    
    public String getSequence(String proteinId, boolean dna){
        if ( !this.proteinIds.contains(proteinId)){
            throw new RuntimeException(String.format("Unknown protein %s for gene %s (%s)", proteinId,gene,proteinIds.toString()));
        }
  
   
        String sequence = proteinIds2Sequence.get(proteinId);
        
        /*
        int mod = sequence.length()%3;
        if ( mod != 0){
            sequence = sequence.substring(0,sequence.length()-(3-mod));
        }
        */
        if ( !dna){
            sequence = GenomicUtils.translate(sequence);
        }
        return sequence;
    }
   
    public POMAGene(Gene gene)
    {
        this(gene,gene.getOrganism().getTaxId(),GenomicUtils.getGeneSeq(gene, gene.getOrganism().getGenome()),true);
    }
    public POMAGene(Gene gene,Integer taxId, String geneSequence, boolean filterIdentical){
        this.gene = gene;
        this.geneSequence = geneSequence;//);
        
        initFromGene(filterIdentical);
        this.taxId = taxId;
        this.geneExons = generateGeneExons();
        this.origGeneExons = generateGeneExons();
        this.alignedRegions = new RegionVector();
    }
    
    public POMAGene(XMLStreamReader xsr) 
    {
        
    }
    
    public POMAGene(Element element)
    {
        String gene_id = XMLUtils.getAttrib(element, "geneId", null, true);
        
        taxId = XMLUtils.getIntAttrib(element, "taxId", null,false);
        this.id = gene_id;
        
        name = element.getAttributeValue("name");
        geneSequence = element.getAttributeValue("gene_seq");
        Element isoforms = element.getChild("Isoforms");

        Organism org = Organism.getOrganism(taxId);

        if ( org != null &&org.getAnnot().containsGeneId(gene_id))
        {
            try{
                this.gene = org.getAnnot().getGene(gene_id);
            }catch(Exception e)
            {
                synchronized(System.err)
                {
                    throw new RuntimeException("Could not load:" + gene_id,e);
                    
                }
            }
        }
      
 
        if( geneSequence == null  && gene != null)
        {
            geneSequence = GenomicUtils.getGeneSeq(gene, gene.getOrganism().getGenome());
        }
        
        if ( geneSequence == null)
        {
            throw new RuntimeException("Insuffient data for init. gene");
        }
        
        if (isoforms == null)
        {
            initFromGene(true);
        }else{
            for(Object childObj : isoforms.getChildren() ) 
            {
                if ( childObj instanceof Element)
                {
                    String id = ((Element) childObj).getAttributeValue("id");
                    String sequence = ((Element) childObj).getAttributeValue("sequence");
                    RegionVector rv = new RegionVector(((Element) childObj).getChild(RegionVector.ELEM_NAME));

                    this.proteinIds.add(id);
                    this.proteinIds2RegionVector.put(id, rv);
                    this.proteinIds2Sequence.put(id, sequence);
                }
            }
        }
        this.geneExons = generateGeneExons();
        this.origGeneExons = generateGeneExons();
        this.alignedRegions = new RegionVector();
    }


    public void clear() {
        this.geneSequence = null;
        this.geneExons = null;
        this.origGeneExons = null;
        this.alignedRegions = null;
        this.proteinIds = new HashSet<String>();
        this.proteinIds2RegionVector = new HashMap<>();
        this.proteinIds2Sequence = new HashMap<>();
    }
    public void update()
    {
        this.geneExons = generateGeneExons();
        this.origGeneExons = generateGeneExons();
        this.alignedRegions = new RegionVector();
        
    }
    private void initFromGene(boolean filterIdentical)
    {
        HashMap<RegionVector,Protein> rvAll = new HashMap<>();
        if ( gene.proteins.values().size() == 0)
            throw new RuntimeException("Gene g:" + gene.id + " has no proteins");
        for(Protein p : gene.proteins.values())
        {
          //  if ( gene.getOrganism().getTaxId() == 9606 && !p.parentTranscript.tags.contains("CCDS") )
          //      continue;
            
            
           // if ( p.parentTranscript.tags.contains("mRNA_end_NF") ||  p.parentTranscript.tags.contains("mRNA_start_NF"))
           //   continue;
            
          //  if ( !p.parentTranscript.source.equals("protein_coding"))
          //      continue;
            
            RegionVector rv = p.getCodingRegionVector(false);
            if (filterIdentical &&  rvAll.containsKey(rv)){
                log.debug("Redundant coding exons for gene %s. Protein: %s (%s); same as: %s (%s)",gene.id,p.id,rv,rvAll.get(rv).id,rvAll.get(rv).getCodingRegionVector(false) );
                continue;
            }
            rvAll.put(rv, p);


            String dna =  GenomicUtils.getSplicedSeq(geneSequence, rv, true);
            String as = GenomicUtils.translate(dna);
            if ( as.endsWith("X"))
            {
                as = as.substring(0,as.length()-1);
                dna = dna.substring(0,dna.length()-3);
                
               // RegionVector corrected = new RegionVector();
               
                List<Region1D> tmp = new ArrayList<>();
                int remove =3;
                for(int i =  rv.getRegions().size()-1;i>=0;i--)
                {
                    Region1D exon = rv.get(i);
                    if ( remove > 0)
                    {
                        int left = remove-exon.getLength();
                                
                        if ( left <0 )
                        {
                            tmp.add(new Region1D(exon.getX1(),exon.getX2()-remove));
                            remove=0;
                        }else{
                            remove = left;
                        }
                        
                      //  ;;Region1D correctedExon 
                    }else{  
                        tmp.add(exon);
                    }
                }
                Collections.reverse(tmp);
                RegionVector corrected =new RegionVector();
                corrected.addRegions(tmp);
                
                String dna_c =  GenomicUtils.getSplicedSeq(geneSequence, corrected, true);
                String as_c = GenomicUtils.translate(dna_c);
                if (!dna_c.equals(dna))
                {
                    System.out.println(as_c);
                    System.out.println(as);
                    throw new RuntimeException("Programming bug.!");
                }
                rv = corrected;
            }
            this.proteinIds.add(p.id);
            this.proteinIds2RegionVector.put(p.id, rv);
            this.proteinIds2Sequence.put(p.id, dna);
        }
        
    }
    
    public RegionVector generateGeneExons(){
        RegionVector ret = new RegionVector();
        List<Region1D> exons = new ArrayList<Region1D>();
        for(String proteinId : getProteins()){
            for(Region1D exon : this.proteinIds2RegionVector.get(proteinId).getRegions()){
                exons.add(new Region1D(exon.getX1(),exon.getX2()));
            }
        }
        Collections.sort(exons);
        ret = new RegionVector();
        if ( exons.size() == 0)return ret;
        ret.addRegion(exons.get(0));
        for(int i = 1 ; i< exons.size();i++){
            if ( ret.getX2() < exons.get(i).getX1()){
                ret.addRegion(exons.get(i));
            }else {
                int newX2= Math.max(ret.getLastRegion().getX2(),exons.get(i).getX2());
                ret.getLastRegion().setX2(newX2);
                ret.setX2(newX2);
            }

        }
        log.debug("Gene exons: %s",geneExons);

        return ret;
    }
    public Organism getOrganism(){
        if ( gene != null)
            return gene.getOrganism();
        
        Organism lookup = Organism.getOrganism(taxId);
        
        if ( lookup != null)
            return lookup;
        
        return null;
    }
    @Override
    public String toString(){
        return this.getId();
    }

  
    public void addToAlign(Region1D e){
        this.geneExons = extend(geneExons,e);
        this.alignedRegions = extend(alignedRegions,e);
        
    }
    
    public static RegionVector extend(RegionVector rv, Region1D e)
    {
        RegionVector ret = new RegionVector();
        int before = rv.getX1();
        if ( rv.size() == 0) 
        {
            ret.addRegion(e);
            return ret;
        }
        
        List<Region1D> regions = new ArrayList<Region1D>();
        regions.addAll(rv.getRegions());
        regions.add(e);       
        Collections.sort(regions);
       
        for(Region1D r : regions)
        {
            if ( ret.size() == 0) 
            {
                ret.addRegion(r);
                continue;
            }
            if ( r.getX2() <= ret.getX2()) 
                continue;
            
            Region1D last = ret.getLastRegion();
            
            if ( last.getOverlap(r)>0 /*|| r.getX2() == last.getX1() || r.getX1() == last.getX2()*/ )
            {
                Region1D newR = new Region1D(last.getX1(),r.getX2(),null);
                
                ret.getRegions().remove(ret.size()-1);
                ret.setX2(last.getX1()-1);
                ret.addRegion(newR);
                continue;
            }
            
            ret.addRegion(r);
        }
        
        if ( ret.getX1() > before)
        {
            throw new RuntimeException(String.format("Failed to add (%s) into (%s) -> result: %s",e,rv,ret) );
        }
        
        if ( ret.getRegions().size() == 0) ret.addRegion(e);
        return ret;
        

    }

    public String getId() 
    {
        if ( id != null)
            return id;
        if ( gene != null)
            return gene.id;
        return "NA";
    }
    public String getName() {
        if ( name != null && name.length()>0)
            return name;
        
        if ( gene != null)
        {
            String name = gene.getName();
            if ( name == null) name = gene.id;
            return name;
       }
        if ( id != null)
            return id;
        return "NA";
    }

    public int getNonCodingAS(String asSeq) {
        int nonCoding = 0;
        for(int i = 0 ; i <asSeq.length()-1; i++)
        {
            if ( asSeq.charAt(i) =='X') 
                nonCoding++;
        }
        return nonCoding;
    }


/*
    TODO: caused partial order exception ... !!
    @Override
    public int hashCode()
    {
        if ( id == null || id.length() == 0) 
            return this.hashCode();
        return this.id.hashCode();
    }
    /*
    @Override
    public boolean equals(Object o)
    {
        if ( o instanceof POMAGene)
        {
            return ((POMAGene) o).id.equals(this.id);
        }
        return false;
    }
*/

}
