package isoformtransfer.isar;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Vector;
import java.util.zip.GZIPOutputStream;

import javax.imageio.ImageIO;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.jdom.Element;

import genome.Organism;
import genome.RegionAlignment;
import isoformtransfer.isar.PartialOrderMultipleAlignment.DebugConfiguration.TimeLog;
import isoformtransfer.isar.PartialOrderMultipleAlignment.Region.Anchor;
import isoformtransfer.isar.oracle.Oracle;
import isoformtransfer.isar.oracle.Oracle.Suggestion;
import lmu.utils.FRuntimeException;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.ObjectGetter;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.SimpleOptionParser;
import lmu.utils.StringUtils;
import lmu.utils.XMLUtils;
import lmu.utils.partialorder.PartialOrderIntGraph;
import lmu.utils.partialorder.PartialOrderRegionGraph;
import lmu.utils.partialorder.PartialOrderSetGraph;
import lmu.utils.partialorder.RegionMatchInfoGetter;
import lmu.utils.tuple.Tuple3;
import lmu.utils.tuple.Tuple4;
import ortholog.TaxonomyNode;
import ortholog.TaxonomyNode.SpeciesDistance;

public class PartialOrderMultipleAlignment
{
    
    public static String POMA_XML_STRING = "PartialOrderMulAlignment";
    
    Logger log = LogConfig.getLogger();
   
    //parameters
    private boolean scoreAnchor = false;
    
    private double MIN_SEQ_ID=0.40f;
    private int MIN_MATCHES=10;
    private double MIN_COVERAGE=0.50f;
    private int MIN_BP_IN_REGION=7;

    private boolean slower = false;

    private DebugConfiguration dbConf = new DebugConfiguration();
    
    private UnmatchedStrategy unmatchedStrategy;
    
    //internal partial order graph
    private PartialOrderSetGraph<POMAGene> porg = null;

    //consistently matched regions
    private List<Region> matchedRegions = new ArrayList<Region>();

    //list of genes included in the ISAR
    private List<POMAGene> genes;
    
    //the alignment
    private List<HashSet<Pair<POMAGene, Integer>>>  ali = null;
    
    private String pomaID;

//    private boolean debug = false;
    
    private UnmappedInformation unmappedInformation =new UnmappedInformation();
    
    static public class UnmappedInformation
    {
        private HashMap<Pair<POMAGene,POMAGene>, Vector<Pair<Region1D, Region1D>>> unMatchedRegionsCache = new   HashMap<Pair<POMAGene,POMAGene>,Vector<Pair<Region1D, Region1D>>>();
        
        
        public  Vector<Pair<Region1D, Region1D>> getRegions(Pair<POMAGene, POMAGene> p)
        {
            return unMatchedRegionsCache.get(p);
        }
        public  Vector<Pair<Region1D, Region1D>> getRegions(POMAGene g1, POMAGene g2)
        {
            return this.getRegions(new Pair<POMAGene,POMAGene>(g1,g2));
        }
        
        public void setUnmappedRegions(POMAGene g1, POMAGene g2, Vector<Pair<Region1D, Region1D>> regions )
        {
            unMatchedRegionsCache.put(new Pair<POMAGene,POMAGene>(g1,g2), regions);
        }

        public void clear() {
            unMatchedRegionsCache = new   HashMap<Pair<POMAGene,POMAGene>,Vector<Pair<Region1D, Region1D>>>();
        }
    }
    public enum UnmatchedStrategy
    {
        ONLY_BASIC,
        ONLY_EXTENDED,
        BOTH;
    }
    public enum DebugStat
    {
        NORMAL,
        NONE,
        PRESENTATION,
        DEBUG;
    }
    enum Operation{
        UPDATE_UNMATCH,
        MATCH,
        INIT_UNMATCH,
        APPLY_ORACLE,
        INSERT,
        FINISH_ORACLE,
        FINISH_ITERATION,
        FINISH,
        START;
    }
    
    class DebugConfiguration
    {
        private File outputDir ;
        
        private DebugStat dbStat;
        
        int matchCount=0;
 
       // PrintWriter timeLogInfo = null;
        
        public DebugConfiguration()
        {
            this.dbStat = DebugStat.NONE;
            this.outputDir  = null;
        }
        
        public DebugConfiguration(DebugStat dbStat,File outputDir,boolean timeLog   )
        {
            if ( outputDir == null && timeLog)
            {
                throw new RuntimeException("Time-log without outputDir");
            }
            
            this.outputDir = outputDir;
            this.dbStat= dbStat;
      
        }
        class TimeLog
        {
            long currentTotal = 0;
            
            Long currentStart = null;
            
            Operation operation = null;
            String msg = null;
            
            public TimeLog(Operation operation, String msg)
            {
                this.operation = operation;
                this.msg = msg;
            }
            
            public void start()
            {
                if ( currentStart != null) throw new RuntimeException("Not yet stopped");
                
                
                currentStart=System.currentTimeMillis();
            }
            public void stop()
            {
                if ( currentStart == null) throw new RuntimeException("Not yet started");
                
                Long diff = System.currentTimeMillis()-currentStart;
                currentTotal+=diff;
                currentStart = null;
            }
            
        }
        HashMap<String,TimeLog> logs = new HashMap<>();
   
        public TimeLog timeLog(Operation operation, String msg,boolean start)
        {
            TimeLog log = null;
            if( !logs.containsKey(operation)) {
                logs.put(msg, (log = new TimeLog(operation,msg)));
            }
            if ( start)
                log.start();
            else
                log.stop();
            
            return log;
        }
        public void writeTimeLog()
        {
            if ( outputDir == null)
                return;
            File f = new File(outputDir + "/" + pomaID + ".time");
            if ( !f.canWrite())
                return;
            PrintWriter pw = FileUtils.getWriter(f);
            
            pw.print("Operation\tMsg\tTime(ms)\n");
            for(Entry<String, TimeLog> e : logs.entrySet())
            {
                pw.printf("%s\t%s\t%d\n",e.getValue().operation,e.getValue().msg,e.getValue().currentTotal);
            }
            pw.close();
        }
        public void process(Operation operation)
        {
            if ( operation == Operation.MATCH) matchCount++;
            
            /*
            if ( operation == Operation.FINISH && timeLogInfo != null)
            {
                writeTimeLog();
            }
            */
            if ( dbStat == DebugStat.NONE)
                return;
            
            if ( dbStat == DebugStat.DEBUG)
            {
                if ( operation == Operation.MATCH) return;
                
                PartialOrderMultipleAlignment.this.writeFile(outputDir,pomaID +"");
                
                return;
            }
            
            if ( dbStat == DebugStat.NORMAL)
            {
                if (  operation != Operation.FINISH_ORACLE && operation != Operation.FINISH_ITERATION)
                    return;
                PartialOrderMultipleAlignment.this.writeFile(outputDir,pomaID +"");
                
                return;
            }
           
            if ( dbStat == DebugStat.PRESENTATION)
            {
                if ( operation == Operation.MATCH)
                    PartialOrderMultipleAlignment.this.writeFile(outputDir,pomaID + "-" +matchCount,true,false );
                else
                    PartialOrderMultipleAlignment.this.writeFile(outputDir,pomaID + "-" +matchCount );
                
                return;
            }
            
       
        }
    }
    public void writeTimeLog()
    {
        this.dbConf.writeTimeLog();
    }
   
    
    public PartialOrderRegionGraph<POMAGene> getPOA() {
        if ( porg instanceof PartialOrderRegionGraph )
            return (PartialOrderRegionGraph<POMAGene>)this.porg;
        
        throw new UnsupportedOperationException();
        
    }
    
    public int getNumberOfIsoforms() {
        int ret = 0;
        for(POMAGene g : this.getGenes())
        {
            ret+=g.getProteins().size();
        }
        return ret;
    }
    
    public List<Pair<POMAGene,String>> getProteins() {
        List<Pair<POMAGene,String>> r = new ArrayList<>();
        for(POMAGene gene : genes)
        {
            for(String p : gene.getProteins())
            {
                r.add(Pair.create(gene,p));
            }
            
        }
                
        return r;
    }
    
    
    public List< HashSet<Pair<POMAGene,Integer>>> getCompleteMapping() {
        
        TimeLog l = this.dbConf.timeLog(Operation.INSERT, "Complete matching", true);
        
        PartialOrderIntGraph<POMAGene> tmp = null;
        
        if ( this.porg instanceof PartialOrderIntGraph ) 
        {
            tmp = (PartialOrderIntGraph<POMAGene>)this.porg;
        }else{
            tmp = new PartialOrderIntGraph<POMAGene>();
            
            for(POMAGene pg : genes)
                tmp.addRegionSource(pg, pg.geneSequence.length());
        }
        
        List<Tuple4<POMAGene,Integer,POMAGene,Integer>> additional = new ArrayList<>();
        for(Region r : this.matchedRegions)
        {
            boolean k = true;
         //   System.out.println(r);
            int srcStart = r.getSrcRegion().getX1();
            int dstStart = r.getDstRegion().getX1();
            
            int srcEnd = r.getSrcRegion().getX2();
            int dstEnd = r.getDstRegion().getX2();
            
            int srcEnd_f = r.ra.src.get(0).getX2();
            int dstEnd_f = r.ra.dst.get(0).getX2();
            for (Pair<Integer,Integer> p: r.ra.ali.alignment)
            {
                
                int base1 = srcStart+(p.getFirst()*3);
                int base2 = dstStart+(p.getSecond()*3);
                
                for (int i=0; i<3; i++)
                {
                  
                    
                    if ( base1 == srcEnd_f || base2 == dstEnd_f){
                        
           //             System.out.println("HERE");
                        for (int j=i; j<3; j++)
                        {
                            if ( base1 >= srcEnd_f)
                                base1 = r.ra.src.get(1).getX1();
                            if ( base2 >= dstEnd_f)
                                base2 = r.ra.dst.get(1).getX1();
                            
                            additional.add(Tuple4.create(r.g1,base1,r.g2, base2));
                            
                            base1++;
                            base2++;
                        }
                        
                        break;
                    }
                    
                    if ( base1 > srcEnd || base2 > dstEnd)
                       additional.add(Tuple4.create(r.g1,base1,r.g2, base2));
                    else{
                        boolean ret =tmp.match(r.g1, base1, r.g2, base2);
                        /*
                        if ( k){
                            System.out.println(base1  + " " + base2);
                            k= false;
                        
                            System.out.println(ret);
                        }
                        */
                    }
                    base1++;
                    base2++;
                }
            
            }
            
        }
        //buggy: ommit!
        for(Tuple4<POMAGene, Integer, POMAGene, Integer> t : additional)
        {
         //  boolean c_ret = tmp.match(t.get0(), t.get1(), t.get2(), t.get3());
        }
            
        List< HashSet<Pair<POMAGene,Integer>> > mapping = new ArrayList<>();
        
        Iterator<HashSet<Pair<POMAGene, Integer>>> it = tmp.getGraphIterator();
        
        while(it.hasNext())
        {
            mapping.add(it.next());
        }
        l.stop();
        return mapping;
    }
    public void reportUnmatched(PrintWriter pw, boolean basic) {
        int idLength = 0;
        for(POMAGene gene : genes)
        {
            idLength = Math.max(idLength,gene.getId().length()+2);
            Vector<String> exons = ObjectGetter.map(gene.geneExons.getRegions(), (Region1D exon)-> exon.getX1()+"-"+exon.getX2());
            pw.print(String.format("#%s:%s\n",gene.getId(),StringUtils.join(",",exons) ));
            
            
            for(String p : gene.getProteins())
            {
                idLength = Math.max(idLength,p.length()+6);
                exons = ObjectGetter.map(gene.getRegionVector(p).getRegions(), (Region1D exon)-> exon.getX1()+"-"+exon.getX2());
                pw.print(String.format("#%s:%s\n",p,StringUtils.join(",",exons) ));
                
            }
        }
        
        this.updateUnMappedRegions(basic);
        
        
        int rId = 0;
        pw.flush();
        for(int i = 0 ; i< genes.size(); i++)
        {
            for(int j = i+1 ; j< genes.size(); j++)
            {
                Vector<Pair<Region1D, Region1D>> regions = this.unmappedInformation.getRegions(genes.get(i), genes.get(j));
                
                
                HashSet<String> s1 = new HashSet<>();
                HashSet<String> s2 = new HashSet<>();
                
                
                for(Pair<Region1D, Region1D> r : regions)
                {
                    
                    Region1D r1 = r.getFirst();
                    Region1D r2 = r.getSecond();
                    
                    pw.printf("\nRegion: %d (Genes: %s-%s Region: %s-%s) \n",rId,genes.get(i).getId(),genes.get(j).getId(),r1.toString(),r2.toString());
                    
                    for(String p1 : genes.get(i).getProteins())
                    {
                        for(String p2 : genes.get(j).getProteins())
                        {
                            RegionVector rv1 = genes.get(i).getProteinCodingIntersecting(p1, r1, false);
                            RegionVector rv2 = genes.get(j).getProteinCodingIntersecting(p2, r2, false);
                            
                            if ( rv1.getCoveredLength() > 2 && rv2.getCoveredLength() > 2)
                            {
                                String as1 = genes.get(i).getSequence(rv1, false);
                                String as2 = genes.get(j).getSequence(rv2, false);
                                
                                if ( s1.contains(as1) && s2.contains(as2))
                                    continue;
                                
                                s1.add(as1);
                                s2.add(as2);
                                pw.printf("\nCoding Region (Proteins: %s-%s) \n",p1,p2);
                                pw.printf("---SEQUENCE---\n");
                                pw.printf("%s %s\n",p1,as1);
                                pw.printf("%s %s\n",p2,as2);
                                pw.printf("---CODING GENOMIC POSITIONS---\n");
                                pw.printf("%s %s\n", p1, StringUtils.join(",", rv1.getRegions(),  (Region1D reg) -> reg.getX1() + "-" + reg.getX2()));
                                pw.printf("%s %s\n", p2, StringUtils.join(",", rv2.getRegions(),  (Region1D reg) -> reg.getX1() + "-" + reg.getX2()));
                                
                            }
                        }
                    }
                    rId++;
                }
            }
            pw.flush();
        }
        pw.flush();
    }
    
    public List<POMAGene> getGenes(int taxId)
    {
        return ObjectGetter.filter(getGenes(), (POMAGene gene) -> gene.getTaxId().equals(taxId));
    }
    
    public List<POMAGene> getGenes(Organism o)
    {
        return getGenes(o.getTaxId());
    }
    
    public List<POMAGene> getGenes()
    {
        return genes;
    }
    public String getPOMAID()
    {
        return pomaID;
    }
    public List<Region> getMatchedRegions()
    {
        return matchedRegions;
    }
    public void setDebugConfiguration(File outputDir,DebugStat dbStat,boolean timeLog)
    {
        this.dbConf = new DebugConfiguration(dbStat,outputDir,timeLog);
    }
    
    
    public static  HashMap<Tuple3<POMAGene,POMAGene,Integer>,Integer> convertAlignment(List<HashSet<Pair<POMAGene, Integer>>> alignment)
    {
        HashMap<Tuple3<POMAGene,POMAGene,Integer>,Integer> ret = new HashMap<>();
        
        for(HashSet<Pair<POMAGene, Integer>> ali : alignment)
        {
            for(Pair<POMAGene, Integer> g1 : ali)
            {
                for(Pair<POMAGene, Integer> g2 : ali)
                {
                    Tuple3<POMAGene, POMAGene, Integer> key = Tuple3.create(g1.getFirst(), g2.getFirst(), g1.getSecond());
                    ret.put(key,g2.getSecond());
                }
            }
        }
        return ret;
    }
    int lastMatchedRegionsSize = 0;
    public List<HashSet<Pair<POMAGene, Integer>>> getAlignment()
    {
        
        boolean forceNew = false;
        if ( lastMatchedRegionsSize != this.matchedRegions.size())
            forceNew = true;
        this.lastMatchedRegionsSize = this.matchedRegions.size();
        
        return getAlignment(forceNew);
    }
    /**
     * Return the positions mappings for a source gene (from) to a target gene (to) 
     * @param poma
     * @param from
     * @param to
     * @param ali
     * @return
     */
    public static HashMap<Integer, Integer> getProjectedMapping( POMAGene from, POMAGene to, List<HashSet<Pair<POMAGene, Integer>>> ali)
    {
        HashMap<Integer, Integer> m  = new HashMap<>();
        for (HashSet<Pair<POMAGene, Integer>> a : ali)
        {
            Pair<POMAGene, Integer> src = ObjectGetter.first(ObjectGetter.filter(a, (Pair<POMAGene, Integer> p) -> p.getFirst() == from));
            Pair<POMAGene, Integer> dst = ObjectGetter.first(ObjectGetter.filter(a, (Pair<POMAGene, Integer> p) -> p.getFirst() == to));

            if (src == null || dst == null)
                continue;

            m.put(src.getSecond(), dst.getSecond());
        }
        return m;
    }
       
    public HashMap<Integer, Integer> getProjectedMapping( POMAGene from, POMAGene to)
    {
        return PartialOrderMultipleAlignment.getProjectedMapping(from,to,this.getAlignment());
    }
    
    public POMAGene getGene(String geneId) {
        for(POMAGene gene : genes)
        {
            if ( gene.getId().equals(geneId))
                return gene;
        }
        log.warn("Cannot find gene %s",geneId);
        return null;
    }
    
    public List<HashSet<Pair<POMAGene, Integer>>> getAlignment(boolean force_new)
    {
        if ( force_new || ali == null)
        {
            /*
            HashMap<POMAGene,Integer> c = new HashMap<>();
            for(POMAGene pg : this.getGenes())
            {
                c.put(pg, 0);
            }
            
            Comparator<PartialNode<HashSet<Pair<POMAGene,Integer>>>> cmp = new Comparator<PartialNode<HashSet<Pair<POMAGene,Integer>>>>(){

                @Override
                public int compare(
                        PartialNode<HashSet<Pair<POMAGene, Integer>>> o1,
                        PartialNode<HashSet<Pair<POMAGene, Integer>>> o2) {
                   
                    int c1 = 0;
                    for(Pair<POMAGene, Integer> d : o1.data)
                    {
                        c1+=c.get(d.getFirst());
                    }
                    int c2 = 0;
                    for(Pair<POMAGene, Integer> d : o2.data)
                    {
                        c2+=c.get(d.getFirst());
                    }
                    if ( c2==c1)
                    {
                        return o1.data.size()-o2.data.size();
                    }
                    
                   // System.out.println(o1 + " " + o2 + ":" + c2 + " " + c1);
                    return c2-c1;
                }

          
                
            };
            
            GraphSorter<HashSet<Pair<POMAGene,Integer>>> g = new GraphSorter<HashSet<Pair<POMAGene,Integer>>>(cmp){

                @Override
                public void update(PartialNode<HashSet<Pair<POMAGene, Integer>>> node) {
                    if ( node.data == null)
                        return;
                   // System.out.println("NEXT:" + node);
                    HashSet<POMAGene> inc = new HashSet<>();
                    
                    for(Pair<POMAGene, Integer> d : node.data)
                    {
                        MapBuilder.update(c, d.getFirst(), 1);
                        inc.add(d.getFirst());
                    }
                    for(POMAGene pg : getGenes())
                    {
                        if ( inc.contains(pg))
                            continue;
                        
                        c.put(pg, 0);
                    }
                }
            };
            */
            
            Iterator<HashSet<Pair<POMAGene, Integer>>> it = this.porg.getGraphIterator();
            List<HashSet<Pair<POMAGene, Integer>>> ret = new ArrayList<HashSet<Pair<POMAGene, Integer>>>();
            while (it.hasNext())
            { 
                HashSet<Pair<POMAGene, Integer>> current = it.next();
                ret.add(current);
            }
            ali = ret;
        }
        return ali;
    }
    /*
    public PartialOrderRegionGraph<POMAGene> getPOA()
    {
        return porg;
    }
     */
    

    public static class Region implements Comparable<Region>
    {
        public static String ELEM_NAME="region";
        
        Logger log = LogConfig.getLogger();
        
        RegionAlignment ra;

        public POMAGene g1;
        public POMAGene g2;
        
        public String p1;
        public String p2;

        boolean fromMultipleAlignment;
        boolean fromSplicedAlignment;

        double seqId;
        double lengthScore;
        
        double score;
        Anchor anchor = Anchor.NONE;
        
        enum Anchor
        {
            NONE, LEFT, RIGHT, BOTH;
        }
        enum ExonBoundary{
            LEFT, RIGHT, BOTH, NONE;
        }
        
        RegionVector srcInput;
        RegionVector dstInput;

        String s1;
        String s2;

        
        Integer matches = null;
        
        //GenomicProteinAlignment gpa;

        
        public ExonBoundary getExonBoundary(boolean src)
        {
            Region1D s = src?getSrcRegion():getDstRegion();
            
            boolean begin = false;
            boolean end = false;
            boolean found = false;
            for(Region1D exon : src?srcInput.getRegions():dstInput.getRegions())
            {
                if ( exon.contains(s.getX1()) && exon.contains(s.getX2()-1))
                {
                    if ( Math.abs(exon.getX1()-s.getX1()) < 3)
                        begin = true;

                    if ( Math.abs(exon.getX2() -s.getX2()) < 3)
                        end = true;

                    found = true;
                    break;
                }
            }

       
            
            if (begin && end ) return ExonBoundary.BOTH;
            else if (begin ) return ExonBoundary.LEFT;
            else if (end ) return ExonBoundary.RIGHT;
            return ExonBoundary.NONE;
        }
        public double getBoundaryScore()
        {
            ExonBoundary srcConstraint = this.getExonBoundary(true);
            ExonBoundary dstConstraint =  this.getExonBoundary(false);
            float boundaryScore = 0.25f;
            
            if ( srcConstraint.equals(ExonBoundary.NONE) && dstConstraint.equals(ExonBoundary.NONE) ) //no boundary match
                boundaryScore=0.25f;
            else if ( srcConstraint.equals(ExonBoundary.BOTH) && dstConstraint.equals(ExonBoundary.BOTH) ) //both perfect
                boundaryScore=1f;
            else if ( (srcConstraint.equals(ExonBoundary.BOTH) || dstConstraint.equals(ExonBoundary.BOTH)) && (!srcConstraint.equals(ExonBoundary.NONE) &&  !dstConstraint.equals(ExonBoundary.NONE))  ) //one both, other left/ or right
                boundaryScore=0.9f;
            
            else if ( (srcConstraint.equals(ExonBoundary.LEFT) && dstConstraint.equals(ExonBoundary.LEFT)) || (srcConstraint.equals(ExonBoundary.RIGHT) &&  dstConstraint.equals(ExonBoundary.RIGHT))  ) //left/right aligned
                boundaryScore=0.75f;
            
            else if ( (srcConstraint.equals(ExonBoundary.BOTH) || dstConstraint.equals(ExonBoundary.BOTH))   ) //one perfect; other none
                boundaryScore=0.75f;
            else{
                boundaryScore=0.5f; 
            }
            return boundaryScore;
        }
        public boolean isFromMultipleAlignment()
        {
            return fromMultipleAlignment;
        }
        public boolean isFromPairwiseAlignment() {
            return (!fromMultipleAlignment && !fromSplicedAlignment);
        }
        public boolean isFromSplicedAlignment()
        {
            return fromSplicedAlignment;
        }
        public void setAnchor(Anchor anchor)
        {
            this.anchor = anchor;
        }
        public Region(Element element, List<POMAGene> genes)
        {
            String g1 = element.getAttributeValue("g1");
            String g2 = element.getAttributeValue("g2");

            for (POMAGene g : genes)
            {
                if ( g.getId().equals(g1)) this.g1 = g;
                if ( g.getId().equals(g2)) this.g2 = g;

            }
            if ( this.g1 == null || this.g2 == null) {
                log.debug("could not find gene ids:" +g1 + " " + g2);
                return;
                //throw new RuntimeException();
            }

            seqId= Float.valueOf(element.getAttributeValue("seqId"));
            lengthScore= Float.valueOf(element.getAttributeValue("lengthScore"));

            score= Float.valueOf(element.getAttributeValue("score"));

            p1 = element.getAttributeValue("p1");
            p2 = element.getAttributeValue("p2");
            
            fromMultipleAlignment = element.getAttributeValue("fromMultipleAlignment").equals("1")?true:false;
            fromSplicedAlignment =element.getAttributeValue("fromSplicedAlignment").equals("1")?true:false;
            
            srcInput = new RegionVector(element.getChild("srcInput"));
            dstInput = new RegionVector(element.getChild("dstInput"));

            Element e = element.getChild("ra");
            if (e == null)
            {
                throw new FRuntimeException("no ra found in elem: ra %s", XMLUtils.getAttribs(e));
            }
            ra = new RegionAlignment(e);
            matches = XMLUtils.getIntAttrib(element, "matches", -1);
            
            s1 = this.g1.getSequence(srcInput, false);
            s2 = this.g2.getSequence(dstInput, false);
            
            if ( element.getAttributeValue("anchor") != null)
            {
                String anchor = element.getAttributeValue("anchor");
                this.anchor = Anchor.valueOf(anchor);
            }
            
            
        }
        public Element toXML()
        {
            Element element = new Element(ELEM_NAME);

            element.setAttribute("g1", g1.getId());
            element.setAttribute("g2", g2.getId());
            
            element.setAttribute("p1", p1);
            element.setAttribute("p2", p2);
            
            element.setAttribute("fromMultipleAlignment",fromMultipleAlignment?1+"":0+"");
            element.setAttribute("fromSplicedAlignment",fromSplicedAlignment?1+"":0+"");

            element.setAttribute("seqId", String.format("%.5f",seqId));
            element.setAttribute("lengthScore", String.format("%.5f",lengthScore));
            element.setAttribute("score", String.format("%.5f",score));
            element.setAttribute("max_coverage",String.format("%.2f",getMaxCoverage()));
            element.setAttribute("matches",this.getMatches() +"");
            
            element.setAttribute("anchor",this.anchor.name());
            
            element.getChildren().add(srcInput.getXML("srcInput"));
            element.getChildren().add(dstInput.getXML("dstInput"));

            element.getChildren().add(ra.getXML("ra"));
            //   element.s

            return element;
        }
        public int getMatches()
        {
            if ( matches == null)
                matches = ra.getNumId(s1, s2);
            return matches;
        }
    
        public Region(POMAGene g1, String s1,String p1,POMAGene g2,String s2,String p2,RegionAlignment ra,RegionVector srcInput, RegionVector dstInput, double seqId)
        {
            this.p1 = p1;
            this.p2  = p2;
            this.ra = ra;
            this.g1 = g1;
            this.g2 = g2;
            this.srcInput  = srcInput;
            this.dstInput  = dstInput;
            this.seqId = seqId;
            this.s1 = s1;
            this.s2 = s2;
        }
        public boolean annoated(boolean src)
        {
            if ( src && g1.isAnnotated(this.getRegion(src)))  return true;
            if ( !src && g2.isAnnotated(this.getRegion(src)))  return true;

            return false;
        }

        private Region1D getRegion(boolean src)
        {

            Region1D r = src?ra.src.get(0):ra.dst.get(0);
            int mod =(r.getLength() %3);

            int diff=mod;
            if ( mod == 0) diff = 2;
            else if ( mod == 2) diff = 1;
            else
                diff=0;
            r = new Region1D(r.getX1(),r.getX2()-diff,null);

            return r;

        }

        public Region1D getSrcRegion()
        {
            return getRegion(true);
        }
        public Region1D getDstRegion()
        {
            return getRegion(false);
        }
        public String getAlignment()
        {
            return String.format("%s-%s: Score: %.5f (SeqId: %.2f, Length-score:%.2f Anchor-score:%.2f Exonboundary-score:%.2f),%s %s",this.p1,this.p2,score,seqId,lengthScore,this.getAnchorScore(),this.getBoundaryScore(),fromMultipleAlignment?"FromMA:":"PairWise:",ra.getFormattedProtAli(g1.getId(), s1, g2.getId(),s2));
        }
        public String toString()
        {
            return this.getAlignment();
        }

        @Override
        public int compareTo(Region o)
        {
            return -1*Double.valueOf(score).compareTo(o.score);
        }
        public int getCompleteLength(boolean src)
        {
            int length =0;
            for (Region1D r : src?ra.src.getRegions():ra.dst.getRegions())
            {
                length+=r.getLength();
            }

            return length;
        }
        public RegionVector getSrc(POMAGene gene, String proteinId)
        {
            if (! gene.getProteins().contains(proteinId)) 
                throw new RuntimeException(String.format("Unknown protein: %s", proteinId));
            
            return gene.getRegionVector(proteinId);// gene.proteins.get(protein).getCodingRegionVector(false);
        }
        
        public float getMaxCoverage()
        {
            int l1 = this.getSrcRegion().getLength();
            int l2 = this.getDstRegion().getLength();

            float coverage1=0;
            if (! p1.equals("Predicted"))
            {
                List<Region1D> intersecting1 =  getSrc(g1,p1).getIntersectingRegions(this.getSrcRegion());
                coverage1 =(float)l1/(float)intersecting1.get(0).getLength();
            }

            float coverage2 =0;

            if (! p2.equals("Predicted"))
            {
                List<Region1D> intersecting2 = getSrc(g2,p2).getIntersectingRegions(this.getDstRegion());
                coverage2 = (float)l2/(float)intersecting2.get(0).getLength();

            }

            
            return Math.max(coverage1,coverage2);
        }

        public double getAnchorScore() {
            float anchorScore = 0.5f;
            if ( this.anchor.equals(Anchor.BOTH))
                anchorScore = 1f;
            else if ( this.anchor.equals(Anchor.RIGHT) || this.anchor.equals(Anchor.LEFT))
                anchorScore = 0.75f;
            
            return anchorScore;
        }


    }

    public boolean isAcceptable(Region r) {

        /*
        if ( this.anchor.equals(Anchor.BOTH) && this.getSrcRegion().getLength() < 9 && this.getDstRegion().getLength() < 9 )
        {
            return true;
        }
          */  
        if (  r.seqId < MIN_SEQ_ID)
        {
            log.debug("SeqId too low: %s",r.getAlignment());
            return false;
        }
        
        if (r.getSrcRegion().getLength() < MIN_BP_IN_REGION || r.getDstRegion().getLength() < MIN_BP_IN_REGION )
        {
            log.debug("Ali too short: %s",r.getAlignment());
            return false;
        }
        
        if (  r.getMatches() < MIN_MATCHES && r.getMaxCoverage() < MIN_COVERAGE)
        {
            log.debug("Alignment too short: %d (%.2f)", r.getMatches(),r.getMaxCoverage());
            return false;
        }
        
        return true;
    }
    /*
    private void printAli(PrintStream pw)
    {
        pw.println("Alignment:");
        
        Iterator<HashSet<Pair<POMAGene, Integer>>> it = this.porg.getGraphIterator();
        while (it.hasNext())
        {
            pw.println(it.next());
        }
        pw.flush();
    }
*/

    public void printHistory(PrintWriter pw,boolean printAligned)
    {
        int k = 1 ; 
        pw.write("#{");
        for (POMAGene ag : genes)
        {
            if ( k > 1) pw.write(", ");
            pw.write(ag.getId() + "=" + ag.geneSequence.length());
            k++;
        }
        pw.write("}\n");
        
   /*
        for (Region h : matchedRegions)
        {
            pw.write(String.format("(%s,%s,((%d-%d),(%d-%d)))\n", h.g1.getId(),h.g2.getId(), h.getSrcRegion().getX1() ,h.getSrcRegion().getX2(), h.getDstRegion().getX1(),h.getDstRegion().getX2() ));
        }
        for(Tuple3<POMAGene, POMAGene, Pair<Integer, Integer>> p : points)
        {
            pw.write(String.format("(%s,%s,((%d-%d),(%d-%d)))\n", p.get0().getId(),p.get1().getId(), p.get2().getFirst() ,p.get2().getFirst()+1, p.get2().getSecond(),p.get2().getSecond()+1));
        }
     */   
        for( Tuple4<POMAGene, Region1D, POMAGene, Region1D> p : tmp)
        {
            pw.write(String.format("(%s,%s,((%d-%d),(%d-%d)))\n", p.get0().getId(),p.get2().getId(), p.get1().getX1() ,p.get1().getX2(), p.get3().getX1(),p.get3().getX2()));
        }
        /*
        if ( printAligned)
        {
            for( HashSet<Pair<POMAGene, Integer>> p : this.getAlignment(true))
            {
                List<Pair<POMAGene,Integer>> tmp = new ArrayList<>(p);
                for(int i = 0 ; i< tmp.size(); i++)
                {
                    for(int j = i+1 ; j< tmp.size(); j++)
                    {
                        Pair<POMAGene,Integer> first = tmp.get(i);
                        Pair<POMAGene,Integer>  second = tmp.get(i);
                                
                        String f = (String.format("(%s,%s,((%d-%d),(%d-%d)))\n", first.getFirst(),second.getFirst(), first.getSecond() ,first.getSecond()+1, second.getSecond(),second.getSecond()+1 ));
                        pw.write(f);
                        //  System.err.print(f);
                    }
                }
       //         System.err.println(p);
            }
         //   return;
        }
        */
        pw.flush();
    }

    public void updateUnMappedRegions(boolean basic)
    {
        unmappedInformation.clear();
        long l1 = System.currentTimeMillis();
        
        log.info("Update unmapped regions basic: %b", basic );
        
        TimeLog initUnmatch = dbConf.timeLog(Operation.INIT_UNMATCH, "Basic:" + basic, true);
        
        RegionMatchInfoGetter rmi = null;
        if ( porg instanceof PartialOrderRegionGraph )
        {
            rmi = ((PartialOrderRegionGraph) porg).getRegionMatchInfo(basic, slower);
        }else{
            rmi = porg.getRegionMatchInfo(basic);
        }
        
        initUnmatch.stop();
        
        unmappedInformation.clear();
        
        TimeLog updateUnmatch = dbConf.timeLog(Operation.UPDATE_UNMATCH, "Basic:" + basic,true);
        
        for (int  i = 0 ; i< genes.size(); i++)
        {
            for (int  j = i+1 ; j< genes.size(); j++)
            {
                log.debug("Update %s and %s",genes.get(i).getId(),genes.get(j).getId());
                
                Vector<Pair<Region1D, Region1D>> notMapped = null;
                try
                {
                    notMapped = rmi.getUnMatchedRegions(genes.get(i),genes.get(j));
                }
                catch (Exception e)
                {
                    this.printHistory(new PrintWriter(System.err),false);
                    throw new RuntimeException(e);
                }
                unmappedInformation.setUnmappedRegions(genes.get(i), genes.get(j), notMapped);

            }    
        }
        updateUnmatch.stop();
        
        long sec = (System.currentTimeMillis()-l1); 
        log.info("Update took: %d ms",sec);
    }
    List<Tuple4<POMAGene,Integer,POMAGene,Integer>> boundaries = new ArrayList<>();
    
    private boolean match(Region r)
    {
        log.debug("Add: %s-%s",r.getSrcRegion(),r.getDstRegion());
        if ( r.g1.equals(r.g2))
        {
            throw new RuntimeException(String.format("Cannot match against itsel\n%s",r));
        }
        
        
        boolean ret = false;
        try
        {
            ret = porg.match(r.g1, r.getSrcRegion(), r.g2,r.getDstRegion() );
            if ( ret)  
            {
                r.g1.addToAlign(r.ra.src.get(0));
                r.g2.addToAlign(r.ra.dst.get(0));

                matchedRegions.add(r);
            }

        }
        catch (Exception e)
        {
            System.err.println(r.g1.getId()  + " " + r.g2.getId() + " "+r.getSrcRegion() + " " + r.getDstRegion() + " " + r.g1.geneSequence.length() + " " + r.g2.geneSequence.length() );
            PrintWriter pw = new PrintWriter(System.out);
            this.printHistory(pw,true);
            pw.flush();
            throw new RuntimeException(e);
        }

        
        return ret;
    }
    List<Tuple4<POMAGene,Region1D,POMAGene,Region1D>> tmp = new ArrayList<>();
    private Pair<Integer,Integer> getMatching(Region re, int pos, boolean src)
    {
        int src_start = re.getSrcRegion().getX1(); //only one exon
        int dst_start = re.getDstRegion().getX1();

        int currentSrc = src_start;
        int currentDst = dst_start;
        
        if ( src && !re.getSrcRegion().contains(pos) ) return null;
        if ( !src && !re.getDstRegion().contains(pos) ) return null;
        
        //System.out.println(pos + "  " + re.getSrcRegion() + " " + re.getDstRegion() + " " + src);
        for (Pair<Integer, Integer> p : re.ra.ali.alignment)
        {

            currentSrc=src_start+(p.first*3);
            currentDst=dst_start+(p.second*3);
            for (int i = 0 ; i< 3 ; i++)
            {
                if ( src &&currentSrc+i==pos||!src && currentDst+i==pos ) {
                    return new Pair<Integer,Integer>(currentSrc+i,currentDst+i);
                }
            }


        }
        log.debug("Position %d in %s not found",pos,(src?re.getSrcRegion():re.getDstRegion()));
        return null;

    }
 
    private Set<Pair<Integer, Integer>> getReleventMatchings(Region region)
    {
        HashSet<Pair<Integer, Integer>>  ret = new HashSet<Pair<Integer, Integer>> ();

        for(Integer point : region.g1.getGeneRelevantPoints() ) 
        {
            Pair<Integer, Integer> map = getMatching(region,point,true);
            if (map != null)
                ret.add(map);
        }
        
        for(Integer point : region.g2.getGeneRelevantPoints() ) 
        {
            Pair<Integer, Integer> map = getMatching(region,point,false);
            if (map != null)
                ret.add(map);
        }

        log.debug("Points: %s (region: %s-%s)",ret,region.getSrcRegion(),region.getDstRegion());


        return ret;
    }

    public static double harmonicMean(List<Float> values,List<Float> factors)
    {
        float sum = 0;
        float sumi = 0;

        for (int i = 0 ; i< values.size() ; i++)
        {
            sum+=factors.get(i);
            sumi+=(factors.get(i)/values.get(i));
        }

        return sum/sumi;
    }
    public static double harmonicMean(List<Float> values)
    {
        List<Float> factors = new ArrayList<Float>();
        ObjectGetter.apply(values, (Float v)-> { factors.add(1f);});

        return harmonicMean(values,factors);
    }
    public static double harmonicMean(Float ... values)
    {
        return harmonicMean(Arrays.asList(values));
    }
    public static double harmonicMean(List<Float> factors,Float ... values)
    {
        return harmonicMean(Arrays.asList(values),factors);
    }


    private void updateScore(List<Region> regions)
    {

        List<Float> factors = new ArrayList<Float>();
        
        //
        factors.add(0.4f);
        factors.add(0.1f);
        
        //
        factors.add(0.10f);
        factors.add(0.40f);
        
        
        for (Region r : regions)
        {
            
            if ( scoreAnchor)
            {
                Pair<Region,Region> anchors = getAnchors(r);
                
                if ( anchors.first == null && anchors.second == null)
                {
                    r.setAnchor(Anchor.NONE);
                }else if ( anchors.first != null && anchors.second != null){
                    r.setAnchor(Anchor.BOTH);
                }
                else if ( anchors.first != null )
                {
                    r.setAnchor(Anchor.LEFT);
                }else if ( anchors.second != null )
                {
                    r.setAnchor(Anchor.RIGHT);
                }
            }
        
            /*
            System.out.println(r);
            System.out.println(r.getExonBoundary(true) + " "  + r.getExonBoundary(false));
            System.out.println(r.srcInput + " " + r.dstInput + " " + boundaryScore);
            */
            
            int length = Math.max(r.getSrcRegion().getLength(),r.getDstRegion().getLength());
//            int max =  r.g1.geneExons.getCoveredLength()+r.g2.geneExons.getCoveredLength();
            int max = 0;
            
            for(Region1D exon : r.g1.geneExons.getRegions())
            {
                max = Math.max(max, exon.getLength());
            }
            for(Region1D exon : r.g2.geneExons.getRegions())
            {
                max = Math.max(max, exon.getLength());
            }
            
            r.lengthScore = (float)length/(float)max;

            if ( !isAcceptable(r))
                r.score = 0;
            else
                r.score = harmonicMean(factors,(float)r.seqId,(float)r.lengthScore,(float)r.getAnchorScore(),(float)r.getBoundaryScore());
            
            
        }
        
        Collections.sort(regions);
        if ( this.dbConf.dbStat.equals(DebugStat.DEBUG))
        {
            System.out.println("------ALL----");
            for(Region r: regions)
            {
          //      System.out.println(r.seqId + " " + r.lengthScore + " " + r.getAnchorScore()  + " " + r.getBoundaryScore() + "=>" + r.score + "(" + r.isAcceptable() + ")");
                System.out.println(r);
            }
            System.out.println("-----END-----");
            
        }

    }
    
    private Pair<Region,Region> getAnchors(Region r)
    {
        Region left = null;
        Region right = null;
        
        for(Region matched : this.matchedRegions)
        {
            if ( matched.p1.equals(r.p1) && matched.p2.equals(r.p2))
            {
                if (matched.ra.src_prot_region.getX2() ==  r.ra.src_prot_region.getX1() && matched.ra.dst_prot_region.getX2() ==  r.ra.dst_prot_region.getX1() )
                {
                    
                        left = matched;
                    
         
                }
            }
            
            if ( matched.p2.equals(r.p1) && matched.p1.equals(r.p2))
            {
                if (matched.ra.dst_prot_region.getX2() ==  r.ra.src_prot_region.getX1() && matched.ra.src_prot_region.getX2() ==  r.ra.dst_prot_region.getX1() )
                {
                        left = matched;
         
                }
            }
            
            //
            if ( matched.p1.equals(r.p1) && matched.p2.equals(r.p2))
            {
                if (matched.ra.src_prot_region.getX1() ==  r.ra.dst_prot_region.getX2() && matched.ra.dst_prot_region.getX1() ==  r.ra.dst_prot_region.getX2() )
                {
                        right = matched;
         
                }
            }
            
            if ( matched.p2.equals(r.p1) && matched.p1.equals(r.p2))
            {
                if (matched.ra.dst_prot_region.getX1() ==  r.ra.src_prot_region.getX2() && matched.ra.src_prot_region.getX1() ==  r.ra.dst_prot_region.getX2() )
                {
                        right = matched;
         
                }
            }
        }

        return Pair.create(left, right);
    }

    public List<Region> generateToAlignList(Suggestion suggestion, boolean fromMultipleAlignment, boolean fromSplicedAlignment)
    {
        ArrayList<Region> ret = new ArrayList<Region>();
        for ( RegionAlignment ra : suggestion.gpa.cdnaAli)
        {
            Region r = new Region(
                    suggestion.g1,
                    suggestion.gpa.getSrcSequence(),
                    suggestion.pId1,
                    suggestion.g2,
                    suggestion.gpa.getDstSequence(),
                    suggestion.pId2,
                    ra,
                    suggestion.gpa.getSrc(),
                    suggestion.gpa.getDst(),
                    ra.localSeqId(suggestion.gpa.getSrcSequence(), 
                    suggestion.gpa.getDstSequence())/100
            );
            
            
            
            //no meaningful lengths
           
            
            r.fromSplicedAlignment = fromSplicedAlignment;
            r.fromMultipleAlignment = fromMultipleAlignment;
            ret.add(r);

        }
        return ret;
    }

    class OracleStat
    {
        int total;

        int valid;
        int notConsidered;
        int invalid;
        
        public OracleStat()
        {
            
        }
        public String toString()
        {
            return String.format("Added %d from %d regions (quality too low: %d, invalid: %d)",this.valid,this.total,this.notConsidered,this.invalid);
        }
        public OracleStat(int total, int valid,int invalid,int notConsidered)
        {
            this.total = total;
            if ( valid+notConsidered+invalid != total) 
            {
                throw new RuntimeException("Oracle stat invalid");
            }
            this.valid = valid;
            this.invalid = invalid;
            this.notConsidered = notConsidered;
        }
    }
    public OracleStat merge(PartialOrderMultipleAlignment inPoma) {
        int total= 0;
        int valid = 0;
        int notConsidered = 0;
        int invalid= 0;
        
        for(Region r : inPoma.getMatchedRegions())
        {
        
            POMAGene g1 = this.getGene(r.g1.getId());
            POMAGene g2 = this.getGene(r.g2.getId());
            
            if ( g1 == null)
            {
                log.warn("Unknown gene id:" + r.g1.getId());
                continue;
            }
            if ( g2 == null)
            {
                log.warn("Unknown gene id:" + r.g2.getId());
                continue;
            }
            
            if (! g1.getProteins().contains(r.p1))
            {
                log.warn("Unknown protein id:" + r.p1 + " for gene " + r.g1.getId());
                continue;
            }
            if (! g2.getProteins().contains(r.p2))
            {
                log.error("Unknown protein id:" + r.p2 + " for gene " + r.g2.getId());
                continue;
            }
            total++;
            
            Region rNew = new Region(g1,r.s1,r.p1,g2,r.s2,r.p2,r.ra,r.srcInput,r.dstInput,r.seqId);
            boolean ret = this.match(rNew);
            if ( ret) 
                valid++;
            else
                invalid++;
        }
        
        return new OracleStat(total,valid,invalid,notConsidered);
    }

    public OracleStat applyOracle(Oracle oracle)
    {
        LinkedList<Region> regions = new LinkedList<Region>();
        
        TimeLog applyOracle = dbConf.timeLog(Operation.APPLY_ORACLE, oracle.getName(),true);
        
        List<Suggestion> sugs = null;
        try{
            sugs = oracle.generateMappings(genes, unmappedInformation);
        }catch(Exception e){
            applyOracle.stop();
           // e.printStackTrace();
            log.fatal("Error applying oracle %s. %s",oracle.getName(),e.getMessage());
            log.debug(e);
            return new OracleStat();
        }
        
        HashSet<Tuple4<POMAGene,POMAGene,Region1D,Region1D>> generated = new HashSet<Tuple4<POMAGene,POMAGene,Region1D,Region1D>>();
        log.info("Create regions");
        for(Suggestion sug : sugs)
        {
            if ( sug == null) continue;
            
            List<Region> rs = generateToAlignList(
                    sug,
                    sug.oracleStrategy==Oracle.OracleStrategy.MULTIPLE_ALIGNMENT,
                    sug.oracleStrategy==Oracle.OracleStrategy.SPLICED_ALIGNMENT);
                
            for(Region r : rs){
             
                Tuple4<POMAGene, POMAGene, Region1D, Region1D> t = Tuple4.create(r.g1,r.g2, r.getSrcRegion(), r.getDstRegion());
                
                if ( generated.contains(t)) {
                    continue;
                }
                regions.add(r);
                generated.add(t);
            }
        }
        applyOracle.stop();
        
        if( regions.size() == 0)
        {
            return new OracleStat();
        }
        updateScore(regions);
        
        int total = regions.size();;
        int valid=0;
        int notConsidered = 0;
        int invalid = 0;
        log.info("Insert regions (%d)",regions.size());
        
        TimeLog insertLog = dbConf.timeLog(Operation.INSERT, "Number of regions:" + regions.size(),true);
 
        while(regions.size() > 0)
        {
            Region top = regions.getFirst();
            if ( !isAcceptable(top)){
                notConsidered++;
                regions.removeFirst();
                continue;
            }
          
            boolean match =  this.match(top);
            this.dbConf.process(Operation.MATCH);
            
            if ( match)
            {
                valid++;
            }else{
                invalid++;
            }
          
            regions.remove(0);
        }
        insertLog.stop();

        
        return new OracleStat(total,valid,invalid,notConsidered);
    }
    
    public void check(){
        
        
        
        for(int i = 0 ; i< genes.size(); i++){
            int starts = 0 ;
            int ends = 0;
            for(String proteinId : genes.get(i).getProteins()  )
            {
                RegionVector rv = genes.get(i).getRegionVector(proteinId);
                for(Region1D exon : rv.getRegions()){
                    for(HashSet<Pair<POMAGene, Integer>> a : this.getAlignment() ) {
                        for(Pair<POMAGene, Integer> pair : a){
                            if ( pair.getFirst() == genes.get(i) && pair.getSecond() == exon.getX1()) starts++;
                            if ( pair.getFirst() == genes.get(i) && pair.getSecond() == exon.getX2()) ends++;
                            
                        }
                    }
                }
            }
            System.out.println(genes.get(i) + " " + starts + " " + ends  );
        }
    }
    
    public void createGraph(Oracle... oracles  )
    {
        TimeLog createGraph = dbConf.timeLog(Operation.START, "TOTAL GRAPH", true);
        
        OracleStat stat =null;
        log.info("Start computation with MIN_SEQ_ID/MIN_MATCHES/MIN_COVERAGE/UPDATE_STRATEGY: %.2f/%d/%.2f/%s",this.MIN_SEQ_ID,this.MIN_MATCHES,this.MIN_COVERAGE,this.unmatchedStrategy);
        this.dbConf.process(Operation.START);
        
        List<Boolean> settings = new ArrayList<>();
        if ( this.unmatchedStrategy.equals(UnmatchedStrategy.ONLY_BASIC))
            settings.add(true);
        else if ( this.unmatchedStrategy.equals(UnmatchedStrategy.ONLY_EXTENDED))
            settings.add(false);
        else{
            settings.add(true);
            settings.add(false);
        }
        
        HashSet<Oracle> toSkipOracles = new HashSet<>();
        for(boolean basic : settings)
        {
            boolean updateUnmatch = true;
            
            for(Oracle oracle : oracles)
            {
                if ( toSkipOracles.contains(oracle))
                    continue;
                
                log.info("Start oracle %s (basic: %b)",oracle.getName(),basic);
                
                int it=1;

                do
                {
                    log.info("Iteration: %d",it);
                    
                    
                    if( updateUnmatch){
                        this.updateUnMappedRegions(basic);
                        this.dbConf.process(Operation.UPDATE_UNMATCH);
                    }
                        
                    stat = applyOracle(oracle);
                    
                    it++;
                    log.info(stat.toString());
                    
                    if (basic == false && (stat.total > 0 && stat.valid+stat.notConsidered == 0) )
                    {
                     //   this.printHistory(new PrintWriter(System.err));
                        log.warn("Generated only invalid mappings for basic=false (check: %s)",pomaID);
                    }

                   
                    updateUnmatch=stat.valid>0;;
                    this.dbConf.process(Operation.FINISH_ITERATION);
                }while(stat.valid>0 && oracle.isRecursiveOracle());
                
                if ( !oracle.isRecursiveOracle())
                    toSkipOracles.add(oracle);
                    
                log.info("No further mappings generated");
                this.dbConf.process(Operation.FINISH_ORACLE);
                
            }
        }
        createGraph.stop();
        this.dbConf.process(Operation.FINISH);
        
    }

    public Element toXML()
    {
        Element root = new Element(POMA_XML_STRING);
        root.setAttribute("id", pomaID+"");
        Element genes = new Element("genes");
        for (POMAGene ag : this.genes)
        {
            genes.addContent(ag.toXML());
        }
        root.getChildren().add(genes);
        Element regions = new Element("regions");

        for (Region r : this.getMatchedRegions())
        {
            regions.getChildren().add(r.toXML());
        }
        root.getChildren().add(regions);


        return root;
    }
    

    public int registerPOMAGene(POMAGene pg)
    {
        int k  = 0;

        for (String proteinId : pg.getProteins())
        {
            k++;
            log.debug(proteinId + "\t" + pg.getRegionVector(proteinId));
        }
        int length = pg.geneSequence.length(); //Math.max(pg.geneExons.getX2()+1, pg.getLength()+1);
        log.debug(pg.getId() + "\t" + pg.geneExons + "\t" +  length);
        porg.addRegionSource(pg, length);
        this.genes.add(pg);
        //addGenePoints(pg);
        
        return k;
    }
    
    public double getMIN_SEQ_ID() {
        return MIN_SEQ_ID;
    }
    public void setMIN_SEQ_ID(double MIN_SEQ_ID) {
        this.MIN_SEQ_ID = MIN_SEQ_ID;
    }
    public int getMIN_MATCHES() {
        return MIN_MATCHES;
    }
    public void setMIN_MATCHES(int MIN_MATCHES) {
        this.MIN_MATCHES = MIN_MATCHES;
    }
    public double getMIN_COVERAGE() {
        return MIN_COVERAGE;
    }
    public void setMIN_COVERAGE(double MIN_COVERAGE) {
        this.MIN_COVERAGE = MIN_COVERAGE;
    }
    public int getMIN_BP_IN_REGION() {
        return MIN_BP_IN_REGION;
    }
    public void setMIN_BP_IN_REGION(int MIN_BP_IN_REGION) {
        this.MIN_BP_IN_REGION = MIN_BP_IN_REGION;
    }
    
    private void initPorg()
    {
        if ( unmatchedStrategy.equals(UnmatchedStrategy.ONLY_BASIC))
        {
            porg = new PartialOrderIntGraph<POMAGene>();
        }else{
            porg = new PartialOrderRegionGraph<POMAGene>();
        }
        
    }
    
    private PartialOrderMultipleAlignment()
    {
        unmatchedStrategy = UnmatchedStrategy.ONLY_BASIC;
        initPorg();
    }
    
    public PartialOrderMultipleAlignment(File xmlFile)
    {
        this(xmlFile,UnmatchedStrategy.ONLY_BASIC,false);
    }
    public PartialOrderMultipleAlignment(File xmlFile, UnmatchedStrategy  unmatchedStrategy,boolean skipFormSplicedAlignment)
    {
        this(xmlFile,unmatchedStrategy,null,skipFormSplicedAlignment);
    }
    
    public PartialOrderMultipleAlignment(File xmlFile, UnmatchedStrategy  unmatchedStrategy,Collection<String> geneIds, boolean skipFormSplicedAlignment ) 
    {
        log.info("Read ISAR: %s", xmlFile);
        this.unmatchedStrategy = unmatchedStrategy;
        initPorg();
        this.setDebugConfiguration(null,DebugStat.NONE,false);
        try{
            BufferedReader reader = FileUtils.getReader(xmlFile);
            
            initFromXML(XMLUtils.getReader(reader),geneIds==null?null:new HashSet<String>(geneIds),skipFormSplicedAlignment,false);
            reader.close();
        }catch(Exception e)
        {
            e.printStackTrace();
            throw new RuntimeException("Could not read XML file");
        }
    }
    
 /*
    public PartialOrderMultipleAlignment(Collection<Gene> g,String pomaID,boolean debug)
    {
        this(g,null,pomaID,debug);
    }
   */ 
    public static PartialOrderMultipleAlignment getEmptyPOMA(File xmlFile)
    {
        PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment();
    
        try{
            BufferedReader reader = FileUtils.getReader(xmlFile);
            poma.initFromXML(XMLUtils.getReader(reader),null,false,true);
            reader.close();
        }catch(Exception e)
        {
            throw new RuntimeException("Could not read XML file");
        }
        return poma;
    }
    
    public PartialOrderMultipleAlignment(Collection<POMAGene> g,UnmatchedStrategy unmatchedStrategy, String pomaID)
    {

        this.unmatchedStrategy = unmatchedStrategy;
        
        this.genes = new ArrayList<POMAGene>(g.size());
        this.pomaID = pomaID;
        
        initPorg();
        
        int k = 0;
        for (POMAGene pg : g)
        {

            k += registerPOMAGene(pg);
           
        }
        log.info("Number of genes: %d; Number of proteins: %s",genes.size(),k);
    }
    /*
    private void addGenePoints(POMAGene gene)
    {
        HashSet<Integer> releventPoints = new HashSet<Integer>();
        
        for(String proteinId : gene.getProteins() )
        {
            for(Region1D r : gene.getRegionVector(proteinId).getRegions())
            {
                releventPoints.add(r.getX1());
                releventPoints.add(r.getX2()-1);
                
            }
        }
        
        for(Integer point : releventPoints)
        {
            porg.addPoint(gene, point);
            gene.getGeneRelevantPoints().add(point);
        }
    }
    */
    
    
    
    public static   List<HashSet<Pair<POMAGene,Integer>>> getCompletMapping(File pomaXML ) 
    {
        File completMatching = new File(pomaXML + ".complete");
        return getCompletMapping(pomaXML,null,completMatching,false);
    }
    public static   List<HashSet<Pair<POMAGene,Integer>>> getCompletMapping(File pomaXML,PartialOrderMultipleAlignment poma  ) 
    {
        File completMatching = new File(pomaXML + ".complete");
        return getCompletMapping(pomaXML,poma,completMatching,false);
    }
    
    public static  List<HashSet<Pair<POMAGene,Integer>>>  getCompletMapping(File pomaXML,  PartialOrderMultipleAlignment poma, File completMatching, boolean forceNew  ) 
    {
        //poma = (poma==null?new PartialOrderMultipleAlignment():poma);
        
        Logger log = LogConfig.getLogger();
        if (!completMatching.exists() || forceNew)
        {
            if ( poma == null)
            {
                poma = new PartialOrderMultipleAlignment(pomaXML);
            }
            log.info("Create complete mapping:" + completMatching + "");
            PrintWriter pw = FileUtils.getWriter(completMatching);
            for(HashSet<Pair<POMAGene, Integer>> ali : poma.getCompleteMapping())
            {
                pw.printf("%s\n",StringUtils.join(" ", ali, (Pair<POMAGene, Integer> a) ->  new String(a.getFirst() + ":" + a.getSecond()) ));
            }

            pw.close();
        }
       
        List<HashSet<Pair<POMAGene,Integer>>> ali = new ArrayList<>();
        
        HashSet<String> relevantGenes = new HashSet<>();
        for(POMAGene pg : poma.getGenes())
        {
            relevantGenes.add(pg.getId());
        }
        
        for(String line : FileUtils.getLines(completMatching) ) 
        {
            HashSet<Pair<POMAGene,Integer>> h = new HashSet<>();
            for(String pair : line.split(" ") ) 
            {
                String[] v = pair.split(":");
                String geneId = v[0];
                
                if ( !relevantGenes.contains(geneId))
                    continue;
                
                Integer pos = Integer.valueOf(v[1]);

                POMAGene pg = poma.getGene(geneId);
                h.add(Pair.create(pg,pos));
            }
            ali.add(h);
        }
        return ali;
    
    }
    private void initFromXML(XMLStreamReader xsr,Set<String> geneIds,boolean skipFromSplicedAlignment, boolean empty) throws XMLStreamException
    {
        genes = new ArrayList<>();
        if ( geneIds!=null &&  geneIds.size() == 0) geneIds = null;
        /*
        HashSet<String> t = new HashSet<>();
        
        t.add("ENSG00000174903");
        t.add("ENSMMUG00000031202");
        
        t.add("ENSMUSG00000020149");
        t.add("SPBC1703.10");
        t.add("YFL038C");
        t.add("FBgn0016700");
        t.add("WBGene00004266");
        t.add("ENSGALG00000008783");
        t.add("ENSTNIG00000004511");
        */
        int total=0;
        int valid =0;
        //System.out.println();
        while (xsr.hasNext())
        {
            int event = xsr.next();
 
            if (event!=XMLStreamConstants.START_ELEMENT)
                continue;
            
            String le = xsr.getLocalName();
            if ( le.equals(POMAGene.ELEM_NAME))
            {
                Element element = XMLUtils.readElemFromStream(xsr, false);
                POMAGene pg = new POMAGene(element);
                 
                if ( geneIds != null && !geneIds.contains(pg.getId()))
                    continue;
                registerPOMAGene(pg);
            }
            if ( le.equals(Region.ELEM_NAME))
            {
                if ( empty) return;
                Element element = XMLUtils.readElemFromStream(xsr, false);

                
                Region r = new Region(element,genes);
                /*
                if ( r.g1.getId().equals("ENSBTAG00000003134") )
                {
                    System.out.println(r.getSrcRegion());
                }
                if ( r.g2.getId().equals("ENSBTAG00000003134") )
                {
                    System.out.println(r.getDstRegion());
                }
                */
                if ( r.g1 == null || r.g2 == null)
                    continue;
                
                if ( geneIds!= null && ( !geneIds.contains(r.g1.getId()) || !geneIds.contains(r.g2.getId())))
                    continue;

                
                if (  skipFromSplicedAlignment && r.fromSplicedAlignment)
                    continue;
                
                boolean match = this.match(r);
                if (! match)
                {
                    log.debug("Could not add %s %s",r.getSrcRegion().toString(),r.getDstRegion().toString());
                }
                else
                {
                    valid++;
                }
               
            }
            if ( le.equals(PartialOrderMultipleAlignment.POMA_XML_STRING))
            {
                this.pomaID = xsr.getAttributeValue(null, "id");
            }
        }
        
      //  System.out.println();
        log.debug("Added %d from %d",valid,total);
        
        
        
       // return gs;
    }
    
    public static List<Integer> getOrderedOrgs(Organism fromOrg, TaxonomyNode tnroot) 
    {
        List<SpeciesDistance> sds = TaxonomyNode.getDistances(fromOrg, tnroot);
        List<Integer> orderedOrgs = new ArrayList<Integer>();
        ObjectGetter.apply(sds, (SpeciesDistance sd) -> {
            for(Organism o : sd.organisms)
            {
                orderedOrgs.add(o.getTaxId());
            }
        });
        
        return orderedOrgs;

    }
    
    /*
    public static void generate(TaxonomyNode tnroot,File outputDir, MultipleAlignmentAlgo algo,SimpleOptionParser transferAlgoCmd,List<File> oggs, boolean overwrite,boolean fast,DebugStat dbStat, boolean debug) throws Exception{
        Logger log = LogConfig.getLogger();
        
        for (File ogFile : oggs)
        {
            
            
            String ogId = ogFile.getName().substring(0, ogFile.getName().lastIndexOf("."));
            if ( !ogFile.exists())
                throw new RuntimeException("Cannot find ogg:" + ogFile);
           
           // File oggDir = ogFile.getParentFile(); 
            
            log.info("Process OGG: %s (%s)", ogId,ogFile);
            
            File outFile = new File(outputDir+ "/" + ogId + ".xml.gz");

            log.info("Output file: %s",outFile);
            if ( overwrite == false && (outFile.exists() || new File(outputDir+ "/" + ogId + ".xml").exists()))
            {
                log.info("POMA (output file) already exists");
                continue;
            }
            OrthologGeneGroup ogg = new OrthologGeneGroup(ogFile);
            log.info("Start gene:%s",ogg.getStartGene());
            List<Integer> orderedOrgs = getOrderedOrgs(ogg.getStartGene().getOrganism(),tnroot);
            List<Gene> genes = new ArrayList<Gene>(ogg.getRelevantGenes());
            Collections.sort(genes,Comparator.comparing((Gene g)->
            {
                for(int i = 0 ; i< orderedOrgs.size(); i++)
                {
                    if ( orderedOrgs.get(i).equals(g.getOrganism().getTaxId()) )
                        return i;
                }
               return 0;
            }));

            
            PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(genes,ogId,debug);
                    
            File mulFile = new File(outputDir + "/" + ogId + ".mul");
            
            MultipleAlignmentOracle msaOra = null;
            if ( algo != null)
                msaOra = new MultipleAlignmentOracle(mulFile,algo,poma.getGenes(),orderedOrgs);
            
            SplicedAlignmentOracle pwSplicedAlignedOra = null;
            if (transferAlgoCmd.isOptionSet("transfer_algo") )
            {
                TransferAlgo tranferAlgo = new IsoformTransferSettings.TransferAlgoCommand(transferAlgoCmd).transferAlgo;
                pwSplicedAlignedOra = new SplicedAlignmentOracle(1,tranferAlgo);
                
            }
            log.info("Multiple alignment oracle:" + msaOra);
            
            PairwiseAlignmentOracle pwASAlignedOra = new PairwiseAlignmentOracle(4);
           
            log.info("Spliced alignment oracle:" + pwASAlignedOra);
            
            PrintWriter pw  = null;
            if ( dbStat == DebugStat.DEBUG && pwSplicedAlignedOra != null)
            {
                pw = FileUtils.getWriter(new File(outputDir + "/" + ogId + ".ali.log"));
                pwSplicedAlignedOra.setLogPrinter(pw);
                
            }
            poma.setDebugConfiguration(outputDir,dbStat ,true);
            
            MultiplePairwiseOracle asOracles = new MultiplePairwiseOracle(msaOra,pwASAlignedOra);
            
            if ( pwSplicedAlignedOra != null)
                poma.createGraph(fast,asOracles,pwSplicedAlignedOra);
            else
                poma.createGraph(fast,asOracles);    
            poma.writeFile(outputDir);
            if ( dbStat == DebugStat.DEBUG)
            {
                pw.close();
            }
            
            PartialOrderMultipleAlignment.getCompletMapping(outFile, poma);
            
          //  new PartialOrderMultipleAlignment(outFile);
            
        //    poma.printHistory(new PrintWriter(System.err));

        }
    }
*/
    
  


    /*
    public static void generateStat(File pomaDir, int threads) throws Exception
    {
        //ConcurrentHashMap<Integer,Integer> isoforms2Sp = new ConcurrentHashMap<Integer,Integer>();
        //ConcurrentHashMap<Integer,Integer> genes2Sp = new ConcurrentHashMap<Integer,Integer>();
      
        class Stat
        {
            Set<String> isoforms = new HashSet<>();

            Set<String> genes = new HashSet<>();

            Set<String> genesInISAR = new HashSet<>();
            Set<String> isoformsForGenesInISAR = new HashSet<>();
        
        }
        
        for(Organism o : Organism.getAllOrganisms())
        {
            o.getAnnot();
            o.getGenome();
        }
        
        HashMap<Organism,Stat> stats = new HashMap<>();
        
        PrintWriter pw = new PrintWriter(System.out);
        AtomicInteger pomas = new AtomicInteger();
        AtomicInteger gcount = new AtomicInteger();
        AtomicInteger icount= new AtomicInteger();
        
        
        List<File> files = Arrays.asList(pomaDir.listFiles());

      //  HashMap<Organism,Set<String>> org2Genes = new HashMap<>();
      //  HashMap<Organism,Set<String>> org2Isoforms = new HashMap<>();
        
        for(Organism o : Organism.getAllOrganisms())
        {
         //   System.out.println("Check" + o.getName());
            Set<String> iso = new HashSet<>();
            Set<String> genes = new HashSet<>();
         
            ThreadUtils.runLambdas( o.getAnnot().getGeneNames(), (String geneId) ->
            {
                Gene gene = o.getAnnot().getGene(geneId);
                if ( gene.proteins.values().size() == 0) 
                    return;

                POMAGene pg = new POMAGene(gene);
                
                synchronized(genes)
                {
                    iso.addAll(pg.getProteins());
                    genes.add(geneId);
                }
            },threads);
            
            Stat stat = new Stat();
            stat.genes = genes;
            stat.isoforms = iso;
            stats.put(o, stat);
            System.out.println(o.getName() + "\t" + genes.size() + " " + iso.size());
        }
        
        ThreadUtils.runLambdas(files, (File xml) -> { 
        
            if ( !xml.getName().endsWith(".xml") && !xml.getName().endsWith(".xml.gz"))
                return;
            if ( pomas.incrementAndGet() %100== 0)
            {
                pw.printf("Processed %d pomas (genes: %d; isoforms: %s)\n",pomas.get(),gcount.get(),icount.get());
                pw.flush();
            }
           
            PartialOrderMultipleAlignment poma = PartialOrderMultipleAlignment.getEmptyPOMA(xml);
            //poma = new PartialOrderMultipleAlignment(xml);
            HashMap<Organism,Integer> o = new HashMap<Organism,Integer>();
            for(POMAGene ag : poma.genes)
            {
            
                gcount.incrementAndGet();
                icount.addAndGet(ag.getProteins().size());
                
                synchronized(stats)
                {
                    stats.get(ag.getOrganism()).genesInISAR.add(ag.getId());
                    stats.get(ag.getOrganism()).isoformsForGenesInISAR.addAll(ag.getProteins());
                }
               
            }
           
     
        },threads);
        
        for(Entry<Organism, Stat> e : stats.entrySet())
        {
            pw.printf("%s %d %d %d %d\n", 
                    e.getKey().getName(),
                    e.getValue().genes.size(),e.getValue().isoforms.size(),
                    e.getValue().genesInISAR.size(),
                    e.getValue().isoformsForGenesInISAR.size()
                    );
            pw.flush();
        }
        pw.flush();

    }
      
    /*
    private boolean writeFile(File outputDir)
    {
        return writeFile(outputDir,pomaID+"",true,true);
    }
    */
    private boolean writeFile(File outputDir, String name)
    {
        return writeFile(outputDir,name,true,true);
    }
    
    public boolean writeImage(File imgFile,boolean small,boolean showIsoforms)
    {
        log.info("Write: %s",imgFile);
        boolean debug = this.dbConf.dbStat.equals(DebugStat.DEBUG);
        try{
            
            ImageIO.write(Drawer.drawAlignment(this,small, showIsoforms), "png",imgFile);
        }catch(Exception e){
           // this.printAli(System.err);
            
            e.printStackTrace(System.err);
            log.warn("Could not write image file %s (%s)",imgFile,e.getMessage());
            return false;
        }
        return true;
    }
    public void writePOMA(File pomaXML)
    {
       
        log.info("Write: %s",pomaXML);
        
        try
        {
            GZIPOutputStream zip = new GZIPOutputStream(new FileOutputStream(pomaXML));
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(zip, "UTF-8"));
            
            PrintWriter pw = new PrintWriter(writer);
            
            pw.println("<" + POMA_XML_STRING + " id=\"" + pomaID+"\">" );
            pw.println("<genes>");
            
            for (POMAGene ag : this.genes)
            {
                pw.println(XMLUtils.prettyXML(ag.toXML()));
            }
            pw.println("</genes>");
            pw.println("<regions>");
            

            for (Region r : this.getMatchedRegions())
            {
                pw.println(XMLUtils.prettyXML(r.toXML()));
            }
            pw.println("</regions>");
            pw.println("</" + POMA_XML_STRING + ">");
            
            //XMLUtils.writeXML(this.toXML(),  pw);
            pw.close();
            writer.close();
            zip.close();
        }
        catch(Exception e){
            throw new RuntimeException(String.format("Could not write POMA-XML (poma ID:%s",pomaID),e);
        }
    }
    
    private boolean writeFile(File outputDir, String name,boolean img,boolean xml){
        boolean ret = true;
        
        if ( xml)
        {
            File pomaXML = new File(outputDir + "/" + name + ".xml.gz");
            writePOMA(pomaXML);
        }
        if( img)
        {
            File imgFile = new File(outputDir + "/" + name + ".png");
            ret = writeImage(imgFile,false,true);
        }
        
        return ret;
    }
   
    
    private static List<File>  getOggs( File oggList,File oggDir) throws Exception{
        List<File> ret= new ArrayList<>();
        BufferedReader br = new BufferedReader(new FileReader(oggList));

        String line;
        while ((line = br.readLine()) != null)
        {
            String id = line.split("\\s+")[0].trim();
            
            File ogFile  =new File("id");
            
            if ( !ogFile.exists())
                ogFile = new File(oggDir + "/" + id + ".gz");
            if (! ogFile.exists())
                 ogFile = new File(oggDir + "/" + id + ".xml.gz");
            if (! ogFile.exists())
                ogFile = new File(oggDir + "/" + id + ".xml");
            if (! ogFile.exists())
                ogFile = new File(oggDir + "/" + id );
            
            if ( !ogFile.exists())
            {
                System.err.printf("No ogg for: %s (skip!)",id );
                continue;
            }
            
            ret.add(ogFile);
        }
        br.close();
        return ret;
    }
    
    public static void writePairwisePOMAImage(File pomaXML, File outDir) throws Exception
    {
        PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(pomaXML);
        Drawer.grid = false;
        
        for(int i = 0 ; i< poma.genes.size(); i++)
        {
            for(int j = i+1 ; j< poma.genes.size(); j++)
            {
                
                BufferedImage img = Drawer.drawAlignment(poma,false, true, poma.getGenes().get(i),poma.getGenes().get(j));
                File outFile = new File(String.format("%s/%s-%s.png",outDir.toString() , poma.genes.get(i).getId() ,poma.genes.get(j).getId() ));
                ImageIO.write(img, "png", outFile);
            }
             
        }
    }
    
    public static SimpleOptionParser getSimpleOptionParser(SimpleOptionParser cmdLine)
    {
        if( cmdLine == null){
            cmdLine = new SimpleOptionParser();
        }
        
        cmdLine.addSimpleOptions("tree","organism","ogg_dir","num_threads","mul_executable","mul_type");
        cmdLine.setFile("organism");
        cmdLine.setInt("num_threads");
        cmdLine.setDir("ogg_dir");
        cmdLine.setFile("mul_executable");
        cmdLine.setDefault("num_threads", "4");
        cmdLine.getOption("mul_type").setHelp("Multiple alignment algorithm PRRN or MAFFT");
        cmdLine.setDefault("mul_type", "PRRN");
        cmdLine.getOption("mul_executable").setHelp("Path to the multiple alignment executable");
        cmdLine.setOptional("mul_executable","mul_type","ogg_dir");

        return cmdLine;
    }
    /*
    public static void main(String[] args) throws Exception  {
        SimpleOptionParser cmdLine = IsoformTransferSettings.TransferAlgoCommand.prepareCmdLine();
        getSimpleOptionParser(cmdLine);
        
        cmdLine.addSimpleOptions("output_dir","compute_all","compute","fast","stat","complete","draw","print","debug","comp_complete");
        cmdLine.setDir("output_dir");
        cmdLine.setFile("compute_all","compute","draw","print");
        cmdLine.setDir("stat");
        cmdLine.setOptional("compute_all","compute","output_dir","stat","draw","print","comp_complete");
        cmdLine.getOption("complete").setHelp("Match every aligned position");
        cmdLine.getOption("fast").setHelp("Use only unmatched=basic setting");
        cmdLine.setOptional("transfer_algo","transfer_algo_executable");
        cmdLine.setSwitches("fast","complete","debug");
        
        LogConfig lc = LogConfig.configureLog(null, true);

        if (!OptionParser.parseParams(args,true, false, true,  cmdLine,lc))
            return;

        Logger log = LogConfig.getLogger();


        Organism.readOrganisms(cmdLine.getFile("organism"));

        List<Organism> orgs = new ArrayList<Organism>(Organism.getAllOrganisms());
        
        File taxf = cmdLine.getFile("tree");
        Taxonomy tax = new Taxonomy(taxf);
        
        TaxonomyNode tnroot = tax.buildSparseTaxonomy(orgs, true);
        
        boolean complete = cmdLine.isSet("complete");
        if ( cmdLine.isSet("complete")) 
        {
            log.error("-complete swtich is not supported anymore");
            System.exit(1);
        }
        
        boolean fast = cmdLine.isSet("fast");
        
        MultipleAlignmentAlgo algo = null;
        if ( cmdLine.isOptionSet("mul_executable"))
        {
            String mulType = cmdLine.getValue("mul_type");
            
            if ( mulType.equals("PRRN"))
                algo = new PRRN(cmdLine.getFile("mul_executable"),false,10);
            
            else if ( mulType.equals("MAFFT"))
                algo = new MAFFT(cmdLine.getFile("mul_executable"));
            
            else{
                throw new RuntimeException("Unknown mul_type:" +mulType);
            }
        }


        if ( cmdLine.isOptionSet("print"))
        {
            new Formater(cmdLine.getFile("print")).printPairwise(new PrintWriter(System.out),false,true);;
        }
        if ( cmdLine.isOptionSet("comp_complete"))
        {
            
            if ( cmdLine.getFile("comp_complete").getName().endsWith(".xml.gz"))
            {
                File pomaXML = cmdLine.getFile("comp_complete");
                
                PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(pomaXML);
                PartialOrderMultipleAlignment.getCompletMapping(pomaXML, poma);
                return;
            }
            for(String line : FileUtils.getLines(cmdLine.getFile("comp_complete")))
            {
                File pomaXML = new File(line);
                
                if ( !pomaXML.exists())
                    throw new RuntimeException(line + " does not exist");
                
                PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(pomaXML);
                PartialOrderMultipleAlignment.getCompletMapping(pomaXML, poma);
            }
            
        }
        if ( cmdLine.isOptionSet("compute_all"))
        {
            
            if (! cmdLine.isOptionSet("ogg_dir"))
            {
                throw new RuntimeException("ogg_dir not set");
            }
            File oggDir = cmdLine.getFile("ogg_dir");

            Drawer.grid = true;
            File out = cmdLine.getFile("output_dir");
            
            if (! out.isDirectory())
            {
                throw new RuntimeException("output_dir is not a valid dir (or not set)");
            }
            File inFile = cmdLine.getFile("compute_all");

            List<File> oggs = getOggs(inFile,oggDir);
            
            
            
            generate(tnroot, out,algo,cmdLine,oggs,false,fast,DebugStat.NONE,false);
            
           
        }

        if ( cmdLine.isOptionSet("compute"))
        {
            Drawer.grid = false;
            File out = cmdLine.getFile("output_dir");
            
            if (! out.isDirectory())
            {
                throw new RuntimeException("output_dir is not a valid dir (or not set)");
            }
            
            List<File> oggs = new ArrayList<File>();
            File ogFile = cmdLine.getFile("compute");
                    
            oggs.add(ogFile);
            generate(tnroot,out,algo,cmdLine,oggs,true,fast,DebugStat.NORMAL,cmdLine.isSet("debug"));
         
            if ( cmdLine.isSet("debug"))
            {
                        
                String ogId = ogFile.getName().substring(0, ogFile.getName().lastIndexOf("."));
                
                File isar = new File(out+ "/" + ogId + ".xml.gz");
                PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(isar);
                
                List<HashSet<Pair<POMAGene, Integer>>> comp = PartialOrderMultipleAlignment.getCompletMapping(isar, poma,true);
                
                Formater f = new Formater(poma,comp);
                f.printPairwise(new PrintWriter(System.out),false,true);;
                f.printISAR(new PrintWriter(System.out), Organism.getOrganism(9606), tnroot);
            }
        }
        
        if ( cmdLine.isOptionSet("stat"))
        {
            generateStat(cmdLine.getFile("stat"),cmdLine.getInt("num_threads"));
        }
        if ( cmdLine.isOptionSet("draw"))
        {
            writePairwisePOMAImage(cmdLine.getFile("draw"),cmdLine.getFile("output_dir"));
        }
        
       log.info("FINISHED.");
    }
    */
}
