package isoformtransfer.isar.oracle;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import algo.alignment.Alignment;
import genome.GenomicProteinAlignment;
import isoformtransfer.isar.POMAGene;
import isoformtransfer.isar.PartialOrderMultipleAlignment.UnmappedInformation;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.Pair;
import lmu.utils.RegionVector;
import ortholog.GeneStructureProteinPath;
import protein.algo.MultipleProteinAlignment;
import protein.algo.MultipleProteinAlignment.MSAFormat;
import protein.algo.external.MultipleAlignmentAlgo;

public class MultipleAlignmentOracle implements Oracle {
    Logger log = LogConfig.getLogger();
    
    private List<Pair<String,String>> geneProteins = null;
    private MultipleAlignmentAlgo algo;
    
    private File[] mulResFiles = null;
    public MultipleAlignmentOracle(File... mulResFiles)
    {
        this.mulResFiles = mulResFiles;
    }
    
    
    public MultipleAlignmentOracle(File mulResFile,MultipleAlignmentAlgo algo,Collection<POMAGene> genes,List<Integer> orgs)
    {
       this.mulResFiles = new File[1];
       this.mulResFiles[0] = mulResFile;
       this.algo = algo;
       if ( mulResFile == null || !mulResFile.exists() || mulResFile.length() == 0)
       {
           if (  orgs != null)
           {
               log.info("Select candidate by best path");

               BestPath gspp = new BestPath(genes,orgs);
               geneProteins = gspp.getBestPath();
               geneProteins = gspp.extend(geneProteins);    
           }else{
               geneProteins  = new ArrayList<>();
               log.info("Select candidate by size");
                
               for(POMAGene g : genes)
               {
                   String longest = null;
                   Integer size = null;
                   for(String proteinId : g.getProteins())
                   {
                       Integer pSize = g.getSequence(proteinId, true).length();
                       if ( longest == null ||  pSize > size )
                       {
                           longest = proteinId;
                           size=pSize;
                       }
                   }
                   geneProteins.add(Pair.create(g.getId(), longest));
               }
           }
       }
    }
    public static class BestPath extends GeneStructureProteinPath{
        public BestPath(Collection<POMAGene> genes,List<Integer> orgs){
            for(POMAGene g : genes)
            {
                /*
                HashSet<String> pIds = new HashSet<>();
                for(String proteinId : g.getProteins()){
                    if ( g.getNonCodingAS(proteinId) < 5)
                        continue;
                    pIds.add(proteinId);
                }
                if  (pIds.size() == 0) continue;
                */
                if ( !org2Gene.containsKey(g.getTaxId()))
                    org2Gene.put(g.getTaxId(),new ArrayList<String>()); 

                org2Gene.get(g.getTaxId()).add(g.getId()); 

                List<String> proteins = new ArrayList<>();
                for(String p : g.getProteins()){

                    protein2Seq.put(p,g.getSequence(p, false));
                    proteins.add(p);
                }
                geneId2ProteinId.put(g.getId(),proteins);
            }
            int k = 0;
            this.orderedOrgs = new ArrayList<>(orgs);
            for(int i = 0; i< orderedOrgs.size(); i++){
                if ( !this.org2Gene.containsKey(orderedOrgs.get(i)) )
                {
                    continue;
                }
                if ( this.startOrg == null){
                    this.startOrg = orderedOrgs.get(i);
                }
                this.org2Index.put(orderedOrgs.get(i), i);
                k++;
            }
    
            generateOptimalPaths();
        }
    }
    
    public POMAGene getPOMAGene(List<POMAGene> genes,String geneId)
    {
        for (int i = 0 ; i< genes.size();i++)
        {
            if ( genes.get(i).getId().equals(geneId) ) return genes.get(i);
        }
        return null;
    }
    
    public List<Suggestion> generateMappings(List<POMAGene> genes, MultipleProteinAlignment mpa) {
        HashMap<Integer,POMAGene> mpaTOGene = new HashMap<Integer,POMAGene>();
        List<Suggestion> gpas = new ArrayList<Suggestion>();
        for (int i = 0 ;i < mpa.idSequences.size(); i++)
        {
            
            Pair<String, String> idSeq = mpa.idSequences.get(i);
           
            POMAGene gene = null;
            
            for (int j = 0; j< genes.size() ; j++)
            {

                if (genes.get(j).getProteins().contains(idSeq.first))
                {
                    gene =genes.get(j);
                    break;
                }
            }

            if ( gene == null)
            {
                throw new RuntimeException(String.format("Protein %s not included in POMAG",idSeq.first));
            }
            if (  !idSeq.second.equals(gene.getSequence(idSeq.first,false)))
            {
                throw new RuntimeException(String.format("Protein %s included with different sequence in POMAG",idSeq.first));

            }
            mpaTOGene.put(i, gene);
        }
        log.info("Generate suggestions");
        int size = mpa.idSequences.size();
        for (int i = 0; i< size; i++)
        {
            for (int j = i+1; j< size; j++)
            {
                POMAGene g1 = mpaTOGene.get(i);
                POMAGene g2 = mpaTOGene.get(j);
                if ( g1 == g2)
                    continue;
                
                
                Alignment ali =mpa.createPairwiseAlignments(i,j);
                
                String id1 = mpa.idSequences.get(i).first;
                String id2 = mpa.idSequences.get(j).first;
                
                RegionVector exon_i= mpaTOGene.get(i).getRegionVector(id1);
                RegionVector exon_j= mpaTOGene.get(j).getRegionVector(id2);
                
                GenomicProteinAlignment gpa = new GenomicProteinAlignment(
                                                                         exon_i,
                                                                         exon_j,
                                                                         g1.getSequence(id1, false),
                                                                         g2.getSequence(id2, false),
                                                                         g1.getSequence(id1, true),
                                                                         g2.getSequence(id2, true),
                                                                         ali,5);
                
                gpas.add(new Suggestion(mpaTOGene.get(i),mpa.idSequences.get(i).first,mpaTOGene.get(j),mpa.idSequences.get(j).first,gpa,OracleStrategy.MULTIPLE_ALIGNMENT));
                
                // List<Region> regions = this.generateToAlignList(mpaTOGene.get(i), mpaTOGene.get(j), gpa, true,false); 
               // this.addToAlign(regions,false);
            }
        }
        log.info("Generated: %d new alignments",gpas.size());
        
        return gpas;
    }
    @Override
    public List<Suggestion> generateMappings(List<POMAGene> genes, UnmappedInformation unmappedInformation) {
        List<Suggestion> gpas = new ArrayList<Suggestion>();
        
        List<MultipleProteinAlignment> mpas = new ArrayList<>();
        if ( geneProteins != null)
        {
            List<Pair<String,String>> id2Sequence = new ArrayList<Pair<String,String>>();
            List<RegionVector> exons = new ArrayList<RegionVector>();
    
            log.info("MSA with: %d sequences",geneProteins.size());
            for (int i = 0 ; i< geneProteins.size(); i++)
            {
                String p = geneProteins.get(i).getSecond();;
                POMAGene gene =getPOMAGene(genes,geneProteins.get(i).getFirst());
                
                if ( gene == null) 
                    throw new RuntimeException("Unknown gene:" +geneProteins.get(i));
                
                id2Sequence.add(new Pair<String,String>(p,gene.getSequence(p, false)));
                exons.add(gene.getRegionVector(p));
            }
            
            MultipleProteinAlignment mpa = new MultipleProteinAlignment(id2Sequence,exons,algo);
            mpa.writeAlignment(mulResFiles[0],MSAFormat.CLUSTAL,70);
            mpas.add(mpa); 
        }else{
//            log.info("Read multiple ali. from file: %s", mulResFile);
            for(File f: mulResFiles)
            {
                mpas.add(new MultipleProteinAlignment(f));
            }
        }
        for(MultipleProteinAlignment mpa: mpas)
        {
            gpas.addAll(generateMappings(genes,mpa));
        }
        
        return gpas;
    }





    @Override
    public String getName() {
        return String.format("MSA (%s)",this.algo!=null?this.algo.getClass().getSimpleName():Arrays.asList(this.mulResFiles));
    }


    @Override
    public boolean isRecursiveOracle() {
        return false;
    }

}
