package isoformtransfer.isar.oracle;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Vector;

import algo.alignment.Alignment;
import genome.GenomicProteinAlignment;
import genome.GenomicUtils;
import isoformtransfer.isar.POMAGene;
import isoformtransfer.isar.PartialOrderMultipleAlignment.UnmappedInformation;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.ThreadUtils;
import lmu.utils.tuple.Tuple4;
import protein.algo.ProteinAligner;

public class PairwiseAlignmentOracle implements Oracle {
    Logger log = LogConfig.getLogger();
    
    HashSet<Tuple4<POMAGene,Region1D,POMAGene,Region1D>> consideredUnMappedRegions = new HashSet<Tuple4<POMAGene,Region1D,POMAGene,Region1D>>();

    HashSet<Tuple4<POMAGene,RegionVector,POMAGene,RegionVector>> consideredAlignments = new  HashSet<Tuple4<POMAGene,RegionVector,POMAGene,RegionVector>>();

    PrintWriter pwDebug = null;
    int threads=1;
    
    public PairwiseAlignmentOracle(int threads)
    {
        this.threads=threads;
    }
    public PairwiseAlignmentOracle()
    {
    }
    




    
    protected List<Suggestion> generateRegions(POMAGene ag1,String p1, Region1D g1NotMapped,POMAGene ag2,String p2,  Region1D g2NotMapped )
    {
        List<Suggestion>  ret = new ArrayList<Suggestion> ();

        log.debug("Generate alignment between %s %s (%s-%s)",p1,p2,g1NotMapped,g2NotMapped);

        RegionVector p1Rec = ag1.getProteinCodingIntersecting(p1,g1NotMapped,true);
        RegionVector p2Rec = ag2.getProteinCodingIntersecting(p2,g2NotMapped,true);

        if ( p1Rec.getCoveredLength() == 0 || p2Rec.getCoveredLength() == 0) return ret;

        Tuple4<POMAGene,RegionVector,POMAGene,RegionVector> t = Tuple4.create(ag1, p1Rec, ag2, p2Rec);
        
        synchronized(t)
        {
            if ( consideredAlignments.contains(t) ) 
            {
                log.debug("Skip %s-%s",p1Rec,p2Rec);
                return ret;
            }
            consideredAlignments.add(t);
        }
        
        String p1Seq = GenomicUtils.getSplicedSeq(ag1.geneSequence, p1Rec, true);
        String p2Seq = GenomicUtils.getSplicedSeq(ag2.geneSequence, p2Rec, true);
        
        String as1 = GenomicUtils.translate(p1Seq);
        String as2 = GenomicUtils.translate(p2Seq);
        
        if ( as1.length() == 0 || as2.length() == 0)
            return ret;
        
        log.debug("Align: %s-%s (%s-%s",as1,as2,p1Rec,p2Rec);
       
        ProteinAligner aligner = new ProteinAligner("local","dayhoff", -12, -1);
        Alignment ali = aligner.align(as1, as2);

        float numId = ali.getNumId(as1, as2);
        float seqId = numId/ (0.5f * (float)(as1.length()+as2.length()));

        log.debug("Alignment of AS1: %s and AS2: %s (seq id: %.2f",as1,as2,seqId);
        GenomicProteinAlignment gpa = new GenomicProteinAlignment(p1Rec,p2Rec,GenomicUtils.translate(p1Seq),GenomicUtils.translate(p2Seq),p1Seq,p2Seq,ali,5);

        Suggestion sug = new Suggestion(ag1,p1,ag2,p2,gpa,OracleStrategy.PROTEIN_AS_ALIGNMENT);
        ret.add(sug);
        
        return ret;
    }
    
    private List<Suggestion> generate(POMAGene g1, POMAGene g2, UnmappedInformation unmappedInformation)
    {
        List<Suggestion> ret = new ArrayList<Suggestion>();

        
        Pair<POMAGene,POMAGene> p =new Pair<POMAGene,POMAGene>(g1, g2);
        Vector<Pair<Region1D, Region1D>> notMapped = unmappedInformation.getRegions(p);

        
        for (Pair<Region1D, Region1D> n : notMapped)
        {
            Tuple4<POMAGene,Region1D,POMAGene,Region1D> t = Tuple4.create(g1,n.getFirst(), g2, n.getSecond());
            
            synchronized(consideredUnMappedRegions){
                if (consideredUnMappedRegions.contains(t)) {
                    log.debug("Already considered: %s",n);
                    continue;
                }
            }
            
            for (String proteinId1 : g1.getProteins())
            {

                for (String proteinId2 : g2.getProteins())
                {
                    List<Suggestion> sugs = generateRegions(g1,proteinId1,n.first,g2,proteinId2,n.second);
    
                    ret.addAll(sugs);
                }
            }
         
            synchronized(consideredUnMappedRegions){
                consideredUnMappedRegions.add(t);
            }
        }
        return ret;
    }
    
    @Override
    public List<Suggestion> generateMappings(List<POMAGene> genes, UnmappedInformation unmappedInformation) {
        List<Suggestion> ret = new ArrayList<Suggestion>();

        log.info("Generate alignments (type:%s; threads: %d)",this.getName().toString(),threads);
        for (int i = 0 ; i< genes.size(); i++)
        {
            POMAGene g1 = genes.get(i);
            List<Integer> index = new ArrayList<Integer>();
            
            //results in better performance 
            if ( threads == 1)
            {
                for (int j = i+1 ; j< genes.size(); j++)
                    ret.addAll(generate(g1,genes.get(j),unmappedInformation));
                
                continue;
            }
            for (int j = i+1 ; j< genes.size(); j++) index.add(j);
            
            ThreadUtils.runLambdas(index, (Integer j) ->
            {
                POMAGene g2 = genes.get(j);
                synchronized(ret)
                {
                    ret.addAll(generate(g1,g2,unmappedInformation));
                }
               

            },threads);
        }
        log.info("Generated: %d new alignments",ret.size());
        
        return ret;
    }

    @Override
    public String getName() {
        return "Dayhoff freeshit pairwise alignment";
    }

    public void setLogPrinter(PrintWriter pwDebug) {
        this.pwDebug = pwDebug;
        
    }
    @Override
    public boolean isRecursiveOracle() {
        return true;
    }
    
}
