package isoformtransfer.isar.oracle;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Vector;

import algo.alignment.Alignment;
import genome.GenomicProteinAlignment;
import genome.GenomicUtils;
import isoformtransfer.InferredExon;
import isoformtransfer.algo.TransferAlgo;
import isoformtransfer.isar.POMAGene;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.tuple.Tuple4;
import protein.algo.ProteinAligner;

public class SplicedAlignmentOracle extends PairwiseAlignmentOracle {

    //TransferAlgo algo;
    
    TransferAlgo[] algoInstances;
    int currentAlgoIndex = 0;
    
    public TransferAlgo getTransferAlgo()
    {
        return algoInstances[(currentAlgoIndex++)%algoInstances.length];
    }
    HashSet<Tuple4<String,RegionVector,String,Region1D>> consideredAlignments = new HashSet<Tuple4<String,RegionVector,String,Region1D>>();

    public SplicedAlignmentOracle(TransferAlgo refAlgo) {
        this(1,refAlgo);
    }
    public SplicedAlignmentOracle(int threads,TransferAlgo refAlgo) {
        if ( threads < 1) throw new RuntimeException("threads < 1");
        super.threads = threads;
        
        algoInstances = new TransferAlgo[threads];
        algoInstances[0] = refAlgo;
        for(int i = 1 ; i< threads; i++)
        {
            try
            {
                Method cloneMethod = refAlgo.getClass().getMethod("clone", refAlgo.getClass());
                if ( cloneMethod == null)
                    throw new Exception(refAlgo.getClass() + " does not implement clone()");
                
                TransferAlgo algo  = (TransferAlgo) cloneMethod.invoke(refAlgo);
                
                algoInstances[i] = algo;
            }catch(Exception e)
            {
                throw new RuntimeException(e);
            }
        }
    }
    
    private boolean intronOverlap(RegionVector exons, Region1D notMapped)
    {
       // RegionVector exons = p.getCodingRegionVector(false);
        if ( notMapped.getX1() < exons.getX1() || notMapped.getX2() > exons.getX2()) return true;
        
        for(Region1D intron : exons.getInverseRegionVector().getRegions())
        {
            
            //if ( notMapped.getOverlap(intron) > 0) return true;
        }
        return false;
    }
    
    private List<Region1D> getIntronPart(RegionVector annotatedExons, Region1D notMapped)
    {
        ArrayList<Region1D> ret = new ArrayList<>();
        
        if ( notMapped.getX1() < annotatedExons.getX1() || notMapped.getX2() > annotatedExons.getX2()) {
            ret.add(notMapped);
            return ret;
        }
        return annotatedExons.getInverseRegionVector().getIntersectingRegions(notMapped, true);
        
    }
    
    
    @Override
    protected List<Suggestion> generateRegions(POMAGene ag1,String p1, Region1D g1NotMapped,POMAGene ag2,String p2,  Region1D g2NotMapped)
    {
        List<Suggestion>  ret = new ArrayList<Suggestion> ();
        
        log.debug("Generate alignment between %s %s (%s-%s)",p1,p2,g1NotMapped,g2NotMapped);

        RegionVector p1Rec = ag1.getProteinCodingIntersecting(p1,g1NotMapped,true);
        RegionVector p2Rec = ag2.getProteinCodingIntersecting(p2,g2NotMapped,true);
        
        if ( p1Rec.getCoveredLength() == 0 && p2Rec.getCoveredLength() == 0) {
            return ret;
        }
        
        Tuple4<String,RegionVector,String,Region1D> t1 = Tuple4.create(ag1.getId(), p1Rec , ag2.getId(), g2NotMapped);
        Tuple4<String,RegionVector,String,Region1D> t2 = Tuple4.create(ag2.getId(), p2Rec , ag1.getId(), g1NotMapped);
        
        
        String p1Seq = GenomicUtils.getSplicedSeq(ag1.geneSequence, p1Rec, true);
        String p2Seq = GenomicUtils.getSplicedSeq(ag2.geneSequence, p2Rec, true);

        String as1 = GenomicUtils.translate(p1Seq);
        String as2 = GenomicUtils.translate(p2Seq);
        
        log.debug("Align: %s-%s (%s-%s",as1,as2,p1Rec,p2Rec);

        boolean c1 =false; 
        boolean c2 =false;
        
        
        synchronized(consideredAlignments)
        {
            c1 = consideredAlignments.contains(t1);
            c2 = consideredAlignments.contains(t2);
            
            
        }
        List<Region1D> g1Relevant = getIntronPart(ag1.origGeneExons,g1NotMapped);
        List<Region1D> g2Relevant = getIntronPart(ag2.origGeneExons,g2NotMapped);
        
        if ( !c1 && as1.length() >= 5 && g2Relevant.size() > 0  ){
           
            for(Region1D nm : g2Relevant)
            {
                if ( nm.getLength() < 15) continue;
                
                ret.add(generateSplicedAlignment(ag1,p1,ag2,p1Rec,p1Seq,as1,nm));
                
            }
            synchronized(consideredAlignments)
            {
                consideredAlignments.add(t1);
            }
        
        }
        if ( !c2 && as2.length() >= 5 && g1Relevant.size() > 15 ){
            
            for(Region1D nm : g1Relevant)
            {
                if ( nm.getLength() < 15) continue;
                
                ret.add(generateSplicedAlignment(ag2,p2,ag1,p2Rec,p2Seq,as2,nm));
              
            }
            synchronized(consideredAlignments)
            {
                consideredAlignments.add(t2);
            }
        }
  
        return ret;
    }
    private Suggestion generateSplicedAlignment(POMAGene source,String sourceP,POMAGene target,RegionVector sourceRV,String cDNA, String AS, Region1D targetRegion)
    {
        
        TransferAlgo algo = getTransferAlgo();//new TransferJob.TransferAlgoCommand(transferAlgoCmd).transferAlgo;
        
        String DNA = target.geneSequence.substring(targetRegion.getX1(), targetRegion.getX2());
        algo.setToGene(null, DNA);
        log.debug("Performe spliced alignment");
        
        if ( this.pwDebug != null)
        {
            pwDebug.printf("%s %s %s %s %s %d %s %d %s\n",
                    this.getName(), 
                    source.getId(),
                    sourceP,
                    target.getId(),
                    sourceRV,
                    sourceRV.getCoveredLength(),
                    targetRegion,
                    targetRegion.getLength(),
                    AS
            );
            pwDebug.flush();
        }
        
        Vector<InferredExon> inferred = algo.transfer(null, cDNA, AS);
        if ( inferred == null || inferred.size() == 0) return null;
        
        RegionVector rv = new RegionVector();
        for(InferredExon exon : inferred)
        {
            rv.addRegion(exon.target.getFirst()+targetRegion.getX1(),exon.target.getSecond()+targetRegion.getX1()-1,exon);
        }
      //  System.out.println(AS + " " + sourceRV + " " + sourceP.getCodingRegionVector(false) + " " + targetRegion);
        String predictedCDNA = null;
        try{
            predictedCDNA = GenomicUtils.getSplicedSeq(target.geneSequence, rv, true);
        }catch(Exception e){
           /*
            System.err.println(">DNA:\n" + DNA);
            System.err.println(">AS:\n" + AS);
            System.err.println(">CDNA:\n" + cDNA);
            */
            System.err.println(targetRegion);
            System.err.println(rv);
            System.err.println(target.geneSequence.length());
            throw new RuntimeException(e);
        }
        String predictedAS = GenomicUtils.translate(predictedCDNA);
        
        
        ProteinAligner aligner = new ProteinAligner();
        Alignment ali = aligner.align(AS, predictedAS);

        float numId = ali.getNumId(AS, predictedAS);
        float seqId = numId/ (0.5f * (float)(predictedAS.length()+AS.length()));
        
        GenomicProteinAlignment gpa = new GenomicProteinAlignment(sourceRV,rv,AS,predictedAS,cDNA,predictedCDNA,ali,5);
        return new Suggestion(source,sourceP,target,"Predicted",gpa,OracleStrategy.SPLICED_ALIGNMENT);
    }
    @Override
    public String getName() {
        return "SPALN2 spliced-aligner oracle";
    }
}
