package isoformtransfer.isar.tools;

import java.io.File;
import java.io.PrintWriter;

import genome.GTFAnnotation;
import genome.Gene;
import genome.GenomeSequenceExtractor;
import genome.GenomicUtils;
import isoformtransfer.isar.ISAR;
import isoformtransfer.isar.POMAGene;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.Logger.LOGLEVEL;
import lmu.utils.OptionParser;
import lmu.utils.Region1D;
import lmu.utils.SimpleOptionParser;

public class InputGenerator {
    
    public static void main(String[] args) throws Exception{
        SimpleOptionParser cmdLine = new SimpleOptionParser();
        
        cmdLine.addSimpleOptions("gtf","gene_ids","fasta_dir","taxId","filter_identical","out_dir");
        cmdLine.setFile("gtf");
        cmdLine.setDir("fasta_dir","out_dir");
        cmdLine.setInt("taxId");
        cmdLine.setOptional("taxId");
        cmdLine.setSwitches("filter_identical");
        cmdLine.getOption("gtf").setHelp("path to Gene Transfer File (gtf)");
        cmdLine.getOption("fasta_dir").setHelp("path to genome FASTA dir");
        cmdLine.getOption("taxId").setHelp("taxonomy id e.g. 9606 for human");
        cmdLine.getOption("gene_ids").setHelp("comma separated list of gene identifiers");
        cmdLine.getOption("filter_identical").setHelp("filters proteins with identical coding exon annotations");
        cmdLine.getOption("out_dir").setHelp("folder for output files (for each specified gene identifier a separate .fasta file is created)");
        cmdLine.setDefault("taxId", "-1");
        
        ISAR.printISARInfo(System.err);
        
        LogConfig lc = LogConfig.configureLog(null, true);

        if (!OptionParser.parseParams(args,true, false, true,  cmdLine))
            return;

        Logger log = LogConfig.getLogger();
        log.setLevel(LOGLEVEL.OFF);
        log.info("Read GTF");
        
        Integer taxId = cmdLine.getInt("taxId");
        
        GTFAnnotation gtf = new GTFAnnotation(cmdLine.getFile("gtf"));
        GenomeSequenceExtractor genome  = new GenomeSequenceExtractor(cmdLine.getFile("fasta_dir")); ;
        File outDir = cmdLine.getFile("out_dir");
        
        for(String geneId : cmdLine.getValue("gene_ids").split(","))
        {
            Gene gene = gtf.getGene(geneId);
            if (gene == null)
            {
                System.err.println("Unknown gene id:" + geneId);
                continue;
            }
            File f= new File(outDir + "/" + gene.id + ".fasta");
            PrintWriter pw = lmu.utils.FileUtils.getWriter(f);
            writeISAROutput(gene,genome,taxId,cmdLine.isSet("filter_identical"),pw);
            pw.close();
            System.out.printf("%s written.\n",f.toString());       
        }
     
    }
    public static void writeISAROutput(Gene gene,GenomeSequenceExtractor genome , Integer taxId,boolean filterIdentifal, PrintWriter pw)
    {
        String dna = GenomicUtils.getGeneSeq(gene, genome);
        pw.printf(">%s %s %d %s:%d-%d:%d\n",gene.id,gene.name!=null?gene.name.replaceAll(" ", "."):gene.id,taxId,gene.chr,gene.start,gene.end,gene.strand?1:0);
        pw.println(getWrapped(dna,80));
        
        POMAGene pg = new POMAGene(gene,taxId,dna,filterIdentifal);
        for(String proteinId : pg.getProteins())
        {
            pw.println(">" + proteinId);

            for(Region1D exon :pg.getRegionVector(proteinId).getRegions() )
            {
                pw.println(";" + exon.getX1() + " " + exon.getX2());
            }

            pw.println(getWrapped(pg.getSequence(proteinId, true),80));    
        }
      
        pw.flush();
    }
    private static String getWrapped(String in, int line_size)
    {
        StringBuffer ret =new StringBuffer();
        
        for(int i = 0 ; i< in.length();i+=line_size)
        {
            ret.append(in.substring(i,Math.min(i+line_size,in.length())) + "\n");
        }
        return ret.toString().substring(0,ret.length()-1);
    }
}
