package lmu.utils;

import java.io.*;
import java.sql.*;
import java.util.*;

import lmu.utils.filter.FlagFilter;

import org.jdom.*;

import com.csvreader.CsvWriter;


/** best way to build a data table from a vector of some objects:
 * 
 *  let's say we have some class Sample
 * 
 *  class Sample
 *  {
 *      public int m1();
 *      public String m2();
 *
 *  }
 * 
 *  and a Vector<Sample> v
 *  and we want to build a DataTable where we can retrieve the original Sample class at position i
 *  but have the sortable,filterable fields m1, m2, m2length, info where info is m2() +":"+Math.abs(m1)
 * 
 *  
 *  DataTable<Sample> dt = DataTable.buildTable(v,
 *                              DataTable.buildHeader("m1", (Sample s) -> s.m1()).
 *                              add("m2", (Sample s) -> s.m2()).
 *                              add("m2length", (Sample s) -> s.m2().length()).
 *                              add("info", (Sample s) -> s.m2() + ": " + Math.abs(s.m1()))
 *                              );
 * 
 *   
 *  the original data (Sample) can be retrieved by the methods getInData() and getInDataByIdx() //filtered or not
 * 
 * (alternatively one can directly build a table from a vector of Tuple assigning names to it's fields)
    
 */
public class DataTable<A> 
{
    Logger log = LogConfig.getLogger();
    
    public interface TableGetter
    {
        public DataTable getTable(Collection data);
    }

    public static class Field
    {
        int idx;
        String name;
        Object value;
        Field(int i, String name)
        {
            idx=i;  this.name=name;
        }

        public Object get()
        {
            return value;
        }

    }
    public static class TupleGetter
    {       
        Tuple result_tuple=null;
        Tuple getter_tuple=null;
        Field[] fields = null;
        DataTable dt;

        TupleGetter(DataTable dt, String ... fields)
        {
            this.dt=dt;
            getter_tuple = new Tuple(fields.length);
            this.fields  = new Field[fields.length];
            for (int i=0; i< fields.length; i++)
            {
                DataHeader dh = dt.header.findHeader(fields[i]);
                if (dh==null)
                {
                    throw new FRuntimeException("cannot find field: %s", fields[i]);
                }
                getter_tuple.set(i, dh.getRefObj());
                this.fields[i]=new Field(i, fields[i]);
            }
            
            result_tuple = new Tuple(fields.length);
        
        }

        public Field getField(String name)
        {
            for (int i=0; i<fields.length; i++)
            {
                if (!fields[i].name.equals(name))
                    continue;
                
                return fields[i];
            }
            throw new RuntimeException("cannot find field: "+name);
        }

        //returns Tuple and saves data in fields
        public Tuple get(ObjectGetter og)
        {  
            for (int i=0; i<getter_tuple.cardinality(); i++)
            {
                Object o = ((ReflectedObject)getter_tuple.get(i)).get(og);
                fields[i].value=o;
                result_tuple.set(i, o);
            }
            return result_tuple;
        }
        public Tuple get(int pos)
        {
            return get(dt.getDataAt(pos));

            
        }
    }


    public interface RecordInfo<A>
    {
        public A get(String header_name);
        public A getByIdx(int idx);
    }


    public static class RecordGetter<A> implements RecordInfo<A>
    {
        TupleGetter tg;

        public RecordGetter(TupleGetter tg)
        {
            this.tg = tg;
        }

        public A get(String header_name)
        {
            return (A)tg.getField(header_name).value;
        }

        public A getByIdx(int idx)
        {
            return (A)tg.get(idx).value;
        }
    }
    public interface IndexMapper
    {
    	public Integer getIndex(RecordGetter record);
    }
    
    public interface ClassConverter<ConvertedClass>
    {
        public ConvertedClass convert(Tuple t);
    }
    
    public<ConvertedClass> Vector<ConvertedClass> convert(ClassConverter<ConvertedClass> c,  String ... header_fields)
    {
        String [] fields = (header_fields == null || header_fields.length == 0) ? StringUtils.convert(header.name2header.keySet()) : header_fields;

        TupleGetter tg = getTupleGetter(fields);
        return ObjectGetter.map(IteratorUtils.getIntIterator(0, getSize()),
                                (Integer idx) -> c.convert(tg.get(idx)));

    }
    public DataTable extendByEntityMapping(IndexMapper mapper, String index_field, String ... header_fields)
    {
        return extendByEntityMapping(mapper, index_field, true, header_fields);
    }
    
    public DataTable extendByEntityMapping(IndexMapper mapper, String index_field, boolean filterUnmapped, String ... header_fields)
    {
    	RecordGetter rg;
    	if(header_fields==null || header_fields.length==0)
    	{
    		rg = new RecordGetter<>(getTupleGetter(StringUtils.convert(header.name2header.keySet())));
    	}
    	else
    	{
    		rg = new RecordGetter(getTupleGetter(header_fields));
    	}
    	Vector<Integer> index = new Vector<>();
    	for(int i=0; i<data.size(); i++)
    	{
    		//saves data at row i internally
    		rg.tg.get(i);
    		index.add(mapper.getIndex(rg));
    	}
    	addData(index_field, new ReflectedObject(true), index);
    	if(!filterUnmapped)
    	    return this;
    	Vector<Integer> included_rows = ObjectGetter.filter(IteratorUtils.getIntIterator(0, getSize()), (Integer idx) -> index.get(idx)!=null); 
    	return filterByIdx(ObjectGetter.toSet(included_rows));
    	/*
    	DataTable filtered = DataTable.buildTable(ObjectGetter.map(included_rows, (Integer idx) -> data.get(idx)), DataTable.buildHeader("idx",new ReflectedObject(true)));
    	for(DataHeader h : header.headers)
    	{
    	    filtered.header.addHeader(h.getName(), h.getRefObj());
    	}
    	DataTable filtered =DataTable.buildTable( included_rows, DataTable.buildHeader(index_field, (Integer idx)->index.get(idx)));
    	
    	for(String h : header.getHeaderNames())
    	{
    	    filtered.addData(h, new ReflectedObject(true), ObjectGetter.map(included_rows, (Integer idx)->getFieldVector(h).get(idx)));
    	}
    	System.out.println("constructing new datatable took: "+(System.currentTimeMillis()-start)+" ms");
    	return filtered;
    	 */
    }
    
    public static class SimpleTableGetter implements TableGetter
    {
        Object defs[] = null;
        Vector<Pair> header_defs = new Vector<Pair>();
        
        public SimpleTableGetter()
        {

        }

        public void addHeader(String label, Object def)
        {
            header_defs.add(new Pair(label, def));
        }

        public SimpleTableGetter(Object ... definitions)
        {
            defs = definitions;
        }
        public DataTable getTable(Collection entities)
        {
            if (defs==null)
            {
                if (header_defs.size()==0)
                {
                    throw new RuntimeException("no headers specified yet!");
                }
                defs=new Object[header_defs.size()*2];
                int idx=0;
                for (Pair p : header_defs)
                {
                    defs[idx]=p.getFirst();
                    defs[idx+1]=p.getSecond();
                    //System.out.printf("get table add next header: %d:  %s on\n", idx, defs[idx]);
                    idx+=2;
                    
                }
                



            }
            return DataTable.createTable(entities, defs);
        }
    }


    public Vector<ReflectedObject> getGetters(String ... names)
    {
        HashMap<String, ReflectedObject>m=header.getReflectedObjectLookup();
        Vector<ReflectedObject> v = new Vector<ReflectedObject>();
        for (String n : names)
        {
            ReflectedObject ro = m.get(n);
            if (ro==null)
            {
                throw new RuntimeException("unknown header: "+n+" valid headers are: "+m.keySet());
            }
            v.add(ro);
        }
        return v;
    }
    ObjectGetter.ObjectGetterHeader<A> o_header = null;
    public DataHeaderInfos  header;
    //public Vector<ObjectGetter> data;
    public Vector<ObjectGetter<A>> data;

    public Vector<ObjectGetter<A>> filtered_data = null; //new Vector<>();
    

    static final String HINFOS="_headerinfos";
    static final String DINFOS="_data";

    static final String DATA_IDX_FIELD="idx";
    static final String HEADER_IDX_FIELD="field_idx";
    static final String NAME_FIELD="fieldname";
    static final String HEADER_LABEL_FIELD="label";
    static final String HEADER_CONVERTER_NAME="converter";

    /** writes two tables:
     * one for the header infos and one for the real
     * data  ${table_prefix}_headerinfos ${table_prefix}_data
     * 
     * if table_prefix is null it creates a random prefix itself
     * 
     */

    public static void dropTables(Connection con, String table_prefix)
    {
        if (table_prefix==null)
            return;

        String lastq = "";
        try
        {
        
            Statement st = con.createStatement();
            st.execute(lastq = "drop  table if exists "+table_prefix+HINFOS+";");
            st.execute(lastq = "drop  table if exists "+table_prefix+DINFOS+";");
        
        }
        catch(SQLException e)
        {
            
            throw new RuntimeException("sql error:"+e.getMessage()+" at dropping data tables last command:"+lastq, e);
        }

    }
    public String saveToDb(Connection con, String table_prefix)
    {
        return saveToDb(con,table_prefix, null);
    }

    public String saveToDb(Connection con, String table_prefix, File mysql_nfs_tmpdir)
    {
        
        Vector<SimpleTypeConverter.SQLConverter> converter = new Vector<SimpleTypeConverter.SQLConverter>();
        
        Vector<ReflectedObject> getter = new Vector<ReflectedObject>();

        String lastq = null;
        try
        {
        
            Locale.setDefault(Locale.UK);
            Statement st = con.createStatement();
    
            dropTables(con,table_prefix);
                        
            if (table_prefix==null)
            {
                table_prefix=SQLUtils.generatePrefix(st, 10, HINFOS);
            }
    
            st.execute("create table "+table_prefix+HINFOS+"("+HEADER_IDX_FIELD+" int, "+NAME_FIELD+" text, "+HEADER_LABEL_FIELD+" text, "+HEADER_CONVERTER_NAME+" text);");
            PreparedStatement ps = con.prepareStatement("insert into "+table_prefix+HINFOS+ " values (?,?,?,?);");
    
            Vector<String> header_names = new Vector<String>();
            StringBuffer create_buff = new StringBuffer();
            
            
            create_buff.append("create table "+table_prefix+DINFOS+" ("+DATA_IDX_FIELD+" int primary key");
            StringBuffer insert_buff=new StringBuffer();
            insert_buff.append("insert into "+table_prefix+DINFOS+" values (?");
            int field_idx=0;
            for (DataHeader dh: header.headers)
            {
                log.info("header: %s",dh.name);
                if (!dh.visible)
                {
                    continue;
                }
    
                field_idx++;                            
                String sql_fieldname="field"+field_idx;
                SimpleTypeConverter.SIMPLETYPE type = null;
                for (ObjectGetter og : data)
                {
                    if (og==null)
                        continue;
                    
                    Object o = dh.get(og);
                    if (o==null)
                        continue;
                    
                    type=SimpleTypeConverter.get(o);
    
                    if (type==null)
                    {
                        throw new RuntimeException("cannot save table as no simple type converter is known to "+o.getClass());
                    }
                    break;
                }
                ps.setInt(1, field_idx);
                ps.setString(2, sql_fieldname);
                ps.setString(3, dh.getName());
                ps.setString(4, type.getName());
                ps.execute();
    
                getter.add(dh.getRefObj());
                converter.add(type.getSQLConverter());
                create_buff.append(", "+sql_fieldname+" "+type.getSQLConverter().getSQLType());
                insert_buff.append(",?");
                //System.out.printf("header: %s type: %s\n", dh.getName(), type);
            }
    
            insert_buff.append(");");
            create_buff.append(");");
    
            //System.out.println("create buff: "+create_buff.toString());
            //System.out.println("insert buff: "+insert_buff.toString());
            
            st.execute(create_buff.toString());
    
//            System.out.println("create: "+create_buff.toString());

            if (mysql_nfs_tmpdir!= null)
            {
                File tmpf = null;
                try
                {
                    tmpf = File.createTempFile("dXXXx", ".datatable", mysql_nfs_tmpdir);
                }
                catch(IOException ie)
                {
                    throw new FRuntimeException("i/o error: %s at creating tmpfile for writing datatable", ie.getMessage(), ie);
                }
                
                PrintWriter pw = FileUtils.getWriter(tmpf);
                for (int i=0; i<getSize(); i++)
                {
                    ObjectGetter og = data.get(i);
        
                    pw.printf("%d", i);

                    //ps.setInt(1, i);
        
                    for (int j=0, fi=2; j<getter.size(); j++,  fi++)
                    {
                        Object o = header.getHeader(j).get(og);
                        /*System.out.println("next object: "+o );
                        if (o != null)
                        {
                            System.out.println("class : "+o.getClass());
                        }
                        */
                        if (o == null || (o instanceof Double && Double.isNaN((Double)o)))
                        {
                            //System.out.println("skipping");
                            //pw.printf("\t");
                            pw.printf("\t\\N");
                            continue;
                        }
                        
                        pw.printf("\t%s", header.getHeader(j).getFormatted(og));
                        
                    }
                    pw.println();
                }
                pw.close();


                st.execute(lastq = "load data infile '"+tmpf.getAbsolutePath()+"' into table "+ table_prefix+DINFOS +" fields terminated by '\\t';");
                tmpf.delete();
                return table_prefix;

            }
            ps = con.prepareStatement(insert_buff.toString());
    
            for (int i=0; i<getSize(); i++)
            {
                ObjectGetter og = data.get(i);
    
                ps.setInt(1, i);
    
                for (int j=0, fi=2; j<getter.size(); j++,  fi++)
                {
                    converter.get(j).set(ps, fi, getter.get(j).get(og));
                }
                lastq = ""+ps;
                ps.execute();
            }
            ps.close();
        }
        catch(SQLException e)
        {
            
            throw new RuntimeException("sql error:"+e.getMessage()+" at saving data table last command:"+lastq, e);
        }
        
        return table_prefix;
    }

    public void writeXML(Element e)
    {
        Element dt_element = new Element("datatable");
        org.apache.commons.io.output.ByteArrayOutputStream os = new org.apache.commons.io.output.ByteArrayOutputStream();
        this.writeCSV(os);
        CDATA content = new CDATA(os.toString());

        dt_element.addContent(content);
        e.addContent(dt_element);
    }

    public static DataTable readFromXML(Element e)
    {
        //System.out.println(XMLUtils.prettyXML(e));
        //File nfs_dir = config.ConfigManager.getNFSDir();
        //File tmp = new File("/home/g/grubermar/Dokumente/ContextManProj/tmp.data");
        //tmpfile = FileUtils.createTmpFile(nfs_dir, "syns_", "tsv");
        //DataTable res = null;
        File tmp = FileUtils.createTmpFile(null, "dt_pre_load", ".tmp");
            //tmp = File.createTempFile("tmp", ".data"); 
            //tmp.deleteOnExit();
        PrintWriter pw = FileUtils.getWriter(tmp);
        List l = e.getContent();
        for (Object c : l)
        {
            if (c instanceof org.jdom.CDATA)
            {
                CDATA cdat = (CDATA)c;
                pw.printf(cdat.getText());
                break;
            }
        }
        pw.close();
        DataTable res = new DataTable();
        res.readDataFile(tmp);
        return  res;
    }

    public static DataTable readFromDb(Connection con, String table_prefix)
    {
        try
        {
        
            Statement st = con.createStatement();
    
            ResultSet header_rs = st.executeQuery("select * from "+ table_prefix+HINFOS+" order by "+HEADER_IDX_FIELD);

            if (!header_rs.first())
            {
                throw new RuntimeException("cannot find header table : "+table_prefix+HINFOS+" for data table prefix: "+table_prefix);
            }
            header_rs.beforeFirst();
            
            Vector<String> header_names= new Vector<String>();
            Vector<String> header_fields= new Vector<String>();

            Vector<SimpleTypeConverter.SQLConverter> converters = new Vector<SimpleTypeConverter.SQLConverter>();
            Vector<Integer> field2index = new Vector<Integer>();

            HashMap<String, Integer> fieldname2idx = new HashMap<String, Integer>();

            while (header_rs.next())
            {
                
                String fn = header_rs.getString(NAME_FIELD);
                fieldname2idx.put(fn, header_names.size());
                field2index.add(null);
                converters.add(SimpleTypeConverter.SIMPLETYPE.get(header_rs.getString(HEADER_CONVERTER_NAME)).getSQLConverter());

                header_names.add(header_rs.getString(HEADER_LABEL_FIELD));
                header_fields.add(fn);
                
                

            }
            
            ObjectGetter.ObjectGetterHeader oh = ObjectGetter.createHeader();
            oh.headers=StringUtils.convert(header_names);
    
            for (int i=0; i<oh.headers.length; i++)
            {
                oh.addField(oh.headers[i], i);
            }
            

            ResultSet data_rs = st.executeQuery("select * from "+ table_prefix+DINFOS);

            if (!data_rs.first())
            {
                throw new RuntimeException("cannot find data table : "+table_prefix+DINFOS+" for data table prefix: "+table_prefix);
            }
            data_rs.beforeFirst();

            

            ResultSetMetaData rmsd = data_rs.getMetaData();
            for (int i=0; i<rmsd.getColumnCount(); i++)
            {
                Integer fieldindex=fieldname2idx.get(rmsd.getColumnName(i+1));
                if (fieldindex==null)
                {
                    continue;
                }
                field2index.set(fieldindex, i+1);
            }

            
            int datasize=0;
            while (data_rs.next())
            {
                oh.addData();
                for (int i=0; i<field2index.size(); i++)
                {
                    oh.getData(datasize).add(converters.get(i).get(data_rs, field2index.get(i)));
                }
                datasize++;
            }

            DataTable dt = new DataTable();
            dt.header = new DataHeaderInfos();
            dt.o_header = oh;
            dt.header.init(oh);
            dt.data=oh.data;
            return dt;
        }
        catch(SQLException e)
        {
            throw new RuntimeException("sql error:"+e.getMessage()+" at loading data table from prefix: "+table_prefix, e);
        }
    }


    public DataTable(DataTable dt)
    {
        o_header = dt.o_header;
        header=dt.header;
        data=dt.data;

        filtered_data=dt.filtered_data;
    }

    public ObjectGetter<A> getDataByIdx(int idx)
    {
        return data.get(idx);
    }
    
    public interface Converter<B>
    {
        public B convert(ObjectGetter og);
    }
    
    public<B> Vector<B> convert(Converter<B> c)
    {
        return convert(c, false);
    }
    public<B> Vector<B> convert(Converter<B> c, boolean filtered)
    {
        Vector<ObjectGetter<A>> d = (filtered && filtered_data!=null) ? filtered_data : data;
        return ObjectGetter.map(d, (ObjectGetter<A> og) -> c.convert(og));
    }
    public A getInDataByIdx(int idx)
    {
        return getInDataByIdx(idx, false);
    }

    public A getInDataByIdx(int idx, boolean in_filtered)
    {
        final Vector<ObjectGetter<A>> v = (in_filtered)?filtered_data:data;

        return v.get(idx).getInData();
    }
    
    public void sortByHeader(String fieldname, boolean asc)
    {
    	ObjectGetter.sortBy(data, getGetter(fieldname),asc);
    }


    public static class MultiColSorter
    {
        Vector<Pair<ReflectedObject, Boolean>> sort_infos = new Vector<>();

        DataTable dt;

        public MultiColSorter(DataTable dt, String fn, boolean asc)
        {
            this.dt = dt;
            sort_infos.add(Pair.create(dt.getGetter(fn), asc));
        }

        public MultiColSorter add(String fn, boolean asc)
        {
            sort_infos.add(Pair.create(dt.getGetter(fn), asc));
            return this;

        }

        int check_dir(int c, boolean asc)
        {
            return (asc) ? c : -c;
        }

        int compare(ObjectGetter og1, ObjectGetter og2, ReflectedObject ro, boolean asc)
        {
            Object o1 = ro.get(og1);
            Object o2 = ro.get(og2);
            if (o1 == null && o2 == null)
                return 0;
            
            
            if (o1 == null || o2 == null)
                return check_dir((o1 == null) ? -1 :  1, asc);

            if ((o1 instanceof Comparable) && (o2 instanceof Comparable))
            {
                Comparable c1 = (Comparable)o1;
                Comparable c2 = (Comparable)o2;
                return check_dir(c1.compareTo(c2), asc);
            }
            String s1 = "" + o1;
            String s2 = "" + o2;
            return check_dir(s1.compareTo(s2), asc);
        }

        int compare(ObjectGetter og1, ObjectGetter og2)
        {
            int diff = 0;
            for (Pair<ReflectedObject, Boolean> p : sort_infos)
            {
                if(0 != (diff = compare(og1, og2, p.getFirst(), p.getSecond())))
                    return diff;
            }
            return diff;
        }

        public void sort()
        {
            Collections.sort(dt.data, (og1, og2) -> compare((ObjectGetter)og1, (ObjectGetter)og2));
        }
    }

    /**
     * sort the data table by multiple columns. This method returns a MultiColSorter, to which further columns 
     * that should be sorted can be added. The table is sorted by the first column, if this column contains equal
     * entries those are sorted by the second column and so on. A typical call would be. 
     * dt.multiColumnSort("field1",true).add("field2",false).sort()
     * @param fn name of the column by which it should be sorted
     * @param asc sorted in ascending order
     * @return MultiColSorter to which further columns can be added before its sort method is called
     */
    public MultiColSorter multiColumnSort(String fn, boolean asc)
    {
        return new MultiColSorter(this, fn, asc);
    }

    public Collection<A> getInData()
    {
        return getInData(false);
    }
    public Collection<A> getInData(boolean in_filtered)
    {
    	Vector<A> indata = new Vector<A>();

        final Vector<ObjectGetter<A>> v = (in_filtered)?filtered_data:data;
    	for(ObjectGetter<A> og : v)
    	{
    	    indata.add(og.getInData());
    	}
    	return indata;
    }
    
    public interface DataFilterListener
    {
        public void filterChanged(FilterSet new_filter, int new_size);
    }

    Vector<FilterSet> saved_filters = new Vector<FilterSet>();
    

    FilterSet quick_filter = new FilterSet("quickfilter");
    
    FilterSet current_filter = null;

    Vector<DataFilterListener> listeners = new Vector<DataFilterListener>();

    public ObjectGetter getDataAt(int idx)
    {
        if (current_filter!=null)
        {
            return filtered_data.get(idx);
        }
        return data.get(idx);
    }
    public void addDataFilterListener(DataFilterListener dfl)
    {
        listeners.add(dfl);
    }

    public void removeDataFilterListener(DataFilterListener dfl)
    {
        listeners.remove(dfl);
    }

    public boolean isFiltered()
    {
        return current_filter!=null;
    }

    public int getFilteredSize()
    {
        return (filtered_data!=null)? filtered_data.size() : getSize();
    }

    public Vector<ObjectGetter<A>> getFilteredData()
    {
        return (filtered_data!=null)?filtered_data:data;
    }
    void notifyListeners()
    {
        FilterSet f = current_filter;
        int ns=(f==null)?data.size():filtered_data.size();
        
        
        for (DataFilterListener dfl: listeners)
        {
            dfl.filterChanged(f,ns);
        }
    }
    public static class Filter
    {
        Vector<ObjectGetter> explicite_filter = null;
        public String name;
        public String expression;
        
        public boolean restrict_preceeding_results=true;
        public int resulting_rows=-1;
    }

    public static class FilterSet
    {
        String name;
        int num_resulting_rows=-1;
        Vector<Filter> filters = new Vector<Filter>();

        public FilterSet(String name)
        {
            this.name=name;
        }

        public FilterSet(String set_name, String filter_name, String expression, boolean use_or)
        {
            this.name=set_name;
            addFilter(filter_name, expression, use_or);
        }
        
        public void clear()
        {
            filters.clear();
        }

        public void addFilter(Vector<ObjectGetter> explicit , boolean use_or)
        {
            Filter f = new Filter();
            f.explicite_filter = explicit;
            f.restrict_preceeding_results=!use_or;
            filters.add(f);
        }
        public void addFilter(String name, String expression, boolean use_or)
        {
            Filter f = new Filter();
            f.name=name;
            f.expression=expression;
            f.restrict_preceeding_results=!use_or;
            filters.add(f);
        }

        
    }

    public DataTable<A> filterByIdx(Set<Integer> idx)
    {
//        DataTable dt = new DataTable(this.header.getHeaderNames());
        DataTable dt = new DataTable(ObjectGetter.map(this.header.headers,(DataHeader dh)-> dh.getName()));
        for (int i : idx)
        {
            ObjectGetter<A> row = getDataAt(i);
            dt.addRecordWithRawData(row.getInData(), row.data.toArray());
        }
        return dt;
    }
    
    public DataTable getFilteredTable(Collection<A> data)
    {
        HashSet<Integer> indeces = new HashSet<>();
        HashSet h = ObjectGetter.toSet(data);
        for (int i=0; i<getSize(); i++)
        {
            if (!h.contains(this.data.get(i).getInData()) && !h.contains(this.data.get(i)))
                continue;

            indeces.add(i);
        }
        return filterByIdx(indeces);
        
    }

    public DataTable getFilteredTable(Set<Tuple> filter, boolean reset_first,String...headers)
    {
        DataTable dt = new DataTable(this.header.getHeaderNames());

        for (ObjectGetter<A> row : filterSet(filter, reset_first, headers))
        {
            dt.addRecordWithRawData(row.getInData(), row.data.toArray());
        }
        return dt;
    }
    
    public DataTable getFilteredTableFromCurrentFilter()
    {
        DataTable dt = new DataTable(this.header.getHeaderNames());
        
        for (ObjectGetter<A> row : getFilteredData())
        {
            dt.addRecordWithRawData(row.getInData(), row.data.toArray());
        }
        return dt;


    }

    public DataTable getFilteredTable(String header, Set<Object> filter, boolean reset_first)
    {
        DataTable dt = new DataTable(this.header.getHeaderNames());

        for (ObjectGetter<A> row : filterSet(header, filter, reset_first))
        {
            dt.addRecordWithRawData(row.getInData(), row.data.toArray());
        }
        return dt;
    }

    public Vector<ObjectGetter<A>> filterSet(String header, Set<Object> filter, boolean reset_first)
    {
    	Vector<ObjectGetter<A>> filtered = new Vector<>();
    	Vector<ObjectGetter<A>> tofilter = reset_first ? data : filtered_data;
    	System.out.println(filter);
    	for(ObjectGetter<A> og : tofilter)
    	{
//    		System.out.println(og.get(header));
    		if(filter.contains(og.get(header)))
    		{
    			filtered.add(og);
    		}
    	}
    	return filtered_data = filtered;
    }

    public Vector<ObjectGetter<A>> filterSet(Set<Tuple> filter, boolean reset_first,String...headers)
    {
    	Vector<ObjectGetter<A>> filtered = new Vector<>();
    	Vector<ObjectGetter<A>> tofilter = reset_first ? filtered_data : data;
    	TupleGetter tg = getTupleGetter(headers);
        for(ObjectGetter<A> og : tofilter)
    	{
    		if(filter.contains(tg.get(og)))
    		{
    			filtered.add(og);
    		}
    	}
    	return filtered_data = filtered;
    }
    
    public void addAggregateColumn(NumUtils.AGGREGATE agg, String aggHeader, boolean ignore, String ... headers)
	{
		Vector<Number> aggCol = new Vector<>();
		if(ignore)
		{
			headers = getAllOtherHeaders(headers);
		}
		
		for(int i=0; i<data.size(); i++)
		{
			ObjectGetter og = data.get(i);
			Vector<Number> numbers = new Vector<>();
			for(String h : headers)
			{
			    Object o = og.get(h);
			    if(o!=null)
			    {
			        numbers.add((Number)o);
			    }
			}
			aggCol.add(NumUtils.reduce(numbers, agg));
		}
		addData(aggHeader, new ReflectedObject(true), aggCol);
	}
    
    public String[] getAllOtherHeaders(String...headers)
    {
    	String[] heads = new String[header.headers.size()-headers.length];
    	HashSet<String> excluded = new HashSet<>();
    	for(String h : headers)
    	{
    		excluded.add(h);
    	}
    	int i=0;
    	for(String h : header.getHeaderNames())
    	{
    		if(excluded.contains(h))
    			continue;
    		heads[i] = h;
    		i++;
    	}
		return heads;
    }
    public DataTable collapseTable(NumUtils.AGGREGATE agg, Vector<String> headerLabels ,Vector<Vector<String>> headers)
    {
        if(headerLabels.size()!=headers.size())
        {
            ExceptionManager.throwRuntimeException("headerLabels and headers have to have the same length");
        }
       DataTable coll = null;
       for(int j=0; j<headers.size(); j++)
       {
           Vector<String> tobecollapsed  = headers.get(j);
           if(tobecollapsed.size()>1)
           {
               Vector<Number> aggCol = new Vector<>();
               for(int i=0; i<data.size(); i++)
               {
                   ObjectGetter og = data.get(i);
                   Vector<Number> numbers = new Vector<>();
                   for(String h : tobecollapsed)
                   {
                       numbers.add((Number)og.get(h));
                   }
                   aggCol.add(NumUtils.reduce(numbers, agg));
               }
               if(coll==null)
               {
                   coll = DataTable.buildTable(aggCol, DataTable.buildHeader(headerLabels.get(j),new ReflectedObject(true)));
               }
               else
               {
                   coll.addData(headerLabels.get(j),new ReflectedObject(true),aggCol);
               }
           }
           else
           {
               if(coll==null)
               {
                   coll = DataTable.buildTable(getFieldVector(tobecollapsed.get(0)), DataTable.buildHeader(headerLabels.get(j),new ReflectedObject(true)));
               }
               else
               {
                   coll.addData(headerLabels.get(j),new ReflectedObject(true),getFieldVector(tobecollapsed.get(0)));
               }
           }
       }
       return coll;
    }
    
    public void saveQuickFilter(String name)
    {
        if (quick_filter.filters.size()==0)
        {
            return; //nothing to save
        }
        FilterSet fs=new FilterSet( name);
        fs.filters.addAll(quick_filter.filters);
        fs.num_resulting_rows=quick_filter.num_resulting_rows;
        saveFilter(fs,name);
    }
    
    public void saveFilter(FilterSet fs, String name)
    {
        if (name!=null)
        {
            fs.name=name;
        }
        saved_filters.add(fs);
        
    }

    public void clearFilter()
    {
        current_filter=null;
        if (filtered_data != null)
        {
            filtered_data.clear();
        }
        filtered_data = null;

        quick_filter.clear();
        notifyListeners();
    }
    
    public void quickFilter(String expression, boolean reset_first)
    {
    	quickFilter(expression, reset_first, false);
    }
    
    public void quickFilter(String expression, boolean reset_first, boolean include_null)
    {
        if (reset_first)
        {
            quick_filter.clear();
        }

        if (expression == null || expression.length() == 0)
        {
            clearFilter();
            setFilter((Expression)null, null, include_null);
            return;
        }

        quick_filter.addFilter("filter",expression,false);
        setFilter(quick_filter, include_null);
        
    }
    
    public String getSetSeparator(boolean is_or)
    {
        return (is_or)?"||":"&&";
    }
    
    public void setFilter(FilterSet fs)
    {
    	setFilter(fs,false);
    }

    void setFilter(Expression exp, Vector<ReflectedObject> getters, boolean include_null)
    {
        if (filtered_data == null)
        {
            filtered_data = new Vector<>();
        }
        filtered_data.clear();

        Object[] o_data = (exp == null) ? null : new Object[getters.size()];

        for (ObjectGetter o : data)
        {
            if(!FlagFilter.filter(flagfilters, o))
                continue;

            if (exp == null)
            {
                filtered_data.add(o);
                continue;
            }

            for (int i = 0; i < getters.size(); i++)
            {
                o_data[i]=o;
            }

            if(exp.check_null(o_data))
            {
                if(include_null)
                {
                    filtered_data.add(o);
                }
                continue;
            }

            //System.out.println("o_data: "+ObjectGetter.map(IteratorUtils.getIntIterator(0, o_data.length), (Integer idx) -> getters.get(idx).get(o_data[idx])));

            if(!((Boolean)exp.eval(o_data)))
                continue;

            //System.out.println("matches!");
            filtered_data.add(o);
        }

        notifyListeners();

    }
    public void setFilter(FilterSet fs, boolean include_null)
    {

        
        StringBuffer pre_expr=new StringBuffer();

        HashMap<String, ReflectedObject> hlookup=header.getReflectedObjectLookup();
        Vector<String> order = new Vector<String>();
        Vector<ReflectedObject> getters = new Vector<ReflectedObject>();
        
        for (Map.Entry<String,ReflectedObject> e: hlookup.entrySet())
        {
            order.add(e.getKey());
            getters.add(e.getValue());
        }
        Object[] o_data = new Object[order.size()];

        for (Filter f : fs.filters)
        {                                                                  
            String sep=getSetSeparator(f.restrict_preceeding_results);
            if (pre_expr.length()>0)
            {
                String pre = "( "+pre_expr.toString()+" ) ";
                pre_expr.setLength(0);
                pre_expr.append(pre);
                pre_expr.append(sep);
            }
            pre_expr.append('(');
            pre_expr.append(f.expression);
            pre_expr.append(')');

            if (f.resulting_rows<0)
            {

                Expression exp = new Expression(pre_expr.toString(), hlookup);
                exp.setOrder(order);
                
                //TestExpression te = new TestExpression(pre_expr.toString(), hlookup,true);

                                
                long filt_start=System.currentTimeMillis();
                
                f.resulting_rows=0;
                int null_objects = 0;
                for (ObjectGetter o : data)
                {
                    for (int i=0; i<getters.size();i++)
                    {
                        o_data[i]=o;
                    }
                    if(exp.check_null(o_data))
                    {
                    	if(include_null)
                    	{
                    		f.resulting_rows++;
                    	}
                    	null_objects++;
                    	continue;
                    }
                    //System.out.println("o_data: "+ObjectGetter.map(IteratorUtils.getIntIterator(0, o_data.length), (Integer idx) -> getters.get(idx).get(o_data[idx])));
                    //System.out.println("o data: "+o+" \n\n+" + pre_expr.toString());
                    if(!(Boolean)(exp.eval(o_data)))
                        continue;
                    f.resulting_rows++;
                }
                long filt_end=System.currentTimeMillis();
                log.debug("filtering with %s took: %2.3f sec\n", pre_expr.toString(), (filt_end-filt_start)/1000.0);
                log.debug("got "+f.resulting_rows+" "+null_objects+" were null");
            }
        }
        current_filter = fs;
        if (filtered_data == null)
        {
            filtered_data = new Vector<>();
        }
        
        Expression exp = new Expression(pre_expr.toString(), hlookup);
        exp.setOrder(order);
        setFilter(exp, getters, include_null);
        current_filter.num_resulting_rows = filtered_data.size();



        
    }


    public static DataTable createTable(Collection data, TableRenderInfos  table_infos)
    {
        Vector<String> headers = table_infos.getHeaderNames();
        
        Object[] infos = new Object[headers.size()*2];
        for (int i=0,j=0; i<headers.size(); i++, j+=2)
        {
            infos[j]=headers.get(i);
            infos[j+1]=ReflectedObject.apply(table_infos, "getCellInfo", i, new ReflectedObject(true));
        }

        
        return DataTable.createTable(data, infos);
    }
    
    
    /** header definitions is a list of two tuples: expression and header name
     * if header name is null then the expression is taken as a header
     */
    public static DataTable createTable(Collection data, Object ... header_definitions)
    {

        Vector<String> header_names = new Vector<String>();

        Vector<ReflectedObject> ref_objs = new Vector<ReflectedObject>();

        
        DataTable dt = new DataTable();
        dt.o_header  = ObjectGetter.createHeader();
        
        Vector<ReflectedObject> v = new Vector<ReflectedObject>();

        for (int i=0; i<header_definitions.length; i+=2)
        {
            

            String hn = (String)header_definitions[i];

            ReflectedObject ro = null;

            if (header_definitions[i+1] instanceof ReflectedObject)
            {
                ro=(ReflectedObject)header_definitions[i+1];
            }
            else
            {
                
                String expr = (String)header_definitions[i+1];
                if (hn==null)
                {
                    hn=expr;
                }
                ro = new ReflectedObject(expr);
            }
            v.add(ro);
            dt.o_header.addField(hn, ro);
            //System.out.println("add data table header: "+hn+" ro: "+ro);
            Iterator it = data.iterator();
            int j=-1;
            while (it.hasNext())
            {
                j++;
                Object o =it.next();
                if (o==null)
                {
                    continue;
                }
                //System.out.printf("\tdata[%d]=%s\n", j,ro.get(o));
                if (j>=10)
                {
                    break;
                }
            }
            
                               
            
        }
        dt.o_header.updateHeaders();
        for (Object o : data)
        {
            dt.o_header.addData(o);
            /*System.out.printf("add next data\n");
            for (int i=0; i<v.size(); i++)
            {
                System.out.printf("\t%d: %s\n", i, v.get(i).get(o));
            }
            */

        }
        
        
        

        dt.header = new DataHeaderInfos();

        int i=0;
        for (String h: dt.o_header.headers)
        {
            dt.header.addHeader(h, new ReflectedObject("get", i++));
        }
        //dt.header.init(dt.o_header);
        dt.data=dt.o_header.getAllData();
        return dt;
    }

    public int getNumHeaders()
    {
        return header.headers.size();
    }
    public Object getFieldData(int i, String fieldname)
    {
        return o_header.getFieldValue(fieldname, i);
    }
               
    public int getSize()
    {
        return data.size();
    }

    
    /** returns a data table contianing the same header information
     * but without any data
     */
    public DataTable initEmptyTable()
    {
        DataTable dt = new DataTable();
        dt.o_header = o_header.clone(false);
        dt.header = new DataHeaderInfos();
        dt.header.init(dt.o_header);
        dt.data= dt.o_header.getAllData();
        return dt;

    }

    public DataTable filterAndExtend(DataTable othertable, String eqfield_name1, String eqfield_name2)
    {
        return filterAndExtend(othertable, eqfield_name1, eqfield_name2, false);
    }
    /** filters the current table so that only records where othertable
     * has also a record are kept, and these are then extended
     * by all the other headers in othertable
     */
    public DataTable filterAndExtend(DataTable othertable, String eqfield_name1, String eqfield_name2, boolean omitDuplicated)
    {

        HeaderGetterManager<Integer> hgm = DataTable.buildHeader(header.getHeader(0).name,  (Integer idx) -> ((ObjectGetter)getDataAt(idx)).get(0));
        log.debug("start headers with: "+header.getHeader(0).name);
        
        
        HashMap<Object, HashSet<Integer>> rowmapping = new HashMap<>();
        HashMap<Integer,Integer> rowidxmapping = new HashMap<>();

        for (int i=1; i<header.getHeaderNames().size(); i++)
        {
            final int col = i;
            hgm = hgm.add(header.getHeader(i).name, (Integer idx) -> ((ObjectGetter)getDataAt(idx)).get(col));
        }

        log.debug(header.getHeaderNames()+" merged by "+eqfield_name1);
        log.debug(othertable.header.getHeaderNames()+" merged by "+eqfield_name2);
        ReflectedObject ro = othertable.getGetter(eqfield_name2);
        for (int i=0; i<othertable.getSize(); i++)
        {
            Object o = ro.get(othertable.getDataAt(i));
            HashSet<Integer> hs = rowmapping.get(o);
            if (hs == null)
            {
                rowmapping.put(o, hs = new HashSet<>());
            }
            hs.add(i);
        }
        
        Vector<Integer> indices2 = new Vector<>();
        DataHeader header1 = header.getHeader(eqfield_name1);
        if(header1==null)
        {
            ExceptionManager.throwRuntimeException("no header with name "+eqfield_name1);
        }

        DataHeader header2 = othertable.header.getHeader(eqfield_name2);
        //add all other fields from othertable
        for (DataHeader dh : othertable.header.headers)
        {
            if (dh==header2)
                continue;
            
            indices2.add(dh.col_index);
        }

        for (int i=0; i<indices2.size(); i++)
        {
            final int cidx = indices2.get(i);
            String hn = othertable.header.getHeader(cidx).name;
            int suffix=1;
            while (header.findHeader(hn) != null)
            {
                suffix++;
                //throw new RuntimeException("multi header: "+hn+"!");
                hn += "_"+suffix;
            }
            
            hgm = hgm.add(hn, (Integer idx) -> ((ObjectGetter)othertable.getDataAt(rowidxmapping.get(idx))).get(cidx));
        }
            

        for (int i=0; i<data.size(); i++)
        {
            ObjectGetter owndata = data.get(i);
            Object own_info = header1.get(owndata);
            if (own_info==null)
            {
                continue; 
            }

            HashSet<Integer> otherrows = rowmapping.get(own_info);
            if (otherrows==null || otherrows.size() == 0)
            {
                continue;
            }
            if (otherrows.size()!=1)
            {
                if(omitDuplicated)
                    continue;
                throw new RuntimeException("expected exactly one info for "+own_info+" (otherwise cannot uniqually extend) got : "+otherrows.size());
            }
            rowidxmapping.put(i, otherrows.iterator().next());
        }
        
        

        return buildTable(rowidxmapping.keySet(), hgm);
    }
    
    
    public interface ColumnInfoGetter<A>
    {
        public A getInfo(ObjectGetter og);
    }

    public<A> DataTable addColumn(String columnname, ColumnInfoGetter<A> cg)
    {
        Vector<A> col_data = new Vector<>();
        for (int i=0; i< getSize(); i++)
        {
            col_data.add(cg.getInfo(getDataAt(i)));
        }
        addData(columnname, new ReflectedObject(true), col_data);
        return this;
    }
    /** extends the current table so that all it contains also the headers of othertable
     * for records that are only contained in one table the headers of the other table are filled with null values
     * only records that are null in the eqfield are omitted
     */
    public DataTable extend(DataTable othertable, String eqfield_name1, String eqfield_name2)
    {

        HashMap filterLookup=othertable.getMultiLookup(eqfield_name2);

        DataTable extended = initEmptyTable();
        
        DataHeader header1 = header.getHeader(eqfield_name1);
        DataHeader header2 = othertable.header.getHeader(eqfield_name2);

        Vector<DataHeader> extend_headers = new Vector<DataHeader>();

        int ngetters = o_header.headers.length;
        //add all other fields from othertable
        for (DataHeader dh : othertable.header.headers)
        {
            if (dh==header2)
                continue;
            
            
            extend_headers.add(dh);
            ReflectedObject ro = extended.o_header.addField(dh.name, ngetters++); 
            
            extended.header.addHeader(dh.name, ro);
        }
        
                
        extended.o_header.updateHeaders();

        HashSet<Object> eqFieldDataMerged = new HashSet<>();
        for (int i=0; i<data.size(); i++)
        {
            ObjectGetter owndata = data.get(i);
            Object own_info = header1.get(owndata);
            if (own_info==null)
            {
                continue; 
            }

            Vector<ObjectGetter> hitinfos = (Vector<ObjectGetter>)filterLookup.get(own_info);
            if (hitinfos==null)
            {
            	//TODO add null getters here
                continue;
            }
            if (hitinfos.size()!=1)
            {
                throw new RuntimeException("expected exactly one info for "+own_info+" (otherwise cannot uniqually extend) got : "+hitinfos.size());
            }
            ObjectGetter merged_getter = extended.o_header.addData();
            
            ObjectGetter extend_data = hitinfos.get(0);
            merged_getter.data.addAll(owndata.data);
            
            for (DataHeader dh : extend_headers)
            {
                merged_getter.data.add(dh.get(extend_data));
            }
            eqFieldDataMerged.add(own_info);
        }
        
        HashMap filterLookupOwn=getMultiLookup(eqfield_name1);
        for (int i=0; i<othertable.data.size(); i++)
        {
            ObjectGetter otherdata = (ObjectGetter)othertable.data.get(i);
            Object other_info = header2.get(otherdata);
            
            if (other_info==null || eqFieldDataMerged.contains(other_info))
            {
            	System.out.println("skip");
                continue; 
            }

            Vector<ObjectGetter> hitinfos = (Vector<ObjectGetter>)filterLookupOwn.get(other_info);
            if (hitinfos==null)
            {
            	//TODO add null getters here
            }
            if (hitinfos.size()!=1)
            {
                throw new RuntimeException("expected exactly one info for "+other_info+" (otherwise cannot uniqually extend) got : "+hitinfos.size());
            }
            ObjectGetter merged_getter = extended.o_header.addData();
            
            ObjectGetter extend_data = hitinfos.get(0);
            merged_getter.data.addAll(otherdata.data);
            
            for (DataHeader dh : extend_headers)
            {
                merged_getter.data.add(dh.get(extend_data));
            }
        }
        
        return extended;
    }
    
    
    public<A> void extendTable(Vector<A> data, HeaderGetterManager<A> hgm)
    {
        //dt.addData("in_peptome", new ReflectedObject(true), in_peptome);
        String headers[] = hgm.names.toArray(new String[0]);

        Getter[] ga = ObjectGetter.map(headers, (String s) -> hgm.getGetter2Header(s)).toArray(new Getter[0]);

        //System.out.printf("extend by %d headers and %d gettters\n", headers.length, ga.length);
        for (int i=0; i<headers.length; i++)
        {
            String n = headers[i];
            final Getter g = ga[i];
            Vector edata = ObjectGetter.map(data, (a) -> g.get(a));
            addData(n, new ReflectedObject(true), edata);
        }
    }
    
    
    public void addData(ObjectGetter og, boolean check)
    {
        if (check)
        {
            String last_tested=null;
            try
            {
                //check if all getter work
                for (DataHeader dh : header.headers)
                {
                    last_tested=dh.name;
                    dh.get(og);
                }

            }
            catch(Exception e)
            {
                throw new RuntimeException("invalid data added, error at getting header: "+last_tested, e);
            }
            
        }
        data.add(og);
    }
    public void writeCSV(File f)
    {
        try
        {
            FileOutputStream fos = new FileOutputStream(f);
            writeCSV(fos);
            fos.close();
        }
        catch(IOException e)
        {
            throw new RuntimeException("i/o error: "+e.getMessage()+" while writing data table to csv file:"+f.getAbsolutePath(), e);
        }
    }
    public void writeCSV(OutputStream os)
    {
        try
        {
            
            CsvWriter w = new CsvWriter(os, '\t', java.nio.charset.Charset.defaultCharset());
    
            Vector<DataHeader> towrite = new Vector<DataHeader>();
            for (DataHeader h: header.headers)
            {
                if(!h.visible)
                    continue;
    
                w.write(h.name);
                towrite.add(h);
                //System.out.printf("add headeR: %s\n", h.name);
            }
            w.endRecord();
                
            for (ObjectGetter o : getFilteredData())
            {
                for (DataHeader dh: towrite)
                {
                    w.write(dh.getFormatted(o));
                }
                w.endRecord();
            }
            
            w.close();
        }
        catch(IOException e)
        {
            throw new RuntimeException("i/o error: "+e.getMessage()+" while writing data table to csv stream", e);
        }

            
    }
    
    public Vector getFieldVector(String field)
    {
    	return getFieldVector(field,false);
    }
    
    public Vector getFieldVector(String field, boolean filtered)
    {
        return getColumn(field,filtered);
        /*
        ReflectedObject ro = getGetter(field);
        Vector rv = new Vector();
        if(filtered)
        {
        	for (Object og: filtered_data)
   	        {
   	            rv.add(ro.get(og));
   	        }
        }
        else
        {
	        for (Object og: data)
	        {
	            rv.add(ro.get(og));
	        }
        }
        return rv;
        */
    }
    public HashMap getMultiLookup(String field)
    {
        
        return ObjectGetter.getMultiLookup(data, getGetter(field), null);
    }

    public<B> HashMap<B, ObjectGetter<A>> getLookup(String field)
    {
        
        ReflectedObject ro = o_header.getFieldInfo(field);

        return ObjectGetter.buildMap(data, (ObjectGetter<A> og) -> (B)ro.get(og), (ObjectGetter<A> og) -> og);
    }
    
    public HashMap getMapping(String field1, String field2)
    {
        ReflectedObject key_getter = getGetter(field1);
        ReflectedObject value_getter = getGetter(field2);
        HashMap rv = new HashMap();

        for (ObjectGetter og: data)
        {
            rv.put(key_getter.get(og), value_getter.get(og));
        }
        return rv;
    }
    
    public DataTable()
    {

    }

    /** creates a data table with the headers as names
     * and creates the objectgetter headers
     * after that one can add records by addrecord, simply as vector of some mixed objects or Object ...
     * 
     */
    public DataTable(String ... headers)
    {
       o_header = new ObjectGetter.ObjectGetterHeader<A>();
       data = o_header.data;

       LogConfig.getLogger().debug("headers: %s", StringUtils.join(",", headers));
       header = new DataHeaderInfos();
       int nheaders=0;
       for (String n : headers)
       {
           
           o_header.addField(n, nheaders++);
           
       }
       o_header.updateHeaders();

       header.init(o_header);
    }
    public DataTable(Collection<String>  headers)
    {
    	this(headers.toArray(new String[headers.size()]));
    }
    
    
    public ObjectGetter addRecord(Vector<Object> data)
    {
        if (o_header.headers.length!=data.size())
        {
            throw new RuntimeException("invalid record size, header has "+o_header.headers.length+" record has :"+data.size());
        }
        ObjectGetter<A> og = o_header.addData();
        og.data=data;
        return og;
    }
    public ObjectGetter addRecord(Object ... data)
    {
        Vector<Object> vo = new Vector<Object>();
        for (Object o : data)
        {
            vo.add(o);
        }
        return addRecord(vo);
    }

    public void addRecordWithRawData(A raw_data, Object ... data)
    {
        Vector<Object> vo = new Vector<Object>();
        for (Object o : data)
        {
            vo.add(o);
        }
        ObjectGetter og =  addRecord(data);
        og.setInData(raw_data);
    }
    public ReflectedObject getGetter(String field)
    {
        return o_header.getFieldInfo(field);
    }

    public interface Getter<B>
    {
        public Object get(B o);
    }

    public interface Header2Getter
    {
        public Getter getGetter2Header(String n);
    }


    
    public static class HeaderGetterManager<A> implements Header2Getter
    {
        Logger log = LogConfig.getLogger();
        Vector<String> names = new Vector<>();

        HashMap<String, Getter<A>> h2g = new HashMap<>();

        public HeaderGetterManager(String n, Getter<A> g)
        {
            add(n,g);
        }

        public HeaderGetterManager<A> add(String n, Getter<A> g)
        {
            //log.info("header getter adds next header: >%s<", n);
            names.add(n);
            h2g.put(n,g);
            return this;
        }

        public HeaderGetterManager<A> add(String n, ReflectedObject refobj)
        {
            return add(n, (A a) -> refobj.get(a));
        }
            
       public HeaderGetterManager<A> add(String n, String expression)
        {
            return add(n, new ReflectedObject(expression));
        }

        public Getter getGetter2Header(String n)
        {
            return h2g.get(n);
        }
    }
    public static<A> HeaderGetterManager<A> buildHeader(String n, String expression)
    {
        return buildHeader(n, new ReflectedObject(expression));
    }

    public static<A> HeaderGetterManager<A> buildHeader(String n, ReflectedObject refobj)
    {
        return buildHeader(n, (A a) -> refobj.get(a));
    }


    public static<A extends Object> HeaderGetterManager<A> buildHeader(String n, Getter<A> g)
    {
        return new HeaderGetterManager<A>(n, g);
    }

    
    
    public static DataTable<Tuple> buildTable(Vector<Tuple> data, String ... headers)
    {
        HeaderGetterManager<Tuple> hgm = new HeaderGetterManager<Tuple>(headers[0], (Tuple t) -> t.get(0));
        for (int i=1; i<headers.length; i++)
        {
            final int idx = i;
            hgm.add(headers[i], (Tuple t) -> t.get(idx));
        }
        return buildTable(data, hgm, headers);
    }

    public static<A> DataTable<A> buildTable(Iterator<A> data, HeaderGetterManager<A> hgm)
    {
        return buildTable(data, hgm, hgm.names.toArray(new String[0]));
    }

    public static<A> DataTable<A> buildTable(Collection<A> data, HeaderGetterManager<A> hgm)
    {
        return buildTable(data, hgm, hgm.names.toArray(new String[0]));
    }

    public static<A> DataTable<Tuple<A>> buildFromTuples(Vector<Tuple<A>> data, String ... headers)
    {

        DataTable<Tuple<A>> dt = new DataTable<Tuple<A>>(headers);
        for (Tuple<A> t : data)
        {
            Object[] td = t.values();
            dt.addRecordWithRawData(t, td);

        }
        return dt;

    }
    public static<A> DataTable<A> buildTable(Collection<A> data, Header2Getter hg, String ... headers)
    {
        return buildTable((data==null) ? new SingleIterator<A>(null) : data.iterator(), hg, headers);
    }
    public static<A> DataTable<A> buildTable(Iterator<A> dataIt, Header2Getter hg, String ... headers)
    {
        DataTable<A> dt = new DataTable<A>(headers);
        
        Getter[] ga = ObjectGetter.map(headers, (String s) -> hg.getGetter2Header(s)).toArray(new Getter[0]);

        while (dataIt.hasNext())
        {
            A a = dataIt.next();
            Object[] td = ObjectGetter.map(ga, (Getter g) -> g.get(a)).toArray();
            dt.addRecordWithRawData(a, td);
        }
        return dt;
    }

    public<B> DataTable<A> addData(String label, Getter<B> g , Vector<B> data)
    {
        ReflectedObject ro = new ReflectedObject((B o) -> g.get(o));
        addData(label,ro,data);
        return this;
    }
    /** returns tihs for chaining */
    public DataTable addData(String label, String expression, Vector data)
    {
        ReflectedObject ro = new ReflectedObject(expression);
        addData(label,ro,data);
        return this;
    }        
    /** returns tihs for chaining */
    
    public interface TupleGetterTransformer<B>
    {
        public B transform(Tuple t);
    }
    public<B> DataTable addData(String label, TupleGetter tg, TupleGetterTransformer<B> tgt)
    {
        Vector<B> v = ObjectGetter.map(IteratorUtils.getIntIterator(0, getSize()), (Integer idx) -> tgt.transform(tg.get(idx)));
        addData(label, new ReflectedObject(true), v);
        return this;

    }
    public DataTable addData(String label, ReflectedObject ro, Vector data)
    {
        if(o_header==null)
        {
            o_header = ObjectGetter.createHeader();
            ReflectedObject nro = o_header.addField(label, ro);
            for (int i=0; i<data.size(); i++)
            {
                ObjectGetter<A> og = o_header.addData();
                og.add(data.get(i));
            }
            return this;
            
        }
        
        ReflectedObject nro = o_header.addData(label, ro, data);
        header.addHeader(label, nro);
        return this;
    }

    /** add new data to the table merged by the fields*/
    public void merge(ObjectGetter.ObjectGetterHeader  n_header, String eq_field_from, String eq_field_to)
    {
        throw new RuntimeException("not yet implemented!");
    }

    public static DataTable readFile(File f)
    {
        return readFile(f,null);
    }
    public static DataTable readFile(File f, String sep, String nullstr)
    {
        DataTable dt = new DataTable();
        dt.readDataFile(f, true, true, sep, -1, nullstr);
        return dt;
    }

    public static DataTable readFile(File f, String sep)
    {
        return readFile(f,sep,null);
    }

    static public<A> DataTable<A> readFromCSVBuffer(Reader reader, String sep)
    {
        DataTable<A> dt = new DataTable<A>();
        DataFile<A> df = new DataFile<A>();
        df.setContainsHeader(true);
        df.setSeparator(sep);
        
        dt.data = df.read(null, reader);

        dt.header = new DataHeaderInfos();
        if (dt.data.size()==0)
        {
            
            for (String s : df.getHeaders())

            {
                dt.header.addHeader(s);
            }
        }
        else
        {
            
            dt.header.init( dt.o_header = ObjectGetter.getHeader(dt.data));
        }
        
        return dt;
    }

    HashSet<String> comment_signs = new HashSet<>();

    /**
     * set comment Strings, lines that start with this String are not read
     * has to be called before readDataFile is called
     */
    public void setCommentSigns(String ... comments)
    {
        ObjectGetter.apply(comments, (s) -> comment_signs.add(s));
    }

    /** reads-in a headered-csv with auto detected separator */
    public void readDataFile(File f)
    {
        readDataFile(f,true,true, null);
    }

    /**
     * 
     * @param f file that should be read
     * @param csv should the csv reader be used. If it is used, escaping of quoted cells is done, but there might be other problems
     * @param gotheader does the cell contain a header. If csv is true, the first line is always taken as header independent of this value
     * @param sep cell separator
     */
    public void readDataFile(File f, boolean csv, boolean gotheader, String sep)
    {
        readDataFile(f,csv,gotheader,sep, -1);
    }
    public void readDataFile(File f, boolean csv, boolean gotheader, String sep, int maxdatalines2read)
    {
        readDataFile(f,csv,gotheader,sep,maxdatalines2read, null);
    }
    public void readDataFile(File f, boolean csv, boolean gotheader, String sep, int maxdatalines2read, String nullstr)
    {
        readDataFile(f, null, csv, gotheader, sep, maxdatalines2read, nullstr);
    }

    public void readDataFile(File f, Reader r, boolean csv, boolean gotheader, String sep, int maxdatalines2read, String nullstr)
    {
        DataFile<A> df = new DataFile<A>();
        df.setContainsHeader(gotheader);
        df.setCSV(csv);
        df.setSeparator(sep);
        df.setPeakSize(maxdatalines2read);
        if (comment_signs.size() != 0)
        {
            ObjectGetter.apply(comment_signs, (s) -> df.addCommentSigns(s));
        }
        data = df.read(f,r);
        
        
        header = new DataHeaderInfos();
        if (data.size()==0)
        {
            
            for (String s : df.getHeaders())
            {
                header.addHeader(s);
            }
        }
        else
        {
            
            header.init(o_header=ObjectGetter.getHeader(data));
        }
    }
    
    public void convertToString(String headername)
    {
        int idx = header.getHeader(headername).col_index;

        for (int i=0; i<getSize(); i++)
        {
            ObjectGetter og = getDataAt(i);
            og.set(idx, ""+og.get(idx));
        }

    }
    public DataTable getUniq(String ... uniq_fields)
    {
        return getUniqAndAggregate(null, uniq_fields);
    }

    public DataTable getUniqAndAggregate(HashMap<String, Vector<NumUtils.AGGREGATE>> field2aggregate, String ... uniq_fields)
    {
        HashMap<Tuple, Vector<Integer>> t2idx = new HashMap<>();

        final int N = uniq_fields.length;
        int col_idx[] = new int[N];
        String[] headernames = header.getHeaderNames().toArray(new String[0]);
        final int M = headernames.length;
        int backmap[] = new int[M];
        int tuple_builder[] = new int[M];
        Arrays.fill(backmap, -1);

        HashSet<Integer> collapsed = new HashSet<>();
        Vector<Pair<Integer, NumUtils.AGGREGATE>> aggr_func = new Vector<>();

        Vector<String> newnames = new Vector<>();

        for(int i=0; i<N; i++)
        {
            String s = uniq_fields[i];
            newnames.add(s);
            col_idx[i] = header.getHeader(s).col_index;
            backmap[col_idx[i]] = i;
        }

        newnames.add("UNIQCOUNT");
        int additional_fields = 0;
        NumUtils.AGGREGATE defag = NumUtils.AGGREGATE.FIRST;
        if (field2aggregate != null)
        {
            
            for (Map.Entry<String, Vector<NumUtils.AGGREGATE>> e : field2aggregate.entrySet())
            {
                String hn = e.getKey();
                final int idx = header.getHeader(hn).col_index;

                collapsed.add(idx);
                for (NumUtils.AGGREGATE ag : e.getValue())
                {
                    aggr_func.add(new Pair<Integer, NumUtils.AGGREGATE>(idx, ag));
                    newnames.add(ag+"("+hn+")");
                }
                additional_fields += e.getValue().size();
            }
        }
        for (int j=0; j<M; j++)
        {
            if (backmap[j] >= 0)
                continue;
            
            if (collapsed.contains(j))
                continue;

            
            newnames.add(header.getHeader(j).getName());
            aggr_func.add(new Pair<Integer, NumUtils.AGGREGATE>(j, defag));
        }
        

        
        
        for (int i=0; i<getSize(); i++)
        {
            ObjectGetter og = getDataAt(i);
            Tuple t = new Tuple(N);
            for (int j=0; j<N; j++)
            {
                t.set(j, og.get(col_idx[j]));
            }
            Vector<Integer> v = t2idx.get(t);
            if (v == null)
            {
                t2idx.put(t, v = new Vector<Integer>());
            }
            v.add(i);

        }
        Vector<Tuple> aggrdata = new Vector<>();

        final int NM = newnames.size();
        
        for (Map.Entry<Tuple, Vector<Integer>> e : t2idx.entrySet())
        {
            Tuple t = e.getKey();
            Tuple nt = new Tuple(NM);
            Vector<Integer> v = e.getValue();
            for (int j=0; j<t.cardinality(); j++)
            {
                nt.set(j, t.get(j));
            }

            int nidx = t.cardinality();

            nt.set(nidx++, v.size());
            for(Pair<Integer, NumUtils.AGGREGATE> p : aggr_func)
            {
                final int colidx = p.getFirst();
                if (p.getSecond() == NumUtils.AGGREGATE.FIRST)
                {
                    nt.set(nidx++, getDataAt(v.iterator().next()).get(colidx));
                    continue;
                }
                nt.set(nidx++, NumUtils.reduce(ObjectGetter.map(v, (Integer idx) -> (Number)getDataAt(idx).get(colidx)), p.getSecond()));
            }
            aggrdata.add(nt);
        }
        //System.out.printf("data table reduced %d -> %d by uniq: %s\n", getSize(), t2idx.size(), StringUtils.join(" , ", uniq_fields));
        return DataTable.buildTable(aggrdata, newnames.toArray(new String[0]));
    }

    public DataTable getUniqAndAggregateByAggrName(HashMap<String, Vector<String>> field2aggregate, String ... uniq_fields)
    {

        HashMap<String, Vector<NumUtils.AGGREGATE>> rv = new HashMap<>();
        for(Map.Entry<String,Vector<String>> e : field2aggregate.entrySet())
        {
            Vector<NumUtils.AGGREGATE> v = new Vector<>();
            for (String s : e.getValue())
            {
                v.add(NumUtils.AGGREGATE.get(s));
            }

            rv.put(e.getKey(), v);
        }
        return getUniqAndAggregate(rv, uniq_fields);
        
    }
    
    public boolean isNumeric(String name)
    {
        return o_header.isNumericHeader(name);
    }

    public boolean isNumeric(int field_idx)
    {
       
        return o_header.isNumericHeader(field_idx);
        //header.getHeader(field_idx).get
    }

    public TupleGetter getTupleGetter(String ... fields)
    {
        return new TupleGetter(this, fields);
    }
    
    public static class Sample
    {
        String s = StringUtils.generateRandomString(1+(int)(10*Math.random()));
        int v = -500+(int)(Math.random() * 1000);
        int v2 = (int)(Math.sin(Math.random()) * 1000);

       public int m1()
       {
           return v;
       }
       public String m2()
       {
           return s;
       }
       public int m3()
       {
           return v2;
       }
         
 
   }
  
    public Vector getColumn(String column, boolean filtered)
    {
        return ObjectGetter.map((filtered) ? filtered_data : data, (ObjectGetter<A> og) -> og.get(column));
    }
    
    public static void test()
    {
        
        Vector<Sample> v = ObjectGetter.create(10000, () -> new Sample());

        DataTable<Sample> dt = DataTable.buildTable(v,
                               DataTable.buildHeader("m1", (Sample s) -> s.m1()).
                               add("m2", (Sample s) -> s.m2()).
                               add("m2length", (Sample s) -> s.m2().length()).
                               add("info", (Sample s) -> s.m2() + ": " + Math.abs(s.m1())).
                               add("m3", (Sample s) -> s.m3())
                              );

        for (String s : dt.header.getHeaderNames())
        {
            System.out.println("numeric header: "+s+" "+dt.isNumeric(s));
        }

        DataTable dt2 = dt.getUniqAndAggregateByAggrName(MapBuilder.build("m1", ObjectGetter.buildVector("max", "min", "avg", "median")).
                                                            add("m3", ObjectGetter.buildVector("max", "min", "median", "avg"))
                                                         
                                                         , "m2length");
        dt.getUniqAndAggregate(null, "m2length", "m1");
        lmu.utils.swing.PagedDataTable.getDetailViewFrame(dt2, null , null);
    }
     

    

    public DataTable restrict(String ... headers)
    {
        Vector<Integer> indices = new Vector<>();
        for (String s : headers)
        {
            DataHeader dh = header.findHeader(s);
            if (dh == null)
            {
                throw new RuntimeException("cannot find header: "+ s+ " known headers: " + header.getHeaderNames());
            }

            indices.add(dh.col_index);
        }

        HeaderGetterManager<Integer> hgm = DataTable.buildHeader(headers[0],  (Integer idx) -> ((ObjectGetter)getDataAt(idx)).get(indices.get(0)));

        for (int i=1; i<indices.size(); i++)
        {
            hgm = hgm.add(headers[i], (Integer idx) -> ((ObjectGetter)getDataAt(idx)).get(indices.get(0)));
        }


        return buildTable(IteratorUtils.getIntIterator(0, getSize()), hgm);

    }

    public void setVisibleHeaders(String ... names)
    {
        HashSet<String> vis = ObjectGetter.toSet(names);

        HashSet<String> invis = ObjectGetter.filterToSet(header.getHeaderNames(), (s) -> !vis.contains(s));

        for (String iv : invis)
        {
            header.findHeader(iv).setVisible(false);
        }
        for (String s : vis)
        {
            header.findHeader(s).setVisible(true);
        }
    }

    public void truncate(int size)
    {
        data = VectorUtils.slice(data, 0, size);
        o_header.data = VectorUtils.slice(o_header.data, 0, size);
    }

    public static void testbuff()
    {
        BufferPrintWriter bpw = new BufferPrintWriter();
        bpw.printf("name\tv1\tv2\tv3\n");
        for (int i=0; i<100; i++)
        {
            bpw.printf("%s\t%d\t%2.4f\t%2.2f\n", StringUtils.generateRandomString(10), (int)(Math.random()*1000), Math.random() * 20, Math.random()*100);
        }
        DataTable dt = DataTable.readFromCSVBuffer(new StringReader(bpw.getBuffer()), "\t");

        dt.writeCSV(new File("/tmp/test.tsv"));
    }


    
    Vector<FlagFilter> flagfilters = new Vector<>();
    
    public FlagFilter addFlagFilter(String header, String splitter, boolean include)
    {
        FlagFilter ff =  new FlagFilter(header, splitter, include);
        flagfilters.add(ff);
        return ff;
    }

    public static void main (String args[]) 
    {
        SimpleOptionParser cmd = new SimpleOptionParser("i", "f", "r", "test", "o", "xml", "merge", "f1", "f2", "fields", "testbuff");
        cmd.setFile("i", "merge");
        cmd.setOptional("f", "i", "o", "r", "xml", "merge", "f1", "f2", "fields");
        cmd.setSwitches("test", "xml", "testbuff");
        cmd.setRequiredSubParamSets(new String[]{"i"}, new String[]{"test"}, new String[]{"testbuff"});
        
        if (!cmd.parseParams(args,true,false, true, cmd))
            return;
        
        if (cmd.isSet("testbuff"))
        {
            testbuff();
            return;
        }
        if (cmd.isSet("test"))
        {
            test();
            return;
        }
        DataTable dt = new DataTable();
        dt.readDataFile(cmd.getFile("i"));
        System.out.println("read headers: "+dt.header.getHeaderNames()+" data : "+dt.getSize());

        if (cmd.isOptionSet("merge"))
        {
            DataTable dt2 = DataTable.readFile(cmd.getFile("merge"));
            HashMap<Object, ObjectGetter> lookup =  dt2.getLookup(cmd.getValue("f2"));
            String fields[] = FileUtils.getRecordSplitter(",").split(cmd.getValue("fields"));
            Vector input = dt.getFieldVector(cmd.getValue("f1"));
            System.out.println(input);

            for (String f : fields)
            {
                Vector data = new Vector();
                for (Object o : input)
                {
                    ObjectGetter og = lookup.get(o);
                    data.add((og == null) ? null : og.get(f));
                }
                System.out.printf("f: %s\n%s\n", f, data);

                dt.addData(f, new ReflectedObject(true), data);
                
            }
            dt.writeCSV(cmd.getFile("o"));
            return;
        }
        if (cmd.isSet("xml"))
        {
            System.out.println("was in xml section");
            Element e = new Element("test");
            dt.writeXML(e);
            System.out.println("created following xml element:");
            System.out.println(XMLUtils.prettyXML(e));

            DataTable test_dt = DataTable.readFromXML(e.getChild("datatable"));
            Element test_e = new Element("test2");
            test_dt.writeXML(test_e);
            System.out.println(XMLUtils.prettyXML(test_e));
            return;
        }
        if (cmd.isOptionSet("f"))
        {
            dt.quickFilter(cmd.getValue("f"), true);
            //java.net.URL(Encoder|Decoder)

            Vector<Double> v = ObjectGetter.convert(dt.getFieldVector("cond2", true));

            System.out.println("filtered ids: "+ v);
            System.out.println("filtered ids: "+ ObjectGetter.map(v, (Double d) -> d+ 10));
            
            System.out.println("all: " +dt.getSize()+ "filtered: "+dt.getFilteredSize());
        }
        if (cmd.isOptionSet("r"))    
        {
            long t1 = System.currentTimeMillis();
            String fields[]  = cmd.getValue("r").split(",");
            DataTable dr = dt.restrict(fields);
            long t2 = System.currentTimeMillis();
            System.out.printf("restrict took: %2.2f sec\n", (t2-t1)/1000.0);
            HashSet<Integer> h = ObjectGetter.toSet(IteratorUtils.getIntIterator(0, dr.getSize(), 3));
            System.out.printf("check filter %d/%d\n", h.size(), dr.getSize());
            t1 = System.currentTimeMillis();
            DataTable fr = dt.filterByIdx(h);
            t2 = System.currentTimeMillis();
            System.out.printf("idx filter took %2.2f sec\n", (t2-t1)/1000.0);

            
            dr.writeCSV(cmd.getFile("o"));
            
        }
    }
}
