package lmu.utils;

import java.io.*;
import java.nio.charset.Charset;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.compress.compressors.CompressorException;
import org.apache.commons.compress.compressors.CompressorStreamFactory;


public class FileUtils
{

    public interface EntryWriter
    {
        public void write(PrintWriter pw);
    }

    public static class MultiSplitRecordIterator extends RecordIterator
    {
        char[] splitters = null;
        Vector<Integer> splits = new Vector<>();

        public MultiSplitRecordIterator(char... splitters)
        {
            this.splitters = splitters;
        }

        public void setLine(String line)
        {
            super.setLine(line);
            splits.clear();

            final int LN = (line == null) ? 0 : line.length();
            index = 0;

            if (LN == 0) return;

            for (char sp: splitters)
            {
                int start = -1;
                while (true)
                {
                    start = line.indexOf(sp, start + 1);
                    if (start < 0) break;
                    splits.add(start);
                }

            }
            splits.add(line.length());
            Collections.sort(splits);

        }

        public boolean next()
        {
            if (index >= splits.size()) return false;

            pos = next_pos + 1;
            next_pos = splits.get(index++);
            return true;
        }
    }

    final static String[] EMPTY_STRING_ARRAY = new String[0];

    public interface StringSplitter
    {
        public String[] split(String line);

    }
    

    public static StringSplitter getRecordSplitter(String sepstr)
    {
        return new RecordIterator(sepstr);
    }

    public static StringSplitter getTokenizer(String skipchars)
    {
        return new Tokenizer(skipchars);
    }

    public static class Tokenizer implements StringSplitter
    {
        HashSet<Character> skipper = new HashSet<>();

        Tokenizer(String skipchars)
        {
            for (int i = 0; i < skipchars.length(); i++)
            {
                skipper.add(skipchars.charAt(i));
            }
        }

        public String[] split(String line)
        {
            Vector<String> v = new Vector<>();
            StringBuffer sb = new StringBuffer();
            for (int i = 0; i <= line.length(); i++)
            {
                boolean toskip = (i == line.length() || skipper.contains(line.charAt(i)));
                if (toskip)
                {
                    if (sb.length() > 0)
                    {
                        v.add(sb.toString());
                    }
                    sb.setLength(0);
                    continue;
                }

                sb.append(line.charAt(i));

            }

            return v.toArray(new String[0]);
        }
    }

    public static class RecordIterator implements StringSplitter
    {
        String line = null;
        int pos = 0;
        char sep = '\t';
        String sepstr = null;
        int index = -1;
        int next_pos = -1;
        String record = null;

        public RecordIterator()
        {
            this('\t');
        }

        public RecordIterator(String sepstr)
        {
            this.sepstr = sepstr;
        }

        public RecordIterator(char sep)
        {
            setSeparator(sep);
        }

        public void setLine(String line)
        {
            this.line = line;
            index = -1;
            pos = 0;
            record = null;
            next_pos = -1;
        }

        public Vector<String> getFieldVector()
        {
            Vector<String> v = new Vector<>();
            while (next())
            {
                v.add(getRecord());
            }
            return v;
        }

        public String[] getFields()
        {
            return getFieldVector().toArray(EMPTY_STRING_ARRAY);
        }

        public static String[] split(String splitter, String line)
        {
            return (new RecordIterator(splitter)).split(line);
        }

        public static String[] split(char splitter, String line)
        {
            return (new RecordIterator(splitter)).split(line);
        }

        public String[] split(String line)
        {
            setLine(line);
            return getFields();

        }

        public Vector<String> splitVector(String line)
        {
            setLine(line);
            return getFieldVector();

        }

        public void setSeparator(char sep)
        {
            this.sep = sep;
        }

        public int getRecordIndex()
        {
            return index;
        }

        public double getDoubleRecord()
        {
            return Double.parseDouble(getRecord());
        }

        public int getIntRecord()
        {
            return Integer.parseInt(getRecord());
        }

        public String getRecord()
        {
            return line.substring(pos, next_pos);
        }

        public boolean next()
        {
            pos = next_pos + 1;

            if (line == null) return false;

            if (pos >= line.length())
            {
                record = null;
                return false;
            }
            if (sepstr != null)
            {
                next_pos = line.indexOf(sepstr, pos);
            } else
            {
                next_pos = line.indexOf(sep, pos);
            }

            next_pos = (next_pos >= 0) ? next_pos : line.length();
            index++;
            return true;
        }

        public Vector<String> getRecords()
        {
            Vector<String> v = new Vector<>();

            while (next())
            {
                v.add(getRecord());
            }
            return v;
        }

        public HashSet<String> getRecordSet()
        {
            HashSet<String> h = new HashSet<>();

            while (next())
            {
                h.add(getRecord());
            }
            return h;
        }
    }

    public static String toContainingDir(File oFile)
    {
        // turns ./out.fq into /home/users/bla/data/out.fq
        String sAbsPath = oFile.getAbsolutePath();
        // creates /home/users/bla/data
        return sAbsPath.substring(0, sAbsPath.lastIndexOf(File.separator));
    }

    public static File toContainingDirFile(File oFile)
    {
        return new File(FileUtils.toContainingDir(oFile));
    }

    public static String getRelativePath(File base, File relative)
    {
        File common_root = relative;

        while (common_root.getParentFile() != null && !common_root.equals(base))
        {
            common_root = common_root.getParentFile();
        }
        if (common_root.getParentFile() == null)
        {return relative.getAbsolutePath(); }
        return relative.getAbsolutePath().substring(common_root.getAbsolutePath().length());

    }

    public static File[] getFiles(File dir, Pattern pattern)
    {

        return dir.listFiles(new lmu.utils.FileNamePatternFilter(pattern));
    }

    public static File[] getFiles(File dir, String pattern)
    {

        return dir.listFiles(new lmu.utils.FileNamePatternFilter(pattern));
    }

    public static HashSet<String> buildComments(char... comments)
    {
        HashSet<String> rv = new HashSet<String>();
        for (char ch: comments)
        {
            rv.add("" + ch);
        }
        return rv;
    }

    /**
     * expects a list of two tuples: header name, format
     * and returns the joined header and format str
     */
    static public Pair<String, String> getHeaderAndFormat(String sep, String... infos)
    {
        StringBuffer header = new StringBuffer();
        StringBuffer format = new StringBuffer();

        for (int i = 0; i < infos.length; i += 2)
        {
            if (header.length() > 0)
            {
                header.append(sep);
                format.append(sep);
            }
            header.append(infos[i]);
            String fmt = infos[i + 1];
            if (fmt.length() == 0 || fmt.charAt(0) != '%')
            {throw new RuntimeException("invalid format: " + fmt + " for header: " + infos[i]); }
            format.append(fmt);
        }
        header.append('\n');
        format.append('\n');

        return new Pair<String, String>(header.toString(), format.toString());
    }

    /**
     * expects a list of two tuples header name, format
     * and writes the header and returns the format string
     */
    static public String writeHeader(PrintWriter pw, String sep, String... infos)
    {
        if (pw == null)
        {return null; }
        Pair<String, String> h = getHeaderAndFormat(sep, infos);
        pw.print(h.getFirst());
        return h.getSecond();
    }

    static public PrintWriter getWriter(File f)
    {
        return getWriter(f, false);
    }

    static public PrintWriter getWriter(File f, boolean append)
    {
        return getWriter(f, false, false);
    }

    static public PrintWriter getWriter(File f, boolean append, boolean gzip)
    {
        if (f == null) return null;

        try
        {
            if (gzip)
            {
                if (append)
                {throw new RuntimeException("Cannot append to gz"); }

                return new PrintWriter(new BufferedWriter(new OutputStreamWriter(new GZIPOutputStream(new java.io.BufferedOutputStream(
                                                                                                          new FileOutputStream(f))))));

            }

            return (!append) ? new PrintWriter(f) : new PrintWriter(new BufferedWriter(new FileWriter(f, true)));
        } catch (FileNotFoundException e)
        {
            throw new RuntimeException("cannot create file " + f.getAbsolutePath() + " for writing, path non existing!", e);
        } catch (IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while creating writer for %s", ie, ie.getMessage(), f.getAbsolutePath());
        }

    }

    static public PrintWriter getWriter(String format, Object... args)
    {
        return getWriter(null, format, args);
    }

    static public PrintWriter getWriter(File parent, String format, Object... args)
    {
        String fname = String.format(format, args);
        return getWriter((parent == null) ? new File(fname) : new File(parent, fname));

    }

    public static Vector<String[] > readMultiFieldCol(File f, String main_separator, int col, String second_separator)
    {
        Pattern sp = Pattern.compile(second_separator);
        Vector<String[] > v1 = readCols(f, main_separator, col);
        Vector<String[] > rv = new Vector<String[] >();
        for (String[] s: v1)
        {
            if (s[0] == null)
            {
                rv.add(null);
                continue;
            }
            rv.add(sp.split(s[0]));
        }
        return rv;
    }

    public static Vector readConvertedCol(File f, String separator, int col, StringConverter cc)
    {
        Vector<String[] > v = readCols(f, separator, col);
        Vector rv = new Vector();
        for (String[] s: v)
        {
            if (s[0] == null)
            {
                rv.add(null);
                continue;
            }
            rv.add(cc.convert(s[0]));
        }
        return rv;
    }

    public static Vector readSimpleConvertedCol(File f, String separator, int col, String typename)
    {
        return readConvertedCol(f, separator, col, SimpleTypeConverter.getConverter(typename, false));
    }

    public static Vector<Integer> readIntCol(File f, String separator, int col)
    {
        return (Vector<Integer>)readSimpleConvertedCol(f, separator, col, "int");
    }

    public static Vector<Double> readDoubleCol(File f, String separator, int col)
    {
        return (Vector<Double>)readSimpleConvertedCol(f, separator, col, "double");
    }

    public static Vector<Float> readFloatCol(File f, String separator, int col)
    {
        return (Vector<Float>)readSimpleConvertedCol(f, separator, col, "float");
    }

    public static Vector<Long> readLongCol(File f, String separator, int col)
    {
        return (Vector<Long>)readSimpleConvertedCol(f, separator, col, "long");
    }

    public static Vector<String> readCol(File f, String separator, int col)
    {
        Vector<String[] > v1 = readCols(f, separator, col);
        Vector<String> rv = new Vector<String>();
        for (String[] s: v1)
        {
            rv.add(s[0]);
        }
        return rv;
    }

    public static HashMap readMap(File f, String separator, int c1, int c2, String key_conv_name, String value_conv_name)
    {
        return readMap(f, separator, c1, c2, SimpleTypeConverter.getConverter(key_conv_name, true),
                       SimpleTypeConverter.getConverter(value_conv_name, true));
    }

    public static HashMap readMap(File f, String separator, int c1, int c2, StringConverter key_converter, StringConverter value_converter)
    {
        Vector<String[] > v = readCols(f, separator, c1, c2);
        HashMap rv = new HashMap();
        for (String[] sp: v)
        {
            System.out.println("will convert: " + sp[0] + ":" + sp[1]);
            Object k = key_converter.convert(sp[0]);
            Object val = value_converter.convert(sp[1]);
            rv.put(k, val);
        }
        return rv;
    }

    /**
     * @param f
     * @param separator
     * @param c1
     * @param c2
     * @param key_conv_name
     * @param value_conv_name
     * @param hasTitle
     * @return
     */
    public static HashMap readMap(File f, String separator, int c1, int c2, String key_conv_name, String value_conv_name, boolean hasTitle)
    {
        return readMap(f, separator, c1, c2, SimpleTypeConverter.getConverter(key_conv_name, true),
                       SimpleTypeConverter.getConverter(value_conv_name, true), hasTitle);
    }

    public static HashMap readMap(File f, String separator, int c1, int c2, StringConverter key_converter, StringConverter value_converter,
                                  boolean hasTitle)
    {
        Vector<String[] > v = readCols(hasTitle, f, separator, c1, c2);
        HashMap rv = new HashMap();
        for (String[] sp: v)
        {
// System.out.println("will convert: " + sp[0] + ":" + sp[1]);
            Object k = key_converter.convert(sp[0]);
            Object val = value_converter.convert(sp[1]);
            rv.put(k, val);
        }
        return rv;
    }

    public interface EntryConverter<A>
    {
        public String convert(A a);
    }

    static public <A> void writeMap(File f, String separator, HashMap<String, A> m)
    {
        writeMap(f, separator, m, (A a) -> "" + a);
    }

    static public <A> void writeMap(File f, String separator, HashMap<String, A> m, EntryConverter<A> sc)
    {
        writeMap(f, separator, m, (String s) -> s, sc);
    }

    static public <A, B> void writeMap(File f, String separator, HashMap<A, B> m, EntryConverter<A> sc, EntryConverter<B> sc2)
    {
        PrintWriter pw = getWriter(f);
        ObjectGetter.apply(m.entrySet(), (Map.Entry<A, B> e) -> pw.printf("%s%s%s\n", sc.convert(e.getKey()), separator, sc2.convert(e.getValue())));
        pw.close();
    }

    /*
     * public static void writeMap(File f, String separator,
     * HashMap<String,String> m)
     * {
     * PrintWriter pw = getWriter(f);
     * ObjectGetter.apply(m.entrySet(), (Map.Entry<String, String> e) ->
     * pw.printf("%s%s%s\n", e.getKey(), separator, e.getValue()));
     * pw.close();
     * }
     */

    public static HashMap<String, String> readMap(File f, String separator, int c1, int c2)
    {
        HashMap<String, Vector<String>> multimap = readMultiMap(f, separator, c1, c2);
        HashMap<String, String> rv = new HashMap<String, String>();
        for (Map.Entry<String, Vector<String>> e: multimap.entrySet())
        {
            if (e.getValue().size() > 1)
            {throw new RuntimeException("map is not well defined in " + f.getAbsolutePath() + " got values: "
                                        + e.getValue() + " for " + e.getKey()); }
            rv.put(e.getKey(), e.getValue().get(0));
        }
        return rv;
    }

    public static HashMap<String, Vector<String>> readMultiMap(File f, String separator, int c1, int c2)
    {
        StringConverter def = SimpleTypeConverter.getConverter("string", false);
        Vector<Pair> v = readPairs(f, separator, c1, c2, def, def);
        return (HashMap<String, Vector<String>>)ObjectGetter.getMultiLookup(v, "getFirst()", "getSecond()");
    }

    /**
     * @author wolf
     * @param f
     * @param separator
     * @param col1
     * @param col2
     * @param second_seperator
     * @return works only if seperator 1 and 2 are different!!!
     */
    public static HashMap<String, String[] > readMultiMapMultiFieldCol(File f, String separator, int c1, int c2, String second_separator)
    {
        StringConverter def = SimpleTypeConverter.getConverter("string", false);
        Vector<Pair> v = readPairs(f, separator, c1, c2, def, def);
        HashMap<String, Vector<String>> result = (HashMap<String, Vector<String>>)ObjectGetter.getMultiLookup(v, "getFirst()", "getSecond()");
        HashMap<String, String[] > return_result = new HashMap<String, String[] >();
        Pattern sp = Pattern.compile(second_separator);
        for (Iterator<String> it_string = result.keySet().iterator(); it_string.hasNext();)
        {
            String key = it_string.next();

            Vector<String> secondPart = result.get(key);

            return_result.put(key, sp.split(secondPart.get(0)));
        }
        return return_result;
    }

    // wolf
    /**
     * for one seperator it creates a hash with the first column as key and the
     * others as String[] value - no check if key is unqiue
     */
    public static HashMap readMap(File f, String separator, int c1, boolean hasTitle, Integer... c2)
    {

        int[] all_cols = new int[c2.length + 1];
        all_cols[0] = c1;
        for (int i = 0; i < c2.length; i++)
        {
            all_cols[i + 1] = c2[i];
        }
        Vector<String[] > v = readCols(hasTitle, f, separator, all_cols);
        HashMap rv = new HashMap();
        for (String[] sp: v)
        {
            // System.out.println("will convert: " + sp[0] + ":" + sp[1]);
            Object k = sp[0];
            Object[] vals = new Object[sp.length - 1];
            for (int j = 1; j < sp.length; j++)
            {
                Object val = sp[1];
                vals[j - 1] = val;
            }
            rv.put(k, vals);
        }
        return rv;
    }

    static StringConverter DEFAULT_CONVERTER = SimpleTypeConverter.getConverter("string", false);

    public static Vector<File> readPaths(File f)
    {
        return ObjectGetter.map(readSet(f), (String s) -> new File(s));
    }

    public static HashSet<String> readSet(File f)
    {
        return readSet(f, "\\s+", 0);
    }

    public static HashSet<String> readSet(File f, String separator, int col)
    {
        return (HashSet<String>)readSet(f, separator, col, (StringConverter)null);
    }

    public static HashSet readSet(File f, String separator, int col, String typename)
    {
        return readSet(f, separator, col, SimpleTypeConverter.getConverter(typename, true));
    }

    public static HashSet readSet(File f, String separator, int col, StringConverter sc)
    {
        Vector<String[] > v1 = readCols(f, separator, col);
        HashSet rv = new HashSet();
        for (String[] sp: v1)
        {
            if (sc == null)
            {
                rv.add(sp[0]);
                continue;
            }
            rv.add(sc.convert(sp[0]));
        }
        return rv;
    }

    public static Vector<Pair> readPairs(File f, String separator, int c1, int c2)
    {
        return readPairs(f, separator, c1, c2, DEFAULT_CONVERTER, DEFAULT_CONVERTER);
    }

    public static Vector<Pair> readPairs(File f, String separator, int c1, int c2, StringConverter sc1, StringConverter sc2)
    {
        Vector<String[] > v = readCols(f, separator, c1, c2);
        Vector<Pair> rv = new Vector<Pair>();
        for (String[] s: v)
        {

            rv.add(new Pair(sc1.convert(s[0]), sc2.convert(s[1])));
        }
        return rv;
    }

    public static Vector<Object[] > readConvertedCols(File f, String separator, Object... cols)
    {
        return readConvertedCols(f, null, separator, cols);
    }

    public static Vector<Object[] > readConvertedCols(File f, HashSet<String> line_comments, String separator, Object... cols)
    {
        Vector<Integer> colvec = new Vector<Integer>();
        Vector<StringConverter> converters = new Vector<StringConverter>();

        for (int i = 0; i < cols.length; i += 2)
        {
            colvec.add((Integer)cols[i]);
            Object o = cols[i + 1];
            if (o instanceof StringConverter)
            {
                converters.add((StringConverter)o);
                continue;
            }
            if (o instanceof String)
            {
                converters.add(lmu.utils.SimpleTypeConverter.getConverter((String)o, true));
                continue;
            }

            throw new RuntimeException("don't know how to use this as a converteR: " + o);
        }
        int[] coa = new int[colvec.size()];
        for (int i = 0; i < colvec.size(); i++)
        {
            coa[i] = colvec.get(i);
        }

        Vector<String[] > data = readCols(f, line_comments, separator, coa);

        Vector<Object[] > rv = new Vector<>();
        for (int i = 0; i < data.size(); i++)
        {
            try
            {

                String[] sp = data.get(i);
                Object[] d = new Object[sp.length];
                for (int j = 0; j < sp.length; j++)
                {
                    d[j] = converters.get(j).convert(sp[j]);
                }
                rv.add(d);
            } catch (Exception e)
            {

                throw new RuntimeException("error:" + e.getMessage() + " at data: " + i + " " + StringUtils.join(" ", data.get(i))
                                           + "  et reading file: " + f.getAbsolutePath(), e);
            }
        }
        return rv;
    }

    public static int getNumLines(File f)
    {
        Integer ln = 0;
        Iterator<String> it = getLineIterator(f);
        while (it.hasNext())
        {
            ln++;
            it.next();
        }
        return ln;
    }

    public static LineIterator getLineIterator(File f)
    {
        return new LineIterator(f);
    }

    public static LineIterator getLineIterator(BufferedReader reader)
    {
        return new LineIterator(reader);
    }

    public static Vector<String> readNLines(File f, int n)
    {
        Vector<String> v = new Vector<>();
        try
        {
            BufferedReader bfr = new BufferedReader(new FileReader(f));
            String l = null;
            for (int i=0; i< n && null != (l = bfr.readLine()); i++, v.add(l));
        }
        catch (IOException ie)
        {
            throw new FRuntimeException("error: %s at reading %d lines from %s", ie, ie.getMessage(), n, f.getAbsolutePath());
        }
        return v;
    }
    public static Iterator<String[] > getNLineIterator(File f, int ln)
    {
        return new NLineIterator(f, ln);
    }

    public static Vector<String> getLines(File f)
    {
        Iterator<String> it = getLineIterator(f);
        Vector<String> v = new Vector<>();

        while (it.hasNext())
        {
            v.add(it.next());
        }
        return v;
    }

// wolf
    public static Vector<String> getGrepLines(File f, String pattern, boolean case_sensitive, boolean wholeWord)
    {
        Iterator<String> it = getLineIterator(f);
        Vector<String> v = new Vector<>();

        while (it.hasNext())
        {
            String line = it.next();
            if (!case_sensitive)
            {
                line = line.toLowerCase();
                pattern = pattern.toLowerCase();
            }

            if (line.indexOf(pattern) != -1)
            {
                if (wholeWord)
                {
                    Pattern wholeWordpattern = Pattern.compile("\\b" + pattern + "\\b");
                    Matcher m = wholeWordpattern.matcher(line);
                    if (m.find())
                    {

                        v.add(line);
                    }
                } else
                {
                    v.add(line);
                }
            }
        }
        return v;
    }

    // wolf
    public static Vector<String> getGrepLines(File f, boolean case_sensitive, Vector<String> patterns, boolean wholeWord)
    {
        Iterator<String> it = getLineIterator(f);
        Vector<String> v = new Vector<>();

        while (it.hasNext())
        {
            String line = it.next();
            if (!case_sensitive)
            {
                line = line.toLowerCase();
            }
            for (String p: patterns)
            {
                String pat = p;
                if (!case_sensitive)
                {
                    pat = p.toLowerCase();
                }

                if (line.indexOf(pat) != -1)
                {
                    if (wholeWord)
                    {
                        Pattern wholeWordpattern = Pattern.compile("\\b" + pat + "\\b");
                        Matcher m = wholeWordpattern.matcher(line);
                        if (m.find())
                        {

                            v.add(line);
                        }
                    } else
                    {
                        v.add(line);
                    }
                }
            }
        }
        return v;
    }

    // wolf
    public static HashMap<Integer, String> getNumberedLines(File f)
    {
        Iterator<String> it = getLineIterator(f);

        int i = 0;
        HashMap<Integer, String> num2line = new HashMap<Integer, String>();
        while (it.hasNext())
        {
            String line = it.next();

            num2line.put(i, line);

            i++;
        }
        return num2line;
    }

    public static Vector<String[] > readCols(File f, String separator, int... cols)
    {
        return readCols(f, null, separator, cols);
    }

    public static Vector<String[] > readCols(boolean hasTitle, File f, String separator, int... cols)
    {
        return readCols(hasTitle, f, null, separator, cols);
    }

    /**
     * reads the requested cols from file - if a line containing less cols, the
     * method will return null
     * to the corresponding field
     */
    public static Vector<String[] > readCols(File f, HashSet<String> comments, String separator, int... cols)
    {
        return readCols(getFieldSetIterator(f, comments, separator), comments, separator, cols);
    }

    /**
     * reads the requested cols from file - if a line containing less cols, the
     * method will return null
     * to the corresponding field
     */
    public static Vector<String[] > readCols(boolean hasTitle, File f, HashSet<String> comments, String separator, int... cols)
    {
        Iterator<String[] > it = getFieldSetIterator(f, comments, separator);
        return readCols(hasTitle, it, comments, separator, cols);
    }

    public static Vector<String[] > readCols(boolean hasTitle, Iterator<String[] > it, HashSet<String> comments, String separator, int... cols)
    {
        Vector<String[] > rv = new Vector<String[] >();
        if (hasTitle)
        {
            it.next();
        }
        while (it.hasNext())
        {
            String[] sp = it.next();
            String[] nf = new String[cols.length];

            for (int i = 0; i < cols.length; i++)
            {

                int ci = cols[i];
                if (ci >= sp.length)
                {
                    nf[i] = null;
                    continue;
                }
                nf[i] = sp[cols[i]];
            }
            rv.add(nf);
        }
        return rv;
    }

    public static Vector<String[] > readCols(Iterator<String[] > it, HashSet<String> comments, String separator, int... cols)
    {
        Vector<String[] > rv = new Vector<String[] >();
        while (it.hasNext())
        {
            String[] sp = it.next();
            String[] nf = new String[cols.length];

            for (int i = 0; i < cols.length; i++)
            {

                int ci = cols[i];
                if (ci >= sp.length)
                {
                    nf[i] = null;
                    continue;
                }
                nf[i] = sp[cols[i]];
            }
            rv.add(nf);
        }
        return rv;
    }

    public static Vector<String[] > readCols(File f, HashSet<String> comments, String separator, Collection<String> headerNames)
    {
        Iterator<String[] > it = getFieldSetIterator(f, comments, separator);
        int[] cols = new int[headerNames.size()];
        if (it.hasNext())
        {
            String[] sp = it.next();
            Iterator<String> head_it = headerNames.iterator();
            int h = 0;
            while (head_it.hasNext())
            {
                String header = head_it.next();
                boolean foundHeader = false;
                for (int i = 0; i < sp.length; i++)
                {
                    if (sp[i].equals(header))
                    {
                        foundHeader = true;
                        cols[h] = i;
                        break;
                    }
                }
                if (!foundHeader)
                {
                    ExceptionManager.throwRuntimeException("unable to match header name " + header);
                }
                h++;
            }
        }
        return readCols(it, comments, separator, cols);
    }

    public static Iterator<String[] > getFieldSetIterator(Reader r, String pat)
    {
        return getFieldSetIterator(r, null, pat);

    }

    public static Iterator<String[] > getFieldSetIterator(Reader r, HashSet<String> comments, RecordIterator rit)
    {
        return new FileFieldSetIterator(new File("unknown"), comments, new BufferedReader(r), null, rit);
    }

    public static Iterator<String[] > getFieldSetIterator(Reader r, HashSet<String> comments, String pat)
    {
        RecordIterator rit = null;
        String sep = null;
        if (pat.length() == 1)
        {
            rit = new RecordIterator(pat.charAt(0));
            sep = pat;
        }
        return new FileFieldSetIterator(new File("unknown"), comments, new BufferedReader(r), pat, rit);

    }

    public static Vector<File> getRecursive(File dir)
    {
        Vector<File> rv = new Vector<File>();
        LinkedList<File> q = new LinkedList<File>();
        q.add(dir);
        while (!q.isEmpty())
        {
            File f = q.poll();
            if (f.isFile())
            {
                rv.add(f);
                continue;
            }
            for (String sf: f.list())
            {
                q.add(new File(f, sf));

            }

        }
        return rv;
    }

    public static class FileLookup
    {
        HashMap<String, File> lookup = new HashMap<String, File>();

        public FileLookup(File d)
        {
            Vector<File> v = getRecursive(d);
            File t = null;
            for (File f: v)
            {
                if ((t = lookup.get(f.getName())) != null)
                {
                    System.err.println("name: " + f.getName() + " is multiple used in dir: " + d + " old: " + t.getAbsolutePath() + " new: "
                                       + f.getAbsolutePath());
                    continue;
                }
                lookup.put(f.getName(), f);
            }
            System.out.printf("%s found %d files\n", d.getName(), lookup.size());
        }

        public File get(String s)
        {
            return lookup.get(s);
        }
    }

    public static Iterator<File> getFilePathIterator(File f)
    {
        return getFilePathIterator(f, "\t", 0);
    }

    public static Iterator<File> getFilePathIterator(File f, String pat, int col)
    {
        return new FilePathIterator(getFieldSetIterator(f, pat), col);
    }

    public static Iterator<String[] > getFieldSetIterator(File f, String pat)
    {
        return new FileFieldSetIterator(f, pat);

    }

    public static Iterator<String[] > getFieldSetIterator(File f, HashSet<String> comments, RecordIterator rit)
    {
        return new FileFieldSetIterator(f, comments, rit);
    }

    public static Iterator<String[] > getFieldSetIterator(File f, HashSet<String> comments, String pat)
    {
        return new FileFieldSetIterator(f, comments, pat);

    }

    public static BufferedReader getReader(File f)
    {
        return getReader(f, java.nio.charset.Charset.defaultCharset());

    }

    public static BufferedReader getReader(File f, java.nio.charset.Charset cs)
    {
        return new BufferedReader(new InputStreamReader(getInputStream(f), cs));

    }

    public static void writeSet(PrintWriter pw, Set data)
    {
        for (Object o: data)
        {
            pw.println("" + o);
        }
    }

    public static void copy(File inF, File outF)
    {
        try
        {

            final long L = inF.length();
            final int BUFSIZE = 64 * 1024;
            byte[] data = new byte[BUFSIZE];

            RandomAccessFile raf = new RandomAccessFile(inF, "r");
            RandomAccessFile out = new RandomAccessFile(outF, "rw");

            for (long i = 0; i < L; i += BUFSIZE)
            {
                int read = raf.read(data, 0, BUFSIZE);
                out.write(data, 0, read);
            }

            raf.close();
            out.close();

            /*
             * FileOutputStream fos = new FileOutputStream(outF);
             * OutputStreamWriter osw = new OutputStreamWriter(fos);
             * BufferedReader bfr = new BufferedReader(new
             * InputStreamReader(getInputStream(inF)));
             * int chr=-1;
             * while (-1!=(chr=bfr.read()))
             * {
             * osw.write(chr);
             * }
             * bfr.close();
             * osw.close();
             */

        } catch (IOException e)
        {
            throw new RuntimeException("error: " + e.getMessage() + " while copy binary data from file: " + inF.getAbsolutePath() + " => "
                                       + outF.getAbsolutePath(), e);
        }
    }

    public static void binCopy(File in, File out)
    {
        copy(in, out);
        // writeBinaryToFile(getBinaryData(in), out);
    }

    public static File createTmpFile(File dir, String prefix, String suffix)
    {
        return createTmpFile(dir, prefix, suffix, false);
    }

    public static File createTmpFile(File dir, String prefix, String suffix, boolean delete_on_exit)
    {
        try
        {
            File f = File.createTempFile(prefix, suffix, dir);
            if (delete_on_exit)
            {
                f.deleteOnExit();
            }
            return f;
        } catch (IOException e)
        {
            throw new FRuntimeException("i/o error: %s while creating tmpfile in %s", e, e.getMessage(), dir);
        }
    }

    public static void writeString(String s, File f)
    {
        PrintWriter pw = getWriter(f);
        pw.print(s);
        pw.close();
    }

    public static void writeBinaryToFile(byte[] data, File f)
    {
        try
        {
            if (f.exists())
            {
                f.delete();
            }
            RandomAccessFile raf = new RandomAccessFile(f, "rw");
            raf.write(data);
            raf.close();
            // System.out.printf("wrote %d bytes to %s fsize: %d\n",
            // data.length, f.getName(), f.length());
        } catch (IOException e)
        {
            throw new RuntimeException("error: " + e.getMessage() + " while writing binary data to file: " + f.getAbsolutePath(), e);
        }
    }

    public static byte[] getURLBinary(String url)
    {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        try
        {
            java.net.URL u = new java.net.URL(url);
            InputStream is = u.openStream();
            DataInputStream dis = new DataInputStream(is);
            final int chunksize = 1024;
            byte[] chunk = new byte[chunksize];
            int nread = 0;

            while (0 <= (nread = dis.read(chunk)))
            {
                baos.write(chunk, 0, nread);
            }
            is.close();
            baos.close();

        } catch (java.net.MalformedURLException mue)
        {
            throw new RuntimeException("mailformed url: " + url);

        } catch (IOException ioe)
        {
            throw new RuntimeException("i/o error: " + ioe.getMessage() + " while reading from url: " + url, ioe);

        }
        return baos.toByteArray();
    }

    public static StringBuffer getURLBuffer(String url)
    {

        StringBuffer sb = new StringBuffer();
        try
        {
            java.net.URL u = new java.net.URL(url);
            InputStream is = u.openStream();
            BufferedReader bfr = new BufferedReader(new InputStreamReader(is));
            String line = null;

            while (null != (line = bfr.readLine()))
            {
                sb.append(line);
                sb.append('\n');
            }
            bfr.close();
        } catch (java.net.MalformedURLException mue)
        {
            throw new RuntimeException("mailformed url: " + url);

        } catch (IOException ioe)
        {
            throw new RuntimeException("i/o error: " + ioe.getMessage() + " while reading from url: " + url, ioe);

        }
        return sb;
    }

    public static StringBuffer getContentBuffer(byte[] data)
    {
        ByteArrayInputStream bais = new ByteArrayInputStream(data);
        StringBuffer sb = new StringBuffer();
        try
        {
            BufferedReader bfr = new BufferedReader(new InputStreamReader(bais));
            String line;
            while (null != (line = bfr.readLine()))
            {
                sb.append(line);
                sb.append('\n');
            }
            bfr.close();

        } catch (IOException e)
        {
            throw new RuntimeException("error: " + e.getMessage() + " while reading content from bytes (" + data.length + ")", e);
        }
        return sb;

    }

    public static StringBuffer getContentBuffer(Reader reader)
    {
        StringBuffer sb = new StringBuffer();
        try
        {
            BufferedReader bfr = new BufferedReader(reader);

            String line;
            while (null != (line = bfr.readLine()))
            {
                sb.append(line);
                sb.append('\n');
            }
            bfr.close();

        } catch (IOException e)
        {
            throw new RuntimeException("error: " + e.getMessage() + " while reading content", e);
        }
        return sb;
    }

    public static StringBuffer getContentBuffer(File f)
    {
        BufferedReader reader = FileUtils.getReader(f);
        return getContentBuffer(reader);

    }

    public static byte[] getBinaryData(File f)
    {
        try
        {

            byte[] data = new byte[(int)f.length()];
            RandomAccessFile raf = new RandomAccessFile(f, "r");
            raf.readFully(data);
            raf.close();
            return data;

        } catch (IOException e)
        {
            throw new RuntimeException("error: " + e.getMessage() + " while reading binary data from file: " + f.getAbsolutePath(), e);
        }
    }

    public static InputStream getInputStream(File f)
    {
        try
        {
            BufferedInputStream bis = new BufferedInputStream(new FileInputStream(f));
            
            try
            {
                return new java.util.zip.GZIPInputStream(bis);

            } catch (IOException e)
            { /* silent error */
                bis.close();
                //e.printStackTrace();
            }

            try
            {
                return new CompressorStreamFactory().createCompressorInputStream(new BufferedInputStream(bis));
            } catch (CompressorException ce)
            { /* silent error */
                bis.close();
                //ce.printStackTrace();
            }

            return new BufferedInputStream(new FileInputStream(f));

        } catch (IOException ie)
        {
            throw new RuntimeException("error:" + ie.getMessage() + " while getting input stream from " + f.getAbsolutePath(), ie);
        }

    }

    public static class FilePathIterator implements Iterator<File>
    {
        Iterator<String[] > ffsi;
        int col = 0;

        public FilePathIterator(Iterator<String[] > ffsi, int col)
        {
            this.ffsi = ffsi;
            this.col = col;
        }

        public boolean hasNext()
        {
            return ffsi.hasNext();
        }

        public void remove()
        {
            ffsi.remove();
        }

        public File next()
        {
            String[] sp = ffsi.next();
            return new File(sp[col]);
        }
    }

    public static class NLineIterator implements Iterator<String[] >
    {

        int nlines;
        String[] data;
        LineIterator lit;

        public NLineIterator(File f, int n)
        {

            lit = new LineIterator(f);
            nlines = n;
            data = new String[n];
        }

        public boolean hasNext()
        {
            return lit.hasNext();
        }

        public String[] next()
        {
            for (int i = 0; i < nlines; i++)
            {
                data[i] = lit.next();
            }
            return data;
        }

    }

    public static class LineIterator implements Iterator<String>
    {
        FileFieldSetIterator fake;

        public LineIterator(File f)
        {
            fake = new FileFieldSetIterator(f, null, (String)null);
        }

        public LineIterator(BufferedReader bfr)
        {
            fake = new FileFieldSetIterator(null, null, bfr, (String)null, null);
        }

        public boolean hasNext()
        {
            return fake.hasNext();
        }

        public String next()
        {
            String[] n = fake.next();
            if (n == null) return null;
            return n[0];
        }

        public void remove()
        {
            fake.remove();
        }

        public void close()
        {
            fake.close();
        }
    }

    public static class FileFieldSetIterator implements Iterator<String[] >
    {
        File f;
        BufferedReader bfr;
        RecordIterator rit = null;
        java.util.regex.Pattern pattern;
        HashSet<String> comments = null;
        String next[];
        String[] dummy = new String[1];
        Vector<String> records = new Vector<>();

        int ln = 0;
        String line;

        public void close()
        {
            if (bfr == null) return;

            try
            {

                System.err.printf("%s closed\n", f);
                bfr.close();
                bfr = null;
            } catch (IOException e)
            {
                throw new RuntimeException("error:" + e.getMessage() + " while closing file: " + f.getAbsolutePath(), e);
            }
        }

        public FileFieldSetIterator(File f, String pat)
        {
            this(f, null, pat);
        }

        public FileFieldSetIterator(File f, RecordIterator rit)
        {
            this(f, null, rit);
        }

        public FileFieldSetIterator(File f, HashSet<String> comments, String pat)
        {
            this(f, comments, new BufferedReader(new InputStreamReader(getInputStream(f))), pat, null);

        }

        public FileFieldSetIterator(File f, HashSet<String> comments, RecordIterator rit)
        {
            this(f, comments, new BufferedReader(new InputStreamReader(getInputStream(f))), null, rit);

        }

        public FileFieldSetIterator(File f, HashSet<String> comments, BufferedReader bfr, String pat, RecordIterator rit)
        {
            this.f = f;
// System.err.printf("will open: %s\n", f);
            this.comments = comments;
            if (rit != null)
            {
                this.rit = rit;
            } else
            {
                pattern = (pat == null) ? null : java.util.regex.Pattern.compile(pat);
            }

            this.bfr = bfr;
            readNext();
        }

        void readNext()
        {
            if (bfr == null) return;

            next = null;
            try
            {
                while (next == null)
                {

                    line = bfr.readLine();
                    if (line == null)
                    {
                        bfr.close();
// System.err.printf("%s closed\n", f);
                        bfr = null;
                        return;
                    }
                    ln++;
                    if (comments != null)
                    {
                        boolean is_comment = false;
                        for (String c: comments)
                        {
                            if (line.startsWith(c))
                            {
                                is_comment = true;
                                break;
                            }
                        }
                        if (is_comment)
                        {
                            continue;
                        }
                    }
                    if (rit != null)
                    {
                        next = rit.split(line);
                        continue;

                    }
                    if (pattern == null)
                    {
                        dummy[0] = line;
                        next = dummy;
                        continue;
                    }
                    next = pattern.split(line);
                }
            } catch (IOException e)
            {
                throw new RuntimeException("I/O error: " + e.getMessage() + " at line: " + ln + " from file " + f.getAbsolutePath(), e);
            }

        }

        public boolean hasNext()
        {
            return next != null;
        }

        public String[] next()
        {
            String[] rv = next.clone();
            readNext();
            return rv;
        }

        public void remove()
        {
            throw new RuntimeException("not implemented!");
        }
    }

    public static String getPartialPath(File f, int start)
    {
        return getPartialPath(f, start, null);
    }

    public static String getPartialPath(File f, int start, Integer end)
    {

        String sp[] = f.getAbsolutePath().split(f.separator);
        return StringUtils.join(f.separator, StringUtils.slice(sp, start, (end == null) ? sp.length : end));
    }

    public static class Converter
    {
        String[] sp;

        public Converter set(String[] sp)
        {
            this.sp = sp;
            return this;
        }

        public int length()
        {
            return sp.length;
        }

        public String toString(int idx)
        {
            if (sp.length <= idx) return null;

            return sp[idx];
        }

        public int toInt(int idx)
        {
            return toInt(idx, null);
        }

        public int toInt(int idx, Integer default_value)
        {
            try
            {
                return Integer.parseInt(sp[idx]);
            } 
            catch (NumberFormatException nfe)
            {
                if (default_value != null) 
                    return default_value;

                throw nfe;
            }
        }


        public long toLong(int idx)
        {
            return toLong(idx, null);
        }

        public long toLong(int idx, Long default_value)
        {
            try
            {
                return Long.parseLong(sp[idx]);
            } 
            catch (NumberFormatException nfe)
            {
                if (default_value != null) 
                    return default_value;

                throw nfe;
            }
        }

        public boolean toBoolean(int idx)
        {
            return toBoolean(idx, null);
        }

        public boolean toBoolean(int idx, Boolean default_value)
        {
            try
            {
                return Boolean.parseBoolean(sp[idx]);
            } 
            catch (NumberFormatException nfe)
            {
                if (default_value != null) 
                    return default_value;

                throw nfe;
            }
        }



        public static double toDbl(String[] sp, int idx)
        {
            return SimpleTypeConverter.toDouble(sp[idx]);
            //return Double.parseDouble(sp[idx]);
        }

        public double toDbl(int idx)
        {
            return SimpleTypeConverter.toDouble(sp[idx]);
            //return Double.parseDouble(sp[idx]);
        }

    }

    public static Converter converter = new Converter();

    public static class HeaderedReader
    {
        Vector<UPair<String>> name2alias = new Vector<>();

        FileUtils.StringSplitter splitter = null;
        public FileUtils.Converter conv = new FileUtils.Converter();
        HashMap<String, Integer> n2info = new HashMap<>();

        String src = null;

        public HeaderedReader(String split)
        {
            if (split == null) return;

            splitter = FileUtils.getRecordSplitter(split);
        }

        public HeaderedReader setSrc(String src)
        {
            this.src = src;
            return this;
        }

        public HeaderedReader add(String name)
        {
            return add(name, name);
        }

        public HeaderedReader add(String name, String alias)
        {
            name2alias.add(UPair.createU(name, alias));

            return this;
        }

        public HashMap<String, String> getMap()
        {
            return ObjectGetter.buildMap(n2info.keySet(), (s) -> getString(s));
        }

        public void init(String line)
        {
            if (splitter == null) throw new FRuntimeException("no splitter configured, cannot split header for %s", src);

            init(splitter.split(line));
        }

        public HeaderedReader add(int pos, String alias)
        {
            n2info.put(alias, pos);
            return this;
        }

        public void init(String[] header)
        {
            for (UPair<String> u: name2alias)
            {
                int i = StringUtils.getIndex(header, u.getFirst());
                if (i < 0) throw new FRuntimeException("header: >%s< not found in %s!", u.getFirst(), src);

                n2info.put(u.getSecond(), i);
            }
        }

        public HeaderedReader setRecord(String line)
        {
            if (splitter == null) throw new FRuntimeException("no splitter configured, cannot split line for %s", src);

            return setRecord(splitter.split(line));
        }

        public HeaderedReader setRecord(String[] record)
        {
            conv.set(record);
            return this;
        }

        int getIdx(String alias)
        {
            Integer idx = n2info.get(alias);
            if (idx == null) 
                throw new FRuntimeException("unknown alias '%s' for %s. known aliases: %s n2infos: %s", alias, src, 
                                            ObjectGetter.map(name2alias, (up) -> up.getSecond()), n2info.keySet());

            return idx;
        }

        public int getInt(String alias)
        {
            return getInt(alias, null);
        }

        public int getInt(String alias, Integer def_value)
        {
            return conv.toInt(getIdx(alias), def_value);
        }

        public double getDbl(String alias)
        {
            return conv.toDbl(getIdx(alias));
        }

        public String getString(String alias)
        {
            return conv.toString(getIdx(alias));
        }

    }

    public static HeaderedReader getHeaderedReader(String split, File f)
    {
        return getHeaderedReader(split, f.getAbsolutePath());
    }

    public static HeaderedReader getHeaderedReader(String split)
    {
        return getHeaderedReader(split, (String)null);
    }

    public static HeaderedReader getHeaderedReader(String split, String src)
    {
        HeaderedReader hr = new HeaderedReader(split);
        if (src != null)
        {
            hr.setSrc(src);
        }

        return hr;
    }

    /*
     * public static void main (String args[])
     * {
     * copy(new File(args[0]), new File(args[1]));
     * }
     */

}
