package lmu.utils;

import java.util.*;

public class IteratorUtils
{
    
    public interface ChildGetter<A>
    {
        public Collection<A> getChildren(A e);
    }
    
    public interface Filter<A>
    {
        public boolean filter(A a);
    }

    public static class FilterIterator<A> implements Iterator<A>
    {

        PeakableIterator<A> it;
        Filter<A> filter;
        A n = null;

        public FilterIterator(Iterator<A> it, Filter<A> filter)
        {
            this.it = new PeakableIterator<A>(it);
            this.filter = filter;
            readNext();
        }

        public boolean hasNext()
        {
            return it.hasNext();
        }

        void readNext()
        {
            while(it.hasNext() && !filter.filter(it.peak()))
            {
                it.next();
            }

        }

        public A next()
        {
            A rv = it.next();
            readNext();
            return rv;
        }
    }


    static public<A> boolean containsFiltered(Iterator<A> it, Filter<A> f)
    {
        Iterator<A> fit = new FilterIterator<A>(it, f);
        return fit.hasNext();
    }
    
    static class TreeIterator<A> implements Iterator<A>
    {
        LinkedList<A> q = new LinkedList<>();
        ChildGetter<A> ch;
        
        public TreeIterator(Iterator<A> it, ChildGetter<A> ch)
        {
            ObjectGetter.addAll(q, it);
            this.ch = ch;
        }
        public TreeIterator(Collection<A> r, ChildGetter<A> ch)
        {
            q.addAll(r);
            this.ch = ch;
        }
        public TreeIterator(A r, ChildGetter<A> ch)
        {
            q.add(r);
            this.ch = ch;
        }

        public boolean hasNext()
        {
            return !q.isEmpty();
        }

        public A next()
        {
            if (!hasNext())
                return null;
            
            A rv = q.poll();
            Collection<A> c;
            if (null  != (c = ch.getChildren(rv)))
            {
                q.addAll(c);
            }
                
            
            return rv;
        }
    }

    public static<A> Iterator<A> filter(Iterator<A> it, Filter<A> f)
    {
        return new FilterIterator<A>(it, f);
    }
    
    public static<A> Iterator<A> getTreeIterator(A root, ChildGetter<A> ch)
    {
        return new TreeIterator<A>(root, ch);
    }


    public static<A> Iterator<A> getTreeIterator(Collection<A> roots, ChildGetter<A> ch)
    {
        return new TreeIterator<A>(roots, ch);
    }

    public static<A> Iterator<A> getTreeIterator(Iterator<A> roots, ChildGetter<A> ch)
    {
        return new TreeIterator<A>(roots, ch);
    }

        
    public static<A> Iterator<A> getDAGIterator(A root, ChildGetter<A> ch)
    {
        HashSet<A> visited = new HashSet<>();
        visited.add(root);
        return new TreeIterator<A>(root, (A a) -> { Set<A> v = ObjectGetter.filterToSet(ch.getChildren(a), (A a2) -> !visited.contains(a2));  visited.addAll(v); return (Collection<A>)v; });
    }


    
    public static<A> Iterator<A> getDAGIterator(Collection<A> roots, ChildGetter<A> ch)
    {
        HashSet<A> visited = new HashSet<>();
        visited.addAll(roots);
        return new TreeIterator<A>(roots, (A a) -> { Set<A> v = ObjectGetter.filterToSet(ch.getChildren(a), (A a2) -> !visited.contains(a2));  visited.addAll(v); return (Collection<A>)v; });
    }


    public static<A> Vector<A> toVector(Iterator<A> it, int maxn)
    {
        Vector<A> rv = new Vector<>();
        for (int i=0; (maxn < 0 || i<maxn) && it.hasNext(); i++)
        {
            rv.add(it.next());
        }
        return rv;
    }

    static class IIterator implements Iterator<Integer>
    {
        int start = 0;
        int end = 0;
        int idx = 0;
        int step = 1;

        public IIterator(int s, int e)
        {
            this(s,e, 1);
        }
        public IIterator(int s, int e, int step)
        {
            idx = start = s; end = e;
            this.step = step;
        }

        public boolean hasNext()
        {
            return idx != end;
        }

        public Integer next()
        {
            int rv = idx;
            idx += step;;
            return rv;
        }

    }
    
    static class DIterator implements Iterator<Double>
    {
        double start = 0;
        double end = 0;
        double idx = 0;
        double step = 1;

        public DIterator(double s, double e)
        {
            this(s,e, 1);
        }
        public DIterator(double s, double e, double step)
        {
            idx = start = s; end = e;
            this.step = step;
        }

        public boolean hasNext()
        {
            return idx < end;
        }

        public Double next()
        {
            double rv = idx;
            idx+=step;;
            return rv;
        }

    }
    
    static public<A> Vector<A> getRange(Iterator<A> it, int from, int to)
    {
        Vector<A> v = new Vector<>();
        int idx = 0;
        while (it.hasNext())
        {
            A a = it.next();
            if (idx < from)
            {
                idx++;
                continue;
            }
            if (idx>to)
                break;
            
            v.add(a);
            idx++;
        }
        return v;
    }

    static public Iterator<Integer> range(Collection c)
    {
        return getIntIterator(c.size());
    }

    static public Vector<Integer> rangev(Collection c)
    {
        return ObjectGetter.toVector(getIntIterator(c.size()));
    }

    static public Vector<Integer> rangev(int length)
    {
        return ObjectGetter.toVector(getIntIterator(length));
    }
    
    static public Iterator<Integer> getIntIterator(int length)
    {
        return getIntIterator(0,length, 1);
    }
    static public Iterator<Integer> getIntIterator(int from, int to)
    {
        return getIntIterator(from,to, 1);
    }
    static public Iterator<Integer> getIntIterator(int from, int to, int step)
    {
        return new IIterator(from, to, step);
    }
    
    public static Iterator<Double> getDoubleIterator(double from, double to, double step)
    {
        return new DIterator(from, to, step);
    }


    public static class EndlessIterator<A> implements Iterator<A>
    {
        Vector<A>  v;
        int pos = 0;

        EndlessIterator(Vector<A>  v)
        {
            this.v = v;
        }
        public boolean hasNext()
        {
                return true;
        }

        public A next()
        {
            pos = (pos < v.size()) ? pos : 0;
            return v.get(pos++);
        }
    }


    static public<A> Iterator<A> getEndlessIterator(A ... values)
    {
        return new EndlessIterator<A>(ObjectGetter.toVector(values));
    }
    static public<A> Iterator<A> getEndlessIterator(Iterator<A> it)
    {
        return new EndlessIterator<A>(ObjectGetter.toVector(it));
    }
                                                            
    static public<A> Iterator<A> getEndlessIterator(Collection<A> c)
    {
        return new EndlessIterator<A>(ObjectGetter.toVector(c));
    }

    

}

