package lmu.utils;

import java.util.*;

/**
    - copy of pattern conversion from log4j - 
 * 
   A flexible layout configurable with pattern string.
   
   This code is known to have synchronization and other issues
   which are not present in org.apache.log4j.EnhancedPatternLayout.
   EnhancedPatternLayout should be used in preference to PatternLayout.
   EnhancedPatternLayout is distributed in the log4j extras companion.

   <p>The goal of this class is to {@link #format format} a {@link
   LoggingEvent} and return the results as a String. The results
   depend on the <em>conversion pattern</em>.

   <p>The conversion pattern is closely related to the conversion
   pattern of the printf function in C. A conversion pattern is
   composed of literal text and format control expressions called
   <em>conversion specifiers</em>.

   <p><i>You are free to insert any literal text within the conversion
   pattern.</i>

   <p>Each conversion specifier starts with a percent sign (%) and is
   followed by optional <em>format modifiers</em> and a <em>conversion
   character</em>. The conversion character specifies the type of
   data, e.g. category, priority, date, thread name. The format
   modifiers control such things as field width, padding, left and
   right justification. The following is a simple example.

   <p>Let the conversion pattern be <b>"%-5p [%t]: %m%n"</b> and assume
   that the log4j environment was set to use a PatternLayout. Then the
   statements
   <pre>
   Category root = Category.getRoot();
   root.debug("Message 1");
   root.warn("Message 2");
   </pre>
   would yield the output
   <pre>
   DEBUG [main]: Message 1
   WARN  [main]: Message 2
   </pre>

   <p>Note that there is no explicit separator between text and
   conversion specifiers. The pattern parser knows when it has reached
   the end of a conversion specifier when it reads a conversion
   character. In the example above the conversion specifier
   <b>%-5p</b> means the priority of the logging event should be left
   justified to a width of five characters.

   The recognized conversion characters are

   <p>
   <table border="1" CELLPADDING="8">
   <th>Conversion Character</th>
   <th>Effect</th>

   <tr>
     <td align=center><b>c</b></td>

     <td>Used to output the category of the logging event. The
     category conversion specifier can be optionally followed by
     <em>precision specifier</em>, that is a decimal constant in
     brackets.

     <p>If a precision specifier is given, then only the corresponding
     number of right most components of the category name will be
     printed. By default the category name is printed in full.

     <p>For example, for the category name "a.b.c" the pattern
     <b>%c{2}</b> will output "b.c".

     </td>
   </tr>

   <tr>
     <td align=center><b>C</b></td>

     <td>Used to output the fully qualified class name of the caller
     issuing the logging request. This conversion specifier
     can be optionally followed by <em>precision specifier</em>, that
     is a decimal constant in brackets.

     <p>If a precision specifier is given, then only the corresponding
     number of right most components of the class name will be
     printed. By default the class name is output in fully qualified form.

     <p>For example, for the class name "org.apache.xyz.SomeClass", the
     pattern <b>%C{1}</b> will output "SomeClass".

     <p><b>WARNING</b> Generating the caller class information is
     slow. Thus, use should be avoided unless execution speed is
     not an issue.

     </td>
     </tr>

   <tr> <td align=center><b>d</b></td> <td>Used to output the date of
     the logging event. The date conversion specifier may be
     followed by a <em>date format specifier</em> enclosed between
     braces. For example, <b>%d{HH:mm:ss,SSS}</b> or
     <b>%d{dd&nbsp;MMM&nbsp;yyyy&nbsp;HH:mm:ss,SSS}</b>.  If no
     date format specifier is given then ISO8601 format is
     assumed.

     <p>The date format specifier admits the same syntax as the
     time pattern string of the {@link
     java.text.SimpleDateFormat}. Although part of the standard
     JDK, the performance of <code>SimpleDateFormat</code> is
     quite poor.

     <p>For better results it is recommended to use the log4j date
     formatters. These can be specified using one of the strings
     "ABSOLUTE", "DATE" and "ISO8601" for specifying {@link
     org.apache.log4j.helpers.AbsoluteTimeDateFormat
     AbsoluteTimeDateFormat}, {@link
     org.apache.log4j.helpers.DateTimeDateFormat DateTimeDateFormat}
     and respectively {@link
     org.apache.log4j.helpers.ISO8601DateFormat
     ISO8601DateFormat}. For example, <b>%d{ISO8601}</b> or
     <b>%d{ABSOLUTE}</b>.

     <p>These dedicated date formatters perform significantly
     better than {@link java.text.SimpleDateFormat}.
     </td>
   </tr>

   <tr>
   <td align=center><b>F</b></td>

   <td>Used to output the file name where the logging request was
   issued.

   <p><b>WARNING</b> Generating caller location information is
   extremely slow and should be avoided unless execution speed
   is not an issue.

   </tr>

   <tr>
   <td align=center><b>l</b></td>

     <td>Used to output location information of the caller which generated
     the logging event.

     <p>The location information depends on the JVM implementation but
     usually consists of the fully qualified name of the calling
     method followed by the callers source the file name and line
     number between parentheses.

     <p>The location information can be very useful. However, its
     generation is <em>extremely</em> slow and should be avoided
     unless execution speed is not an issue.

     </td>
   </tr>

   <tr>
   <td align=center><b>L</b></td>

   <td>Used to output the line number from where the logging request
   was issued.

   <p><b>WARNING</b> Generating caller location information is
   extremely slow and should be avoided unless execution speed
   is not an issue.

   </tr>


   <tr>
     <td align=center><b>m</b></td>
     <td>Used to output the application supplied message associated with
     the logging event.</td>
   </tr>

   <tr>
   <td align=center><b>M</b></td>

   <td>Used to output the method name where the logging request was
   issued.

   <p><b>WARNING</b> Generating caller location information is
   extremely slow and should be avoided unless execution speed
   is not an issue.

   </tr>

   <tr>
     <td align=center><b>n</b></td>

     <td>Outputs the platform dependent line separator character or
     characters.

     <p>This conversion character offers practically the same
     performance as using non-portable line separator strings such as
     "\n", or "\r\n". Thus, it is the preferred way of specifying a
     line separator.


   </tr>

   <tr>
     <td align=center><b>p</b></td>
     <td>Used to output the priority of the logging event.</td>
   </tr>

   <tr>

     <td align=center><b>r</b></td>

     <td>Used to output the number of milliseconds elapsed from the construction 
     of the layout until the creation of the logging event.</td>
   </tr>


   <tr>
     <td align=center><b>t</b></td>

     <td>Used to output the name of the thread that generated the
     logging event.</td>

   </tr>

   <tr>

     <td align=center><b>x</b></td>

     <td>Used to output the NDC (nested diagnostic context) associated
     with the thread that generated the logging event.
     </td>
   </tr>


   <tr>
     <td align=center><b>X</b></td>

     <td> 
     
     <p>Used to output the MDC (mapped diagnostic context) associated
     with the thread that generated the logging event. The <b>X</b>
     conversion character <em>must</em> be followed by the key for the
     map placed between braces, as in <b>%X{clientNumber}</b> where
     <code>clientNumber</code> is the key. The value in the MDC
     corresponding to the key will be output.</p>
     
     <p>See {@link MDC} class for more details.
     </p>
     
     </td>
   </tr>

   <tr>

     <td align=center><b>%</b></td>

     <td>The sequence %% outputs a single percent sign.
     </td>
   </tr>

   </table>

   <p>By default the relevant information is output as is. However,
   with the aid of format modifiers it is possible to change the
   minimum field width, the maximum field width and justification.

   <p>The optional format modifier is placed between the percent sign
   and the conversion character.

   <p>The first optional format modifier is the <em>left justification
   flag</em> which is just the minus (-) character. Then comes the
   optional <em>minimum field width</em> modifier. This is a decimal
   constant that represents the minimum number of characters to
   output. If the data item requires fewer characters, it is padded on
   either the left or the right until the minimum width is
   reached. The default is to pad on the left (right justify) but you
   can specify right padding with the left justification flag. The
   padding character is space. If the data item is larger than the
   minimum field width, the field is expanded to accommodate the
   data. The value is never truncated.

   <p>This behavior can be changed using the <em>maximum field
   width</em> modifier which is designated by a period followed by a
   decimal constant. If the data item is longer than the maximum
   field, then the extra characters are removed from the
   <em>beginning</em> of the data item and not from the end. For
   example, it the maximum field width is eight and the data item is
   ten characters long, then the first two characters of the data item
   are dropped. This behavior deviates from the printf function in C
   where truncation is done from the end.

   <p>Below are various format modifier examples for the category
   conversion specifier.

   <p>
   <TABLE BORDER=1 CELLPADDING=8>
   <th>Format modifier
   <th>left justify
   <th>minimum width
   <th>maximum width
   <th>comment

   <tr>
   <td align=center>%20c</td>
   <td align=center>false</td>
   <td align=center>20</td>
   <td align=center>none</td>

   <td>Left pad with spaces if the category name is less than 20
   characters long.

   <tr> <td align=center>%-20c</td> <td align=center>true</td> <td
   align=center>20</td> <td align=center>none</td> <td>Right pad with
   spaces if the category name is less than 20 characters long.

   <tr>
   <td align=center>%.30c</td>
   <td align=center>NA</td>
   <td align=center>none</td>
   <td align=center>30</td>

   <td>Truncate from the beginning if the category name is longer than 30
   characters.

   <tr>
   <td align=center>%20.30c</td>
   <td align=center>false</td>
   <td align=center>20</td>
   <td align=center>30</td>

   <td>Left pad with spaces if the category name is shorter than 20
   characters. However, if category name is longer than 30 characters,
   then truncate from the beginning.

   <tr>
   <td align=center>%-20.30c</td>
   <td align=center>true</td>
   <td align=center>20</td>
   <td align=center>30</td>

   <td>Right pad with spaces if the category name is shorter than 20
   characters. However, if category name is longer than 30 characters,
   then truncate from the beginning.

   </table>

   <p>Below are some examples of conversion patterns.

   <dl>

   <p><dt><b>%r [%t] %-5p %c %x - %m%n</b>
   <p><dd>This is essentially the TTCC layout.

   <p><dt><b>%-6r [%15.15t] %-5p %30.30c %x - %m%n</b>

   <p><dd>Similar to the TTCC layout except that the relative time is
   right padded if less than 6 digits, thread name is right padded if
   less than 15 characters and truncated if longer and the category
   name is left padded if shorter than 30 characters and truncated if
   longer.

  </dl>

   <p>The above text is largely inspired from Peter A. Darnell and
   Philip E. Margolis' highly recommended book "C -- a Software
   Engineering Approach", ISBN 0-387-97389-3.

   @author <a href="mailto:cakalijp@Maritz.com">James P. Cakalic</a>
   @author Ceki G&uuml;lc&uuml;


   
*/
public class Logger  extends org.apache.log4j.Logger implements java.io.Serializable
{



    public enum LOGLEVEL
    {
        OFF("off", org.apache.log4j.Level.OFF),
        DEBUG("debug", org.apache.log4j.Level.DEBUG),
        TRACE("trace", org.apache.log4j.Level.TRACE),
        INFO("info", org.apache.log4j.Level.INFO),
        FATAL("fatal", org.apache.log4j.Level.FATAL),
        WARN("warn", org.apache.log4j.Level.WARN),
        ERROR("error", org.apache.log4j.Level.ERROR);


        org.apache.log4j.Level level;
        String name;

        LOGLEVEL(String n, org.apache.log4j.Level level)
        {
            name=n;
            this.level=level;
            EnumGetter.add(name, this);
        }

        public String getName()
        {
            return name;
        }

        public static LOGLEVEL get(String n)
        {
            LOGLEVEL l = (LOGLEVEL)EnumGetter.get(n);
            if (l==null)
            {
                throw new RuntimeException("unknown loglevel: "+n+" known levels are: "+EnumGetter.getLookups());
            }
            return l;
        }

    }
    
    interface LogLevelWriter
    {
        void printf(String fmt, Object ... data);
    }


    LogLevelWriter getLevelWriter(LOGLEVEL level)
    {
        switch(level)
        {
            case INFO:
                return (fmt, data) -> info(fmt, data);
            case DEBUG:
                return (fmt, data) -> debug(fmt, data);
            case WARN:
                return (fmt, data) -> warn(fmt, data);
            case ERROR:
                return (fmt, data) -> error(fmt, data);
            case TRACE:
                return (fmt, data) -> trace(fmt, data);
            case FATAL:
                return (fmt, data) -> fatal(fmt, data);
        }

        return null;
    }


    
    static class LogStream extends java.io.OutputStream
    {
        Logger log;
        LogLevelWriter lw;
        StringBuilder mBuf = new StringBuilder();


        public LogStream(Logger log, LOGLEVEL level)
        {
            super();
            this.log = log;
            lw = log.getLevelWriter(level);
        }

        public void write(int b) 
        {
            char ch = (char)b;
            if (ch == '\n') 
            {
                lw.printf(mBuf.toString());
                mBuf.setLength(0);
                return;
            }
            mBuf.append((char) b);
        }
    }


    public java.io.PrintStream getPrintStream(String ln)
    {
        return new java.io.PrintStream(new LogStream(this, LOGLEVEL.get(ln)));
    }

    private static final String FQCN = Logger.class.getName();

    public static class FmtLoggerFactory implements org.apache.log4j.spi.LoggerFactory 
    {

        /**
           The constructor should be public as it will be called by
           configurators in different packages.  */
        public FmtLoggerFactory()
        {
        }


        public Logger makeNewLoggerInstance(String name) 
        {
            return new lmu.utils.Logger(name);
        }

    }



    public static void reloadFactory()
    {
        fmtFactory = new FmtLoggerFactory();
    }

    private static FmtLoggerFactory fmtFactory = new FmtLoggerFactory();

    public  static org.apache.log4j.Logger getLogger(String name) 
    {

        return org.apache.log4j.Logger.getLogger(name, fmtFactory); 
    }

    public String getLevelName()
    {
        org.apache.log4j.Level l = getLevel();

        for (LOGLEVEL ll : LOGLEVEL.values())
        {
            if (l == ll.level)
                return ll.getName();
        }
        return null;
    }

    protected void forcedLog(String fqcn, org.apache.log4j.Priority level, Object message, Throwable t) 
    {

        //System.out.println("will og message: "+message+" FQCN: "+FQCN);
        //System.out.println("invoker: \n"+ExceptionManager.getCallerInfo(1)+"\n\n");
        callAppenders(new org.apache.log4j.spi.LoggingEvent(FQCN, this, level, message, t));
    }

    public Logger()
    {
        super("unnamed");
    }

    public Logger(String name)
    {
        super(name);
    }

    public void info(Object o)
    {
        super.info(o);
    }
    public void info(String msg)
    {
        //System.out.println("will invoke: logger: info:"+msg);
        super.info(msg);
    }
    public void info(Throwable e)
    {
        super.info(ExceptionManager.getStackTrace(e));
    }
    public void info(String msg, Object ... params)
    {
        String imsg = String.format(msg, params);
        super.info(imsg);
    }

    public void debug(Object o)
    {
        super.debug(o);
    }

    public void debug(String msg)
    {
        super.debug(msg);
    }
    public void debug(Throwable e)
    {
        super.debug(ExceptionManager.getStackTrace(e));
    }
    public void debug(String msg, Object ... params)
    {
        debug(false,msg,params);
    }
    public void debug(boolean logcaller, String msg, Object ... params)
    {
        if (!isDebugEnabled())
            return;

        if (logcaller)
        {
            String stack=ExceptionManager.logCallerStack();
            super.debug(stack);
        }
        if (params==null)
        {
            super.debug(msg);
            return;
        }
        String imsg = String.format(msg, params);

        super.debug(imsg);
    }


    public void trace(Object o)
    {
        super.trace(o);
    }
    public void trace(String msg)
    {
        super.trace(msg);
    }
    public void trace(Throwable e)
    {
        super.trace(ExceptionManager.getStackTrace(e));
    }
    public void trace(String msg, Object ... params)
    {
        trace(false, msg,params);
    }
    public void trace(boolean logcaller, String msg, Object ... params)
    {
        if (!isTraceEnabled())
            return;


        if (logcaller)
        {
            String stack=ExceptionManager.logCallerStack();
            super.trace(stack);
        }
        String imsg = String.format(msg, params);
        super.trace(imsg);
    }

    public void fatal(Object o)
    {
        super.fatal(o);
    }
    public void fatal(String msg)
    {
        super.fatal(msg);
    }
    public void fatal(Throwable e)
    {
        super.fatal(ExceptionManager.getStackTrace(e));
    }
    public void fatal(String msg, Object ... params)
    {
        fatal(false,msg,params);
    }
    public void fatal(boolean logcaller, String msg, Object ... params)
    {
        if (logcaller)
        {
            String stack=ExceptionManager.logCallerStack();
            super.fatal(stack);
        }
        String imsg = String.format(msg, params);

        super.fatal(imsg);
    }

    public void error(String msg)
    {
        super.error(msg);
    }
    public void error(Throwable e)
    {
        super.error(ExceptionManager.getStackTrace(e));
    }
    public void error(String msg, Object ... params)
    {
        error(false,msg,params);
    }
    public void error(boolean logcaller, String msg, Object ... params)
    {
        if (logcaller)
        {
            String stack=ExceptionManager.logCallerStack();
            super.error(stack);
        }

        String imsg = String.format(msg, params);
        super.error(imsg);
    }

    public void error(Object o)
    {
        super.error(o);
    }
    public void warn(String msg)
    {
        super.warn(msg);
    }

    public void warn(Throwable e)
    {
        super.warn(ExceptionManager.getStackTrace(e));
    }
    public void warn(boolean logcaller, String msg, Object ... params)
    {
        if (logcaller)
        {
            String stack=ExceptionManager.logCallerStack();
            super.warn(stack);
        }


        String imsg = String.format(msg, params);
        super.warn(imsg);
    }
    public void warn(String msg, Object ... params)
    {
        warn(false,msg,params);

    }

    public void warn(Object o)
    {
        super.warn(o);
    }
    public void setLevel(String n)
    {
        super.setLevel(LOGLEVEL.get(n).level);

    }

    public void setLevel(LOGLEVEL level)
    {
        super.setLevel(level.level);

    }
}
