package lmu.utils;

public class NonBlockingHandler<A> extends Thread
{
    static class JobCancelledItselfException extends RuntimeException
    {

    }

    static public<A> boolean havePartialHandler(NonBlockingHandler<A> handler)
    {
        return handler!=null && handler.havePartialHandler();
    }
    static public<A> void reportPartialResult(NonBlockingHandler<A> handler, A data)
    {
        if (handler == null)
            return;
        
        handler.reportPartialResult(data);
    }
    static public<A> void checkCancelled(NonBlockingHandler<A> handler)
    {
        if (handler == null || !handler.isCancelled())
            return;

        handler.setJobCancelledItself();
    }
    static public<A> void reportStatus(NonBlockingHandler<A> handler, String status, Object ... data)
    {
        if (handler == null)
            return;
        
        handler.reportStatus(status, data);
    }
    public boolean blocking = false;

    A result;
    Exception error;

    public static StatusReporter getStreamReporter(java.io.PrintStream ps)
    {
        return (String s) -> ps.println(s);
    }


    public void setJobCancelledItself()
    {
        throw new JobCancelledItselfException();
    }
    

    public static StatusReporter getLoggerReporter(Logger log, String level)
    {
        switch(Logger.LOGLEVEL.get(level))
        {
            case INFO:
                return (String s) -> log.info(s);
            case DEBUG:
                return (String s) -> log.debug(s);
            case ERROR:
                return (String s) -> log.error(s);
            case TRACE:
                return (String s) -> log.trace(s);
            case WARN:
                return (String s) -> log.warn(s);
            case FATAL:
                return (String s) -> log.fatal(s);
            
                
        }
        return (String s) -> {};
    }

    /** constructor for blocking variant */
    public NonBlockingHandler(StatusReporter status_reporter)
    {
        blocking = true;
        this.statusReporter = status_reporter;
    }

    public NonBlockingHandler(ReadyReporter<A> reporter, StatusReporter status_reporter)
    {
        if (reporter == null)
            throw new RuntimeException("cannot set non blocking handler without reporter"); //actually we coult....  but put than an empty lambda in it
        
        this.reporter = reporter;
        this.statusReporter = status_reporter;
    }

    public boolean havePartialHandler()
    {
        return partialHandler != null;
    }
    public void reportPartialResult(A data)
    {
        if (!havePartialHandler())
            return;
        
        partialHandler.handlePartialResult(data);
    }
    
    public void setPartialHandler(PartialResultHandler<A> handler)
    {
        partialHandler=handler;
    }
    
    public boolean isCancelled()
    {
        return cancelled;
    }

    public Exception getException()
    {
        return job_error;
    }
    public A getResult()
    {
        return result;
    }
    boolean cancelled = false;
    StatusReporter statusReporter;
    ReadyReporter<A> reporter;
    PartialResultHandler<A> partialHandler;
    Exception job_error;

    public void cancel()
    {
        cancelled = true;
        if (reporter!=null)
        {
            reporter.ready(null, null, true);
        }
        
    }

    public void setReady(A result, Exception exception)
    {
        if (cancelled)
            return;
        
        job_error = exception;
        if (reporter!=null)
        {
            reporter.ready(result, exception, false);
        }

    }

    

    public void reportStatus(String fmt, Object ... data)
    {
        reportStatus(String.format(fmt, data));
        NonBlockingHandler.checkCancelled(this);
    }
    public void reportStatus(String status)
    {
        if (statusReporter == null)
            return;
        
        statusReporter.updateStatus(status);
    }

    
    public interface StatusReporter
    {
        public void updateStatus(String status);
    }

    public interface ReadyReporter<A>
    {
        public void ready(A result, Exception error, boolean was_cancelled);
    }

    public interface PartialResultHandler<A>
    {
        public void handlePartialResult(A result);
    }
    
    public interface Job<A>
    {
        public A doIt();
    }

    public NonBlockingHandler()
    {

    }

    Job<A> job;
    

    

    public NonBlockingHandler<A> setJob(Job<A> j)
    {
        this.job = j;
        if (blocking)
        {
            runBlocking();
            return this;
        }
        
        
        start();
        return this;
    }

     
    public void runBlocking()
    {
        try
        {
            result = job.doIt();
            setReady(result, null);
        }
        catch(JobCancelledItselfException je)
        {
            cancelled = true;
            setReady(null, null);
            
        }
        catch(Exception e)
        {
            setReady(null, e);
        }
    }

    public interface ReportingJob<A>
    {
        public A doIt(NonBlockingHandler<A> nbh);
    }

    static public<A> NonBlockingHandler<A> runAsynchron(ReadyReporter<A> reporter, StatusReporter status_reporter, ReportingJob<A> rj)
    {
        NonBlockingHandler<A> handler = new NonBlockingHandler<>(reporter, status_reporter);
        handler.setJob(() -> rj.doIt(handler));

        return handler;
    }
    
    public void run()
    {
        runBlocking();
    }

    public A getResult(boolean throw_exc)
    {
        if (throw_exc && job_error!=null)
        {
            throw new RuntimeException(job_error);
        }
        return result;
    }
}

