package lmu.utils;
import java.util.*;
import org.jdom.*;
import javax.xml.stream.*;

public class Region1D<T> implements Comparable
{
    public final static String ELEM_NAME="r1d";
    int x1, x2;
    T obj;
    
    public Region1D(Pair<Integer,Integer> p)
    {
        this(p, null);
    }
    public Region1D(Pair<Integer,Integer> p, T o)
    {
        this(p.getFirst(), p.getSecond(), o);
    }

    public Region1D(XMLStreamReader xsr) throws XMLStreamException
    {
        x1=XMLUtils.getIntAttrib(xsr, "x1", null);
        x2=XMLUtils.getIntAttrib(xsr, "x2", null);
        reset(x1,x2);
    }

    public Region1D(int x1, int x2)
    {
        this(x1, x2, null);
    }

    public Region1D(int x1, int x2, T o)
    {
        this.obj=o;
        reset(x1,x2);
        
    }
    public Region1D(Region1D<T> r)
    {
        this.obj=r.obj;
        reset(r.x1,r.x2);
    
    }

    public Region1D(Element e)
    {
        x1=XMLUtils.getIntAttrib(e, "x1", null);
        x2=XMLUtils.getIntAttrib(e, "x2", null);
        reset(x1,x2);
    }

    public void writeXML(XMLStreamWriter xsw) throws XMLStreamException
    {
        writeXML(xsw, null);
    }
    public void writeXML(XMLStreamWriter xsw, String elem_name) throws XMLStreamException
    {
        String en = (elem_name!=null)?elem_name:ELEM_NAME;
        xsw.writeEmptyElement(en);
        xsw.writeAttribute("x1", ""+x1);
        xsw.writeAttribute("x2", ""+x2);

    }
    
    public Element getXML(String Name)
    {
        Element e = new Element(Name);
        e.setAttribute("x1", ""+x1);
        e.setAttribute("x2", ""+x2);
        return e;
        
    }
    

    public Element getXML()
    {
        return getXML(ELEM_NAME);
    }
    /**
     * Computes the intersected Region between two region vectors
     * @param other -- a region vector
     * @return the intersected region between this and other, or null
     */
    public Region1D<?> getIntersectedRegion(Region1D<?> other)
    {
    	if ( this.getOverlap(other) == 0)
    	{
    		return null;
    	}
    	
    	int start = Math.max(this.getX1(),other.getX1());
    	int end = Math.min(this.getX2(),other.getX2());
         
    	return new Region1D<Object>(start,end,null);
    }
    
    public static int getIndex(Region1D r, Vector<Region1D> v , boolean throw_exc)
    {
        return getIndex(r, r.getObject(),v,throw_exc);
    }
   
    public static int getIndex(Object o, Vector<Region1D> v, boolean throw_exc)
    {
        return getIndex(null, o,v,throw_exc);
    }
    public static int getIndex(int x1, int x2, Object o, Vector<Region1D> v, boolean throw_exc)
    {
        return getIndex(new Region1D(x1,x2,o), o,v,throw_exc);
    }
    public static int getIndex(int x1, int x2, Vector<Region1D> v , boolean throw_exc)
    {
        return getIndex(x1,x2,null,v,throw_exc);
    }
    public static int getIndex(Region1D r, Object o, Vector<Region1D> v, boolean throw_exc)
    {
        
        if (r==null && o==null)
        {
            throw new RuntimeException("no region, no object ,how to find the requested region???");
        }

        for (int i=0; i<v.size(); i++)
        {
            Region1D totest = v.get(i);

            if (r!=null && r==totest)
            {
                return i;
            }
            if (totest==null)
            {
                continue;
            }
            if (r==null && o==totest.getObject())
            {
                return i;
            }
            if (o==null && r.getX1()==totest.getX1() && r.getX2()==totest.getX2())
            {
                return i;
            }
            
        }
        if (throw_exc)
        {
            throw new RuntimeException("cannot find region:"+r+" and object:" +o+" in "+v);
        }
        return -1;
    }
    
    public Pair<Integer,Integer> toPair()
    {
        return new Pair<Integer,Integer>(x1,x2);
    }
    public void reset(int x1, int x2)
    {
        this.x1=x1; this.x2=x2; 
        if (x2<x1)
        {
            throw new RuntimeException("invalid region x2 must be >= x1!");
        }
        hash_code = -1;

    }
    public double getOverlapNormedOnMin(Region1D d1)
    {
        int ov = getOverlap(d1);
        if (ov<=0) return 0.0;
        
        return ov/(double)Math.min(getLength(), d1.getLength());
    }

    public static int getOverlap(int x11, int x12, int x21, int x22)
    {
        int max1=Math.max(x11, x21);
        int min2=Math.min(x12, x22);
        return  Math.max(0, min2-max1); //-> getOverlap(x,x,x,x) == 0;

    }
    public static int getOverlap(Region1D d1, int x1, int x2)
    {
        return getOverlap(d1.getX1(),d1.getX2(),x1,x2);
        
    }
    public int getOverlap(Region1D d1)
    {
        return getOverlap(this,d1.getX1(), d1.getX2());
        

    }
    public int getOverlap(int x1, int x2)
    {
        return getOverlap(this,x1, x2);
        

    }
    public int getLength()
    {
        return x2-x1;
    }
    public boolean contains(int x)
    {
        return (x1<=x && x2>x);
    }

        
    public boolean contains(Region1D r )
    {
        return contains(r.getX1()) && contains(r.getX2()-1);
    }
    public int getX1()
    {
        return x1;
    }

    public int getX2()
    {
        return x2;
    }

    int hash_code = -1;

    public void setX1(int x1)
    {
        this.x1=x1;
        hash_code = -1;
    }

    public void setX2(int x2)
    {
        this.x2=x2;
        hash_code = -1;
    }


    public Region1D<T> translate(int x)
    {
        return new Region1D<T>(getX1() + x, getX2() + x, getObject());
    }

    public void setObject(T o)
    {
        obj=o;
    }
    public T getObject()
    {
        return obj;
    }

    public int hashCode()
    {
        if (hash_code>=0)
            return hash_code;
        
        return hash_code = (new Pair<Integer,Integer>(x1,x2)).hashCode();
    }
    
    public boolean equals(Object o)
    {
        return (0 == compareTo(o));
    }
    public int compareTo(Object o)
    {
        if (!(o instanceof Region1D))
        {
            if(1==1)throw new RuntimeException("no region1d!");
            return -1;
        }
        Region1D r = (Region1D)o;
        int d = x1-r.x1;
        if (d!=0)return d;
        
        return x2-r.x2;
        
    }

    public String toString()
    {
        return String.format("(%d-%d)", x1,x2);
    }
}
