package lmu.utils;

import java.sql.*;
import java.util.*;

public class SQLUtils
{

    public static int getGeneratedId(PreparedStatement ps)
    {
        try
        {
            ResultSet rs = ps.getGeneratedKeys();

            if (!rs.first())
            {
                throw new RuntimeException("could not generate new id with statement: "+ps);
            }
            return rs.getInt(1);
        }
        catch(SQLException ex)
        {
            throw new RuntimeException("error while getting generated key with statement: "+ps, ex);
        }

    }

    static public String generatePrefix(DbConnection con, int length, String table_suffix)
    {
        try
        {
            return generatePrefix(con.getConnection().createStatement(), length, table_suffix);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("error: %s  while generating prefix with suffix: %s", e, table_suffix);
        }
    }

    static public String generatePrefix(Statement st, int length, String table_suffix)
    {
        try
        {
            String prefix=null;
            while (prefix==null)
            {
                prefix = StringUtils.generateRandomString(length);
                ResultSet rs = st.executeQuery("show tables like '"+prefix+table_suffix+"';");
                if (!rs.first())
                {
                    return prefix;
                }
                prefix=null;
                           
            }
        }
        catch(SQLException e)
        {
            throw new RuntimeException("sql error:"+e.getMessage()+" while generating random prefix for table: "+table_suffix, e);
        }
        return null;
    }
    /** returns null if column not defined, otherwise the sql type id
     */
    public static Integer getDefinedColumnType(Statement st, String table, String colname) 
    {
        if (colname==null)
        {
            throw new RuntimeException("cannot check null-columname in "+table+"!");
        }
        try
        {
            ResultSet rs = st.executeQuery("select * from "+table+" limit 1");
            ResultSetMetaData rsmd = rs.getMetaData();
            //check if given column is defined
            for (int i=0; i<rsmd.getColumnCount(); i++)
            {
                String n = rsmd.getColumnName(i+1);
                if (n.equals(colname))
                {
                    return rsmd.getColumnType(i+1);
                }
                
                
            }

            return null;
        }
        catch(SQLException e)
        {
            throw new RuntimeException("error: "+e.getMessage()+" while checking if column: "+colname+" exists in "+table, e);
        }
    }
    static public Double getDouble(ResultSet rs, String colname) throws SQLException
    {
        double rv = rs.getDouble(colname);
        if (rs.wasNull())
            return null;
        
        return new Double(rv);
    }
    
    static public Double getDouble(ResultSet rs, int colindex) throws SQLException
    {
        double rv = rs.getDouble(colindex);
        if (rs.wasNull())
            return null;
        
        return new Double(rv);
    }


    static public Float getFloat(ResultSet rs, String colname) throws SQLException
    {
        Float rv = rs.getFloat(colname);
        if (rs.wasNull())
            return null;
        
        return new Float(rv);
    }
    
    static public Float getFloat(ResultSet rs, int colindex)throws SQLException
    {
        Float rv = rs.getFloat(colindex);
        if (rs.wasNull())
            return null;
        
        return new Float(rv);
    }


    
    static public void setDouble(PreparedStatement ps, int pos, Double v) throws SQLException
    {
        if (v == null || Double.isNaN(v))
        {
            ps.setNull(pos, java.sql.Types.DOUBLE);
            return;
        }
        ps.setDouble(pos, v);
    }


    static public void setInt(PreparedStatement ps, int pos, Integer v) throws SQLException
    {
        if (v == null)
        {
            ps.setNull(pos, java.sql.Types.INTEGER);
            return;
        }
        ps.setInt(pos, v);
    }

    static public Integer getInteger(ResultSet rs, String colname) throws SQLException
    {
        Integer rv = rs.getInt(colname);
        if (rs.wasNull())
            return null;
        
        return new Integer(rv);
    }
    
    static public Integer getInteger(ResultSet rs, int colindex)
    {
        try
        {
            Integer rv = rs.getInt(colindex);
            if (rs.wasNull())
                return null;
            
            return new Integer(rv);
        }
        catch (SQLException se)
        {
            throw new FRuntimeException("error: %s while fetching col: %d from resultset", se, se.getMessage(), colindex);
        }
    }

    static public Long getLong(ResultSet rs, String colname) throws SQLException
    {
        Long rv = rs.getLong(colname);
        if (rs.wasNull())
            return null;
        
        return new Long(rv);
    }
    
    static public Long getLong(ResultSet rs, int colindex) throws SQLException
    {
        Long rv = rs.getLong(colindex);
        if (rs.wasNull())
            return null;
        
        return new Long(rv);
    }

    public interface RSMapper<A>
    {
        public A map(ResultSet rs) throws SQLException;
    }

    public interface RSApplyer
    {
        public void apply(ResultSet rs) throws SQLException;
    }

    static public void apply(ResultSet rs, RSApplyer rapply)
    {
        try
        {
            while (rs.next())
            {
                rapply.apply(rs);
            }
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while iterating over results", se, se.getMessage());
        }
    }

    static class SQLTransformer<A> implements Iterator<A>
    {
        ResultSet rs;
        RSMapper<A> map;
        A n = null;

        public SQLTransformer(ResultSet rs, RSMapper<A> map)
        {
            this.map = map;
            this.rs = rs;
            move();
        }

        void move()
        {
            n = null;
            try
            {
                if (!rs.next())
                {
                    //auto close the result set
                    try
                    {
                       rs.close();
                    }
                    catch (SQLException se)
                    {
                        throw new FRuntimeException("sql error: %s while closing result set", se, se.getMessage());
                    }
                    return;
                }
                
                n = map.map(rs);
            }
            catch(SQLException se)
            {
                throw new FRuntimeException("sql error: %s while converting entry: %s", se, se.getMessage(), showResults(rs));
            }
        }
        public boolean hasNext()
        {
            return n != null;
        }
        public A next()
        {
            if (!hasNext())
                return null;
            
            A rv = n;
            move();
            return rv;
        }
    }

    static public<A> Iterator<A> transform(ResultSet rs, RSMapper<A> map)
    {
        return new SQLTransformer<A>(rs, map);
    }


    static public<A, C extends Collection<A>> C mapTo(ResultSet rs, RSMapper<A> rmap, C c)
    {
        apply(rs, (ResultSet rrs) -> c.add(rmap.map(rrs)));
        try
        {
            rs.close();
        }
        catch (SQLException se)
        {
            throw new FRuntimeException("sql error: %s while closing result set", se, se.getMessage());
        }
        return c;
    }

    static public<A> Vector<A> map(ResultSet rs, RSMapper<A> rmap)
    {
        return mapTo(rs, rmap, new Vector<A>());
    }

    static public<A> HashSet<A> mapToSet(ResultSet rs, RSMapper<A> rmap)
    {
        return mapTo(rs, rmap, new HashSet<A>());
    }

    static public boolean isEmpty(ResultSet rs)
    {
        try
        {
            boolean rv = rs.first();
            rs.beforeFirst();
            return rv;
        }
        catch(SQLException e)
        {
            throw new RuntimeException("sql error while checking if result set empty: "+ e.getMessage(), e);
        }
    }
    
    static public String showResults(ResultSet rs) 
    {
        try
        {
            
            ResultSetMetaData rsm = rs.getMetaData();
            
            StringBuffer sb = new StringBuffer();
            for (int i=1; i<=rsm.getColumnCount(); i++)
            {
                String v = null;
                switch(rsm.getColumnType(i))
                {
                    case Types.BLOB:
                    case Types.CLOB:
                    case Types.DATALINK:
                    case Types.BINARY:
                        v="<binary";
                        break;
                    default:
                        v=rs.getString(i);
                        break;
                }
                if (sb.length()>0)
                {
                    sb.append(' ');
                }
                sb.append(rsm.getColumnLabel(i)+":"+v);
                
            }
            return sb.toString();
        }
        catch(SQLException e)
        {
            return "error: "+e.getMessage() +" while getting attribs from sql result set";
        }
    }

    static public void reset(ResultSet rs)
    {
        try
        {
            rs.beforeFirst();
        }
        catch(SQLException e)
        {
            throw new RuntimeException("sql error while resetting resultset");
        }
    }

}
