package lmu.utils;

import java.sql.*;
import java.util.Vector;
import java.util.regex.Pattern;

import org.jdom.Element;

public class SimpleTypeConverter implements StringConverter 
{

    public static final String VECTOR_TYPE = "vector";
    public static final String XML_TYPE = "xml_element";


    public static Vector parseVector(Element ate)
    {

        String t = XMLUtils.getAttrib(ate, "type", null, true);

        if (t == null || !t.equals(VECTOR_TYPE))
            return null;
        
        Vector v = new Vector();
        for (Element sube : XMLUtils.getChildren(ate, "elem"))
        {
            String subt = XMLUtils.getAttrib(sube, "type", null, true);

            Vector subv = parseVector(sube);
            if (subv != null)
            {
                v.add(subv);
                continue;
            }
            
            StringConverter sc = SimpleTypeConverter.getConverter(subt, true);
            
            v.add(sc.convert(XMLUtils.getAttrib(sube, "value", null, true)));
        }
            
        
        
        return v;
    }

    public static Element createVectorElement(String elemname, Object o)
    {
        if (!(o instanceof Vector))
            return null;
        
        Vector v = (Vector)o;

        Element e = new Element(elemname);
        e.setAttribute("type", VECTOR_TYPE);

        for (Object vo : v)
        {
            Element subV = createVectorElement("elem", vo);
            if (subV != null)
            {
                e.addContent(subV);
                continue;
            }
            
            SimpleTypeConverter.SIMPLETYPE st = SimpleTypeConverter.get(vo, false);
            
            if (st == null)
            {
                throw new RuntimeException("no converter found to "+vo.getClass().getCanonicalName()+" for value:"+o +" while elems of vector");
            }
            Element subE = new Element("elem");
            subE.setAttribute("type",  st.getName());
            subE.setAttribute("value", ""+ vo);
            e.addContent(subE);
            
        }
        return e;

    }

    
            
            
            

    public interface SQLConverter
    {
        public void set(PreparedStatement ps, int i, Object o) throws SQLException;

        public Object get(ResultSet rs, int i) throws SQLException;

        public Object get(ResultSet rs, String field) throws SQLException;
        
        public String getSQLType();

        
    }
    static class IntSQLConverter implements SQLConverter
    {
        
        public void set(PreparedStatement ps, int i, Object o) throws SQLException
        {
            if(o==null)
            {
                ps.setNull(i, java.sql.Types.INTEGER);
            }
            else
            {    
                ps.setInt(i,(Integer)o);
            }
        }

        public Object get(ResultSet rs, int i) throws SQLException
        {
            Integer v = rs.getInt(i);
            if(rs.wasNull())
                return null;
            return v;
        }

        public Object get(ResultSet rs, String field) throws SQLException
        {
            Integer v= rs.getInt(field);
            if(rs.wasNull())
                return null;
            return v;
        }
        
        public String getSQLType()
        {
            return "int";
        }
    }


    static class DoubleSQLConverter implements SQLConverter
    {
        
        public void set(PreparedStatement ps, int i, Object o) throws SQLException
        {
            
            Double v = (Double)o;
            if (v==null || Double.isNaN(v))
            {
                v = null;
                ps.setNull(i, java.sql.Types.DOUBLE);
                return;
            }

            ps.setDouble(i, v);
        }

        public Object get(ResultSet rs, int i) throws SQLException
        {
            Double d = rs.getDouble(i);
            if(rs.wasNull())
                return null;
            return d;
        }

        public Object get(ResultSet rs, String field) throws SQLException
        {
            Double d = rs.getDouble(field);
            if(rs.wasNull())
                return null;
            return d;
        }
        
        public String getSQLType()
        {
            return "float";
        }
    }


    static class StringSQLConverter  implements SQLConverter
    {
        
        public void set(PreparedStatement ps, int i, Object o) throws SQLException
        {
            
            ps.setString(i,(String)o);
        }

        public Object get(ResultSet rs, int i) throws SQLException
        {
            return rs.getString(i);
        }

        public Object get(ResultSet rs, String field) throws SQLException
        {
            return rs.getString(field);
        }
        
        public String getSQLType()
        {
            return "text";
        }
    }

    static class FloatSQLConverter implements SQLConverter
    {
        
        public void set(PreparedStatement ps, int i, Object o) throws SQLException
        {
            
            ps.setFloat(i,(Float)o);
        }

        public Object get(ResultSet rs, int i) throws SQLException
        {
            return rs.getFloat(i);
        }

        public Object get(ResultSet rs, String field) throws SQLException
        {
            return rs.getFloat(field);
        }
        
        public String getSQLType()
        {
            return "float";
        }
    }

    static class LongSQLConverter implements SQLConverter
    {
        
        public void set(PreparedStatement ps, int i, Object o) throws SQLException
        {
            
            ps.setLong(i,(Long)o);
        }

        public Object get(ResultSet rs, int i) throws SQLException
        {
            return rs.getLong(i);
        }

        public Object get(ResultSet rs, String field) throws SQLException
        {
            return rs.getLong(field);
        }
        
        public String getSQLType()
        {
            return "long";
        }
    }


    static class BooleanSQLConverter implements SQLConverter
    {
    
        public void set(PreparedStatement ps, int i, Object o) throws SQLException
        {
    
            ps.setBoolean(i,(Boolean)o);
        }
    
        public Object get(ResultSet rs, int i) throws SQLException
        {
            return rs.getBoolean(i);
        }
    
        public Object get(ResultSet rs, String field) throws SQLException
        {
            return rs.getBoolean(field);
        }
    
        public String getSQLType()
        {
            return "boolean";
        }
    }
    

    public enum SIMPLETYPE
    {
        INT("int", Integer.class, new IntSQLConverter()),
        BOOLEAN("boolean", Boolean.class, new BooleanSQLConverter()),
        FLOAT("float", Float.class, new FloatSQLConverter()),
        DOUBLE("double", Double.class, new DoubleSQLConverter()),
        LONG("long", Long.class,  new LongSQLConverter()),
        STRING("string", String.class, new StringSQLConverter()),
        CHAR("char", String.class, new StringSQLConverter())
        ;

        String name;

        SQLConverter sql_converter;
        SimpleTypeConverter silent_converter;
        SimpleTypeConverter nervous_converter;

        Class cz;

        SIMPLETYPE(String s, Class cz, SQLConverter sqc)
        {
            name=s;
            
            sql_converter=sqc;
            this.cz=cz;
            EnumGetter.add(s, this);

            silent_converter = new SimpleTypeConverter(this, false);
            nervous_converter = new SimpleTypeConverter(this, true);


        }

        public String getName()
        {
            return name;
        }
        public SQLConverter getSQLConverter()
        {
            return sql_converter;
        }
        public static SIMPLETYPE get(Class c)
        {
            return get(c, false);
        }
        
        static SIMPLETYPE get(Class c, boolean simple)
        {
            return get(c, simple, true);
        }
        static SIMPLETYPE get(Class c, boolean simple, boolean throw_exc)
        {
            for (SIMPLETYPE st: SIMPLETYPE.values())
            {
                if (st.cz==c)
                    return st;
                
            }
            //check if it is a simple type
            //Class sc = (simple)?null:ReflectedObject.getSimpleClass(c);
            
            Class sc = ReflectedObject.getSimpleClass(c);
            //System.out.println("check for "+c+" sc: "+sc);
            if (sc!=null)
            {
                return get(sc, true);
            }
            
            if(!throw_exc)
                return null;

            throw new RuntimeException("unknown converter for "+c.getName());
        }
        public static SIMPLETYPE get(String s)
        {
            return (SIMPLETYPE)EnumGetter.get(s);
        }
    }

    public static SIMPLETYPE get(Object o)
    {
        return get(o, true);
    }
    public static SIMPLETYPE get(Object o, boolean throw_exc)
    {
        if (o==null)
            return null;
        
        return SIMPLETYPE.get(o.getClass(), false, throw_exc);
    }
    public static StringConverter getConverter(Class clazz, boolean throw_exc)
    {

        SIMPLETYPE st = SIMPLETYPE.get(""+clazz);
        if (st==null)
        {
            st = SIMPLETYPE.get(clazz);
        }
        
        return (throw_exc)?st.nervous_converter:st.silent_converter;
    }
    public static StringConverter getConverter(String typename, boolean throw_exc)
    {
        SIMPLETYPE st = SIMPLETYPE.get(typename);
        if (st==null)
        {
            throw new RuntimeException("unknown simpletype name: "+typename+ "known types: "+ EnumGetter.getLookups(SIMPLETYPE.class));
        }
        return (throw_exc)?st.nervous_converter:st.silent_converter;
    }
    SIMPLETYPE st;

    boolean throw_exc=false;

    public SimpleTypeConverter(String st_name)
    {
        this(st_name, false);
    }
    public SimpleTypeConverter(String st_name, boolean throw_exception_if_error)
    {
        this(SIMPLETYPE.get(st_name), throw_exception_if_error);
    }

    public SimpleTypeConverter(SIMPLETYPE st)
    {
        this(st, false);
    }
    public SimpleTypeConverter(SIMPLETYPE st, boolean throw_exception_if_error)
    {
        this.st=st;
        throw_exc=throw_exception_if_error;
    }

    Pattern NOTANUMBER = Pattern.compile("NaN|null|na|NA", Pattern.CASE_INSENSITIVE);
    Pattern INFINITY = Pattern.compile("inf|infty|infinity|maxval", Pattern.CASE_INSENSITIVE);
    Pattern NEG_INFINITY = Pattern.compile("(-|neg|negative)( )?(inf|infty|infinity|minval)", Pattern.CASE_INSENSITIVE);

    static String getNumString(String s)
    {
        if(s.indexOf(',')<0)
            return s;
        return s.replaceAll(",", ".");
    }

    static public Vector<Double> toDouble(String[] s)
    {
        return ObjectGetter.map(s, (String ss) -> doubleValue(ss));
    }


    static public Double doubleValue(Object o)
    {
        if (o == null)
            return null;
        
        
        Class cz = o.getClass();
        if (cz == Double.class)
            return (Double)o;

        if (cz == String.class)
        {
            return (Double)SIMPLETYPE.DOUBLE.silent_converter.convert(getNumString((String)o));
            //return Double.parseDouble(getNumString((String)o));
        }
            
        
        if (cz == Float.class)
            return ((Float)o).doubleValue();

        if (cz == Long.class)
            return ((Long)o).doubleValue();

        if (cz == Integer.class)
            return ((Integer)o).doubleValue();
                
        throw new RuntimeException("could not convert to double :"+ o +" class: "+cz);
    }

    public static Integer toInt(String s, Integer def_value, boolean error_if_not_int, String msg, Object ... infos)
    {
        try
        {
            return Integer.parseInt(s);
        }
        catch(NumberFormatException ex)
        {
            if (error_if_not_int)
                throw new FRuntimeException("%s can not convert >%s< to integer", (msg == null) ? "" : String.format(msg,infos), s);
            
            return def_value;
        }
    }
    
    public Object convert(String s)
    {
        if (s==null)
            return s;
        
        try
        {
            switch(st)
            {
                case BOOLEAN:
                    return Boolean.parseBoolean(s);
                    
                case INT:
                    return Integer.parseInt(s);
                case FLOAT:
                    return Float.parseFloat(getNumString(s));
                case DOUBLE:
                    return Double.parseDouble(getNumString(s));
                case LONG:
                    return Long.parseLong(s);
                case CHAR:
                    return (s==null) ? null: s.charAt(0);
                case STRING:
                    return s;
            }
        }
        catch(NumberFormatException nfe)
        {
            if (NOTANUMBER.matcher(s).matches())
            {
                switch(st)
                {
                    case INT:
                    case LONG:
                        return null;
                    case FLOAT:
                        return Float.NaN;
                    case DOUBLE:
                        return Double.NaN;
                }
                return null;
            }
            if (INFINITY.matcher(s).matches())
            {
                switch(st)
                {
                    
                    case INT:
                        return Integer.MAX_VALUE;
                    case FLOAT:
                        return Float.POSITIVE_INFINITY;
                    case DOUBLE:
                        return Double.POSITIVE_INFINITY;
                    case LONG:
                        return Long.MAX_VALUE;
                }

            }
            if(NEG_INFINITY.matcher(s).matches())
            {
                switch(st)
                {
                    
                    case INT:
                        return Integer.MIN_VALUE;
                    case FLOAT:
                        return Float.NEGATIVE_INFINITY;
                    case DOUBLE:
                        return Double.NEGATIVE_INFINITY;
                    case LONG:
                        return Long.MIN_VALUE;
                }

            }
            if(throw_exc)
            {
                //check if not a number pattern or infinity
                throw new RuntimeException("error at converting "+s+" to "+st.name+" :"+nfe.getMessage(), nfe);
            }
            
        }
        return null;

    }

    

    static public Double toDouble(Object o)
    {
        if (o instanceof Integer)
        {
            return 1.0+(Integer)o;
        }

        if (o instanceof Float)
        {
            return 1.0+(Integer)o;
        }
        if (o instanceof Double)
        {
            return (Double)o;
        }
        if (o instanceof String)
        {
            return (Double)SIMPLETYPE.DOUBLE.silent_converter.convert(getNumString((String)o));
            //return Double.parseDouble((getNumString((String)o).trim()));
        }

        throw new RuntimeException("cannot convert "+o+" to double");
    }
}

