package lmu.utils;

import java.util.*;
public class ThreadUtils
{
    static public abstract class CancellableThread extends Thread
    {
        public abstract void cancel();
    }

    public interface ThreadsFinishedCallback
    {
        public void ready(Collection<? extends Thread> threads);
    }
    public static void waitUntilFinished(Collection<? extends Thread> threads)
    {
        waitUntilFinished(threads, null);
    }
    public static void waitUntilFinished(Collection<? extends Thread> threads, ThreadsFinishedCallback callback)
    {
        waitUntilFinished(threads, false);
    }
    
    public static void waitUntilFinished(Collection<? extends Thread> threads, boolean start)
    {
        waitUntilFinished(threads, start, null);
    }
    public static void waitUntilFinished(Collection<? extends Thread> threads, boolean start, ThreadsFinishedCallback callback)
    {
        waitUntilFinished(threads, start, 500);
    }
    
    public static void waitUntilFinished(Collection<? extends Thread> threads, boolean start, int sleep_time)
    {
        waitUntilFinished(threads,start, sleep_time, null);
    }

    static class ThreadWaiter extends Thread
    {
        boolean start;
        Collection<? extends Thread> threads;
        int sleep_time;
        ThreadsFinishedCallback callback;

        ThreadWaiter(Collection<? extends Thread> threads, boolean start, int sleep_time, ThreadsFinishedCallback callback)
        {
            this.start  = start; this.threads = threads; this.sleep_time = sleep_time; this.callback = callback;
        }

        public void doit(boolean in_thread)
        {
            if (start)
            {
                for (Thread t : threads)
                {
                    t.start();
                }
            }
            int nalive = threads.size();
            boolean running = true;
            while (running)
            {
                running = false;
    
                for (Thread t : threads)
                {
                    if (!t.isAlive())
                        continue;
                    
                    running = true;
                    break;
                }
                if (!running)
                    break;
                
                try
                {
                    if (!in_thread)
                    {
                        Thread.sleep(sleep_time);
                    }
                    else
                    {
                        sleep(sleep_time);
                    }
                    
                }
                catch(InterruptedException iex)
                {
                    break;
                }
            }
            if(callback != null)
            {
                callback.ready(threads);
            }
        }

        public void run()
        {
            doit(true);
        }

        
    }

    @FunctionalInterface
    public interface DataHandlerFactory<E extends Object>
    {
        public DataHandler<E> createHandler();
    }

    @FunctionalInterface
    public interface DataHandler<E extends Object>
    {
        public void handle(E e);
    }

    public static class IteratorThread extends Thread
    {
        Logger log = LogConfig.getLogger();
        int nprocessed = 0;
        Iterator it;
        DataHandler handler;
        int report_after_n;
        public IteratorThread(Iterator it, DataHandler handler, int report_after_n)
        {
            this.it = it; this.handler = handler;
            this.report_after_n = report_after_n;
        }

        public void run()
        {
            if (report_after_n > 0)
            {
                log.info("lambda thread %s started", getId());
            }
            
            while(true)
            {
                Object data = null;
                synchronized(it)
                {
                    if (!it.hasNext())
                        break;
                    
                    data = it.next();
                }
                if (data == null)
                    continue;
                
                handler.handle(data);
                if (report_after_n > 0 )
                {
                    if (++nprocessed % report_after_n == 0)
                    {
                        log.info("%s processed %d data sets", getId(), nprocessed);
                    }
                }
                
            }
            if (report_after_n > 0 )
            {
            
                log.info("%s finished processed %d data sets", getId(), nprocessed);
            }
        }
    }
    public static<E extends Object> void runLambdas(Collection<E> data, DataHandler<E> function, int num_threads)
    {
        Iterator<E> it = data.iterator();
        runLambdas(it, function, num_threads, 0, true, -1, null);
    }
    public static<E extends Object> void runLambdas(Iterator<E> it, DataHandler<E> function, int num_threads)
    {
        runLambdas(it, function, num_threads, 0, true, -1, null);
    }
    
    public static<E extends Object> void runLambdas(Collection<E> data, DataHandler<E> function, int num_threads, int sleep_time, boolean blocking, int report_after_n,  ThreadsFinishedCallback callback)
    {
        Iterator<E> it = data.iterator();
        runLambdas(it, function, num_threads, sleep_time, blocking, report_after_n, callback);
    }

    public static<E extends Object> void runLambdas(Iterator<E> data, DataHandler<E> function, int num_threads, int sleep_time, boolean blocking, int report_after_n,  ThreadsFinishedCallback callback)
    {
        Vector<IteratorThread> threads = new Vector<>();
        for (int i=0; i<num_threads; i++)
        {
            threads.add(new IteratorThread(data, function, report_after_n));
        }
        waitUntilFinished(threads, true, sleep_time, blocking, callback);
    }

    public static<E extends Object> void runLambdas(Collection<E> data, DataHandlerFactory<E> factory, int num_threads, int sleep_time, boolean blocking, int report_after_n,  ThreadsFinishedCallback callback)
    {
        Iterator<E> it = data.iterator();
        runLambdas(it, factory, num_threads, sleep_time, blocking, report_after_n, callback);
    }

    public static<E extends Object> void runLambdas(Iterator<E> data, DataHandlerFactory<E> factory, int num_threads, int sleep_time, boolean blocking, int report_after_n,  ThreadsFinishedCallback callback)
    {
        Vector<IteratorThread> threads = new Vector<>();
        for (int i=0; i<num_threads; i++)
        {
            threads.add(new IteratorThread(data, factory.createHandler(), report_after_n));
        }
        waitUntilFinished(threads, true, sleep_time, blocking, callback);
    }

    public static void waitUntilFinished(Collection<? extends Thread> threads, boolean start, int sleep_time, ThreadsFinishedCallback callback)
    {
        waitUntilFinished(threads, start, sleep_time, true, callback);
    }
    public static void waitUntilFinished(Collection<? extends Thread> threads, boolean start, int sleep_time, boolean blocking, ThreadsFinishedCallback callback)
    {
        ThreadWaiter tw = new ThreadWaiter(threads, start, sleep_time, callback);
        if (blocking)        {
            tw.doit(false);
        }
        else
        {
            tw.start();
        }
   }
}
