package lmu.utils;

import java.util.*;
import java.io.*;

import javax.xml.stream.*;
import org.jdom.*;


public class XmlIndexer
{


    public interface XmlIndexHandler
    {
        public void onHit(String key, long start, long end);
    }

    public static class XmlIndexWriter implements XmlIndexHandler
    {
        PrintWriter pw ;
        File f;
        public XmlIndexWriter(File f) throws IOException
        {
            this.f=f;
            pw=new PrintWriter(f);
        }

        public void onHit(String key, long start, long end)
        {
            pw.printf("%s\t%d\t%d\n", key, start,end);
        }
        public void close() throws IOException
        {
            pw.close();
        }
    }
    public interface XmlIndexKeyExtractor
    {
        public String extractKey(XMLStreamReader xsr) throws XMLStreamException;
    }

    public static void readUntilClosed(XMLStreamReader xsr, String elemname, int num_open) throws XMLStreamException
    {
        while (xsr.hasNext())
        {
            int event = xsr.next();
            
            
            if (event==XMLStreamConstants.START_ELEMENT && xsr.getLocalName().equals(elemname))
            {
                num_open++;
                continue;
            }
            if(event!=XMLStreamConstants.END_ELEMENT || !xsr.getLocalName().equals(elemname))
                continue;
            
            if (--num_open==0) return;
        }
    }

    public static class XmlAttribReader implements XmlIndexKeyExtractor
    {
        String attribname;
        public XmlAttribReader(String attribname)
        {
            this.attribname=attribname;
        }
        public String extractKey(XMLStreamReader xsr) throws XMLStreamException
        {
            String rv = xsr.getAttributeValue(null, attribname);
//            for(int i=0; i<xsr.getAttributeCount(); i++)
//            {
//            	LogConfig.getLogger().info("trying to extract key ("+attribname+") from: "+xsr.getAttributeName(i));
//            }
//            LogConfig.getLogger().info("found value: "+rv);
            readUntilClosed(xsr, xsr.getLocalName(), 1);
            return rv;
        }
    }

    public static class XmlElementTextReader implements XmlIndexKeyExtractor
    {
        String elemname;
        public XmlElementTextReader(String elemname)
        {
            this.elemname=elemname;
        }
        public String toString()
        {
            return "subelemreader with elemname: "+elemname;
        }
        public String extractKey(XMLStreamReader xsr) throws XMLStreamException
        {
            String rv = null;

            String main_elem = xsr.getLocalName();
            int nopen=1;

            String lastname = main_elem;
            while (xsr.hasNext())
            {
                int event = xsr.next();
        
                if (event==XMLStreamConstants.END_ELEMENT && xsr.getLocalName().equals(main_elem))
                {
                    if (--nopen==0)
                    {
                        return null;
                    }
                    continue;
                }
                if (event==XMLStreamConstants.START_ELEMENT && xsr.getLocalName().equals(main_elem))
                {
                    nopen++;

                }

                if(event==XMLStreamConstants.CHARACTERS && lastname.equals(elemname))
                {
                    String key=xsr.getText();
                    readUntilClosed(xsr, main_elem, nopen);
                    return key;
                }
                if(event!=XMLStreamConstants.START_ELEMENT)
                {
                    continue;
                }
                lastname=xsr.getLocalName();
                //System.out.println("got lastname for subelem: "+lastname);
            }
            return null;
            
        }
    }

    public static void index(File xml, String elemname, XmlIndexKeyExtractor extractor, XmlIndexHandler handler)
    {
        index(xml,false, elemname, extractor, handler);
    }

    static class ConcatReader extends Reader
    {
        int pos;

        Reader[] reader;
        int index=0;
        int tread=0;
        Logger log = LogConfig.getLogger();
                          
        ConcatReader(Reader ... reader)
        {
            this.reader=reader;
        }

        public int read(char cbuf[], int off, int len) throws IOException
        {
            //log.info(String.format("request: cbuff. %d off: %d len: %d current reader: %d\n", cbuf.length, off, len, index));
            if (index>=reader.length)
            {
                return -1;
            }
            int nread=0;
            int niter=0;
            while (nread<len && index<reader.length)
            {
                if (++niter>10)
                {
                    throw new RuntimeException(String.format("after %d iteration, still reading  %d bytes from %d TOTAL FOR THIS: %d\n", niter, len-nread, index, tread));
                }
                
                int nnread = reader[index].read(cbuf,off+nread ,len-nread );
                /*if (niter>2)
                {
                    System.out.printf("after %d iteration, still reading  %d bytes from %d TOTAL FOR THIS: %d nnread: %d\n", niter, len-nread, index, tread, nnread);
                }
                */
                //System.out.println("read "+nnread);
                if (nnread<=0  && index<reader.length)
                {
                    reader[index].close();
                    index++;  tread=0;
                }
                if (nnread<=0 && index==reader.length-1)
                {
                    //System.out.printf("end of stream: "+nread);
                    return (nread==0)?((nnread==0)?nnread:-1):nread;
                }
                nread+=nnread;
                tread+=nnread;

                
            }
            pos+=nread;
            //System.out.printf("return %d index: %d/%d\n", nread, index, reader.length);
            return nread;
        }
        /**
         * Closes the stream and releases any system resources associated with
         * it.  Once the stream has been closed, further read(), ready(),
         * mark(), reset(), or skip() invocations will throw an IOException.
         * Closing a previously closed stream has no effect.
         *
         * @exception  IOException  If an I/O error occurs
         */
        public void close() throws IOException
        {
            reader[index].close();
        }
    }

    public static void index(File xml, boolean gzip, String elemname, XmlIndexKeyExtractor extractor, XmlIndexHandler handler)
    {
        XMLStreamReader xsr = null;
        
        org.apache.log4j.Logger log = LogConfig.getLogger();
        
        log.debug("will index xml:"+xml+" with elem: "+elemname+" extractor: "+extractor);
        int buffersize=10000;
        BufferedRandomAccessFile barf=null;
        try
        {
            barf = new BufferedRandomAccessFile(xml, "r", buffersize);
            if (gzip)
            {
                xsr = XMLInputFactory.newInstance().createXMLStreamReader(new java.util.zip.GZIPInputStream(new FileInputStream(xml)));
            }
            else
            {
                //xsr = XMLInputFactory.newInstance().createXMLStreamReader(new FileReader(xml));
                xsr = XMLInputFactory.newInstance().createXMLStreamReader(barf.getReader());
            }
        }
        catch(FileNotFoundException fne)
        {
            throw new RuntimeException("xml file not found: "+xml.getAbsolutePath(), fne);
        }
        catch(XMLStreamException e)
        {
            throw new RuntimeException("error:"+e.getMessage()+" at opening xmlfile "+xml.getAbsolutePath(), e);
        }
        catch(IOException ie)
        {
            throw new RuntimeException("error:"+ie.getMessage()+" at opening xmlfile "+xml.getAbsolutePath(), ie);
        }
        
        

        String HEADER=null;
        int firstreported=-1;
        boolean gotfirstelem=false;
        
        int nhits=0;
        //boolean test=true;
        try
        {
            
            int start=-1;
            String key=null;
            long current=0;
            int current_i=0;
            long last=0;
            
            
            while (xsr.hasNext())
            {
                int event = xsr.next();

                if (event!=XMLStreamConstants.START_ELEMENT || !xsr.getLocalName().equals(elemname) )
                {
                    if (start<0)
                        continue;
                    
                                        
                    int offset =xsr.getLocation().getCharacterOffset();
                    if (firstreported<0)
                    {
                        firstreported=start;
                    }


                    boolean reinit=(offset<0 || offset==Integer.MAX_VALUE);
                    //nhits++;
                    //reinit = (reinit || nhits%10==0);
                    

                    
                    
                    if (reinit)
                    {
                        
                        current+=start;
                        log.info("after "+nhits +" reinit xml by key "+key+" current: "+current);
                        xsr.close();
                        barf.close();

                        barf = new BufferedRandomAccessFile(xml, "r", buffersize);

                        if (HEADER==null)
                        {
                            HEADER = barf.readString(firstreported);
                            System.out.printf("got header: "+HEADER);
                            

                        }
                        barf.seek(current);
                        ConcatReader cr = new ConcatReader(new CharArrayReader(HEADER.toCharArray()), barf.getReader());
                        current-=HEADER.length();
                        xsr = XMLInputFactory.newInstance().createXMLStreamReader(cr);
                        start=-1;
                        continue;
                    }
                    //System.out.printf(nhits+"-th hit I: offset: %d %s %d-%d\n", offset, key, current_i+start,current_i+offset);
                    handler.onHit(key,current+start,current+offset);
                    nhits++;

                    start=-1;
                
                    continue;
                }
                start = xsr.getLocation().getCharacterOffset();
                
                key = extractor.extractKey(xsr);

            }
        }
        catch(XMLStreamException e)
        {
            e.printStackTrace();
            throw new RuntimeException("error:"+e.getMessage()+" while reading from xmlfile "+xml.getAbsolutePath()+
                                       " line: "+xsr.getLocation().getLineNumber()+":"+xsr.getLocation().getColumnNumber(), 
                                       e);
        }
        catch(IOException ie)
        {
            
            throw new RuntimeException("i/o error: "+ie.getMessage()+" while indexing "+xml.getAbsolutePath(), ie);
        }
    }


    public static class IndexCollector implements XmlIndexHandler
    {
        Vector<Pair<String, Pair<Long,Long>> > indeces = new Vector<Pair<String, Pair<Long,Long>> >();

        public void onHit(String key, long start, long end)
        {
            indeces.add(new Pair<String, Pair<Long,Long>>(key, new Pair<Long,Long>(start, end)));
        }
        public Vector<Pair<String, Pair<Long,Long>> > get()
        {
            return indeces;
        }
    }
    public static Vector<Pair<String, Pair<Long,Long>> > getKeyPositionsByAttrib(File xml, String elemname, String attrib )
    {
        IndexCollector ic = new IndexCollector();
        indexByAttrib(xml,elemname,attrib,ic);
        return ic.get();
    }

    public static Vector<Pair<String, Pair<Long,Long>> > getKeyPositionsBySubElem(File xml, String elemname, String subelem )
    {
        IndexCollector ic = new IndexCollector();
        indexBySubElem(xml,elemname,elemname,ic);
        return ic.get();
    }
    public static void indexByAttrib(File xml, String elemname, String attrib, XmlIndexHandler handler)
    {
        index(xml,xml.getName().endsWith(".gz"), elemname,new XmlAttribReader(attrib), handler);
    }

    public static void indexByAttrib(File xml, boolean gzipped, String elemname, String attrib, XmlIndexHandler handler)
    {
        index(xml,gzipped, elemname,new XmlAttribReader(attrib), handler);
    }

    public static void indexBySubElem(File xml, String elemname, String subelem, XmlIndexHandler handler)
    {
        indexBySubElem(xml,xml.getName().endsWith(".gz"),elemname , subelem, handler);
    }

    public static void indexBySubElem(File xml, boolean gzipped, String elemname, String subelem, XmlIndexHandler handler)
    {
        index(xml,gzipped, elemname,new XmlElementTextReader(subelem), handler);
    }


       
    static public StringBuffer extractBuffer(File f, long start, long end)
    {
        return extractBuffer(f,f.getName().endsWith(".gz"),start, end, false);
    }

    static public StringBuffer extractBuffer(File f, boolean gzip, long start, long end, boolean force_seek)
     {
        Logger log = LogConfig.getLogger();
        StringBuffer xbuff = new StringBuffer();
        try
        {
            
            
            
            
            
            char cdata[]=new char[(int)(end-start+1)];
            if (force_seek)
            {
                    
                RandomAccessFile raf = new RandomAccessFile(f, "r");
                raf.seek(start);
                byte bdata[] = new byte[(int)(end-start+1)];
                raf.read(bdata, 0, bdata.length);
                BufferedReader bfr = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(bdata)));
                String line;
                
                while (null!=(line=bfr.readLine()))
                {
                    xbuff.append(line);
                    xbuff.append('\n');
                
                }
                bfr.close();
                /*System.err.printf("extract xbuff l: %d end-start: %d :\n%s\n", xbuff.length(), 
                                  end-start, xbuff.toString());
                                  */
                raf.close();
                return xbuff;

            }
            else
            {
            
                FileInputStream fis = new FileInputStream(f);
                InputStream is = null;

                if(gzip)
                {
                    java.util.zip.GZIPInputStream gi = new java.util.zip.GZIPInputStream(fis);
                    gi.skip(start);
                    is=gi;
                }
                else
                {
                    fis.skip(start);
                    is=fis;
                }
            
                BufferedReader bfr = new BufferedReader(new InputStreamReader(is));
            
                bfr.read(cdata, 0, cdata.length);
                bfr.close();
                is.close();
            }

            BufferedReader bfr = new BufferedReader(new CharArrayReader(cdata));
            String line;
            
            while (null!=(line=bfr.readLine()))
            {
                xbuff.append(line);
                xbuff.append('\n');
            
            }
            bfr.close();
            

        }
        catch(IOException e)
        {
            String err=String.format("error:%s while reading %d-%d from %s\n", e.getMessage(), start, end, f.getAbsolutePath());
            log.warn(err);
            throw new RuntimeException(err, e);
        }
        return xbuff;

    }

    static  public Element extractElement(File xml, long start, long end)
    {
        return extractElement(xml, xml.getName().endsWith(".gz"), start, end);
    }

    
    static  public Element extractElementSeek(File xml, long start, long end )
    {
        StringBuffer xbuff = extractBuffer(xml,false,start,end, true);
        return XMLUtils.makeElement(xbuff.toString());
    }
        
    static  public Element extractElement(File xml, boolean gzip, long start, long end)
    {
        StringBuffer xbuff = extractBuffer(xml,gzip,start,end, true);
//        System.err.println(xbuff.toString());
        return XMLUtils.makeElement(xbuff.toString());

    }

	static java.util.regex.Pattern tabP = java.util.regex.Pattern.compile("\t");
	
    public static class XMLLookup extends FileLookup implements XmlIndexHandler
    {
        Logger log = LogConfig.getLogger();
        String attribName=null;
        String elemName=null;
        public HashMap<String, Pair<Long,Long> > lookups = null;

        public XMLLookup(File f, File indexF)
        {
        	super(f,indexF);
        }
        
        public XMLLookup(File f, File indexF, String attribName, String elemName)
        {
        	this(f,indexF);
        	this.attribName=attribName;
        	this.elemName = elemName;
        }
        

        public void initLookups()
        {
        	if(attribName==null)
                {
                    initLookups(false);
                }
        	else
                {
                    initLookups(elemName, attribName);
                }
        		
                log.debug("after init of lookup it contains "+lookups.size()+" keys");
    		changed=false;
        }
        
        public void initLookups(boolean force)
        {
            if (lookups!=null && !force)
                return;
            
            try
            {
                lookups =new HashMap<String, Pair<Long,Long> >();
                BufferedReader bfr = new BufferedReader(new FileReader(indexF));
                String line=null;
                while (null!=(line=bfr.readLine()))
                {
                    String sp[] = tabP.split(line);
                    if (sp.length != 3)
                    {
                        log.warn("unexpected line: >%s< in indexfile %s", line, indexF);
                        continue;
                    }
                    lookups.put(sp[0], new Pair<Long,Long>(Long.parseLong(sp[1]), Long.parseLong(sp[2])));
                }
                bfr.close();
            }
            catch(IOException e)
            {
                throw new RuntimeException("error:"+e.getMessage()+" while initing xml lookup from "+indexF.getAbsolutePath()+" index and "+f.getAbsolutePath()+" data file", e);
            }
        }
        public void initLookups(String elemName, String attrib)
        {
        	indexByAttrib(f, elemName, attrib, this);
        }

        
        public String getBuffer(String key)
        {
            if(lookups==null || changed)
            {
                initLookups();
            }

            if(lookups==null)
            {
            	ExceptionManager.throwRuntimeException("no keys found for element name "+elemName+" and attribute "+attribName+" in file "+f.getAbsolutePath());
            }
            
            Pair<Long,Long> p = lookups.get(key);
            if (p==null)
            {
                return null;
            }
            StringBuffer xbuff = extractBuffer(f, false, p.getFirst(), p.getSecond(), true);
            return xbuff.toString();
        }
        
        public Element get(String key)
        {
            String buf = getBuffer(key);
            return XMLUtils.makeElement(buf);
        }
        
        public Element get(String key, Pair<Long,Long> p)
        {
            if(p==null)
            {
                return null;
            }
            StringBuffer xbuff = extractBuffer(f, false, p.getFirst(), p.getSecond(), true);
            return XMLUtils.makeElement(xbuff.toString());
        }

		public void onHit(String key, long start, long end)
		{
			if(lookups==null)
			{
				lookups = new HashMap<String, Pair<Long,Long>>();
				LogConfig.getLogger().debug("inited new lookup");
			}
			LogConfig.getLogger().debug("putting "+key+" into lookup. Lookup size is now "+lookups.size());
			lookups.put(key, new Pair<Long,Long>(start, end));
		}
    }

    public static void main (String args[]) 
    {
        SimpleOptionParser cmd = new SimpleOptionParser("index", "xml", "key");
        cmd.setFile("index", "xml");

        if (!OptionParser.parseParams(args, true, false, true, cmd))
            return;
        
        XMLLookup lookup = new XMLLookup(cmd.getFile("xml"), cmd.getFile("index"));

        System.out.println(XMLUtils.prettyXML(lookup.get(cmd.getValue("key"))));

    }
}
