package lmu.utils.dag;

import java.io.*;
import java.sql.*;
import java.util.*;

import lmu.utils.*;
import lmu.utils.obo.*;
import lmu.utils.tuple.Tuple3;

import org.jdom.Element;

public class DAG implements java.io.Serializable
{
    Logger log = LogConfig.getLogger();
    String prefix;
    DbConnection con;

    static DbDefinition node_table = new DbDefinition("dagnodes", "id int primary key", "def longtext", "term char(255)", "name text", "description text", "children longtext", "parents longtext", "synonyms longtext",
                                              "xrefs longtext", DbDefinition.KEYDEFS, "term" );

    public String getPrefix()
    {
        return prefix;
    }

    public DbConnection getCon()
    {
        return con;
    }

    public String getNameSearchQuery()
    {
        return "select id from "+ prefix + node_table.getTableName() +" where name like ?;";
    }
    public void delete()
    {
        if (con == null || prefix == null)
        {
            return;
        }
        delete(con, prefix);
    }

    public static void delete(DbConnection con, String prefix)
    {
        con.execWithError("db dag with prefix: "+prefix,  "drop table if exists "+ prefix + node_table.getTableName()+";");
    }
    public static DAG readFromXML(File f)
    {
        Element root = XMLUtils.getRootElement(f);
        DAG dag = new DAG(XMLUtils.getAttrib(root, "name", null, true),
                          XMLUtils.getAttrib(root, "provider", null, true));

        dag.loadDAGFromXML(XMLUtils.getChildren(root, DAGEntry.ELEM_NAME)
                           .iterator());

        return dag;
    }

    public interface ObjectReader
    {
        public Object readEntry(DAGEntry de);
    }

    ObjectReader objReader = null;

    public void setObjectReader(ObjectReader o)
    {
        objReader = o;
    }

    public Element getXML()
    {
        Element dage = new Element("dag");
        dage.setAttribute("name", name);
        dage.setAttribute("provider", provider);

        Iterator<DAGEntry> it = getEntryIterator();

        while (it.hasNext())
        {
            DAGEntry de = it.next();
        
            Element e = de.toXML();
            dage.addContent(e);
        }
        return dage;

    }

    public String saveToDb(Connection con, String prefix)
    {
        return saveToDb(con,prefix, null);
    }
    public String saveToDb(Connection con, String prefix, File nfs_tmpdir)
    {
        try
        {
            if (prefix == null)
            {
                prefix = this.prefix;
            }
            
            
            
            Statement st = con.createStatement();
            if (prefix != null)
            {
                st.execute("drop table if exists " + prefix+ node_table.getTableName() + ";");
            }
            else
            {
                prefix = SQLUtils.generatePrefix(st, 10, node_table.getTableName());
            }
            this.prefix = prefix;

            String tablename = node_table.create(prefix, st);
            if (nfs_tmpdir != null)
            {
                File tmpf = null;
                try
                {
                    tmpf = File.createTempFile("dag_", ".tsv", nfs_tmpdir);
                }
                catch(java.io.IOException ie)
                {
                    throw new FRuntimeException("i/o error: %s while creating nfs-tmpdir file for dag: %s", ie, ie.getMessage(), this);
                }
                
                PrintWriter pw = FileUtils.getWriter(tmpf);
                                               
                final String LE =  "\n\n;;\n\n";
                pw.printf("%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s%s", -1, "", name, provider, "", 
                          "", "", "", "", LE);
                

                log.info("%s will save %d entries (%d not null)", this, idx2entry.size(), ObjectGetter.filter(idx2entry.values(), (DAGEntry de) -> de!=null).size());
                int nwritten = 0;
                //System.out.println("0::: " +idx2entry.get(0)+" id : " +idx2entry.get(0).id);
                for (DAGEntry de : idx2entry.values())
                {
                    if (de == null)
                    {
                        continue;
                    }
                    Element e = de.toXML();
                    pw.printf("%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s%s", de.id, XMLUtils.prettyXML(e), 
                              de.getTerm(), de.getName(), (de.description == null) ? "" : de.description,
                              StringUtils.joinObjects(",", de.getChildrenIndeces()),
                              StringUtils.joinObjects(",", de.getParentIndeces()),
                              StringUtils.joinObjects("|", de.alt_names),
                              StringUtils.joinObjects("|", de.xrefs),
                              LE
                              );
                    nwritten++;
                    /*
                    System.out.printf("%d: parents: %s children: %s\n", de.id,
                                      StringUtils.joinObjects(",", de.getParentIndeces()),
                              StringUtils.joinObjects(",", de.getChildrenIndeces())
                                      );
                                      */
                                      
                }

                pw.close();
                log.info("%s wrote %d dag entries to %s\n", this, nwritten, tmpf.getAbsolutePath());
                String lastq = "";
                try
                {
                    con.createStatement().execute(lastq = "load data infile '"+tmpf.getAbsolutePath()+"' into table "+ tablename +" fields terminated by '\\t'"+
                                                  " lines terminated by '\\n\\n;;\\n\\n';");
                }
                catch(SQLException se)
                {
                    throw new FRuntimeException("sql error: %s while loading dag %s via infile: %s", se, se.getMessage(), this, tmpf);
                }
                
                
                //System.out.printf("%s\n", lastq);
                //System.out.printf("write %s to %s:%s\n",  tmpf.getAbsolutePath(), provider, name);
                tmpf.delete();
                
                return prefix;
            }
            EPreparedStatement insert = node_table.prepareInsert(prefix, con, "id", "def", "term", "name", "description", "parents", "children", "synonyms", "xrefs");

            int colid = 0;
            insert.setInt(++colid, -1);
            insert.setString(++colid, "");
            insert.setString(++colid, name);
            insert.setString(++colid, provider);
            insert.setString(++colid, "");
            insert.setString(++colid, "");
            insert.setString(++colid, "");
            insert.setString(++colid, "");
            insert.setString(++colid, "");
            insert.execute();
            
            
            
            for (DAGEntry de : idx2entry.values())
            {
                if (de == null)
                {
                    continue;
                }
                saveDAGEntryToDB(de, prefix, con);

            }
        }
        catch (SQLException e)
        {
            throw new RuntimeException("sql error:" + e.getMessage()
                                       + " while saving dag: " + this, e);
        }
        return prefix;

    }
    
    protected void saveDAGEntryToDB(DAGEntry de, String prefix, Connection con)
    {
        saveDAGEntryToDB(de,prefix,con, null, null, true);
    }

    protected void saveDAGEntryToDB(DAGEntry de, String prefix, Connection con, Integer novel_parent, Integer novel_child, boolean add_novel)
    {
        Collection<Integer> v1 = de.getParentIndeces();
        if (novel_parent != null && novel_parent >=0)
        {
            if(add_novel)
            {
                v1.add(novel_parent);
            }
            else
            {
                v1.remove(novel_parent);
            }
            
        }
        Collection<Integer> v2 = de.getChildrenIndeces();
        if (novel_child != null)
        {
            if(add_novel)
            {
                v2.add(novel_child);
            }
            else
            {
                v2.remove(novel_child);
            }
            
        }
        
        deleteDAGEntryInDB(de);

        EPreparedStatement insert = node_table.prepareInsert(prefix, con, "id", "def", "term", "name", "description", "parents", "children", "synonyms", "xrefs");

        int colid = 0;
        insert.setInt(++colid, de.id);
        if (de.id < 0 )
        {
            insert.setString(++colid, "");
            insert.setString(++colid, name);
            insert.setString(++colid, provider);
            insert.setString(++colid, "");
        
            insert.setString(++colid, "");
            insert.setString(++colid, StringUtils.joinObjects(",", v2));
            insert.setString(++colid, "");
            insert.setString(++colid, "");
        }
        else
        {
            Element e = de.toXML(v1, v2);
            insert.setString(++colid, XMLUtils.prettyXML(e));
            insert.setString(++colid, de.getTerm());
            insert.setString(++colid, de.getName());
            insert.setString(++colid, de.description);
        
            insert.setString(++colid, StringUtils.joinObjects(",", v1));
            insert.setString(++colid, StringUtils.joinObjects(",", v2));
            insert.setString(++colid, StringUtils.joinObjects(",", de.alt_names));
            insert.setString(++colid, StringUtils.joinObjects(",", de.xrefs));
        }
        
        try
        {
            insert.execute();

        }
        catch (SQLException se)
        {
            throw new RuntimeException("sql error:" + se.getMessage()
                    + " while saving dagEntry: " + de, se);
        }
    }
    
    protected void updateDAGEntryInDB(DAGEntry de, Integer novel_parent, Integer novel_child, boolean add_novell)
    {
        saveDAGEntryToDB(de, prefix, con.getConnection(), novel_parent, novel_child, add_novell);
    }

    public void removeDAGEntry(DAGEntry del)
    {
        log.info("removeing "+del);
        term2entry.remove(del.getTerm());
        name2entry.remove(del.getName());
        if(del.getObject()!=null)
        {
            object2entry.remove(del.getObject());
        }
        for(Object label : del.labels)
        {
            label2entries.get(label).remove(del);
        }
        idx2entry.remove(del.id);
        for(DAGEntry p : del.getParents())
        {
            p.removeChild(del);
        }
        for(DAGEntry c : del.getChildren())
        {
            log.info("check if it has another parent: "+c);
            if(c.parents.size()==0)
            {
                removeDAGEntry(c);
                continue;
            }
            log.info("just remove reference");
            if(lookup_in_db)
            {
                updateDAGEntryInDB(c, del.id, null, false);
            }
            
        }
        if(lookup_in_db)
        {
            deleteDAGEntryInDB(del);
        }
    }
    
    protected void deleteDAGEntryInDB(DAGEntry de)
    {
        if (de == null || prefix == null || con == null)
            return;
        
        deleteDAGEntryInDB(de, prefix, con.getConnection());
    }
    protected void deleteDAGEntryInDB(DAGEntry de, String prefix, Connection con)
    {
        try
        {
            Statement delete = con.createStatement();
            String sql = "delete from "+prefix+node_table.getTableName()+" where id="+de.id+";";
            delete.execute(sql);
        }
        catch(SQLException se)
        {
            throw new RuntimeException("sql error:" + se.getMessage()
                    + " while deleting dagEntry: " + de, se);
        }
    }
    
    public static DAG readFromObo(String name, String provider, File obo_file,
                                  String root_elem_name)
    {
        return readFromObo(name, provider, obo_file, root_elem_name, null);
    }


    static DAGEntry initEntry(DAG dag, OboElement ch, boolean is_root)
    {
        Logger log = LogConfig.getLogger();
        HashMap<OboField, Vector<String>> entries = ch.getEntries();
        
        DAGEntry de = dag.addDAGEntry(is_root, ch.getName(), ch.getTerm());
        
        for (Iterator<OboField> it_entr = entries.keySet().iterator(); it_entr.hasNext();)
        {
            OboField f = it_entr.next();
            Vector<String> values = entries.get(f);
            if (values == null || values.size() ==0)
                continue;
            
            
            if (f == OboField.DEF)
            {
                de.description = StringUtils.join(" ", values);
                continue;
            }
            
            if (f == OboField.SYNONYM)
            {
                de.alt_names = new Vector<>();
                for (String syn : values)
                {
                    int pre = syn.indexOf('"');
                    if (pre < 0)
                        continue;
                    int pos = syn.lastIndexOf('"');
                    if (pos < 0)
                        continue;
                    
                    if(pre>=pos)
                    {
                        log.warn("invalid synonym: >%s< to %s",syn, de.getTerm());
                        continue;
                    }
                    String cleaned =syn.substring(pre+1, pos);
        
                    de.alt_names.add(cleaned);
                    
                }
                continue;
            }
            
            de.attribs.put(f.toString(), values);
        }
        return de;
    }
    public static DAG readFromObo(String name, String provider, File obo_file,
                                  String root_elem_name, HashMap<OboElement, DAGEntry> mapping)
    {
        Logger log = LogConfig.getLogger();
        DAG dag = new DAG(name, provider);

        OboReader or = new OboReader(obo_file);
        Vector<OboElement> roots = or.readTree();

        log.info("%s:%s got %d roots", name, provider, roots.size());
        if (mapping == null)
        {
            mapping = new HashMap<OboElement, DAGEntry>();
        }

        int skipped = 0;
        for (OboElement oe : roots)
        {
            if (oe.getType() != OboElementType.TERM)
                continue;

            if (root_elem_name != null && !oe.getName().equals(root_elem_name))
            {
                skipped++;
                // System.out.println("skip  :"+oe.getTerm()+" as not root ..:");
                continue;
            }
            DAGEntry root_da = initEntry(dag, oe, true);

            mapping.put(oe, root_da);

            Vector<OboElement> elems = new Vector<>();
            for (OboElement ch : OboReader.getTreeElems(oe))
            {
                elems.add(ch);
                if (oe == ch)
                    continue;
                mapping.put(ch, initEntry(dag, ch, false));
            }

            for (OboElement oee : elems)
            {
                if (oee.children == null)
                    continue;

                DAGEntry dep = mapping.get(oee);

                for (OboElement ch : oee.children)
                {
                    DAGEntry dech = mapping.get(ch);

                    dep.children.add(dech);
                    dech.parents.add(dep);
                }
            }
        }
        log.info("by reading " + root_elem_name + " skipped: "+ skipped + " entries got " + dag.idx2entry.size()+ " objects" +" not null: "+ ObjectGetter.filter(dag.idx2entry.values(), (DAGEntry de) -> de != null).size());

        return dag;
    }

    public void writeObo(PrintWriter pw)
    {
                throw new RuntimeException(
                                  "not yet implemented (check method with BufferedReader)");
        
    }
     /**
     * @param title
     * @param provider
     * @param oboInfile
     * @param terms
     * @return Returns a DAG containing only branches containing nodes which
     *         have been in the term list.
     */
    public static DAG createSlimDAG(String title, String provider,
                                    File oboInfile, HashSet<String> terms)
    {
        /** this method is incorrect */
        throw new RuntimeException("this method is incorrect... talk to gergely...");
        //return null;
        /*
        DAG inDAG = DAG.readFromObo(title, provider, oboInfile, null);

        DAG outDAG = new DAG(title, provider);
        HashSet<DAGEntry> newDAG = new HashSet<DAGEntry>();

        // collect all nodes dependent from the term entries
        for (String act_node : terms)
        {

            DAGEntry child = inDAG.term2entry.get(act_node);
            if (child != null)
            {
                if (newDAG.contains(child))
                {
                    continue;
                }

                newDAG.add(child);

                HashSet<DAGEntry> parents = child.parents;
                if (parents == null || parents.size() == 0)
                {
                    continue;
                }
                // get all parents up to the root for the child
                for (Iterator<DAGEntry> it_de = child.getAscendetsIterator(-1); it_de
                    .hasNext();)
                {
                    DAGEntry part = it_de.next();

                    newDAG.add(part);

                }

            }
        }

        for (DAGEntry entry : newDAG)
        {

            // change parents, children by removing all entries which are not in
            // the slim dag
            HashSet<DAGEntry> kids = new HashSet<DAGEntry>();
            for (DAGEntry och : entry.children)
            {

                if (newDAG.contains(och))
                {
                    DAGEntry newEntry = outDAG.name2entry.get(och.getName());
                    if (newEntry == null)
                    {
                        newEntry = outDAG.addDAGEntry(och);
                    }

                    kids.add(newEntry);
                }
            }

            HashSet<DAGEntry> parents = new HashSet<DAGEntry>();
            for (DAGEntry opar : entry.parents)
            {

                if (newDAG.contains(opar))
                {
                    DAGEntry newEntry = outDAG.name2entry.get(opar.getName());
                    if (newEntry == null)
                    {
                        newEntry = outDAG.addDAGEntry(opar);
                    }

                    parents.add(newEntry);
                }
            }
            DAGEntry newEntry = outDAG.name2entry.get(entry.getName());
            if (newEntry == null)
            {
                newEntry = outDAG.addDAGEntry(entry);
            }

            newEntry.children = kids;
            newEntry.parents = parents;

        }
        return outDAG;
        */
    }

    static class SQLDAGEntryElemIterator implements Iterator<Element>
    {
        ResultSet rs;
        Element nextE = null;

        public SQLDAGEntryElemIterator(ResultSet rs)
        {
            this.rs = rs;
            next();
        }

        @Override
        public boolean hasNext()
        {
            return nextE != null;
        }

        @Override
        public Element next()
        {
            Element rv = nextE;
            nextE = null;
            try
            {
                if (!rs.next())
                {
                    return rv;
                }
                nextE = XMLUtils.makeElement(rs.getString(2));
            }
            catch (SQLException e)
            {
                throw new RuntimeException("sql error: " + e.getMessage()
                                           + " while loading next dag entry xml", e);
            }
            return rv;

        }

        @Override
        public void remove()
        {
            throw new UnsupportedOperationException();
        }
    }

    DAGEntry getSingle(HashMap<String, HashSet<DAGEntry>> l, String key)
    {
        return getSingle(l,key, false);
    }
    
    DAGEntry getSingle(String keyinfo, HashSet<DAGEntry> hs, boolean check_uniq)
    {
        if (hs == null || hs.size() == 0)
            return null;
        if (check_uniq && hs.size() > 1)
            throw new FRuntimeException("%s is not uniq in %s:%s", keyinfo, provider, name);
        
        return hs.iterator().next();
    }

    DAGEntry getSingle(HashMap<String, HashSet<DAGEntry>> l, String key, boolean check_uniq)
    {
        return getSingle(key, l.get(key), check_uniq);
    }

    private void updateLookup(DAGEntry de, HashMap<String, HashSet<DAGEntry>> l, String oldkey, String newkey)
    {
        HashSet<DAGEntry> hs = null;
        if (oldkey !=null )
        {
            hs = l.get(oldkey);
            if (hs != null)
            {
                hs.remove(de);
                if (hs.size() == 0)
                {
                    l.remove(oldkey);
                }
            }
        }

        if (newkey != null)
        {
            hs = l.get(newkey);
    
            if (hs == null)
            {
                l.put(newkey, hs = new HashSet<>());
            }
    
            hs.add(de);
        }
    }
    
    
    private void loadDAGFromXML(Iterator<Element> dagentry_it)
    {
        AutoResizeVector<Vector<Integer>> parents = new AutoResizeVector<>();
        AutoResizeVector<Vector<Integer>> children = new AutoResizeVector<>();

        Vector<DAGEntry> v = new Vector<DAGEntry>();

        maxid = 0;

        while (dagentry_it.hasNext())
        {
            Vector<Integer> ps = new Vector<Integer>();
            Vector<Integer> chs = new Vector<Integer>();
            DAGEntry de = new DAGEntry(dagentry_it.next(), ps, chs);
            de.dag = this;
            idx2entry.put(de.id, de);
            maxid = Math.max(maxid, de.id);
            v.add(de);
            parents.set(de.id, ps);
            children.set(de.id, chs);
            

            if (de.is_root)
            {
                root.children.add(de);
            }
            updateLookup(de, term2entry, null, de.getTerm());
            updateLookup(de, name2entry, null, de.getName());

        }
        maxid++;
        
        
        for (int i = 0; i < v.size(); i++)
        {
            DAGEntry de = v.get(i);
            Vector<Integer> ps = parents.get(de.id);
            Vector<Integer> chs = children.get(de.id);

            for (int p : ps)
            {
                if (p < 0)
                {
                    de.parents.add(root);
                    continue;
                }
                de.parents.add(idx2entry.get(p));
                
                //log.debug("set parent to %s:  %s\n", de.getName(), idx2entry.get(p).getName());
            }

            for (int c : chs)
            {
                log.debug("set child to %s:  %s\n", de.getName(), idx2entry
                          .get(c).getName());
                de.children.add(idx2entry.get(c));
            }
        }
    }

    String tblname = null;
    

    

    public boolean isInDbLookupMode()
    {
        return lookup_in_db;
    }

    public void setLookupInDbMode(boolean b)
    {
        if (this.con == null || this.prefix == null || tblname == null)
        {
            if (!b)
                return; //not in db mode and can never be
            
            throw new RuntimeException("not in db mode");
        }
        lookup_in_db = b;
    }
    public DAG(DbConnection con, String prefix, boolean loadfully)
    {
        this(con.getConnection(), prefix, loadfully);
        this.con = con;
        tblname = prefix + node_table.getTableName();
        lookup_in_db = !loadfully;

    }
    
    public DAG(Connection con, String prefix)
    {
        this(con, prefix, true);
    }
    
    public DAG(DbConnection con, String prefix, boolean loadfully, String name, String provider)
    {
        this(name, provider);
        saveToDb(con.getConnection(), prefix);
        lookup_in_db = !loadfully;
        this.prefix=prefix;
        tblname = prefix + node_table.getTableName();
        this.con=con;
    }
    
    public static boolean dagTableExists(Connection con, String prefix)
    {
        String sql = "show tables like '"+prefix+node_table.getTableName()+"';";
        try
        {
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(sql);
            return  rs.first();
        }
        catch(SQLException se)
        {
            ExceptionManager.handleSQLException(LogConfig.getLogger(), "unable to check for dag table", se);
        }
        return false;
    }

    DAG(Connection con, String prefix, boolean loadfully)
    {
        this.name = null;
        this.prefix = prefix;

        SQLUtils.apply(DbConnection.query(con, "select term, name from " + prefix + node_table.getTableName() +" where id = -1"),
                              (ResultSet rs) -> { this.name = rs.getString(1); this.provider = rs.getString(2);});
        
        root.dag = this;
        root.id = -1;
        //System.out.println("select name, term from " + prefix + node_table.getTableName() +" where id = -1");
        //System.out.println("in dag constructor: name"+name+" prov: "+provider);
        if (name == null)
        {
            throw new RuntimeException("no entry for dag definition!");
        }
        lookup_in_db = !loadfully;

        if (!loadfully)
            return;
        
        ResultSet rs = DbConnection.query(con, "select id, def from " + prefix + node_table.getTableName() +" where id >=0;");
        loadDAGFromXML(new SQLDAGEntryElemIterator(rs));

    }

    public DAG(String name, String provider)
    {
        this.name = name;
        this.provider = provider;
        root.dag = this;
        root.id = -1;

    }

    boolean lookup_in_db = false;
    
    public String provider;
    public String name;
    public HashSet<Object> labels = new HashSet<Object>();
    DAGEntry root = new DAGEntry();

    protected HashMap<Object, DAGEntry> object2entry = new HashMap<>();
    private HashMap<String, HashSet<DAGEntry>> term2entry = new HashMap<>();
    private HashMap<String, HashSet<DAGEntry>> name2entry = new HashMap<>();
    protected HashMap<Object, Vector<DAGEntry>> label2entries = new HashMap<Object, Vector<DAGEntry>>();

    //private AutoResizeVector<DAGEntry> idx2entry = new AutoResizeVector<>();
    int maxid = 0;
    private HashMap<Integer, DAGEntry> idx2entry = new HashMap<>();

    public void clearStructure()
    {
        if (!lookup_in_db)
        {
            for(DAGEntry entry : idx2entry.values())
            {
                entry.children.clear();
                entry.parents.clear();
            }
            return;
        }

        throw new RuntimeException("not yet implemented! (for what do you need it anyways?)");
    }

    public interface Applyer
    {
        void apply(DAGEntry de);
    }

    public void apply(Applyer a)
    {
        ObjectGetter.apply(getEntryIterator(), (DAGEntry de) -> a.apply(de));
    }
    public Iterator<String> getEntryNameIterator()
    {
        if (!lookup_in_db)
            return name2entry.keySet().iterator();

        return SQLUtils.transform(con.query("select name from "+tblname+" where id>=0;"), (ResultSet rs) -> rs.getString(1));
    }

    public Iterator<String> getEntryTermIterator()
    {
        if (!lookup_in_db)
            return term2entry.keySet().iterator();
        
        return SQLUtils.transform(con.query("select term from "+tblname+" where id>=0;"), (ResultSet rs) -> rs.getString(1));
    }
    
    public Iterator<Integer> getEntryIdIterator()
    {
        if (!lookup_in_db)
            return idx2entry.keySet().iterator();
        
        return SQLUtils.transform(con.query("select id from "+tblname+" where id>=0;"), (ResultSet rs) -> rs.getInt(1));
    }
    public Iterator<DAGEntry> getEntryIterator()
    {
        if (!lookup_in_db)
            return idx2entry.values().iterator();
        
        return SQLUtils.transform(con.query("select id from "+tblname+" where id>=0;"), (ResultSet rs) -> getByIdx(rs.getInt(1)));
    }
    
    /**
     * get iterator of basic infos (idx, term and name) for all dag entries
     */
    public Iterator<Tuple3<Integer,String,String>> getInfos()
    {
        return SQLUtils.transform(con.query("select id,term,name from "+tblname+" where id>=0;"), (ResultSet rs) -> Tuple3.create(rs.getInt(1),rs.getString(2),rs.getString(3)));
    }
    
    public Tuple3<Integer,String,String> getInfos(int idx)
    {
        try
        {
            ResultSet rs = con.query("select id,term,name from "+tblname+" where id="+idx+";");
            if(rs.next())
            {
                return Tuple3.create(rs.getInt(1),rs.getString(2),rs.getString(3));
            }
        }
        catch(SQLException e)
        {
            ExceptionManager.handleSQLException(log, "unable to get entry with id "+idx+" from db", e);
        }
        return null;
    }

    public int getNumEntries()
    {
        if (!lookup_in_db)
        {
            //log.info("getNumEntries: "+idx2entry);
            return idx2entry.size();
        }
        //log.info(" db query  for entries..");
        return con.queryInt("select count(1) from "+tblname+" where id>=0;");
        
    }
    protected void addTermReference(String oldterm, String term, DAGEntry de)
    {
        updateLookup(de, term2entry, oldterm, term);
    }

    public DAGEntry getByIdx(int idx)
    {
        return getByIdx(idx, true);
    }
    public DAGEntry getByIdx(int idx, boolean error_if_missing)
    {
        if (!lookup_in_db)
            return idx2entry.get(idx);
        

        if (idx < 0)
            return root;
        
        try
        {
            if (!error_if_missing)
            {
                if(con.isEmpty("select 1 from "+tblname+" where id ="+idx+";"))
                    return null;
            }
            String q = "select def from "+tblname+" where id="+ idx +";";
            String def = con.queryWithError(q, true, "unknown db id: %d in dag: %s", idx, this).getString(1);
                        
            Element e = null;
            //TODO remove
            try
            {
                e= XMLUtils.makeElement(def);
            }
            catch(Exception ex)
            {
                ex.printStackTrace();
                ExceptionManager.throwRuntimeException("unable to load dagentry with id "+idx+" from "+tblname);
            }
            DAGEntry de =  new DAGEntry(e, null, null);
            
            de.dag = this;
            if (objReader != null)
            {
                de.setObject(objReader.readEntry(de));
            }
            return de;
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while getting dag entry: %d from db for %s",  se, se.getMessage(), idx, this);
        }
        
    }

    public void addRoot(DAGEntry de)
    {
        root.addChild(de);
    }
    
    protected void addNameReference(String oldname, String name, DAGEntry de)
    {
        updateLookup(de, name2entry, oldname, name);
    }

    public DAGEntry lookup(String n, String term)
    {
        DAGEntry de = null;

        if (n!=null && (null != (de = lookup(n, false))))
            return de;
        
        return lookup(term,true);
    }
    public DAGEntry lookup(String n, boolean by_term)
    {
        return (by_term) ? getByTerm(n, true, true) : getByName(n);
    }
    
    public DAGEntry getOrCreate(String n, String term, boolean is_root)
    {
        DAGEntry de = lookup(n,term);
        return de==null ? addDAGEntry(is_root, n, term) : de;
    }

    public HashSet<DAGEntry> getByName(String name, boolean only_first)
    {
        if (!lookup_in_db)
            return name2entry.get(name);


        try
        {
            String q = "select def from "+tblname+" where name='"+ name +"'";
            if (only_first)
            {
                q +=" limit 1";
            }

            ResultSet rs  = con.query(q, false, "");
            //ResultSet rs = con.queryWithError(q+";", false, "");
            if (!rs.first())
                return null;
    
            return readEntries(rs);            
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while getting dag entry: %s from db for %s",  se, se.getMessage(), name, this);
        }
    
            
        
    }
    
    public DAGEntry getByName(String name)
    {
        return ObjectGetter.first(getByName(name, true));
    }

    protected Vector<Integer> getLinks(int dbid, boolean parental)
    {
        System.out.println(dbid);
        if (dbid < 0)                        
        {
            if(parental)
                return new Vector<Integer>();
            return ObjectGetter.filter(SQLUtils.map(con.query("select id  from "+tblname+" where parents='';"), (ResultSet rs) -> rs.getInt(1)), (Integer idx) -> idx!=-1);
        }

        try
        {
            
            String q = null;
            String t = (parental) ? "parents":"children";
            ResultSet rs = con.queryWithError("select " + t + "  from "+tblname+" where id="+ dbid +";", false, "unknown db id: %d in dag: %s", dbid, this);
            if(!rs.first())
                return new Vector<Integer>();
            String def = rs.getString(1);
            return StringUtils.convert_int_vector(def, ",");
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while getting dag entry: %d from db for %s",  se, se.getMessage(), dbid, this);
        }


    }

    HashSet<DAGEntry> readEntries(ResultSet rs)
    {
        SQLUtils.reset(rs);
        return SQLUtils.mapToSet(rs, (ResultSet rrs) -> 
                         {
                            DAGEntry de =  new DAGEntry(XMLUtils.makeElement(rrs.getString(1)), null, null);
                            de.dag = this;
                            return de;
                          });
    }
    
    public int getIdForTerm(String term)
    {
        if (!lookup_in_db)
            return getSingle(term2entry, term, true).id;
        
        try
        {
            String q = "select id from "+tblname+" where term='"+ term +"' limit 2";
            
            ResultSet rs  = con.query(q, false, "");
            if (!rs.first())
                ExceptionManager.throwRuntimeException("no unique id for term "+term);
            
            return rs.getInt(1);
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while getting dag entry by term: %s from db for %s",  se, se.getMessage(), term, this);
        }
        
    }

    public DAGEntry getByTerm(String term)
    {
        return getByTerm(term, true, true);
    }

    public DAGEntry getByTerm(String term, boolean only_first, boolean uniq)
    {
        if (!lookup_in_db)
            return getSingle(term2entry, term, uniq);
        
            
        
        try
        {
            String q = "select def from "+tblname+" where term='"+ term +"'";
            if (only_first)
            {
                q +=" limit " + ((uniq) ? "2" : "1");
            }

            
            ResultSet rs  = con.query(q, false, "");
            if (!rs.first())
                return null;
            
            return getSingle(term, readEntries(rs), uniq);
            
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while getting dag entry by term: %s from db for %s",  se, se.getMessage(), term, this);
        }
    
        
    }

    public DAGEntry getByObject(Object o)
    {
        if (lookup_in_db)
        {
            //error - > cannot do as not completely in memory
            throw new FRuntimeException("%s:%s cannot lookup dag entry by object as lookup is done by sql", name, provider);
        }
        return object2entry.get(o);
    }

    public Vector<DAGEntry> getByLabel(Object label)
    {
        if (lookup_in_db)
        {
            //error - > cannot do as not completely in memory
            throw new FRuntimeException("%s:%s cannot lookup dag entry by object as lookup is done by sql", name, provider);
        }
        return label2entries.get(label);
    }





    public DAGEntry getRoot()
    {
        return root;
    }

    public int getSize()
    {
        return getNumEntries();
    }

    private int updateDescendantNum(HashMap<DAGEntry, Integer> map, DAGEntry e)
    {
        Integer v = map.get(e);
        if (v != null)
        {
            return v;
        }
        int ndesc = e.children.size();
        for (DAGEntry de : e.children)
        {
            ndesc += updateDescendantNum(map, de);
        }
        map.put(e, ndesc);
        return ndesc;
    }

    public HashMap<DAGEntry, Integer> getDAG2DescentantNum()
    {
        HashMap<DAGEntry, Integer> rv = new HashMap<>();

        for (DAGEntry de : root.children)
        {
            updateDescendantNum(rv, de);
        }
        return rv;
    }

    

    /**
     * objects are pairs: name and children defs
     */
    public Vector<DAGEntry> createByDefinition(boolean is_root, Object... def)
    {
        Vector<DAGEntry> rv = new Vector<>();
        for (int i = 0; i < def.length; i += 2)
        {
            String name = (String) def[i];
            Object[] children_def = (Object[]) def[i + 1];
            DAGEntry de = getSingle(name2entry,name);
            if (de == null)
            {
                de = addDAGEntry(is_root, name, null);
            }
            rv.add(de);
            if (children_def == null)
            {
                continue;
            }
            Vector<DAGEntry> children = createByDefinition(false, children_def);
            for (DAGEntry chd : children)
            {
                de.addChild(chd);
            }

        }
        return rv;
    }

    public DAGEntry addDAGEntry(boolean is_root, String name, String term, int idx)
    {
        //log.info("add entriy %s:%s idx: %d", name, term, idx);
        DAGEntry da = new DAGEntry();
        da.is_root = is_root;
        da.dag = this;
        da.setName(name);
        da.setTerm(term);
        da.id = idx;
        
        if (!lookup_in_db)
        {
            idx2entry.put(da.id, da);
            maxid = Math.max(maxid, da.id)+1;
            
        }
        
        
        if (is_root)
        {
            root.children.add(da);
        }
        return da;
    }

    int getLastIdx()
    {
        if (!isInDbLookupMode())
        {
            return maxid;
        }
        return con.queryInt("select max(id)+1 from "+tblname);
    }

    public DAGEntry addDAGEntry(boolean is_root, String name, String term)
    {
        DAGEntry da = addDAGEntry(is_root, name,term, getLastIdx());
        maxid = Math.max(da.id + 1, maxid);
        if(isInDbLookupMode())
        {
            saveDAGEntryToDB(da, prefix, con.getConnection());
        }
        return da;
    }
    
    /*
    public DAGEntry addDAGEntry(DAGEntry de)
    {
        addDAGEntry(de, false);
    }

    public DAGEntry addDAGEntry(DAGEntry de, boolean isRoot)
    {
        DAGEntry da = new DAGEntry();
        // da = de;

        da.is_root = isRoot;
        da.dag = this;
        da.setName(de.getName());
        da.setTerm(de.getTerm());
        da.id = idx2entry.size();

        idx2entry.add(da);

        da.children = new HashSet<>();
        da.children.addAll(de.children);
        da.parents = new HashSet<>();
        if (!isRoot)
        {
            da.parents.addAll(de.parents);
        }
        da.attribs = new HashMap<>();
        da.attribs.putAll(de.attribs);
        da.labels = new HashSet<>();
        da.labels.addAll(de.labels);
        if (da.is_root)
        {

            root = da;
        }
        return da;
    }
    

    */

    public void insertDAGEntryInBetween(DAGEntry insert, DAGEntry child,
                                        DAGEntry parent)
    {
        parent.children.remove(child);
        child.parents.remove(parent);
        parent.addChild(insert);
        insert.addChild(child);
    }
    

    void removeLabel(DAGEntry de, Object o)
    {
        Vector<DAGEntry> v = label2entries.get(o);
        if (v == null)
            return;
        v.remove(o);
        if (v.size() == 0)
        {
            label2entries.remove(o);
        }

    }

    public void addLabels(DAGEntry de, Object... labels)
    {
        addLabels(de, new CollectionIterator(labels));
    }

    public void addLabels(DAGEntry de, Collection labels)
    {
        addLabels(de, labels.iterator());
    }

    public void addLabels(DAGEntry de, Iterator labels)
    {
        while (labels.hasNext())
        {
            Object o = labels.next();
            Vector<DAGEntry> v = label2entries.get(o);
            if (v == null)
            {
                label2entries.put(o, v = new Vector<>());
            }
            v.add(de);
        }
    }

    private DAGEntry getEqualEntry(DAGEntry de)
    {
        DAGEntry candidate = getByTerm(de.getTerm());
        if(candidate==null)
            return null;
        if(!candidate.getName().equals(de.getName()) && candidate.getObject().equals(de.getObject()))
                return null;
        return candidate;
    }
    public DAGEntry copySubDAG(DAGEntry target, DAGEntry src)
    {
        DAGEntry nsrc = addDAGEntry(target == root, src.getName(), src.getTerm());
        if (target != root)
        {
            target.addChild(nsrc);
        }
        
        src.cloneInfos(nsrc);
        HashMap<Integer, DAGEntry> m = new HashMap<>();

        log.trace("src id: %d : %s", src.id, src);
        m.put(src.id, nsrc);

        
        Iterator<DAGEntry> it = src.getDescendatsIterator(-1);
        while (it.hasNext())
        {
            DAGEntry de = it.next();
            DAGEntry cloned;
            if(m.containsKey(de.id))
                cloned = m.get(de.id);
            else if((cloned=getEqualEntry(de))==null)
            {
                cloned = addDAGEntry(false, de.getName(), de.getTerm());
                de.cloneInfos(cloned);
            
                log.trace("next descendant: " +de.id+ " parents: "+de.getParentIndeces());
                m.put(de.id, cloned);
            }
            Vector<Integer> pids = ObjectGetter.map(de.getParentIndeces(), (Integer idx) -> idx);

            
            if (de.is_root)
            {
                pids.add(-1);
            }
             
            for (Integer pid : pids)
            {
                DAGEntry cp = m.get(pid);
                
                if (cp == null)
                    continue;
                
                log.trace(" add cloned edge: %d-%d to %d-%d", pid, de.id, cp.id, cloned.id); 
                cp.addChild(cloned);
            }

        }
        
        return nsrc;
    }

    @Override
    public String toString()
    {
        return String.format("dag: %d %s:%s", hashCode(), name, provider);
    }

    public DAGEntry getByPath(boolean create_if_not_there, Collection<String> path)
    {
        return getByNamePath(create_if_not_there, path.iterator());
    }
    public DAGEntry getByNamePath(boolean create_if_not_there, Iterator<String> path)
    {
        DAGEntry de = getRoot();
        boolean first = true;
        while (path.hasNext())
        {
            String n = path.next();
            DAGEntry ch = ObjectGetter.first(ObjectGetter.filter(de.getChildren(), (DAGEntry chde) -> chde.getName().equals(n)));
                
            if (ch != null)
            {
                first = false;
                de = ch;
                continue;
            }

            if (!create_if_not_there)
                return null;
            
            
            ch = addDAGEntry(first, n, null);
                
            first = false;
            de.addChild(ch);
            de = ch;
        }
        return de;
    }

    public DAGEntry getByNamePath(boolean create_if_not_there, String ... path)
    {
         return getByNamePath(create_if_not_there, new CollectionIterator<String>(path));
    }

    public static DAG buildDAGByNamePathes(String name, String provider, Collection<String[]> c)
    {
        return buildDAGByNamePathes(name,provider, c.iterator());
    }

    public static DAG buildDAGByNamePathes(String name, String provider, Iterator<String[]> pathIt)
    {
        DAG d = new DAG(name,provider);

        while (pathIt.hasNext())
        {
            d.getByNamePath(true, pathIt.next());
        }
        return d;
    }

    public static DAG buildDAGByNamePathes(String name, String provider, File f, String record_sep, int record_idx, String path_sep)
    {
        FileUtils.RecordIterator splitter = new FileUtils.RecordIterator(path_sep);
        Iterator<String[]> it = FileUtils.getFieldSetIterator(f, record_sep);
        Vector<String[]> pathes = new Vector<>();
        while (it.hasNext())
        {
            String[] records = it.next();
            if (record_idx >= records.length)
            {
                throw new FRuntimeException("invalid record idx: %d for file %s with separator >%s< got %d records for %s",
                                            record_idx, f, record_sep, records.length, StringUtils.join(record_sep, records));
            }
            String r = records[record_idx];
            String[] sp = splitter.split(r);
            pathes.add(sp);
        }
        return buildDAGByNamePathes(name,provider, pathes);
    }

    public static DAG buildDAGByNamePathes(String name, String provider, DataTable dt, String path_header_name, String path_sep)
    {
        DAG d = new DAG(name,provider);
        FileUtils.RecordIterator splitter = new FileUtils.RecordIterator(path_sep);
        for (int i=0; i<dt.getSize(); i++)
        {
            String sp[] = splitter.split(""+dt.getDataAt(i).get(path_header_name));
            DAGEntry de = d.getByNamePath(true, sp);
            de.setObject(i);
        }
        return d;
    }

    public static Object getTableData(DAGEntry de, DataTable dt, String header, NumUtils.AGGREGATE aggr_function)
    {
        Integer idx = (Integer)de.getObject();
        
        if (idx != null)
            return dt.getDataAt(idx).get(header);
        
        
        if (idx == null && aggr_function == null)
            return null;
    
        Vector data = new Vector<>();
        Iterator<DAGEntry> it = de.getDescendatsIterator(-1);
        while (it.hasNext())
        {
            DAGEntry e = it.next();
            Integer chidx = (Integer)e.getObject();
            if (chidx == null)
                continue;

            data.add(dt.getDataAt(chidx).get(header));

        }
        return NumUtils.reduce(data, aggr_function);
    }
    
    public static void main(String args[])
    {

        SQLConnectionOptionParser sql = new SQLConnectionOptionParser();
        sql.setOptional(true);
        SimpleOptionParser cmd = new SimpleOptionParser("obo", "fromxml", "toxml", "name", "provider", "test", "testadd", "show");
        cmd.setFile("obo", "fromxml");
        cmd.setOptional("obo", "fromxml", "name", "provider", "toxml", "testadd", "show");
        cmd.setSwitches("test");
        cmd.setRequiredSubParamSets(ObjectGetter.toArray("obo"),
                                    ObjectGetter.toArray("fromxml"), ObjectGetter.toArray("test"));
                                    

        if (!OptionParser.parseParams(args, true, false, true, cmd, sql))
            return;

        String prefix = cmd.getOptionalValue("prefix", "testpref");
        String n = cmd.getOptionalValue("name", null);
        String p = cmd.getOptionalValue("provider", null);
        if (cmd.isSet("test"))
        {
            /*mapping.OrganismMapping om = config.ConfigManager.getGeneralOrganismMapping();
            
            DAG d = null;
            if (!om.doesDAGExists(n, p))
            {
                d = new DAG(n, p);
                om.saveDAG(d);
            }
            d = om.getDAG(n, p, false);

            if (cmd.isOptionSet("testadd"))
            {
                String en = cmd.getValue("testadd");
                d.addDAGEntry(true, en, en);
            }
            if (cmd.isOptionSet("show"))
            {
                System.out.println(d.getByName(cmd.getValue("show")));
            }
            
            return;
            */
        }
        DAG d = null;

        if (cmd.isOptionSet("obo"))
        {
            d = readFromObo("test", "test", cmd.getFile("obo"), null);
        }
        if (cmd.isOptionSet("fromxml"))
        {
            d = readFromXML(cmd.getFile("fromxml"));
        }

        PrintWriter pw = cmd.getWriter("toxml");

        XMLUtils.writeXML(d.getXML(), pw);

        pw.close();

    }
}
