package lmu.utils.graph;

import lmu.utils.*;
import java.util.*;

public class GraphUtils
{

    public interface EdgeGetter<A>
    {
        Iterator<A> getEdges(A n);
    }

    static public<A>  HashMap<A, HashSet<A>> getEdges(Iterator<A> nodes, EdgeGetter<A> edgeGetter)
    {
        HashMap<A, HashSet<A>> rv = new HashMap<>();
        while (nodes.hasNext())
        {
            A n = nodes.next();
            rv.put(n, ObjectGetter.toSet(edgeGetter.getEdges(n)));
        }
        return rv;
    }
    
    static public<A> HashMap<A,HashSet<A>> getTransitiveEdges(Iterator<A> nodes, EdgeGetter<A> edgeGetter)
    {
        return getTransitiveEdges(getEdges(nodes, edgeGetter));
    }
    
    static public<A> HashMap<A,HashSet<A>> getTransitiveEdges(HashMap<A, HashSet<A>> edges)
    {

        HashMap<A, HashSet<A>> transE = new HashMap<>();
        transE.putAll(edges);
        int numpre = 0;
        int numafter = NumUtils.sum(transE.values(), (HashSet<A> h) -> h.size());

        while (numafter > numpre)
        {
            numpre = numafter;
            Vector<A> keys = ObjectGetter.toVector(transE.keySet());

            for (A k : keys)
            {
                HashSet<A> toadd = new HashSet<A>();
                for (A  a : transE.get(k))
                {
                    HashSet<A> c = edges.get(a);
                    if (c == null)
                        continue;
                    toadd.addAll(c);
                }
                transE.get(k).addAll(toadd);
            }

            numafter = NumUtils.sum(transE.values(), (HashSet<A> h) -> h.size());
        }

        return transE;
    }


    static class DistanceNode<A>
    {
        A node;
        int distance;
        HashSet<DistanceNode<A>> parents = new HashSet<>();
        HashSet<A> children = new HashSet<>();

        DistanceNode(A n, DistanceNode<A> parent)
        {
            node = n; 
            distance = (parent == null) ? 0 : parent.distance + 1;
        }
    }
    
    static public<A> Vector<A> getShortestPath(A src, A target, HashMap<A,HashSet<A>> edges)
    {
        return getShortestPath(src, target, (A n) -> edges.get(n).iterator());
    }
    
    static public<A> Vector<A> getShortestPath(A src, A target, EdgeGetter<A> edgeGetter)
    {
        ShortestPathInfo<A> spi = calcShortestPaths(src, target, edgeGetter);
        return (spi == null) ? null : spi.getRandomPath();
    }
    

    static class ShortestPathInfo<A>
    {
        A src;
        A target;
        int dist;
        HashMap<A,HashSet<A>> relevant_graph;
        
        ShortestPathInfo(A src, A target, int dist, HashMap<A,HashSet<A>> relevant_graph)
        {
            this.src = src; this.target = target; this.dist = dist;
            this.relevant_graph = relevant_graph;
        }
        public Vector<A> getRandomPath()
        {
            Vector<A> path = new Vector<>();
            path.add(src);

            A current = src;

            while (current != target)
            {
                path.add( current = ObjectGetter.first(relevant_graph.get(current)));
            }
            
            return path;
        }
    }

    static public<A> ShortestPathInfo<A> calcShortestPaths(A src, A target, EdgeGetter<A> edgeGetter)
    {
        Logger log = LogConfig.getLogger();

        DistanceNode<A> start = new DistanceNode<>(src, null);
        boolean got_target = false;
        Vector<DistanceNode<A>> next_level = ObjectGetter.toVector(start);

        HashMap<A, DistanceNode<A>> a2node = new HashMap<>();
        a2node.put(src, start);
        
        int level = 0;
        DistanceNode<A> end = null;
        while (end == null)
        {
            HashMap<A, DistanceNode<A>> new_nodes = new HashMap<>();

            for (DistanceNode<A> dn : next_level)
            {
                Iterator<A> it = edgeGetter.getEdges(dn.node);
                if (it == null)
                    continue;
                
                
                
                while (it.hasNext())
                {
                    A n = it.next();
                    if (a2node.containsKey(n))
                        continue;
                    
                    DistanceNode<A> ndn = new_nodes.get(n);
                    if (ndn == null)
                    {
                        new_nodes.put(n, ndn = new DistanceNode<>(n, dn));
                    }
                    ndn.parents.add(dn);
                }
            }

            level++;
            
            end = new_nodes.get(target);
            log.debug("after level: %d got %d new nodes, known nodes: %d found target: %s", level, new_nodes.size(), a2node.size(), end != null);
            if (end != null || new_nodes.size() == 0)
                break;
            
            a2node.putAll(new_nodes);
            next_level.clear();
            next_level.addAll(new_nodes.values());
        }

        if (end == null)
            return null;
        
        
        int dist = end.distance;
        HashMap<A, HashSet<A>> relevant_edges = new HashMap<>();

        LinkedList<DistanceNode<A>> q = new LinkedList<>();
        q.add(end);

        HashSet<DistanceNode<A>> visited = new HashSet<>();
        while (!q.isEmpty())
        {
            DistanceNode<A> dn = q.poll();
            if(visited.contains(dn))
                continue;

            visited.add(dn);
            

            for (DistanceNode<A> pdn : dn.parents)
            {
                HashSet<A> h = relevant_edges.get(pdn.node);
                if (h == null)
                {
                    relevant_edges.put(pdn.node, h = new HashSet<>());
                }
                h.add(dn.node);
                
                q.add(pdn);
            }
        }

        return new ShortestPathInfo<A>(src, target, dist, relevant_edges);
    }
    
    static public<A> Pair<Integer, HashMap<A,HashSet<A>>> getShortestPaths(A src, A target, EdgeGetter<A> edgeGetter)
    {
        ShortestPathInfo<A> spi = calcShortestPaths(src, target, edgeGetter);
        return (spi == null) ? null : Pair.create(spi.dist, spi.relevant_graph);
        
    }
}

