package lmu.utils.partialorder;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Vector;

import lmu.utils.BufferPrintWriter;
import lmu.utils.CollectionIterator;
import lmu.utils.IteratorUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.NumUtils;
import lmu.utils.ObjectGetter;
import lmu.utils.OptionParser;
import lmu.utils.Pair;
import lmu.utils.SimpleOptionParser;


/** 
 * 
 * 
 *  req features: 
 *              - linking nodes -- equality - all prefixes/suffixes from one are also prefixes suffixes of another
 *              - retrieve ordered and unordered regions
 * 
 * 
 * two modes - 
 * (i) given a set of entities, and incrementally add relations betwenn them
 * (ii) given a comparator and incrementally add entities
 * 
 * 
 *  properties: 
 *  - the method iterator() returns a possible ordering consistent with the partial order
 * 
 *  - the graph has the properties
 *      - nodes connected with start have no other prefixes
 *      - nodes connected with end have no other suffixes
 *      - 
 * 
 *      
 *   - start    - n1   - end
 * 
 *   - start    - n2   - end
 * 
 *  - start    - n3   - end
 * 
 *      n3 > n2 ->
 * 
 *  
 *         -  n1       - end
 *  start  - n2 - n3   - end 
 * 
 * 
 *  n1 < n3
 * 
 *  
 * start - n1  - n3  - end
 *      -  n2  /
 * 
 *  for the iteration whenever we ge visit a node x we have to wait until all prefixes
 *  of x are processed before we can process x
 */

public class PartialOrderGraph<A>
{
    Logger log = LogConfig.getLogger();

    public static class PartialNode<A>
    {
        public A data;
        HashSet<PartialNode<A>> pre = new HashSet<>();
        HashSet<PartialNode<A>> post = new HashSet<>();

        public PartialNode(A data)
        {
            this.data = data;
        }

        boolean isPrefix(PartialNode<A> pn)
        {
            return isAffix(pn, true);
        }
        boolean isSuffix(PartialNode<A> pn)
        {
            return isAffix(pn, false);
        }

        boolean isAffix(PartialNode<A> pn, boolean is_pre)
        {
            HashSet<PartialNode<A>> src = (is_pre) ? pre : post;
            if (src.contains(pn))
                return true;
            
            for (PartialNode<A> ch : src)
            {
                if (ch.isAffix(pn, is_pre))
                    return true;
            }
            return false;
        }

        public String toString()
        {
            return "pn: " +data +" "+pre.size()+" pre "+ post.size()+ " post";
        }
    }

    PartialNode<A> start = new PartialNode<>(null);
    PartialNode<A> end = new PartialNode<>(null);

    HashMap<A, PartialNode<A>> nodes = new HashMap<>();


    public PartialOrderGraph()
    {

    }

    public PartialOrderGraph(A ... data)
    {
        this(new CollectionIterator<A>(data));
    }

    public PartialOrderGraph(Collection<A> c)
    {
         this(c.iterator());
    }
    public PartialOrderGraph(Iterator<A> it)
    {
        while (it.hasNext())
        {
            add(it.next());
        
        }
    }

        
    public static abstract class GraphSorter<A>
    {
        Comparator<PartialNode<A>> comp;
        public abstract void update(PartialNode<A> node);
        
        public GraphSorter(Comparator<PartialNode<A>> comp)
        {
            this.comp = comp;
        }
    }
    
    class POIterator implements Iterator<A>
    {
        class Dependency
        {
            A entry;
            HashSet<A> dependent_on = new HashSet<>();

            public Dependency(A e)
            {
                entry = e;
            }

            boolean allowed()
            {
                return dependent_on.size() == 0;
            }
        }
        
        HashSet<A> visited = new HashSet<>();
        
        HashMap<A, HashSet<Dependency>> dependencies = new HashMap<>();
        
        HashSet<PartialNode<A>> added = new HashSet<>();
        LinkedList<PartialNode<A>> q = new LinkedList<>();

        //Comparator<PartialNode<A>> comp = null;
        GraphSorter<A> gs = null;
        
        
        public POIterator()
        {
            this(null);
        }
        
        public POIterator(GraphSorter<A> gs)
        {
            this.gs = gs;
            q.add(start);
            step();
        }



        A step()
        {
            if ( gs != null)
            {
                Collections.sort(q, gs.comp);
            }
            
            PartialNode<A> n = q.poll();
            A rv = n.data;
            visited.add(rv);
            added.add(n);
            
            if ( gs != null)
                gs.update(n);
            
            //put all successors in the queue
            for (PartialNode<A> pn : n.post)
            {
                if (pn == end)
                    continue;
                
                if (pn == null)
                {
                    throw new RuntimeException("null as postfor  " +n);
                }
                //log.trace("will dependency for data : " +pn.data+" as post from : " + n.data);
                if (null == nodes.get(pn.data))
                {
                    throw new RuntimeException("invalid post! data: " +pn.data+ " is not referenced any more by node: " + n);
                }
                BufferPrintWriter bpw = new BufferPrintWriter();
                bpw.setSkip(!log.isDebugEnabled());
                Dependency d = new Dependency(pn.data);

                for(PartialNode<A> pre_pn : pn.pre)
                {
                    if (visited.contains(pre_pn.data))
                        continue;
                    
                    bpw.printf("add dependency %s dep. on %s\n", pn.data, pre_pn.data);
                    d.dependent_on.add(pre_pn.data);
                }
                if (d.dependent_on.size() == 0)
                {
                    if (!added.contains(rv))
                    {
                        added.add(pn);
                        //log.trace("will add : " +pn);
                        q.add(pn);
                    }
                    
                    continue;
                }
                //log.debug("\n"+bpw.getBuffer());
                //got dependencies - cannot add, but check dependencies
                for (A dep_src : d.dependent_on)
                {
                    HashSet<Dependency> h = dependencies.get(dep_src);
                    if (h == null)
                    {
                        dependencies.put(dep_src, h = new HashSet<>());
                    }
                    h.add(d);
                }
            }

            //resolve dependencies
            HashSet<Dependency> h = dependencies.get(rv);

            if (h == null)                     
                return rv; //no dependecies, no other nodes to add
            
            
            
            for (Dependency d : h)
            {
                d.dependent_on.remove(rv);
                if (!d.allowed())
                    continue;
                
                dependencies.remove(d.entry);
                //all dependencies resolved , add to queue
                PartialNode<A> pn2 = nodes.get(d.entry);
                if (pn2 == null)
                {
                    throw new RuntimeException("null node for dependency: " +d);
                }
                if (!added.contains(pn2))
                {
                    added.add(pn2);
                    //log.trace("will add : " +pn2+" entry was: " + d.entry);
                    q.add(pn2);
                }
                
                
            }
            dependencies.remove(rv);

            return rv;
        }
        public boolean hasNext()
        {
            boolean rv = !q.isEmpty();

            if (rv || dependencies.size() == 0)
                return rv;
            
            BufferPrintWriter bpw = new BufferPrintWriter();


            HashSet<Dependency> dh = new HashSet<>();
            ObjectGetter.apply(dependencies.values(), (HashSet<Dependency> h) -> dh.addAll(h));
            bpw.printf("got %d unresolved dependencies\n", dh.size());
            
            for (Dependency d : dh)
            {
                //bpw.printf("\t%s depends on %s\n", ObjectGetter.mapSorted(d.entry, (Pair<A,Integer> p) -> p, true), d.dependent_on);
            }

            bpw.println("\ndependencies\n");
            for(Map.Entry<A, HashSet<Dependency>> e : dependencies.entrySet())
            {
                bpw.printf("\t%s\n",e.getKey());

                for (Dependency d : e.getValue())
                {
                    bpw.printf("\t\t%s <- %s depends on %s\n", e.getKey() , d.entry, d.dependent_on);
                }
            }
            throw new RuntimeException(bpw.getBuffer());
        }

        public A next()
        {
            return step();
        }
    }

    

    public Iterator<A> iterator()
    {
        return new POIterator();
    }
    
    
    public Iterator<A> iterator(GraphSorter<A> gs)
    {
        return new POIterator(gs);
    }
    
    /** removes the node
     * and transfers all indirect relations pre -> a -> post 
     * to pre -> post
     */
    public PartialNode<A> remove(A a)
    {
        PartialNode<A> n = nodes.get(a);
        ObjectGetter.apply(n.pre, (PartialNode<A> pn) -> { pn.post.remove(n); pn.post.addAll(pn.pre);} );
        ObjectGetter.apply(n.post, (PartialNode<A> pn) -> { pn.pre.remove(n); pn.pre.addAll(pn.post);} );
        nodes.remove(a);
        return n;
    }

    public void add(A ... data)
    {
        for (int i=0; i<data.length; i++)
        {
            add(data[i]);
        }
    }
    public PartialNode<A> add(A a)
    {

        PartialNode<A> n = nodes.get(a);
        if (n != null)
            return n;
        
        nodes.put(a, n = new PartialNode<>(a));
        start.post.add(n);
        end.pre.add(n);
        n.pre.add(start);
        n.post.add(end);
        
        return n;
    }


    public enum ACCEPT_STATE
    {
        ACCEPT,
        CONTAINED,
        INVALID
        ;
    }

    long check_time = 0;
    long collect_time = 0;


    IteratorUtils.ChildGetter<PartialNode<A>> preGetter = (PartialNode<A> pn) -> pn.pre;
    IteratorUtils.ChildGetter<PartialNode<A>> postGetter = (PartialNode<A> pn) -> pn.post;

    class StopChildGetter implements IteratorUtils.ChildGetter<PartialNode<A>>
    {
        boolean pre = true;
        PartialNode<A> stopnode;
        
        StopChildGetter(boolean pre, PartialNode<A> pn)
        {
            this.pre = pre;
            this.stopnode = pn;
        }

        public Collection<PartialNode<A>> getChildren(PartialNode<A> e)
        {
            if (stopnode!= null && e == stopnode)
                return null;
            

            /*Collection<PartialNode<A>> c = (pre) ? e.pre : e.post;
            if (c == null || ObjectGetter.filter(c, (PartialNode<A> pn ) -> pn == null).size() > 0)
            {
                log.warn("error at node: %s pre: %s children: %s\n", e, pre, c);
            }
            */
            
            //System.out.println("pre: " + pre + " next node: " +e);
            //System.out.println((pre) ? e.pre : e.post);
            return (pre) ? e.pre : e.post;
        }
    }
    

    
        
    Iterator<PartialNode<A>> getTransitiveAffixIterator(PartialNode<A> pn, boolean inclusive, boolean pre, PartialNode<A> stopNode)
    {
        StopChildGetter childgetter =new StopChildGetter(pre, stopNode);

        if (!inclusive)
            return IteratorUtils.getDAGIterator((pre) ? pn.pre : pn.post, childgetter); 
        
        return IteratorUtils.getDAGIterator(pn, childgetter); 
    }

    Iterator<PartialNode<A>> getTransitivePrefixIterator(PartialNode<A> pn, boolean inclusive, PartialNode<A> stopNode)
    {
        return getTransitiveAffixIterator(pn,inclusive, true, stopNode);
    }

    Iterator<PartialNode<A>> getTransitiveSuffixIterator(PartialNode<A> pn, boolean inclusive, PartialNode<A> stopNode)
    {
        return getTransitiveAffixIterator(pn,inclusive, false, stopNode);
    }

    HashSet<PartialNode<A>> getTransitiveAffixes(PartialNode<A> pn, boolean inclusive, boolean pre, PartialNode<A> stopNode)
    {
        return ObjectGetter.toSet(getTransitiveAffixIterator(pn, inclusive, pre, stopNode));
    }

    HashSet<PartialNode<A>> getTransitivePrefixes(PartialNode<A> pn, boolean inclusive, PartialNode<A> stopNode)
    {
        return getTransitiveAffixes(pn,inclusive, true, stopNode);
    }

    HashSet<PartialNode<A>> getTransitiveSuffixes(PartialNode<A> pn, boolean inclusive, PartialNode<A> stopNode)
    {
        return getTransitiveAffixes(pn, inclusive, false, stopNode);
    }

    /** checks if there are direct relations starting from any node in pre_prefixes
     * ending inthe post suffix set
     */
    void searchCrossingRelations(HashSet<Pair<PartialNode<A>, PartialNode<A>>> crossing_relations, 
                                 Collection<PartialNode<A>> pre_prefixes,
                                 HashSet<PartialNode<A>> post_suffixes)
    {
        for (PartialNode<A> pref : pre_prefixes)
        {
            for (PartialNode<A> pn : pref.post)
            {
                if (!post_suffixes.contains(pn))
                    continue;
                
                //got pref < smaller ^ pref < pn ^ smaller < bigger ^  bigger < pn
                // change pref <  pn  -> pref < smaller -> bigger -> pn
                crossing_relations.add(Pair.create(pref, pn));
            }
        }
    }

    void fixCrossingRelations(HashSet<Pair<PartialNode<A>, PartialNode<A>>> crossing_relations, PartialNode<A> smallerN, PartialNode<A> biggerN)
    {
        for (Pair<PartialNode<A>, PartialNode<A>> p : crossing_relations)
        {
            PartialNode<A> s = p.getFirst();
            PartialNode<A> e = p.getSecond();

            s.post.remove(e);
            e.pre.remove(s);
            s.post.add(smallerN);
            smallerN.pre.add(s);
            e.pre.add(biggerN);
            biggerN.post.add(e);
        }

    }
    public ACCEPT_STATE addRelation(A smaller, A bigger)
    {
        return addRelation(smaller,bigger, null, null);
    }

    public ACCEPT_STATE addRelation(A smaller, A bigger, PartialNode<A> prefixStopNode, PartialNode<A> suffixStopNode)
    {
        long t1 = System.currentTimeMillis();

        PartialNode<A> smallerN = nodes.get(smaller);
        PartialNode<A> biggerN = nodes.get(bigger);
        
        HashSet<PartialNode<A>> pre_suffixes  =  getTransitiveSuffixes(smallerN, false, suffixStopNode);
        HashSet<PartialNode<A>> post_suffixes =  getTransitiveSuffixes(biggerN, false, suffixStopNode);
                    
        boolean invalid = post_suffixes.contains(smallerN);
        boolean contained = pre_suffixes.contains(biggerN);

        long t2 = System.currentTimeMillis();
        check_time += (t2-t1);
        
        //if (post_a_set.contains(smaller)) //the two inputs are inverse related... 
        if (invalid)
            return ACCEPT_STATE.INVALID;
        
            
        
        //if (pre_suffixes.contains(bigger))
        if (contained)
            return ACCEPT_STATE.CONTAINED; //already ordered
        
        //System.out.printf("collect crossing relations for %s %s\n", smaller, bigger);
        
        HashSet<PartialNode<A>> pre_prefixes  = getTransitivePrefixes(smallerN, false, prefixStopNode);

        /** look for relations "crossing the new relation: i.e. start in the prefixset of pre, and end in the suffix set of post */

        HashSet<Pair<PartialNode<A>, PartialNode<A>>> crossing_relations = new HashSet<>();
        searchCrossingRelations(crossing_relations, pre_prefixes, post_suffixes);

        long t3 = System.currentTimeMillis();
        collect_time += (t3-t2);

        //System.out.printf("got %d crossing relations for %s %s time for cross coll: %2.3f sec for checks: %2.3f secs\n", 
        //                  crossing_relations.size(), smaller, bigger, collect_time/1000.0, check_time/1000.0);
        smallerN.post.remove(end);
        biggerN.pre.remove(start);
        start.post.remove(biggerN);
        end.pre.remove(smallerN);

        fixCrossingRelations(crossing_relations, smallerN, biggerN); 

        smallerN.post.add(biggerN);
        biggerN.pre.add(smallerN);

        /**      (1,2), (2,3), (1,6), (4,5), (5,6)
         * 
         *       1 --> 2 --->3
         *          |
         *          +------+-->6
         *                 |
         *       4 --> 5 --+ 
         * 
         *          (3,4)
         * 
         *  --> all prefixes of 3 (1,2) which have suffixes from 4 (5,6) get instead of them 4 
         *     and all suffixes of 4 (5,6) having prefixes which are prefixes of 3 (1,2) get them
         *    as prefix
         * 
         *      in general:
         * 
         *              n1, n2
         *              (3,4) is a new order - for every x,y   x < n1 ^ y > n2  -> x < n1 < n2 < y
         * 
         *          if n1 was not a suffix on
         * 
         *      3 < 4


        //if ok -> all prefixes of smaller are no longer direct prefixes of bigger 
        // all suffixes of bigger are no longer direct prefixes of bigger


        HashSet<A> suffixes;

        

        /**   1 -> 2, 3 -> 4
         *  2 < 4 ==> 1 -> 2 -> 4
         *                 3 -> 4
         * 
         *  1 < 3 ==>  1 -> 2 -> 4
         *               \     /
         *                  3
         */
        
        

        return ACCEPT_STATE.ACCEPT;

    }

    public void printGraph()
    {
        BufferPrintWriter bpw = new BufferPrintWriter();
        bpw.println("GRAPH");
        Iterator<A> it = iterator();
        bpw.printf("START  post: %s\n",  ObjectGetter.map(start.post, (PartialNode<A> pn) -> pn.data));
        while (it.hasNext())
        {
            
            PartialNode<A> n = nodes.get(it.next());
            bpw.printf("\t%s pre %s post: %s\n", n.data, ObjectGetter.map(n.pre, (PartialNode<A> pn) -> pn.data), ObjectGetter.map(n.post, (PartialNode<A> pn) -> pn.data));
            
        }
        System.out.println(bpw.getBuffer());
    }

    int getNumEdges()
    {
        return NumUtils.sum(ObjectGetter.map(nodes.values(), (PartialNode<A> pn) -> pn.pre.size() + pn.post.size()));
        
        
    }
    /** test methods -> shuffle integers, and record how it's get's sorted  by mergesort
     *  then put these comparisons in an arbitrary order into the partial order graph ->
     *  it should result a perfect order
     */

    static class RecordingComparator implements Comparator<Integer>
    {
        Vector<Pair<Integer,Integer>> recorded = new Vector<>();

        public int compare(Integer i1, Integer i2)
        {
            recorded.add(Pair.create(Math.min(i1,i2), Math.max(i1,i2)));
            return i1.compareTo(i2);
        }
    }

    static void testInts(int num_entries, int nrepeats, boolean verbose, boolean with_sets)
    {
        for (int NRI = 0; NRI < nrepeats; NRI++)
        {
            
            //after num  accepted relations we should have a full ordered list
            Vector<Integer> objects = ObjectGetter.map(IteratorUtils.getIntIterator(0, num_entries), (Integer idx) -> idx);
            Collections.sort(objects);
            RecordingComparator rc = new RecordingComparator();
            Collections.sort(objects, rc);

            
            Collections.shuffle(rc.recorded);
            Vector<Integer> order = null;
            if (with_sets)
            {
                PartiallyOrderedSet pos = new PartiallyOrderedSet();
        
                ObjectGetter.apply(objects, (Integer idx) -> pos.add(idx));
                order = processPairsS(pos, rc.recorded, verbose);
            }
            else
            {
                PartialOrderGraph<Integer> pog = new PartialOrderGraph(objects);

                order = processPairs(pog, rc.recorded, verbose);
            }
            
            System.out.printf("TEST %d/%d FINISHED:\n", NRI+1, nrepeats);
            //System.out.println(order);


        }

    }

    static Vector<Integer> processPairs(PartialOrderGraph<Integer> pog, Vector<Pair<Integer,Integer>> tests, boolean verbose)
    {
        int naccept = 0;
        int nmaxsteps = tests.size();
        int nsteps = 0;
        int ncontained = 0;

        for(Pair<Integer, Integer> key : tests)
        {
            Integer o1 = key.getFirst();
            Integer o2 = key.getSecond();
            
            nsteps++;
            
            if (verbose)
            {
                System.out.printf("next: %d < %d\n",o1, o2);
            }
            
            
            ACCEPT_STATE as = pog.addRelation(o1, o2);
            if (verbose)
            {
                System.out.printf("result of: %d < %d = %s\n",o1, o2, as);
            }
            
            switch(as)
            {
                case INVALID:
                    throw new RuntimeException("eeerrror!");
                case CONTAINED:
                    if (verbose)
                    {
                        System.out.printf("%d < %d already implied\n", o1, o2);
                    }
                    
                    ncontained++;
                    break;
                case ACCEPT:
                    naccept++;
                    break;
            }                                                                                                    
            if (verbose)
            {
                pog.printGraph();
            
                System.out.printf("ordered: %s \n", ObjectGetter.toVector(pog.iterator()));
            }
            
            if (verbose && nsteps % 1_000 == 0)
            {
                System.out.printf("after %d/%d steps, %d nodes, %d edges got %d contained, %d accepted ordered: %s \n", nsteps, tests.size(), 
                                  pog.nodes.size(), pog.getNumEdges(),
                                  ncontained, naccept,ObjectGetter.toVector(pog.iterator()));
            }
            
        }
        if (verbose)
        {
            System.out.printf("after %d steps, got %d contained, %d accepted\n", nsteps, ncontained, naccept);
        }
        Vector<Integer> order = ObjectGetter.toVector(pog.iterator());
        for (int i=1; i<order.size(); i++)
        {
            if (order.get(i-1) >= order.get(i))
            {
                throw new RuntimeException("invalid order: " + order +" at pos: " +i);
            }
        }
        
        return order;
        
    }
        


    static Vector<Integer> processPairsS(PartiallyOrderedSet pog, Vector<Pair<Integer,Integer>> tests, boolean verbose)
    {
        int naccept = 0;
        int nmaxsteps = tests.size();
        int nsteps = 0;
        int ncontained = 0;

        for(Pair<Integer, Integer> key : tests)
        {
            Integer o1 = key.getFirst();
            Integer o2 = key.getSecond();
            
            nsteps++;
            
            if (verbose)
            {
                System.out.printf("next: %d < %d\n",o1, o2);
            }
            
            
            pog.setOrdering(o1, o2);
            
        }
        if (verbose)
        {
            System.out.printf("after %d steps, got %d contained, %d accepted\n", nsteps, ncontained, naccept);
        }
        Vector<Integer> order = ObjectGetter.toVector(pog.iterator());
        for (int i=1; i<order.size(); i++)
        {
            if (order.get(i-1) >= order.get(i))
            {
                throw new RuntimeException("invalid order: " + order +" at pos: " +i);
            }
        }
        
        return order;
        
    }
    


    public static void main (String args[]) throws java.io.IOException
    {
        SimpleOptionParser cmd = new SimpleOptionParser("ne", "iter", "v", "random", "withsets", "transtest", "check");
        cmd.setInt("ne", "iter");
        cmd.setOptional("ne", "iter");
        cmd.setSwitches("v", "random", "withsets", "transtest", "check");
        if (!OptionParser.parseParams(args, true, false, true, cmd))
            return;
        
        
        if (cmd.isSet("check"))
        {
            PartiallyOrderedSet pos = new PartiallyOrderedSet();
            Vector<Integer> objects = ObjectGetter.map(IteratorUtils.getIntIterator(0, cmd.getOptionalInt("ne",10)), (Integer idx) -> idx);
            ObjectGetter.apply(objects, (Integer idx) -> pos.add(objects.get(idx)));
            for (int i=1; i<objects.size(); i++)
            {
                pos.setOrdering(objects.get(i-1), objects.get(i));
            }
            System.out.println(ObjectGetter.toVector(pos.iterator()));
            pos.setOrdering(objects.get(objects.size()-1), objects.get(0));
            System.out.println(ObjectGetter.toVector(pos.iterator()));
            return;
        }
        if (cmd.isSet("transtest"))
        {
            PartiallyOrderedSet pos = new PartiallyOrderedSet();
            Vector<Integer> objects = ObjectGetter.map(IteratorUtils.getIntIterator(0, cmd.getOptionalInt("ne",10)), (Integer idx) -> idx);
            ObjectGetter.apply(objects, (Integer idx) -> pos.add(idx));

            pos.setOrdering(1,2);
            pos.setOrdering(2,3);
            pos.setOrdering(1,4);
            pos.setOrdering(3,6);

            Pair<Integer,Integer> test = Pair.create(1,6);
            Vector<Integer> v = ObjectGetter.toVector(pos.iterator());
            System.out.println("ordering:  " + v);

            for (int i=1; i<v.size(); i++)
            {
                test = Pair.create(v.get(i-1), v.get(i));
                System.out.printf("test : %s result: %s\n", test, pos.hasOrdering(test.getFirst(), test.getSecond()));
            }

            System.out.printf("test : %s result: %s\n", test, pos.hasOrdering(test.getFirst(), test.getSecond()));
            return;
        }
        if (cmd.isSet("random"))
        {
            Vector<Integer> objects = ObjectGetter.map(IteratorUtils.getIntIterator(0, cmd.getOptionalInt("ne",10)), (Integer idx) -> idx);
            PartialOrderGraph<Integer> pog = new PartialOrderGraph(objects);

            PartiallyOrderedSet pos = new PartiallyOrderedSet();
        
            ObjectGetter.apply(objects, (Integer idx) -> pos.add(idx));
            
            java.io.BufferedReader bfr = new java.io.BufferedReader(new java.io.InputStreamReader(System.in));
            String line = null;
            while (null != (line = bfr.readLine()))
            {
                int x = (int)(Math.random()*objects.size());
                int y = x;
                while(x == y)
                {
                    y=(int)(Math.random()*objects.size());
                }
                Pair<Integer,Integer> p = Pair.create(Math.min(x,y), Math.max(x,y));

                if (line.equals("iv"))
                {
                    Pair<Integer,Integer> p2 = Pair.swap(p);
                    p = p2;
                }
                pog.addRelation(p.getFirst(), p.getSecond());

                System.out.println("adding : " +p);
                pog.printGraph();
                pos.setOrdering(p.getFirst(), p.getSecond());
                                   
                System.out.println(ObjectGetter.toVector(pog.iterator()));

                System.out.println("POS");
                System.out.println(ObjectGetter.toVector(pos.iterator()));

                
            }
            return;
        }
        testInts(cmd.getOptionalInt("ne",10), cmd.getOptionalInt("iter", 1), cmd.isSet("v"), cmd.isSet("withsets"));

        
        
    }
}
