package lmu.utils.partialorder;

import java.util.AbstractSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * A set of <code>Object</code>s with pairwise orderings between them.
 * The <code>iterator</code> method provides the elements in
 * topologically sorted order.  Elements participating in a cycle
 * are not returned.
 *
 * Unlike the <code>SortedSet</code> and <code>SortedMap</code>
 * interfaces, which require their elements to implement the
 * <code>Comparable</code> interface, this class receives ordering
 * information via its <code>setOrdering</code> and
 * <code>unsetPreference</code> methods.  This difference is due to
 * the fact that the relevant ordering between elements is unlikely to
 * be inherent in the elements themselves; rather, it is set
 * dynamically accoring to application policy.  For example, in a
 * service provider registry situation, an application might allow the
 * user to set a preference order for service provider objects
 * supplied by a trusted vendor over those supplied by another.
 *
 */
public class PartiallyOrderedSet<A> extends AbstractSet<A> 
{

    // The topological sort (roughly) follows the algorithm described in
    // Horowitz and Sahni, _Fundamentals of Data Structures_ (1976),
    // p. 315.

    // Maps Objects to DigraphNodes that contain them
    private Map<A, DigraphNode<A>> poNodes = new HashMap();

    // The set of Objects
    private Set<A> nodes = poNodes.keySet();

    /**
     * Constructs a <code>PartiallyOrderedSet</code>.
     */
    public PartiallyOrderedSet()
    {
    }

    public int size()
    {
        return nodes.size();
    }

    public boolean contains(Object o)
    {
        return nodes.contains(o);
    }

    /**
     * Returns an iterator over the elements contained in this
     * collection, with an ordering that respects the orderings set
     * by the <code>setOrdering</code> method.
     */
    public Iterator iterator()
    {
        return new PartialOrderIterator(poNodes.values().iterator());
    }

    /**
     * Adds an <code>Object</code> to this
     * <code>PartiallyOrderedSet</code>.
     */
    public boolean add(A o)
    {
        if (nodes.contains(o))
        {
            return false;
        }

        DigraphNode<A> node = new DigraphNode<>(o);
        poNodes.put(o, node);
        return true;
    }

    /**
     * Removes an <code>Object</code> from this
     * <code>PartiallyOrderedSet</code>.
     */
    public boolean remove(Object o)
    {
        DigraphNode<A> node = (DigraphNode<A>)poNodes.get(o);
        if (node == null)
        {
            return false;
        }

        poNodes.remove(o);
        node.dispose();
        return true;
    }

    public void clear()
    {
        poNodes.clear();
    }

    /**
     * Sets an ordering between two nodes.  When an iterator is
     * requested, the first node will appear earlier in the
     * sequence than the second node.  If a prior ordering existed
     * between the nodes in the opposite order, it is removed.
     *
     * @return <code>true</code> if no prior ordering existed
     * between the nodes, <code>false</code>otherwise.
     */
    public boolean setOrdering(A first, A second)
    {
        DigraphNode<A> firstPONode = (DigraphNode<A>)poNodes.get(first);
        
        DigraphNode<A> secondPONode = (DigraphNode<A>)poNodes.get(second);

        secondPONode.removeEdge(firstPONode);
        
        return firstPONode.addEdge(secondPONode);
    }

    /**
     * Removes any ordering between two nodes.
     *
     * @return true if a prior prefence existed between the nodes.
     */
    public boolean unsetOrdering(A first, A second)
    {
        DigraphNode<A> firstPONode = poNodes.get(first);
        DigraphNode<A> secondPONode = poNodes.get(second);

        return firstPONode.removeEdge(secondPONode) || secondPONode.removeEdge(firstPONode);
    }

    /**
     * Returns <code>true</code> if an ordering exists between two
     * nodes.
     */
    public boolean hasOrdering(A preferred, A other)
    {
        DigraphNode<A> preferredPONode = poNodes.get(preferred);
        DigraphNode<A> otherPONode = poNodes.get(other);

        return preferredPONode.hasEdge(otherPONode);
    }
}


