package ortholog;

import genome.*;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import lmu.utils.*;

public class MultiSpeciesGeneMapping
{
    private static String TRANSITIVE_ORTHOLOGS ="Transitiv";
    private static String DIRECT_ORTHOLOGS ="Direct";
    
    Logger log = LogConfig.getLogger();

    HashMap<Integer,Integer> taxid2idx = new HashMap<Integer,Integer>();
    
    OrthMappings[][] om;
    ParalogMappings[] pm;
    
    boolean includeTransitive = true;
    
    Vector<Organism> organism = new Vector<Organism>();
    
    static class Mappings
    {
        HashMap<String, HashSet<String>> src2dst = new HashMap<String, HashSet<String>>();
        
        public  HashMap<String, HashSet<String>> getMappings(){
            return src2dst;
        }
        
        public HashSet<String> getMappings(String src)
        {
            HashSet<String> hs = src2dst.get(src);
            if (hs!=null)
                return hs;
            
            return new HashSet<String>();
        }

        public void addMapping(String from, String to)
        {
            if (from==null || to==null || from.equals(to))
            {
                return;
            }
            HashSet<String> s = src2dst.get(from);
            if (s==null)
            {
                src2dst.put(from,s=new HashSet<String>()); 
            }
            s.add(to);
        }
    }
    static class ParalogMappings extends Mappings
    {
        Organism org;
        int x;      

        public ParalogMappings(int x, Organism o)
        {
            org = o; this.x = x;
        }

        
    }
    static class OrthMappings extends Mappings
    {
        int x; 
        int y;
        Organism org1;
        Organism org2;

        public OrthMappings(int x, int y, Organism o1, Organism o2)
        {
            this.x=x; this.y=y; this.org1=o1; this.org2=o2;
        }
    }


    void readMapping(File mapping, int src_idx)
    {
        readMapping(mapping,null,src_idx);
    }
    Pattern source_db_split_pattern = Pattern.compile(",");

    public void addMapping(HashMap<String,HashSet<String>> oMapping, Organism o1, Organism o2)
    {
        int id1 = taxid2idx.get(o1.getTaxId());
        int id2 = taxid2idx.get(o2.getTaxId());
        
        for( Entry<String, HashSet<String>> e : oMapping.entrySet())
        {
            for(String to : e.getValue())
            {
                om[id1][id2].addMapping(e.getKey(), to);
                om[id2][id1].addMapping(to,e.getKey());
            }
        }
        
        //om[oi][src_idx].addMapping(to, from);
        
    }
    void readMapping(File mapping,Set<OrthologSource> source_databases, int src_idx)
    {
        if (mapping.isDirectory())
        {
            return;
        }
        
        log.debug("next mapping file: "+mapping.getAbsolutePath());
        if (!mapping.exists())
        {
            log.warn("no ortholog mapping found, looked here: "+mapping.getAbsolutePath());
            return;
        }

        ParalogMappings pmi = pm[src_idx];

        Iterator<String[]> it = FileUtils.getFieldSetIterator(mapping, "\\s+");

        Integer t_id = null;
        try{
            t_id = Integer.valueOf(mapping.getName()); 
        }catch(NumberFormatException e)
        {
            throw new RuntimeException(String.format("%s is not a taxId",t_id),e);
        }
        
        int nrelevant = 0;
        int nlines_read=0;
        
        while(it.hasNext())
        {
            String[] tokens = it.next();
            
            nlines_read++;
            String from=null;
            String to = null;
            int target_taxid = -1;
            boolean skip = true;

            if( tokens.length == 4)
            {
                from = tokens[0];
                to = tokens[1];
                target_taxid = t_id;
                
                if ( source_databases != null){
                    for(String t : source_db_split_pattern.split(tokens[2]) ) 
                    {
                        OrthologSource type = null;
                        try{
                           type =OrthologSource.valueOf(t);
                        }catch(Exception e)
                        {
                            throw new RuntimeException(String.format("Invalid ortholog format, or unknown ortholog database %s",t),e);
                        }
                        if ( source_databases.contains(type)) 
                        {
                            skip = false;
                            break;
                        }   
                    }
                }else{
                    skip = false;
                }
            }else if ( tokens.length == 5)
            {
                skip = false;
                
                from = tokens[1];
                to = tokens[2];
                String test = tokens[3];
                target_taxid = Integer.parseInt(test);

                String annot = tokens[4];
                if (!annot.equals(TRANSITIVE_ORTHOLOGS) && !annot.equals(DIRECT_ORTHOLOGS))
                {
                    throw new RuntimeException("No valid annotation type:" + annot);
                }
                
                if ( !includeTransitive && annot.equals(TRANSITIVE_ORTHOLOGS)  )
                    skip = true;
            }else{
                throw new RuntimeException("Unknown ortholog format");
            }
            
            Integer oi = taxid2idx.get(target_taxid);
            
            if (oi==null || skip )
            {
                continue;
            }

            if (src_idx == oi)
            {
                pmi.addMapping(from, to);
                continue;
            }
            nrelevant++;
            om[src_idx][oi].addMapping(from, to);
            om[oi][src_idx].addMapping(to, from);

        }
/*
        for (int j=0; j<om[src_idx].length; j++)
        {
            if (om[src_idx][j]==null)
                continue;
                
            log.info("read %d relevant mappings (%d lines) from  %d genes", nrelevant, nlines_read, om[src_idx][j].src2dst.size());
            
                
        }
*/
    }


    
    public MultiSpeciesGeneMapping(File orthdir, Organism ... orgs)
    {
        this(orthdir,true, null,new CollectionIterator<Organism>(orgs));
    }
    public MultiSpeciesGeneMapping(File orthdir, Collection<Organism> orgs)
    {
        this(orthdir, true,null,orgs.iterator());
    }
    public MultiSpeciesGeneMapping(File orthdir, Iterator<Organism> orgs)
    {
    	  this(orthdir, true,null,orgs);
    }
   
    public MultiSpeciesGeneMapping(File orthdir,boolean includeTransitive, Organism ... orgs)
    {
        this(orthdir,includeTransitive,null, new CollectionIterator<Organism>(orgs));
    }
    public MultiSpeciesGeneMapping(File orthdir,boolean includeTransitive, Collection<Organism>  orgs)
    {
        this(orthdir,includeTransitive,null, orgs.iterator());
    }
    
    public MultiSpeciesGeneMapping(File orthdir,boolean includeTransitive, Iterator<Organism>  orgs)
    {
        this(orthdir,includeTransitive,null, orgs);
    }
    public MultiSpeciesGeneMapping(File orthdir,Set<OrthologSource> sourceOrthologDatabases, Organism ... orgs)
    {
        this(orthdir,false,sourceOrthologDatabases, new CollectionIterator<Organism>(orgs));
    }
    public MultiSpeciesGeneMapping(File orthdir,Set<OrthologSource> sourceOrthologDatabases, Collection<Organism> orgs)
    {
        this(orthdir,false,sourceOrthologDatabases, orgs.iterator());
    }
    
    public MultiSpeciesGeneMapping(File orthdir,boolean includeTransitive,Set<OrthologSource> orthoDb, Iterator<Organism> orgs)
    {
        this.includeTransitive = includeTransitive;
        
    	while (orgs.hasNext())
        {
    	    organism.add(orgs.next());
        }

        pm = new ParalogMappings[organism.size()];
        om = new OrthMappings[organism.size()][organism.size()];

        for (int i=0; i<organism.size(); i++)
        {
            pm[i] = new ParalogMappings(i, organism.get(i));
        
            taxid2idx.put(organism.get(i).getTaxId(), i);
            for (int j=0; j<organism.size(); j++)
            {
                if (i==j)
                    continue;
        
                om[i][j] = new OrthMappings(i, j, organism.get(i), organism.get(j));
            }
        }
        log.info("got taxidmapping: %s (trans.: %b)",taxid2idx,includeTransitive);

        //check if we have the id/id splitted files
        for (int i=0; i<organism.size(); i++)
        {
            for (int j=0; j<organism.size(); j++)
            {
                File mapping = new File(new File(orthdir, ""+organism.get(i).getTaxId()), ""+organism.get(j).getTaxId());
                readMapping(mapping,orthoDb, i);
            }
        }
        
        
        //parse the mappings from non-splitted files (if there)

        for (int src_idx = 0; src_idx < organism.size(); src_idx++)
        {
            File mapping = new File(orthdir, ""+ organism.get(src_idx).getTaxId());
            readMapping(mapping, src_idx);
        }
        
        for (int i=0; i<om.length; i++)
        {
            for (int j=0; j<om.length; j++)
            {
                if (i==j)
                    continue;
                
                OrthMappings omap = om[i][j];
                log.info("%s->%s: %d/%d mapped", omap.org1.getName(), omap.org2.getName(),
                                  omap.src2dst.size(), omap.org1.getAnnot().getNumGenes());
            }
        }

    }

    public Vector<Organism> getOrganisms()
    {
        return organism;
    }
  

    public void addParalogMapping(Organism o, String id1, String id2)
    {
        pm[getOrgIdx(o)].addMapping(id1, id2);

    }

    public int getOrgIdx(Organism o)
    {
        Integer i =  taxid2idx.get(o.getTaxId());
        if (i == null)
        {
            throw new RuntimeException("uncofigured organism: "+o.getName());
        }

        return i;
    }
  
    public HashSet<String> getMappings(int from_idx, int to_idx, String src)
    {
        return (from_idx == to_idx) ? pm[from_idx].getMappings(src) : om[from_idx][to_idx].getMappings(src);
    }

    public HashSet<String> getMappings(Organism from, Organism to, String src)
    {
        if (from == to)
        {
            return pm[getOrgIdx(from)].getMappings(src);
        }
        return om[getOrgIdx(from)][getOrgIdx(to)].getMappings(src);
    }
    
    public HashMap<String,HashSet<String>> getMappings(Organism from, Organism to)
    {
        if (from == to)
        {
            return pm[getOrgIdx(from)].getMappings();
        }
        return om[getOrgIdx(from)][getOrgIdx(to)].getMappings();
    }
    
    
    public static void main (String args[]) 
    {
        Locale.setDefault(Locale.UK);
        LogConfig lc = LogConfig.configureLog(null, true);
        SimpleOptionParser cmd = new SimpleOptionParser("orgxml","ortholog_dir");
        cmd.setDir("ortholog_dir");
        cmd.setFile("orgxml");
        if (!OptionParser.parseParams(args,true, false, true, cmd, lc))
            return;
        
        Organism.readOrganisms(cmd.getFile("orgxml"));

        Set<OrthologSource>os = new HashSet<OrthologSource>();
        os.add(OrthologSource.EnsemblCompara);
        
        MultiSpeciesGeneMapping msgm = new MultiSpeciesGeneMapping(cmd.getFile("ortholog_dir"),false,os,Organism.getOrganisms().iterator());
        
        List<Organism> orgs = new ArrayList<Organism>(Organism.getOrganisms());
        
        for(int i = 0 ; i< orgs.size(); i++){
            for(int j = i+1 ; j< orgs.size(); j++){
                HashMap<String, HashSet<String>> mapping = msgm.getMappings(orgs.get(i), orgs.get(j));
                int k=0;
                for(Entry<String, HashSet<String>> e : mapping.entrySet()) k+=e.getValue().size();
                
                System.out.println(String.format("%s and %s (%d/%d) %d mappings",orgs.get(i),orgs.get(j),orgs.get(i).getTaxId(),orgs.get(j).getTaxId(),k));
            }    
        }
        
    }
}

