package ortholog;

import java.util.HashMap;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.jdom.Element;

import algo.alignment.Alignment;
import genome.Gene;
import genome.GenomeSequenceExtractor;
import genome.GenomicProteinAlignment;
import genome.GenomicUtils;
import genome.Organism;
import genome.Protein;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.RegionVector;
import lmu.utils.XMLUtils;
import protein.algo.ProteinAligner;

public class OrthologInfo                                                                                                                               
{

	Logger log = LogConfig.getLogger();

	public static int MAX_GAP=5;
    public static final String ORTHO_ELEM = "orthinfo";
    
    public Gene src_gene;
    public Gene target_gene;
    public String src_id;
    public String dst_id;
    
    public Protein src_protein;
    public Protein dst_protein;
    
    RegionVector src;
    RegionVector dst;

    String p1seq;
    String p2seq;

    GenomicProteinAlignment genomic_ali = null;
    Alignment ali;
    public double seqid=-1.0;
    int num_id = -1;
    
    public OrthologInfo()
    {

    }
    public Protein getSrcProtein(){
    	return src_protein;
    }
    public Protein getTargetProtein(){
    	return dst_protein;
    }
    
    public String getP1Seq()
    {
        return p1seq;
    }

    public String getP2Seq()
    {
        return p2seq;
    }
    public Organism getSrcOrg()
    {
        return src_gene.organism;
    }

    public Organism getTargetOrg()
    {
        return target_gene.organism;
    }
    public Gene getSrcGene()
    {
        return src_gene;
    }
    public Gene getTargetGene()
    {
        return target_gene;
    }

    public RegionVector getSrc()
    {
        return src;
    }

    public RegionVector getTarget()
    {
        return dst;
    }
    public String getSrcId()
    {
        return src_id;
    }

    public String getTargetId()
    {
        return dst_id;
    }

    public double getSeqId()
    {
        return seqid;
    }
    public double getSeqIdNormedOnAlignmentPart(){
    	int p1Length = 1+ali.alignment.get(ali.alignment.size()-1).first - ali.alignment.get(0).first;
    	int p2Length = 1+ali.alignment.get(ali.alignment.size()-1).second - ali.alignment.get(0).second;
    	
    	double norm = 0.5*(p1Length+p2Length);
    	double ret = num_id/norm;
  
    	return ret;
    }
    
    public double getNormLength()
    {
        return 0.5*(p1seq.length()+p2seq.length());
    }
    
    public double getDNASeqId()
    {
        return getGenomicAlignment().dnaseqid;
    }
    
    public Alignment getAli()
    {
        return ali;
    }
    
    public OrthologInfo(XMLStreamReader xsr, HashMap<String, Gene> genemap) throws XMLStreamException
    {
        src_gene = genemap.get(XMLUtils.getAttrib(xsr, "from", null, true));
        target_gene = genemap.get(XMLUtils.getAttrib(xsr, "to", null, true));

        seqid = XMLUtils.getDoubleAttrib(xsr, "seqid", 0.0);
        
        while (xsr.hasNext())
        {
            int event = xsr.next();
            if (event == XMLStreamConstants.END_ELEMENT && xsr.getLocalName().equals(ORTHO_ELEM))
                return;
            
            if (event!=XMLStreamConstants.START_ELEMENT)
                continue;
            
            
            String le = xsr.getLocalName();
            
            if (le.equals("fromp"))
            {
                src_id = XMLUtils.getAttrib(xsr, "id", null, true);
                p1seq = XMLUtils.getAttrib(xsr, "seq", null, true);
                src = new RegionVector(xsr);
                continue;
            }
            if (le.equals("top"))
            {
                dst_id = XMLUtils.getAttrib(xsr, "id", null, true);
                p2seq = XMLUtils.getAttrib(xsr, "seq", null, true);
                dst = new RegionVector(xsr);
                continue;
            }
            if (le.equals("ali"))
            {
                ali = Alignment.parseAliFromXML(xsr);
                continue;
            }
            if (le.equals("genomicali"))
            {
                genomic_ali  = new GenomicProteinAlignment(xsr);
                continue;
            }
            
            if (le.equals("protein2protein"))
            {
                readProtein2ProteinAlignments(xsr,genemap);
            }
        }
    }
    private void readProtein2ProteinAlignments(XMLStreamReader xsr, HashMap<String, Gene> genemap) throws XMLStreamException
    {
        while (xsr.hasNext())
        {
            int event = xsr.next();
            
            if (event == XMLStreamConstants.END_ELEMENT && xsr.getLocalName().equals(ORTHO_ELEM))
                return;
            
            if (event!=XMLStreamConstants.START_ELEMENT)
                continue;
            
            OrthologInfo oi = new OrthologInfo(xsr,genemap);
            
        }
    }
    public OrthologInfo(Element e, HashMap<String, Gene> genemap)
    {
        src_gene = genemap.get(XMLUtils.getAttrib(e, "from", null, true));
        target_gene = genemap.get(XMLUtils.getAttrib(e, "to", null, true));

        Element fromE = e.getChild("fromp");
        Element dstE = e.getChild("top");

        src = new RegionVector(fromE);
        dst = new RegionVector(dstE);
        src_id = XMLUtils.getAttrib(fromE, "id", null, true);
        dst_id = XMLUtils.getAttrib(dstE, "id", null, true);
        p1seq = XMLUtils.getAttrib(fromE, "seq", null, true);
        p2seq = XMLUtils.getAttrib(dstE, "seq", null, true);

        seqid = XMLUtils.getDoubleAttrib(e, "seqid", 0.0);
        ali = Alignment.parseAliFromXML(e.getChild("ali"));
        
        Element ga = e.getChild("genomicali");
        if (ga!=null)
        {
            genomic_ali  = new GenomicProteinAlignment(ga);
        }
        
    
    }

    public GenomicProteinAlignment getGenomicAlignment()
    {

        if (genomic_ali!=null)
        {
            return genomic_ali;
        }


        
        GenomeSequenceExtractor src_genome = src_gene.organism.getGenome();
        GenomeSequenceExtractor dst_genome = target_gene.organism.getGenome();

        String cdna1 = GenomicUtils.getSplicedSeq(src_gene.chr, src_gene.strand, GenomicUtils.toGlobal(src,src_gene), src_genome, true);

        String check1 = GenomicUtils.translate(cdna1);
        if (!check1.equals(p1seq))
        {
            System.err.println("ERROR INVALID TRANSLATION for "+src_id);
            System.err.println("prot: "+ p1seq+ "\ncheck: "+check1);

        }
        String cdna2 = GenomicUtils.getSplicedSeq(target_gene.chr, target_gene.strand, GenomicUtils.toGlobal(dst, target_gene), dst_genome, true);
        String check2 = GenomicUtils.translate(cdna2);
        if (!check2.equals(p2seq))
        {
            System.err.println("ERROR INVALID TRANSLATION for "+dst_id);
            System.err.println("prot: "+ p1seq+ "\ncheck: "+check2);

        }
        return genomic_ali = new GenomicProteinAlignment(src, dst, p1seq, p2seq, cdna1, cdna2, ali, MAX_GAP);
        
    }
    

    public Element toXML()
    {
        
        Element oie = new Element(ORTHO_ELEM);
        oie.setAttribute("from", src_gene.id);
        oie.setAttribute("from_org", ""+src_gene.organism.getTaxId());
        oie.setAttribute("to_org", ""+target_gene.organism.getTaxId());
        oie.setAttribute("to", target_gene.id);
        
        Element srcE = src.getXML("fromp");
        srcE.setAttribute("id", src_id);
        srcE.setAttribute("seq", p1seq);
        oie.addContent(srcE);

        Element dstE = dst.getXML("top");
        dstE.setAttribute("id", dst_id);
        dstE.setAttribute("seq", p2seq);
        oie.addContent(dstE);

        oie.setAttribute("seqid", ""+seqid);
        

        oie.addContent(ali.getXML());


        oie.addContent(getGenomicAlignment().getXML("genomicali"));
        

        return oie;
    }


    public OrthologInfo reverse()
    {
        OrthologInfo oi = new OrthologInfo();
        oi.src_gene = target_gene;
        oi.target_gene = src_gene;
        oi.src = dst;
        oi.dst = src;
        oi.src_protein = dst_protein;
        oi.dst_protein = src_protein;
        oi.src_id = dst_id;
        oi.dst_id = src_id;
        oi.p1seq = p2seq;
        oi.p2seq = p1seq;
        if ( ali != null)
            oi.ali = ali.reverseAli();
        oi.seqid = seqid;
        
        
        return oi;
    }
    public OrthologInfo(Gene src, String src_id, String src_geneseq, RegionVector src_reg, 
                        Gene target, String target_id, String target_geneseq, RegionVector target_reg, 
                        ProteinAligner aligner)
    {
        src_gene = src;
        target_gene = target;

        this.src_id = src_id;
        this.dst_id = target_id;

        this.src = src_reg;
        this.dst = target_reg;
        p1seq = GenomicUtils.getSplicedSeq(src_geneseq, src_reg, false);
        p2seq = GenomicUtils.getSplicedSeq(target_geneseq, target_reg, false);

        ali = aligner.align(p1seq, p2seq);
        seqid = ali.getNumId(p1seq, p2seq)/getNormLength();
        
    }
    public OrthologInfo(Protein src, Protein target,  ProteinAligner aligner)
    {
        this(src,GenomicUtils.getProteinSeq(src, src.getOrganism().getGenome(),false),target,GenomicUtils.getProteinSeq(target, target.getOrganism().getGenome(),false),aligner);
    }
    public OrthologInfo(Protein src,String srcSequence, Protein target,String targetSequence,  ProteinAligner aligner)
	{
        
		src_gene = src.parentTranscript.parentGene;
		target_gene = target.parentTranscript.parentGene;;
		
		this.src_id = src.id;
		this.dst_id = target.id;
		
		this.src = src.getCodingRegionVector(false);
		this.dst = target.getCodingRegionVector(false);
		p1seq = srcSequence;
		p2seq = targetSequence;
		
		ali = aligner.align(p1seq, p2seq);
		num_id = ali.getNumId(p1seq, p2seq);
		seqid = num_id/getNormLength();
	}
    
    public OrthologInfo(Gene src, Gene target,  HashMap<Protein, String> ps1, HashMap<Protein, String> ps2, ProteinAligner aligner)
    {
        src_gene = src;
        target_gene = target;

        Protein p1 = null;
        Protein p2 = null;
        for (Map.Entry<Protein,String>  e1: ps1.entrySet())
        {
            String s1 = e1.getValue();
            for (Map.Entry<Protein,String>  e2: ps2.entrySet())
            {
                String s2 = e2.getValue();
               // Alignment a = aligner.align(s1, s2);
                //double normed = a.getNumId(s1,s2)/(0.5*(s1.length()+s2.length()));

                OrthologInfo oi = new OrthologInfo(e1.getKey(),s1,e2.getKey(),s2,aligner);
                
                if (oi.getSeqId()<=seqid){
                	continue;
                }
                ali = oi.ali;
                seqid = oi.getSeqId();
                p1 = e1.getKey();
                p2 = e2.getKey();
                

                p1seq = s1;
                p2seq = s2;
            }
        }
        this.src_protein = p1;
        this.dst_protein = p2;
        this.src_id = p1.id;
        this.dst_id = p2.id;
        this.src = p1.getCodingRegionVector(false);
        this.dst = p2.getCodingRegionVector(false);
        
    }
    /**
     * Harmonic mean of sequence  and gene structure identity
     * @return
     */
    public double getSeqGeneStructureId() {
    	return  2*((this.seqid*this.getGeneStructureId())/(this.seqid+this.getGeneStructureId())); 
    }
    
	public double getGeneStructureId() {
		double geneStructureId = this.getGenomicAlignment().getGeneStructureId();
		
    	if(log.isDebugEnabled() && seqid > 0.9 && src.size() == dst.size() &&geneStructureId != 1 ){
    		
    		log.debug(geneStructureId + "\t" + src.size() + "\t" + dst.size() + "\t" + this.seqid );
    		log.debug(this.getGenomicAlignment().getCheckAlignmentInfo());
    		
    	}
		
		return geneStructureId ;
	}
    
    public OrthologInfo(Gene src, Gene target, ProteinAligner aligner)
    {
    	this(src,target, GenomicUtils.getProteinSequences(src, src.organism.getGenome(), false),GenomicUtils.getProteinSequences(target, target.organism.getGenome(), false),aligner);
    }

    public String toString()
    {
        return String.format("%s(%s) -> %s (%s) %2.2f%% (id: %d on %d vs %d) ", 
                          src_gene.id, src_gene.organism.getSymbol(),
                          target_gene.id, target_gene.organism.getSymbol(),
                           100.0*seqid,
                          (ali==null)?0:ali.getNumId(p1seq, p2seq),
                          (p1seq==null)?0:p1seq.length(), (p2seq==null)?0:p2seq.length());
    }

    
}


