package ortholog;

import java.io.*;
import java.util.*;
import java.util.regex.Pattern;

import lmu.utils.*;
import lmu.utils.dag.*;
import lmu.utils.obo.*;
import algo.alignment.Gotoh;

public class Taxonomy
{

    Logger log = LogConfig.getLogger();
    HashMap<String, OboElement> lookup;
    HashMap<Integer, OboElement> tax2elem = new HashMap<Integer, OboElement>();

    HashMap<Integer, Integer> alt2tax2id = new HashMap<Integer, Integer>();

    Gotoh g = new Gotoh();
    static Pattern synP = Pattern.compile("\"([^\"]+)\"");

    File taxfile = null;

    public void extractSubSet(HashSet<Integer> taxids, PrintWriter to)
    {
        HashSet<OboElement> toextract = new HashSet<OboElement>();
        for (int ti: taxids)
        {
            OboElement oe = getTaxElement(ti);
            if (oe == null)
                continue;
            
            
            toextract.add(oe);
        }
        OboReader.extractSubTree(taxfile, to, toextract);
        to.flush();
        
        
    }
    
    public OboElement getRoot()
    {
       return lookup.get("root");
    }
    
    public Taxonomy(File f)
    {
        taxfile = f;
        OboReader taxa = new OboReader(f);
        Vector<OboElement> roots = taxa.readTree();
        lookup = (HashMap<String, OboElement>)ObjectGetter.getLookup(roots, "getName()");
        
        OboElement root = lookup.get("root");
        Collection<OboElement> taxelems = OboReader.getTreeElems(root);
        
        for (OboElement oe: taxelems)
        {
            String[] sp = oe.getTerm().split(":");
               
            if (sp.length<2)
            {
                log.warn("invalid taxonomy term: %s for element: %s", oe.getTerm(), oe.getName());
                continue;
            }
                
            try
            {

                int taxid = Integer.parseInt(sp[1]);
                tax2elem.put(taxid, oe);

                //wolf add alt_ids to taxonomy - makes it variable to older versions
                Vector<String> altids=oe.getEntries(OboField.ALT_ID);
                if (altids == null)
                    continue;
                
                for (String act_altid : altids)
                {
                    String[] parts = act_altid.split(":");
                    if (parts.length < 2)
                    {
                        log.warn("invalid taxonomy term: %s for element: %s", act_altid, oe.getName());
                        continue;
                    }
                    
                    alt2tax2id.put(Integer.parseInt(parts[1]), taxid);
                }
                
                
            }
            catch(NumberFormatException nfe)
            {
                log.warn("invalid taxonomy term: %s for element: %s", oe.getTerm(), oe.getName());
            }
                
            
        }
        
    }
    
    public int getTaxId(OboElement oe)
    {
        String[] sp = oe.getTerm().split(":");
        
        if (sp.length<2)
        {
            throw new FRuntimeException("invalid taxonomy term: %s for element: %s", oe.getTerm(), oe.getName());
        }
        
        try
        {
            return Integer.parseInt(sp[1]);
        }
        catch(Exception e)
        {
            throw new RuntimeException("could not parse tax id from "+sp[1]);
        }

    }
    Vector<Pair<Integer, OboElement>> getParents(OboElement e, OboElement tocheck)
    {
        Vector<Pair<Integer, OboElement>> rv = new Vector<Pair<Integer, OboElement>>();

        HashSet<OboElement> seen = new HashSet<OboElement>();
        LinkedList<Pair<Integer, OboElement>> parents = new LinkedList<>();
        parents.add(new Pair<Integer, OboElement>(0, e));

        while (!parents.isEmpty())
        {
            Pair<Integer, OboElement> p = parents.poll();

            if (tocheck!=null && p.getSecond()==tocheck)
            {
                rv.clear();
                rv.add(p);
                return rv;
            }
            rv.add(p);
            
            for (OboElement oe: p.getSecond().parents)
            {
                if (seen.contains(oe))
                    continue;
                
                parents.add(new Pair<Integer, OboElement>(p.getFirst()+1,oe));
            }
        }

        return rv;

    }
    public OboElement getLCA(OboElement oe1, OboElement oe2)
    {

        Vector<Pair<Integer, OboElement>> p1 = getParents(oe1, oe2);
        Vector<Pair<Integer, OboElement>> p2 = getParents(oe2, oe1);

        HashMap<OboElement,Integer> hm = new HashMap<OboElement,Integer>();
        for (Pair<Integer,OboElement> p : p1)
        {
            hm.put(p.getSecond(), p.getFirst());
        }
        int best = -1;
        OboElement lca = null;

        for (Pair<Integer,OboElement> p : p2)
        {
            Integer o = hm.get(p.getSecond());
            if (o==null)
                continue;
            
            if (best>=0 && best <= o+p.getFirst())
                continue;
            
            best = o+p.getFirst();
            lca = p.getSecond();
        }

        return lca;
    }
    
    public OboElement getParent(int tax)
    {
        OboElement oe = getTaxElement(tax);
        if (oe == null)
            throw new RuntimeException("tax node: "+tax+" is unknown!");
        
        if (oe.parents == null || oe.parents.size() == 0)
            return null;
        
        return oe.parents.get(0);
    }
    public OboElement getTaxElement(int tax)
    {
        OboElement oe = tax2elem.get(tax);
            
        if (oe != null)
            return oe;
        
        Integer mainTaxId = alt2tax2id.get(tax);

        return (mainTaxId == null) ? null : tax2elem.get(mainTaxId);
    
    }
    
    public Vector<Integer> getTaxonomyPath(int fromtax, int totax, HashSet<Integer> overtax)
    {
        HashMap<OboElement, HashSet<Integer>> oe2target = new HashMap<OboElement, HashSet<Integer>>();

        HashSet<Integer> allt = new HashSet<Integer>();
        allt.addAll(overtax);
        allt.add(fromtax);
        allt.add(totax);

        //push from the leafs
        for (int t : allt)
        {
            OboElement oe  = tax2elem.get(t);
            if (oe==null)
                throw new RuntimeException("cannot find taxonomy entry for: "+t);
            
            LinkedList<OboElement> q = new LinkedList<OboElement>();
            q.add(oe);
            while (!q.isEmpty())
            {
                OboElement p = q.poll();

                if (p.parents==null)
                {
                    continue;
                }
                q.addAll(p.parents);

                HashSet<Integer> h = oe2target.get(p);
                if (h==null)
                {
                     oe2target.put(p, h = new HashSet<Integer>());
                }
                h.add(t);
            }
        }

        Vector<Integer> rv = new Vector<Integer>();
        rv.add(fromtax);

        LinkedList<OboElement> q = new LinkedList<OboElement>();
        q.add(tax2elem.get(fromtax));
        HashSet<Integer> seen = new HashSet<Integer>();
        seen.add(fromtax);

        while (!q.isEmpty())
        {
            OboElement p = q.poll();

            HashSet<Integer> h = oe2target.get(p);


            if (h.contains(totax))
            {
                rv.add(totax);
                return rv;
            }

            SetInfo si = new SetInfo(seen, h, SetInfo.EXL_SET_2_FLAG);

            HashSet<Integer> novel = si.exl_set2;
            for (int n: novel)
            {
                rv.add(n);
            }

            seen.addAll(novel);

            if (p.parents!=null)
            {
                q.addAll(p.parents);
            }
            


        }

        return rv;
        
    }

    public DAG buildDAG(TaxonomyNode root)
    {
        DAG d = new DAG("taxonomy", "ncbi");
       
        LinkedList<TaxonomyNode> q = new LinkedList<>();
        
        
        DAGEntry droot = d.addDAGEntry(true, root.getName(), ""+root.getTaxId());
        //DAGEntry droot = d.getRoot();


        //droot.setName(root.getName());
        //droot.setTerm(""+root.getTaxId());
        droot.userObject = root;
        //DAGEntry droot = d.addDAGEntry(true, root.getName(), ""+root.getTaxId());
        //droot.userObject = root;

        
        HashMap<TaxonomyNode, DAGEntry> n2n = new HashMap<TaxonomyNode, DAGEntry>();
        n2n.put(root, droot);

        q.addAll(root.getChildren());

        while (!q.isEmpty())
        {
            TaxonomyNode tn = q.poll();
            
            DAGEntry ne = d.addDAGEntry(false, tn.getName(), ""+tn.getTaxId());
            ne.userObject = tn;

            n2n.put(tn, ne);
            DAGEntry p = n2n.get(tn.getParent());

            p.addChild(ne);

            q.addAll(tn.getChildren());
        }

        return d;
    }

    public TaxonomyNode buildSparseTaxonomyFromTaxIds(Collection<Integer> taxids, boolean collapse)
    {
        return buildSparseTaxonomy(ObjectGetter.buildMap(taxids,(Integer tax) -> (genome.Organism)null), collapse);
    }

    public TaxonomyNode buildSparseTaxonomy(Collection<genome.Organism> orgs, boolean collapse)
    {
        return buildSparseTaxonomy(ObjectGetter.buildReverseMap(orgs,(genome.Organism org) -> org.getTaxId()), collapse);
    }

    public TaxonomyNode buildSparseTaxonomy(HashMap<Integer, genome.Organism> orgs, boolean collapse)
    {
        LinkedList<TaxonomyNode> q = new LinkedList<>();

        HashMap<OboElement,Integer> known = new HashMap<>();
        HashSet<Integer> processed = new HashSet<>();

        for (Map.Entry<Integer,genome.Organism> e: orgs.entrySet())
        {
            genome.Organism o = e.getValue();
            OboElement oe = getTaxElement(e.getKey());
            if (oe == null)
            {
                log.warn("unknown taxid: %d, skipping", e.getKey());
                continue;
            }
            if (known.containsKey(oe))
            {
                log.warn("multi-defined taxnode last taxid: %d  last: %s", e.getKey(), known.get(oe));
                continue;
            }
            known.put(oe, e.getKey());
            
            //wolf
            if(o==null){
                q.add(new TaxonomyNode(null, oe.getName() , e.getKey()));
            }else{
            
            q.add(new TaxonomyNode(o, (o == null) ? ""+e.getKey() : o.getName(), e.getKey()));}
        }

        HashMap<Integer, TaxonomyNode> t2n = new HashMap<>();

        TaxonomyNode root = null;
        
        while(!q.isEmpty())
        {
            TaxonomyNode tin = q.poll();

            
            
            if(processed.contains(tin.getTaxId()))
                continue;
            
            processed.add(tin.taxid);
            t2n.put(tin.taxid, tin);
            
            OboElement p = getParent(tin.taxid);
            if (p == null)
            {
                root = tin;
                continue;
            }
            
            int pi = getTaxId(p);
            TaxonomyNode pn = t2n.get(pi);
            if (pn==null)
            {
                t2n.put(pi, pn = new TaxonomyNode(null, p.getName(), pi));
                
                q.add(pn);
            }
            
            pn.addChild(tin);

            
            
        }
        if (collapse)
        {
            root.collapse();
        }
        

        return root;
        
    }
    
    public static void main (String args[]) 
    {
        SimpleOptionParser cmd = new SimpleOptionParser("tax", "orgxml", "from", "to", "extract","tax_list");
        cmd.setFile("tax", "orgxml","tax_list");
        

        cmd.setOptional("from", "to", "extract","tax_list","orgxml");

        if (!cmd.parseParams(args,true, false, true, cmd))
            return;
        
        

        Taxonomy t = new Taxonomy(cmd.getFile("tax"));

        HashSet<Integer> taxes = new HashSet<Integer>();


        if ( cmd.isOptionSet("orgxml"))
        {
            Collection<genome.Organism> orgs = genome.Organism.readOrganisms(cmd.getFile("orgxml"));

            for (genome.Organism o: orgs)
            {
                taxes.add(o.getTaxId());
            }

        }
        if ( cmd.isOptionSet("tax_list"))
        {
            for(String line : FileUtils.getLines(cmd.getFile("tax_list")))
                taxes.add(Integer.valueOf(line));
        }
        
        if (cmd.isOptionSet("extract"))
        {
            PrintWriter pw = cmd.getWriter("extract");
            t.extractSubSet(taxes, pw);
            pw.close();
            return;
        }

        if (cmd.isOptionSet("from") && cmd.isOptionSet("to"))
        {
            
            genome.Organism fromO = genome.Organism.getOrganism(cmd.getValue("from"));
            genome.Organism toO = genome.Organism.getOrganism(cmd.getValue("to"));
    
            System.out.printf("path %s -> %s\n" , fromO.getSymbol(), toO.getSymbol());
            
    
            Vector<Integer> path = t.getTaxonomyPath(fromO.getTaxId(), toO.getTaxId(), taxes);
    
            for (int p : path)
            {
                genome.Organism o = genome.Organism.getOrganism(p);
                System.out.println("next: "+o.getSymbol());
            }
        }
    }
}
