package ortholog;

import java.io.*;
import java.util.*;

import genome.Organism;

public class TaxonomyNode
{
    String name;
    int taxid;
    TaxonomyNode parent = null;
    Organism org;

    HashSet<TaxonomyNode> children = new HashSet<>();
    
        
    public TaxonomyNode(Organism org, String name, int taxid)
    {
        this.org = org;
        this.name= name;
        this.taxid=taxid;
    }
    
    public Collection<TaxonomyNode> getChildren()
    {
        return children;
    }

    public int getTaxId()
    {
        return taxid;
    }
    public String getName()
    {
        return name;
    }
    
    public TaxonomyNode getParent()
    {
        return parent;
    }
    public String getShortName(){
        String[] name = this.getName().split("\\s+");
        if ( name.length < 2) return name[0];
        return name[0].charAt(0) + "." + name[1].replaceAll("[\\W]|_", "");
    }
    public HashSet<TaxonomyNode> getLeafs()
    {
        HashSet<TaxonomyNode> leafs = new HashSet<TaxonomyNode>();

        LinkedList<TaxonomyNode> q = new LinkedList<TaxonomyNode>();
        q.add(this);

        while (!q.isEmpty())
        {
            TaxonomyNode tn = q.poll();
            if (tn.getChildren().size() == 0)
            {
                leafs.add(tn);
                continue;
            }
            q.addAll(tn.getChildren());
        }
        return leafs;
    }
    public HashMap<Integer,TaxonomyNode> getTaxMap()
    {
        HashMap<Integer,TaxonomyNode> rv = new HashMap<>();

        LinkedList<TaxonomyNode> q = new LinkedList<TaxonomyNode>();
        q.add(this);

        while (!q.isEmpty())
        {
            TaxonomyNode tn = q.poll();
            rv.put(tn.getTaxId(), tn);
            q.addAll(tn.getChildren());
        }
        return rv;

    }
    HashMap<Organism, TaxonomyNode> o2tax = null;

    public HashMap<Organism, TaxonomyNode> getOrganismMapping()
    {
        if (o2tax!=null)
            return o2tax;
        
        
        HashMap<Organism, TaxonomyNode> rv = new HashMap<>();
        LinkedList<TaxonomyNode> q = new LinkedList<TaxonomyNode>();
        q.add(this);

        while (!q.isEmpty())
        {
            TaxonomyNode tn = q.poll();
            if (tn.org!=null)
            {
                rv.put(tn.org, tn);
            }

            q.addAll(tn.children);
        }
        return o2tax = rv;
    }

    boolean hasDescendant(HashSet<TaxonomyNode> tnset)
    {
        for (TaxonomyNode tn : tnset)
        {
            if (!isDescendant(tn))
                continue;
            
            return true;
        }
        return false;
    }
    boolean isDescendant(TaxonomyNode tn)
    {
        if (this == tn || children.contains(tn))
            return true;
        
        
        for (TaxonomyNode ch : children)
        {
            if (!ch.isDescendant(tn))
                continue;
            
            return true;
        }
        return false;
    }
    
    /** returns true if at least two diff children have one node as a leaf*/
    public boolean isJoining(HashSet<TaxonomyNode> nodes)
    {
        int rel_children = 0;

        for (TaxonomyNode ch : children)
        {
            if(!ch.hasDescendant(nodes))
                continue;

            if (++rel_children > 1)
                return true;
        }
        return false;
    }
    public static TaxonomyNode getLCA(TaxonomyNode node1, TaxonomyNode node2)
    {
    	TaxonomyNode lca = node1;
        while (!lca.isDescendant(node2))
        {
            lca = lca.parent;
        }
        
        return lca;
        
        
    }
    public static TaxonomyNode getLCA(Collection<TaxonomyNode> nodes)
    {
        if (nodes.size() == 0)
            return null;
        
        
        Iterator<TaxonomyNode> it = nodes.iterator();

        TaxonomyNode lca = it.next();
        while (it.hasNext())
        {
            TaxonomyNode tn = it.next();
            while (!lca.isDescendant(tn))
            {
                lca = lca.parent;
            }
        }
        return lca;
        
        
    }
    public Organism getOrganism()
    {
        return org;
    }
    boolean hasSingleChild()
    {
        return (children.size()==1);
    }
    
    void collapse()
    {
        Vector<TaxonomyNode> collapsed = new Vector<>();

        for (TaxonomyNode tin : children)
        {
            if (!tin.hasSingleChild())
            {
                collapsed.add(tin);
                continue;
            }

            while (tin.hasSingleChild())
            {
                tin = tin.children.iterator().next();
            }
            collapsed.add(tin);
        }
        children.clear();
        children.addAll(collapsed);
        
        for (TaxonomyNode tin: children)
        {
            tin.parent = this;
            tin.collapse();
        }
    }

    public String toString()
    {
        return getInfo(null);
    }

    public interface InfoGetter<TaxonomyNode>
    {
        public String info(TaxonomyNode t);
    }

    public String getNewick(Vector<TaxonomyNode> order)
    {
        return getNewick(null, order, 1);
    }

    boolean nodepth = true;
    public String getNewick(InfoGetter<TaxonomyNode> info_getter, Vector<TaxonomyNode> order, int depth)
    {

        order.add(this);
        String info = (info_getter == null)? name : info_getter.info(this);
        info = info.replaceAll(" ", "");

        if (children.size() == 0)
        {
            if (nodepth)
            {
                
                if (parent == null)
                    return "("+info+");";
           
                    
                return info;
            }
            if (parent == null)
                return "("+info+":"+(depth)+");";
           
            return info+":"+(depth);
            
        }
            
       StringBuffer sb = new StringBuffer();
       sb.append(info);
       if (parent!=null && !nodepth)
       {
           sb.append(':');

           sb.append(""+(depth));
       }

       sb.append(',');
       sb.append('(');
       boolean first=true;
       for (TaxonomyNode tn : children)
       {
           if (first)
           {
               first = false;
           }
           else
           {
               sb.append(',');
           }
           sb.append(tn.getNewick(info_getter, order, 2*depth));
       }
       sb.append(')');
       
              
       if (parent == null)
       {
           sb.append(';');
       }
       return sb.toString();
    }
    public String getInfo(InfoGetter<TaxonomyNode> info_getter)
    {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        print(pw, "\t", info_getter);
        pw.close();
        
        return sw.getBuffer().toString();
    }

    public void addChild(TaxonomyNode tin)
    {
        children.add(tin);
        tin.parent = this;
    }
    
    public void print(PrintWriter pw)
    {
        print(pw, "\t", null);
    }

    public void print(PrintWriter pw, InfoGetter<TaxonomyNode> info_getter)
    {
        print(pw, "\t", info_getter);
    }

    public void print(PrintWriter pw, String sep, InfoGetter<TaxonomyNode> info_getter)
    {
        
        String info = (info_getter == null) ? "" : info_getter.info(this);
        if (info == null)
            info = "";
        
        pw.printf("%s%s (%d children tax: %d) %s\n", sep, name, children.size(), taxid, info);

        final String nsep ="\t"+sep;
        for (TaxonomyNode child: children)
        {
            child.print(pw, nsep, info_getter);
        }
    }
    
    /**
     * Number of inner nodes to root
     */
    public int getHeight(){
    	int height = 0;
    	TaxonomyNode n = this;
    	
    	while(n != null){
    		n = n.getParent();
    		height++;
    	}
    	return height;
    }
    /**
     * Height of <b>from<b> minus height of LCA(from,to).
     */
    public static int getDistance(TaxonomyNode from,TaxonomyNode to,TaxonomyNode root)
    {
    	return from.getHeight() - TaxonomyNode.getLCA(from, to).getHeight();
    }
    
    
    public static class SpeciesDistance implements Comparable<SpeciesDistance> 
    {
    	public Integer distance;
    	public List<Organism> organisms;
    	public SpeciesDistance(Integer distance)
    	{
    		this.distance = distance;
    		this.organisms = new ArrayList<Organism>();
    	}
    	
		@Override
		public int compareTo(SpeciesDistance o) {
			return this.distance.compareTo(o.distance);
		}
    }
    public static List<SpeciesDistance> getDistances(Organism from, TaxonomyNode root)
    {
    	HashMap<Integer,SpeciesDistance> dist2Sp = new HashMap<Integer,SpeciesDistance>();
		
		TaxonomyNode start = root.getOrganismMapping().get(from);
		if ( start == null) throw new RuntimeException(from + " not in " + root);
		List<Organism> orgs =  new ArrayList<Organism>(root.o2tax.keySet());
		
		Collections.sort(orgs, (Organism o1, Organism o2) -> { return Integer.valueOf(o1.getTaxId()).compareTo(o2.getTaxId()); }); 
		Collections.reverse(orgs);
		
		for(Organism o : orgs)
		{
			TaxonomyNode node = root.getOrganismMapping().get(o);
			
			int distance = TaxonomyNode.getDistance(start,node,root);
			
			if (! dist2Sp.containsKey(distance)) dist2Sp.put(distance, new SpeciesDistance(distance));
			
			dist2Sp.get(distance).organisms.add(o);
			
		}
		List<SpeciesDistance> ret = new ArrayList<SpeciesDistance>(dist2Sp.values());
		Collections.sort(ret);
		
		return ret;
	}

}


