
package fasta;

import java.io.*;
import java.util.zip.ZipEntry;
import java.util.zip.GZIPInputStream;
import lmu.utils.*;
import java.util.*;

public abstract class FastaLikeReader<A, B> implements Iterator<Pair<A,B> >
{
    static final String GZIPSUFFIX=".gz";
    A object;
    
    String header = null;
    Pair<A, B> next_entry=null;

    HeaderParser<A> parser;
    
    BufferedReader bfr = null;
    //File infile=null;
    String info=null;
    boolean is_gzipped=false;
    boolean ignore_init_header=false;

    int ln=0;

    boolean before_first=true;

    
    public FastaLikeReader(HeaderParser<A> parser, File fastafile) 
    {
        this(parser, fastafile, null, false);
    }
    public FastaLikeReader(HeaderParser<A> parser, InputStream is, String info) 
    {
        init(parser,is,info, false);
    }

    public FastaLikeReader(HeaderParser<A> parser, File fastafile, Boolean gzipped, boolean ignore_invalid_header) 
    {
        //infile=fastafile;
        if (gzipped==null)
        {
            gzipped=fastafile.getName().endsWith(GZIPSUFFIX);
        }

        InputStream is = null;
        is_gzipped=gzipped;
        try
        {
        
            FileInputStream fis = new FileInputStream(fastafile);
            is = (gzipped)?new GZIPInputStream(fis):fis;
        }
        catch(IOException ie)
        {
            throw new RuntimeException("i/o error: "+ie.getMessage()+" while starting to read fasta entries from "+fastafile.getAbsolutePath(), ie);
        }

        init(parser,is, fastafile.getAbsolutePath(), ignore_invalid_header);
    }


    public void init(HeaderParser<A> parser, InputStream is, String info, boolean ignore_invalid_header) 
    {
        this.parser=parser;
        bfr = new BufferedReader(new InputStreamReader(is));
        ignore_init_header=ignore_invalid_header;
        
        
    }

    public static void writeFasta(String seq, PrintWriter pw)
    {
        writeFasta(seq,pw,80);
    }

    public static void writeFasta(String seq, PrintWriter pw, final int step)
     {
         final int l = seq.length();
         
         for (int i=0; i<l; i+=step)
         {
             int end=Math.min(i+step, l);
             pw.printf("%s\n", seq.substring(i, end));
         }
     }


    String readNextHeader(boolean ignore_invalid_headers) 
    {
        if (header!=null)
        {
            return header;
        }
        
        if (bfr==null)
            return null;
        
        
        try
        {
            
            while (header==null)
            {
                header=bfr.readLine();
                if (header==null)
                {
                    break;
                }
                ln++;
                if (header.length()==0)
                {
                    continue;
                }
                if (header.charAt(0)!='>')
                {
                    if (ignore_invalid_headers)
                    {
                        header=null;
                        continue;
                    }
                    throw new RuntimeException("error while reading line "+ln+" from: "+info+" gzipped: "+is_gzipped+" expected header, but got: "+header);
                }
            }
            return header;
        }
        catch(IOException e)
        {
            throw new RuntimeException("i/o error: "+e.getMessage()+" at line: "+ln+" while reading fasta from "+info, e);
        }
    }

    public void close()
    {
        if (bfr==null)
            return;

        try
        {
            bfr.close();
        }
        catch(IOException ie)
        {
            throw new RuntimeException("error at closing fasta file: "+info+" gzipped: "+is_gzipped);
        }
        
        bfr=null;
    }

    protected abstract Object initValue();
    protected abstract void updateValue(Object v, String line);
    protected abstract B finishValue(Object v);

    void readNext(boolean ignore_invalid_header) 
    {
        next_entry=null;
        String entry_header = readNextHeader(ignore_invalid_header);
        if (entry_header==null && bfr!=null)
        {
            close();
            
            
        }
        if (entry_header==null)
        {
            return;
        }
        header=null;
        String line=null;
        
        Object v = initValue();

        try
        {
            
            while (null!=(line=bfr.readLine()))
            {
                ln++;
    
                if (line.length()==0)
                    continue;
                
                
                if (line.charAt(0)=='>')
                {
                    header=line;
                    break;
                }
                updateValue(v,line);
                
            }
        }
        catch(IOException e)
        {
            throw new RuntimeException("i/o error: "+e.getMessage()+" at line: "+ln+" while reading fasta from "+info, e);
        }


        next_entry = new Pair<A, B>(parser.parse(entry_header.substring(1)), finishValue(v));

    }

    public boolean hasNext()
    {
        if (before_first)
        {
            readNext(ignore_init_header);
            before_first=false;
        }
        return next_entry!=null;
    }

    public Pair<A,B> next()
    {
        if (next_entry==null)
        {
            return null;
        }
        Pair<A,B> rv = next_entry;
            
        readNext(false);
        
        
        
        return rv;
    }

    public void remove()
    {
        throw new RuntimeException("cannot delete from fasta file!");
    }
    
}
