package fasta;

import java.util.regex.*;
import java.util.*;
import lmu.utils.*;
import java.io.*;

public class SimpleFastaReader extends FastaReader<String>
{
    static class NoConverter implements HeaderParser<String>
    {
        
        public String parse(String fasta_header)
        {
            return fasta_header;
        }
    }
    static NoConverter conv = new NoConverter();

    public SimpleFastaReader(File infile) 
    {
        this(infile, null);
    }
    public SimpleFastaReader(File infile, Boolean gzipped) 
    {
        super(conv, infile, gzipped);
    }
    


    public interface StringTransformator
    {
        public String transform(String s);
    }

    public static class MatchTransformator implements StringTransformator
    {
        java.util.regex.Pattern matchP;
        int pos;

        public MatchTransformator(String regexp, int group)
        {
            matchP = java.util.regex.Pattern.compile(regexp);
            pos=group;
        }
        public String transform(String s)
        {
            Matcher m = matchP.matcher(s);
            if (!m.find())
            {
                throw new RuntimeException("match transformator cannot convert "+s+" to an id with pattern "+matchP.pattern()+" and group "+pos);
            }
            if(m.groupCount()<pos)
            {
                throw new RuntimeException("match transformator cannot convert "+s+" to an id with pattern "+matchP.pattern()+" and group "+pos);
            }
            return m.group(pos);
        }
    }
    public static class SplitStringTransformator implements StringTransformator
    {
        java.util.regex.Pattern splitP;
        int pos;

        public SplitStringTransformator(String pat, int pos)
        {
            splitP = java.util.regex.Pattern.compile(pat);
            this.pos=pos;
        }

        public String transform(String s)
        {
            String sp[] = splitP.split(s);
            if (sp.length<=pos)
            {
                throw new RuntimeException("could not extract fasta key from "+s+" by using "+pos+"-th element after split by "+splitP.pattern());
            }
            return sp[pos];
        }
    }

    public static StringTransformator getTransformator(String pattern, int pos, boolean splitter)
    {
        if (splitter)
        {
            return new SplitStringTransformator(pattern, pos);
        }
        return new MatchTransformator(pattern, pos);
    }

    public static Pair<String, String> readOne(File f, String key, String pattern, int pos, boolean matcher)
    {
        return readOne(f,key, getTransformator(pattern, pos, !matcher));
    }
    public static Pair<String, String> readOne(File f, String key)
    {
        return readOne(f,key,null);
    }


    public static Pair<String, String> readOne(File f, String key, StringTransformator trans)
    {
        SimpleFastaReader sfr = new SimpleFastaReader(f);
        
        while (sfr.hasNext())
        {
            Pair<String,String> p = sfr.next();
            String tocheck=p.getFirst();
            if (trans!=null)
            {
                tocheck = trans.transform(p.getFirst());
            }
            if (!key.equals(tocheck))
            {
                continue;
            }

            sfr.close();
            return p;
        }
        return null;
    }
    public static HashMap<String, String> readAll(File f)
    {
        return readAll(f,null);
    }
    public static HashMap<String, String> readAll(File f, String pattern, int index, boolean matcher)
    {
        return readAll(f, getTransformator(pattern,index, !matcher));
    }
    public static HashMap<String, String> readAll(File f, StringTransformator trans)
    {
        SimpleFastaReader sfr = new SimpleFastaReader(f);
        HashMap<String,String> rv = new HashMap<String,String>();
        int report_num=0;
        while (sfr.hasNext())
        {
            Pair<String,String> p = sfr.next();
            String k=p.getFirst();
            if (trans!=null)
            {
                k = trans.transform(p.getFirst());
            }
            if (false && ++report_num<10)
            {
                System.out.printf("read >%s< (%s) from %s trans :%s\n", k, p.getFirst(), f.getAbsolutePath(), trans);
            }
            
            rv.put(k, p.getSecond());
        }
        return rv;
    }
    public static class CmdLine extends OptionParser<SimpleFastaReader>
    {
        static final int INFILE_INDEX=0;
        static final int FILTER_INDEX=1;
        static final int VERBOSE_INDEX=2;
        boolean verbose=false;
        File fasta;
        public CmdLine()
           {

               registerOneParamInputFileOption(INFILE_INDEX, "fasta", "fastafile");
               registerOneParamBoolOption(VERBOSE_INDEX, "v", verbose);
               

           }

           public boolean parse(int option_index, String[] args, int i, StringBuffer error)
           {
               switch(option_index)
               {
                   case INFILE_INDEX:
                       fasta=new File(args[i+1]);
                       return true;

                   case VERBOSE_INDEX:
                       verbose = Boolean.parseBoolean(args[i+1]);
                       return true;
                   default:
                       return false;
               }

           }


           public String getName()
           {
               return "FastaTest";
           }

           public SimpleFastaReader getObject()
           {
               return new SimpleFastaReader(fasta);           
               
           }

    }

    public static void main (String args[]) throws Exception
    {
        CmdLine cmdline = new CmdLine();
        if(!OptionParser.parseParams(args,true, true, true, cmdline))
            return;

        SimpleFastaReader sf = cmdline.getObject();
        int nentries=0;
        while (sf.hasNext())
        {
            Pair<String,String> n = sf.next();

            if(++nentries%1000==0)
            {
                System.out.println("read "+nentries+" entries");
            }
            if (cmdline.verbose)
            {
                System.out.println("next entry: "+n.getFirst()+" seq length: "+n.getSecond().length());
            }
        }

        System.out.println("fasta file: "+ cmdline.fasta.getAbsolutePath()+" contains "+nentries+" entries");
    }
}
