package genome;

import java.io.*;
import java.nio.channels.NonWritableChannelException;
import java.util.*;
import java.util.regex.*;

import org.mapdb.DB;
import org.mapdb.DBMaker;

import genome.config.GenomicConfigManager;

import lmu.utils.*;
import lmu.utils.tuple.*;


public class GTFAnnotation implements GenomeAnnotationHandler       
{
    public static final String PROTEIN_CODING = "protein_coding";
    Organism organism;

    
    public static HashSet<String> PROTCODING_TYPES = StringUtils.buildSet(
        PROTEIN_CODING,
        "IGC gene", "IGD gene", "IG gene", "IGJ gene", "IGLV gene", 
        "IGM gene", "IGV gene", "IGZ gene", "nonsense mediated decay", 
        "nontranslating CDS", "non stop decay", "polymorphic pseudogene", "TRC gene", "TRD gene", "TRJ gene");

    public static HashSet<String> LNCRNA_TYPES = StringUtils.buildSet("3prime overlapping ncrna", "ambiguous orf", "antisense", "antisense RNA", "lincRNA", 
                                                             "ncrna host", "processed transcript", "sense intronic", "sense overlapping");


    public Organism getOrganism()
    {
        return organism;
    }
    Logger log = LogConfig.getLogger();
    enum GTFHEADER
    {
        CHR(0),
        BIOTYPE(1),
        ENRYTYPE(2),
        START(3),
        END(4),
        SCORE(5),
        STRAND(6),
        FRAME(7),
        ATTRIBUTE(8)
        ;

        int index;
        GTFHEADER(int i)
        {
            index=i;
        }
        
    }

    enum  ATTRIB_TYPE
    {
        GENEID("gene_id"),
        GENE_VERSION("gene_version"),
        TRANSCRIPT_VERSION("transcript_version"),
        EXON_VERSION("exon_version"),
        PROTEIN_VERSION("protein_version"),
        TRANSCRIPT_BIOTYPE("transcript_biotype"),
        TRANSCRIPT_TYPE("transcript_type"),
        TRANSCRIPTID("transcript_id"),
        EXONNUMBER("exon_number"),
        HAVANA_GENE("havana_gene"),
        HAVANA_TRANSCRIPT("havana_transcript"),
        HAVANA_TRANSCRIPT_VERSION("havana_transcript_version"),
        HAVANA_GENE_VERSION("havana_gene_version"),
        
        EXONID("exon_id"),
        GENENAME("gene_name"),
        TRANSCRIPTNAME("transcript_name"),
        PROTEINID("protein_id", "p_id"),
        
        BIOTYPE("gene_biotype"),
        GENE_TYPE("gene_type"),
        GENE_STATUS("gene_status"),
        SEQEDIT("seqedit"),
        TRANSCRIPT_SOURCE("transcript_source"),
        TRANSCRIPT_STATUS("transcript_status"),
        TRANSCRIPT_SUPPORT_LEVEL("transcript_support_level"),
        GENE_SOURCE("gene_source"),
        ONTOLOGY("ont"),
        TAG("tag"),
        CCDSID("ccds_id", "ccdsid"),
        PEPTIDE("peptide")
        ;
        String name;
        String[] names;
        ATTRIB_TYPE(String ... names)
        {
            this.name=names[0];
            this.names = names;
            for (String n : names)
            {
                EnumGetter.add(n, this);
            }
            

        }
        static ATTRIB_TYPE get(String n)
        {
            return (ATTRIB_TYPE)EnumGetter.get(n);
        }

    }
    enum  ENTRY_TYPE
    {
        GENE("gene"),
        TRANSCRIPT("transcript"),
        EXON("exon"),
        CDS("CDS"),
        NCRNA("ncRNA"),
        START_CODON("start_codon"),
        STOP_CODON("stop_codon"),
        UTR("UTR"),
        UTR5P("five_prime_utr"),
        UTR3P("three_prime_utr"),
        SELENOCYSTEINE("Selenocysteine") ; 

        String name;
        ENTRY_TYPE(String n)
        {
            name=n;
            EnumGetter.add(name, this);
        }

        static ENTRY_TYPE get(String n)
        {
            return (ENTRY_TYPE)EnumGetter.get(n);
        }

    }

    static Pattern attribP = Pattern.compile("(\\w+) \"([^\"]+)\";");
    HashMap<String, Gene> id2maingene =new HashMap<String, Gene>();
    //HashMap<String, Protein> id2prot =new HashMap<String, Protein>();
    HashMap<String, HashSet<Gene>> id2genes =new HashMap<String, HashSet<Gene>>();
    //Vector<Gene> genes = new Vector<Gene>();

    int ngenes=0;
    Vector<Pair<String,Integer> > chr2lengths = new Vector<Pair<String,Integer> >();

    HashMap<String, String> prot2seq = null;

    //HashMap<String, StringBuffer> gene2buffer = new HashMap<String, StringBuffer>();
    HashMap<String, HashSet<String>> biotype2gene = new HashMap<String, HashSet<String>>();
    HashMap<String, GeneLookupInfo> gene2buffer = new HashMap<String, GeneLookupInfo>();
    HashMap<String, Vector<GeneLookupInfo>> ordered_infos = null;
    Map<String, HashSet<String>> name2geneid = null;

    public HashMap<String, String> tr2gene = new HashMap<String, String>();
    public HashMap<String, String> p2gene = new HashMap<String, String>();
    String EXON_FMT = "EXON%06d";
    File gtf;
    int biggest_buffer = 0;
    
    
    public static class ChrInfo
    {
        String chr;
        int maxpos = -1;

        Overlap2DDetector od = new Overlap2DDetector();
        Overlap2DDetector ig_od = new Overlap2DDetector();

        public Vector<GeneLookupInfo> ordered;
        static Pair<Double, Double> DIM2 = Pair.create(0.0,1.0);

        public ChrInfo(String chr, Collection<GeneLookupInfo> c)
        {
            this.chr = chr;

            maxpos = NumUtils.max(c, (GeneLookupInfo g) -> g.end) + 1;
            ordered = ObjectGetter.toVector(c);
            Collections.sort(ordered, (GeneLookupInfo gl1, GeneLookupInfo gl2) -> gl1.start - gl2.start);

            GeneLookupInfo last = null;
            int lastx = 0;
            
            Vector<InterGenicRegion> igs = new Vector<>();
            for (GeneLookupInfo gli : ordered)
            {
                if (gli.start > lastx)
                {
                    igs.add(new InterGenicRegion(chr, lastx, gli.start, (last == null) ? null : last.gene, gli.gene));
                }
                if (gli.end > lastx)
                {
                    lastx = gli.end;
                    last = gli;
                }
                
                od.add(Pair.create((double)gli.start, (double)gli.end), DIM2, gli);
                
            }
            igs.add(new InterGenicRegion(chr, lastx, lastx + 1_000_00, (last == null) ? null : last.gene, null));

            ObjectGetter.apply(igs, (InterGenicRegion igr) -> ig_od.add(Pair.create((double)igr.region.getX1(), (double)igr.region.getX2()), DIM2, igr));
            
        }

        public Iterator<InterGenicRegion> getOverlappingIG(Integer start, Integer end)
        {
            return ObjectGetter.getTransformedIterator(
                        ig_od.getOverlaps(
                                (start == null) ? 0.0 : start.doubleValue(),
                                (end == null) ? (double)maxpos : end.doubleValue(),
                                 DIM2.getFirst().doubleValue(),
                                DIM2.getSecond().doubleValue(),
                                null
                                ).iterator(),
                        (Object o) -> (InterGenicRegion)o);
        }
        public Iterator<GeneLookupInfo> getOverlapping(Integer start, Integer end)
        {

            return ObjectGetter.getTransformedIterator(
                        od.getOverlaps(
                                (start == null) ? 0.0 : start.doubleValue(),
                                (end == null) ? (double)maxpos : end.doubleValue(),
                                 DIM2.getFirst().doubleValue(),
                                DIM2.getSecond().doubleValue(),
                                null
                                ).iterator(),
                        (Object o) -> (GeneLookupInfo)o);

        }


    }

    public static class InterGenicRegion
    {
        public String chr;
        public Region1D region;
        
        public String pre_gene;
        public String post_gene;
        
        public InterGenicRegion(String chr, int start, int end, String pre_gene, String post_gene)
        {
            this.chr = chr; 
            region = new Region1D(start, end);
            this.pre_gene = pre_gene;
            this.post_gene = post_gene;
        }
    }

    HashMap<String, ChrInfo> chr_infos = null;

    Set<String> getChrs()
    {
        if (chr_infos != null)
            return chr_infos.keySet();
            
        getChrInfo("tmp");

        return chr_infos.keySet();
    }
    
    public Iterator<Gene> getCloseGenes(Gene g, boolean upstream, boolean strand_specific)
    {
        return new IteratorUtils.FilterIterator<Gene>(getCloseGenes(g.chr, (upstream) ? g.start : g.end, upstream, (!strand_specific) ? null : g.strand), (_g) -> (upstream) ? _g.start > g.start : _g.start < g.start);
    }

    public Iterator<Gene> getCloseGenes(String chr, int position, boolean upstream, Boolean strand)
    {
        ChrInfo ch = getChrInfo(chr);
        if (ch == null)
            return null;
        
        Pair<Boolean, Integer> p = VectorUtils.binarySearch(ch.ordered, position, (g, _p) -> Integer.compare(((upstream) ? g.end : g.start), _p));


        Iterator<Integer> idxIt = (upstream) ? IteratorUtils.getIntIterator(p.getSecond(), ch.ordered.size()) :
                                    IteratorUtils.getIntIterator(p.getSecond() -1 , -1, -1);
                        
        return ObjectGetter.getTransformedIterator(
                new IteratorUtils.FilterIterator<Integer>(idxIt, (_idx) -> strand == null || strand == (ch.ordered.get(_idx)).strand),
                (_idx2) -> getGene(ch.ordered.get(_idx2).gene));
    }


    public Gene getGeneByName(String gene_name)
    {
        if(name2geneid == null)
            throw new FRuntimeException("no gene name lookup given for: gtf file: %s", gtf.getAbsolutePath());

        HashSet<String> gn = name2geneid.get(gene_name);
        if (gn == null)
            return null;

        if (gn.size() > 1)
            throw new FRuntimeException("gene name: >%s< not uniq got mapping genes: %s\n", 
                                        StringUtils.joinObjects("\n\t", ObjectGetter.map(gn, (_gn) -> getGene(_gn).getLocationInfo())));
        
        return getGene(ObjectGetter.first(gn));
    }

    public synchronized ChrInfo getChrInfo(String chr)
    {
        
        if (chr_infos != null)
            return chr_infos.get(chr);

        HashMap<String, HashSet<GeneLookupInfo>> mm = ObjectGetter.buildReverseMultiMap(gene2buffer.values(), (GeneLookupInfo gli) -> gli.chr);

        chr_infos = new HashMap<>();

        for (Map.Entry<String, HashSet<GeneLookupInfo>> e: mm.entrySet())
        {
            chr_infos.put(e.getKey(), new ChrInfo(e.getKey(), e.getValue()));
        }

        return chr_infos.get(chr);
    }

    final static String CHR_PREFIX = "chr";

    static Vector<String> getChrVariations(String chr)
    {
        if (chr == null)
            return new Vector<String>();


        Vector<String> v = ObjectGetter.toVector(chr);

        if (chr.startsWith(CHR_PREFIX))
        {
            v.add(chr.substring(CHR_PREFIX.length()));
        }
        else
        {
            v.add(CHR_PREFIX + chr);
        }
        return v;
    }

    
    public ChrInfo findChrInfo(String chr)
    {
        ChrInfo c = null;
        for (String s  : getChrVariations(chr))
        {
            if (null == (c = getChrInfo(s)))
                continue;
            return c;
        }
        return null;
    }
    
    public Iterator<InterGenicRegion> getInterGenicRegions()
    {
        return getInterGenicRegions(null, null, null);
    }

    public Iterator<InterGenicRegion> getInterGenicRegions(String chr, Integer start, Integer end)
    {
        if (chr == null)
            return new IteratorIterator(ObjectGetter.map(getChrs(), (String _chr) -> getInterGenicRegions(_chr, null, null)));
        
        ChrInfo c = findChrInfo(chr);
        return (c==null) ? null : c.getOverlappingIG(start,end);
    }


    public String getTranscriptBioType(String trid)
    {
        String gn = getGeneIdByTranscriptId(trid);
        if (gn == null)
            return null;

        return getGeneInfo(gn).biotype;
    }
    public Transcript getTranscript(String trid)
    {
        String gn = getGeneIdByTranscriptId(trid);
        if (gn == null)
            return null;

        Gene g = getGene(gn);
        if (g == null)
            return null;
        
        return g.transcripts.get(trid);
    }

    public Iterator<Gene> getAllGeneIterator()
    {
        Vector<GeneLookupInfo> v = ObjectGetter.toSortedVector(gene2buffer.values(), (g1, g2) -> Long.compare(g1.fp, g2.fp), true);
        System.out.printf("fps: %s\n", VectorUtils.slice(ObjectGetter.map(v, (g) -> g.fp), 0, 20));
        return ObjectGetter.getTransformedIterator(v.iterator(), (GeneLookupInfo g) -> getGene(g.gene));

    }

    public Iterator<Gene> getGenes(String chr, Integer start, Integer end)
    {
        if (chr==null)
            return ObjectGetter.getTransformedIterator(getOrderedBufferIterator(), (_gli) -> getGene(_gli.gene));
            
        
        ChrInfo c = findChrInfo(chr);
        if (c == null)
            return null;
        
        Vector<GeneLookupInfo> v = ObjectGetter.toVector(c.getOverlapping(start,end));
        NumUtils.sort(v, (gli) -> gli.fp);

        return ObjectGetter.getTransformedIterator(v.iterator(), (GeneLookupInfo g) -> getGene(g.gene));
    }


    public Iterator<GeneLookupInfo> getOverlappingGeneInfos(String chr, Integer start, Integer end)
    {
        ChrInfo c = findChrInfo(chr);
        return (c==null) ? null : c.getOverlapping(start,end);
    }

    public int getNumProteins()
    {
        return (p2gene==null)?-1:p2gene.size();
    }
    public void setProteinFasta(File fastafile)
    {
        long t1 = System.currentTimeMillis();
        log.info("will read protein fasta: "+fastafile.getAbsolutePath());
        prot2seq = new HashMap<String,String>();
        fasta.FastaReader fr = new fasta.FastaReader(new fasta.RegexGroupParser(java.util.regex.Pattern.compile("([^\\s]+).+")), fastafile);
        while (fr.hasNext())
        {
            Pair<String[], String> p = fr.next();
            prot2seq.put(p.getFirst()[0], p.getSecond());
            
        }
        fr.close();
        long t2 = System.currentTimeMillis();
        log.info("protein fasta read took: %2.3f sec (%s)", (t2-t1)/1000.0, fastafile.getAbsolutePath());
    }
    public String getProteinSeq(String id)
    {
        return getProteinSeq(id,false);
    }
    public String getProteinSeq(String id, boolean null_if_not_configured)
    {
        if(prot2seq == null)
        {
            if (null_if_not_configured)
                return null;
            throw new RuntimeException("no protein sequences is provided in a GTF file handler (add protein fasta file)");
            
        }
            

        return prot2seq.get(id);
    }

    public Iterator<Pair<String,Integer>> getChromosomeLengths()
    {
        return chr2lengths.iterator();
    }

    BufferedRandomAccessFile gtf_raf = null;

    public void finalize()
    {
        if (gtf_raf == null)
            return;
        try
        {
            gtf_raf.close();
        }
        catch(IOException e)
        {
            log.warn("error: %s while closing gtf file: %s", e.getMessage(), gtf.getAbsolutePath());
        }
        
    }

    public Collection<String> getGeneTypes()
    {
        return biotype2gene.keySet();
    }

    public Collection<String> getGenesByType(String biotype)
    {
        return biotype2gene.get(biotype);
    }
    
    public Collection<String> getGeneNames()
    {
        return gene2buffer.keySet();
    }

    public GeneLookupInfo getGeneInfo(String gid)
    {
        return gene2buffer.get(gid);
    }


    public Iterator<GeneLookupInfo> getInfos()
    {
        return gene2buffer.values().iterator();
    }

    public StringBuffer getGeneBuffer(String id)
    {
        //return gene2buffer.get(id);
        try
        {
            GeneLookupInfo gli = gene2buffer.get(id);
            if (gli==null)
                return null;
            
            StringBuffer sb = new StringBuffer();
            //char[] chars = new char[gli.buffsize];
            
            synchronized(gtf_raf)
            {

                gtf_raf.seek(gli.fp);
                //String line;
                while (sb.length() < gli.buffsize)
                {
                    sb.append(gtf_raf.readLine());
                    sb.append('\n');
                }
                
    

            }
            return sb;

        }
        catch(IOException ie)
        {
            throw new RuntimeException("i/o error: "+ie.getMessage()+" while fetching gene: "+id+" from "+gtf.getAbsolutePath());
        }
    }

    boolean no_cache = false;

    public void clearCache()
    {

        id2maingene.clear();
        //id2prot.clear();
        id2genes.clear();

    }

    public void setCaching(boolean b)
    {
        no_cache = !b;
    }

    
    public GenomicRegion getGeneRegion(String id)
    {
        synchronized(id2maingene)
        {
            Gene g =  id2maingene.get(id);
            if (g!=null)
            {
                return new GenomicRegion(g.chr, g.strand, g.start, g.end);
            }
        }
        GeneLookupInfo gli = gene2buffer.get(id);
        if (gli==null)
            return null;
        
        return new GenomicRegion(gli.chr, gli.strand, gli.start, gli.end);
    }
    
	public boolean containsGeneId(String id) {
		return gene2buffer.containsKey(id);
	}
    
    public Gene getGene(String id)
    {
        log.debug("gene: %s will be fetched by thread: %s", id, Thread.currentThread().getId());
        synchronized(id2maingene)
        {
            
            Gene g =  id2maingene.get(id);
            if (g!=null)
            {
                log.debug("gene: %s in memory %d-%d length: %d return to thread: %s", id, g.start, g.end, g.length, Thread.currentThread().getId());            
                
                return g;
            }
                
        }
        
                          
        StringBuffer sb = getGeneBuffer(id);
        log.debug("gene >%s< will be parsed form %d long buffer", id, (sb == null) ? -1 : sb.length());
        if (sb==null)
        {
            log.warn("no buffer found to :>%s<", id);
            int s=0;
            for (String gn : gene2buffer.keySet())
            {
                log.debug(">%s<\n", gn);
                if (++s>=10)
                {
                    break;
                }
            }
            return null;
        }

        
        log.debug("parse gene:  %s into memory thread: %s", id, Thread.currentThread().getId());            
        
        return parseGene(sb);
        
    }

    public String getGeneIdByProteinId(String id)
    {
        return p2gene.get(id);
    }

    public String getGeneIdByTranscriptId(String id)
    {
        return tr2gene.get(id);
    }

    public Protein getProtein(String id)
    {
        String gn = p2gene.get(id);
        if (gn == null)
        {
            return null;
        }
        Gene g = getGene(gn);
        if (g == null)
            return null;
        
        return g.proteins.get(id);
        
    }
    
    Vector<GeneLookupInfo> ordered_buffers = null;

    synchronized Iterator<GeneLookupInfo> getOrderedBufferIterator()
    {
        if (ordered_buffers != null)
            return ordered_buffers.iterator();
        
        ordered_buffers = new Vector<>();
        HashMap<String, HashSet<GeneLookupInfo>> infos = ObjectGetter.buildReverseMultiMap(gene2buffer.values(), (GeneLookupInfo gli) -> gli.chr);
        Vector<String> chrs = ObjectGetter.toVector(infos.keySet());
        Collections.sort(chrs);

        for (String chr : chrs)
        {
            Vector<GeneLookupInfo> v = ObjectGetter.toVector(infos.get(chr));
            NumUtils.sort(v, (gli) -> gli.start);
            ordered_buffers.addAll(v);
        }
        return ordered_buffers.iterator();
    }
    


    public int getNumGenes()
    {
        return getNumGenes(null);
    }
    public int getNumGenes(String biotype)
    {
        if (biotype==null)
        {
            return gene2buffer.size();
        }
        HashSet<String> genes = biotype2gene.get(biotype);
        if (genes==null)
        {
            return 0;
        }
        return genes.size();
    }

    public Iterator<Gene> getGenes()
    {
        return getGenes(null, null, null);
    }
    
    public Vector<Region1D> getGeneRegions(String chr, Integer from, Integer to)
    {
        Iterator<Gene> it = getGenes(chr,from,to);
        Vector<Region1D> v = new Vector<Region1D>();
        while(it.hasNext())
        {
            Gene g = it.next();
            v.add(new Region1D(g.start, g.end, g.id));
        }
        return v;
    }
    
    static class GTFEntry
    {
        String chr;
        int start;
        int end;
        String biotype;
        boolean strand;

        ENTRY_TYPE etype;
        Integer frame = null;

        
        
    }
    void compileGene(Gene g)
    {
        if (g==null)
            return;

        g.compile();
        ngenes++;
        
        if (no_cache)
            return;
        
        synchronized(id2genes)
        {
            MapBuilder.update(id2genes, g.id, g);
        }
        /*
        synchronized(genes)
        {
            genes.add(g);
        }
        */

    }
    
    static FileUtils.StringSplitter TABSP = FileUtils.getRecordSplitter("\t");
    static FileUtils.Converter FIELDC = new FileUtils.Converter();
    String name;
    
    public String getAnnotationName()
    {
        return name;
    }
    /*public GTFAnnotation(File gtf)
    {
        this(gtf, null);
    }
    */

    
    public static class GeneLookupInfo
    {
        public String gene;
        long fp;
        int buffsize;
        public String chr;
        public String biotype;
        public int start;
        public int end;
        public boolean strand;

        GeneLookupInfo()
        {

        }

        public String getChr()
        {
            return chr;
        }

        public String getBioType()
        {
            return biotype;
        }

        public boolean getStrand()
        {
            return strand;
        }

        public int getStart()
        {
            return start;
        }

        public int getEnd()
        {
            return end;
        }
        public GeneLookupInfo clone()
        {
            return new GeneLookupInfo(gene,fp,buffsize,chr,biotype,start,end, strand);
        }

        GeneLookupInfo(String gene, long fp, int buffsize, String chr, String biotype, int start, int end, boolean strand)
        {
            this.gene =gene; this.fp=fp; this.buffsize=buffsize; this.chr=chr; this.biotype=biotype; this.strand=strand;
            this.start=start; this.end=end;
        }
        
        GeneLookupInfo(String line)
        {
            int nexttab=0;
            FIELDC.set(TABSP.split(line));
            
            for (int i=0; i < 8 && i < FIELDC.length(); i++)
            {
                switch(i)
                {
                    case 0:
                        gene = FIELDC.toString(i);
                        break;
                    case 1:
                        biotype = FIELDC.toString(i);
                        break;
                    case 2:
                        chr = FIELDC.toString(i);
                        break;

                    case 3:
                        fp = FIELDC.toLong(i);
                        break;

                    case 4:
                        buffsize  = FIELDC.toInt(i);
                        break;
                    case 5:
                        start  = FIELDC.toInt(i);
                        break;

                    case 6:
                        end  = FIELDC.toInt(i);
                        break;
                    case 7:
                        this.strand  = FIELDC.toBoolean(i);
                        break;


                }
            }
            

        }

        public String toString()
        {
            return String.format("%s(%s:%c) %d-%d %s", gene, chr, GenomicUtils.getStrand(strand), start, end, biotype);
        }

        void write(PrintWriter pw)
        {
            pw.printf("%s\t%s\t%s\t%d\t%d\t%d\t%d\t%s\n", gene, biotype, chr, fp, buffsize, start, end, strand);
        }
    }


    public GTFAnnotation(File gtf)
    {
        this(null, gtf, null, null, null);
    }

    void compileInfos()
    {

        HashMap<String, Integer> chr2length = new HashMap<String, Integer>();
        for (GeneLookupInfo gli: gene2buffer.values())
        {
            biggest_buffer = Math.max(biggest_buffer, gli.buffsize);
            HashSet<String> h = biotype2gene.get(gli.biotype);
            if (h==null)
            {
                biotype2gene.put(gli.biotype, h = new HashSet<String>());
            }
            h.add(gli.gene);
            Integer old = chr2length.get(gli.chr);
            chr2length.put(gli.chr, (old==null)?gli.end:Math.max(old, gli.end));

        }
                    
        for (Map.Entry<String, Integer> e: chr2length.entrySet())
        {
            chr2lengths.add(new Pair<String,Integer>(e.getKey(), e.getValue()));
        }

        


    }

    void create_lookup(File gtf_lookup, File gtf_gene2prot, File gtf_gene2tr)
    {
        if (gtf_lookup == null || gtf_gene2prot == null || gtf_gene2tr == null)
        {
            return;
        }
        try
        {

        	
            PrintWriter pw1 = new PrintWriter(gtf_lookup);

            for(GeneLookupInfo gli : gene2buffer.values())
            {
                gli.write(pw1);
            }
            pw1.close();

            FileUtils.writeMap(gtf_gene2prot, "\t", p2gene);
            FileUtils.writeMap(gtf_gene2tr, "\t", tr2gene);
            

        }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while writing lookup to %s and %s",ie, ie.getMessage(), gtf_lookup.getAbsolutePath(), gtf_gene2prot.getAbsolutePath());
        }
    }

    static final String NAMELOOKUP_KEY = "namelookup";

    void createGeneNameLookup(File outfile)
    {
        DB db = DBMaker.newFileDB(outfile).make();
        name2geneid = (db.createHashMap(NAMELOOKUP_KEY)).make();

        int nadded = 0;
        int nnonames = 0;
        Iterator<Gene> it = getGenes();
        while (it.hasNext())
        {
            Gene g = it.next();
            if (g.name == null)
            {
                //System.err.printf("gene: %s has no name. (so far added: %d no names: %d)\n", g.getLocationInfo(), nadded, nnonames);
                nnonames++;
                continue;
            }
            MapBuilder.update(name2geneid, g.name, g.id);
            nadded++;
        }

        log.info("created gene name lookup got names for %d genes, %d without", nadded, nnonames);
        db.commit();
    }

    void setGeneNameLookup(File lookup)
    {
        try
        {
            DB db = DBMaker.newFileDB(lookup).readOnly().make();
            name2geneid = db.getHashMap(NAMELOOKUP_KEY);
            System.out.printf("set gene name lookup: %s got %d entries\n", lookup, name2geneid.size());
        }
        catch (java.nio.channels.NonWritableChannelException nce)
        {
            log.warn("could not set gene lookup, even if read only mode... Gergely still does not understand why...");
        }
    } 

    boolean init_from_lookup(File gtf_lookup, File gtf_gene2prot, File gtf_gene2tr)
    {
        if (gtf_lookup==null || !gtf_lookup.exists())
        {
            return false;
        }

        if (gtf_gene2prot==null || !gtf_gene2prot.exists())
        {
            return false;
        }

        if (gtf_gene2tr==null || !gtf_gene2tr.exists())
        {
            return false;
        }

        HashMap<String,Integer> chr2length = new HashMap<String,Integer>();

        try
        {
            ngenes = 0;
            BufferedRandomAccessFile bfr = new BufferedRandomAccessFile(gtf_lookup, "r", 30000);
            String line=null;



            while (null!=(line=bfr.readLine()))
            {
                GeneLookupInfo gli = new GeneLookupInfo(line);
                gene2buffer.put(gli.gene, gli);
                    
                ngenes++;
            }
            bfr.close();

        }
        catch(IOException e)
        {
            throw new FRuntimeException("i/o error: %s while reading from gtf index file: %s", e, e.getMessage(),  gtf_lookup.getAbsolutePath());
        }
        
        
        p2gene = FileUtils.readMap(gtf_gene2prot, "\t", 0, 1);
        tr2gene = FileUtils.readMap(gtf_gene2tr, "\t", 0, 1);

        log.info("read %d prot2gene mappings from: %s", p2gene.size(), gtf_gene2prot);
        log.info("read %d tr2gene mappings from: %s", tr2gene.size(), gtf_gene2tr);
        
            
        compileInfos();
        try
        {

            gtf_raf = new BufferedRandomAccessFile(gtf, "r", biggest_buffer);
        }
        catch(IOException e)
        {
            throw new FRuntimeException("i/o error: %s while reading opening gtf  file: %s", e, e.getMessage(),  gtf.getAbsolutePath());
        }
        return true;

    }
    public GTFAnnotation(Organism org, File gtf, File gtf_lookup, File gtf_gene2prot, File gtf_gene2tr)
    {
        this.organism = org;
        this.gtf = gtf;
        name=gtf.getName();
        final int num_fields= GTFHEADER.ATTRIBUTE.index+1;

        long t1 = System.currentTimeMillis();
        if (init_from_lookup(gtf_lookup,gtf_gene2prot, gtf_gene2tr))
        {
            
            long t2 = System.currentTimeMillis();
            log.info(String.format("gtf lookup finishing, took %2.3f sec %d genes to %d ids...", (t2-t1)/1000.0, ngenes, gene2buffer.size()));
            return;
        }

        HashSet<String> comments = new HashSet<String>();
        comments.add("#");
        log.info("start reading file: "+gtf);
        final String id_sep = "_id \"";
        final String gene_str = "gene";
        final String tr_str = "transcript";
        final String prot_str = "protein";
        final String cds_str = "CDS";
        

        String v_suffix = " \"";                                                 
        Vector<String> biotype_prefixes = ObjectGetter.toVector(ATTRIB_TYPE.GENE_TYPE.name + v_suffix, ATTRIB_TYPE.BIOTYPE.name + v_suffix);
        int ln = -1;
        
        int ngenes = 0;

        String lastgid = null;
        StringBuffer sb = new StringBuffer();
        boolean ps_coding = false;
        GeneLookupInfo gli = new GeneLookupInfo();

        try
        {
            
            BufferedRandomAccessFile bfr = new BufferedRandomAccessFile(gtf, "r", 30000);
            //BufferedReader bfr = new BufferedReader(new FileReader(gtf));
            String line=null;
            
            
            
            
            String biotype = null;

            while (null!=(line=bfr.readLine()))
            {
                ln++;

                int nexttab = 0;
                String chr=null;
                biotype = null;
                String type = null;
                int entry_start=-1;
                int entry_end=-1;
                int lasttab = nexttab;
                boolean strand = true;
                
                for (int i=0; i<7; i++)
                {
                    int nt = line.indexOf('\t', nexttab);
                    if (nt<0)
                        break;
                    
                    lasttab = nexttab;
                    nexttab = nt+1;

                    
                    switch(i)
                    {
                        case 0:
                            chr = line.substring(lasttab, nt);
                            break;
                        case 1:
                            biotype = line.substring(lasttab, nt);
                            break;
                        case 2:
                            type = line.substring(lasttab, nt);
                            break;
                        case 3:
                            entry_start = Integer.parseInt(line.substring(lasttab, nt));
                            break;
                        case 4:
                            entry_end = Integer.parseInt(line.substring(lasttab, nt));
                            break;
                        case 6:
                            strand = (line.charAt(nt-1)=='+');
                            break;
                        
                    }
                    
                    

                }
                //check biotype 
                for (String bpref : biotype_prefixes)
                {
                    int idx = line.indexOf(bpref);
                    if (idx < 0)
                        continue;
                    
                    int idxend = line.indexOf('"', idx + bpref.length() + 1);
                    if (idxend < 0)
                        continue;
                    
                    biotype = line.substring(idx + bpref.length(), idxend);
                    
                    break;
                }
                

                //System.out.printf("chr: %s bjiotype: %s type: %s line: %s\n", chr, biotype, type, ObjectGetter.toVector(line.split("\t")));
                String gid=null;
                String trid=null;
                String pid = null;
                int start = 0;
                while (true)
                {
                    int nextid = line.indexOf(id_sep,  start);
                    if (nextid<0)
                        break;
                    
                    int idstart = nextid;
                    
                    while (idstart>=0 && !Character.isWhitespace(line.charAt(idstart)))
                    {
                        idstart--;
                    }

                    int value_start = nextid+id_sep.length();
                    int value_end = value_start;
                    
                    while (value_end<line.length() && line.charAt(value_end)!='"')
                    {
                        value_end++;
                    }
                    
                    String kn = line.substring(idstart+1, nextid);
                    if (gene_str.equals(kn))
                    {
                        gid = line.substring(value_start, value_end);
                    }
                    if (prot_str.equals(kn))
                    {
                        pid = line.substring(value_start, value_end);
                    }
                    if (tr_str.equals(kn))
                    {
                        trid = line.substring(value_start, value_end);
                    }
                    //System.out.printf("%s:%s\n", , line.substring(value_start, value_end));
                    start = value_end;
                }
                /*Matcher m = idp.matcher(line);
                int start=0;
                while (m.find(start))
                {
                    start = m.end();
                }
                */
                if (lastgid!=null && !gid.equals(lastgid))
                {
                	
                    ngenes++;
                    long fp = bfr.getFilePointer();
                    fp=fp-line.length()-1;
                    fp-=sb.length();
                    gli.buffsize = sb.length();
                    gli.fp = fp;
                    gene2buffer.put(lastgid, gli.clone());
                    sb.setLength(0);
                    ps_coding = false;
                }
                if (lastgid==null || !gid.equals(lastgid))
                {
                    gli.gene = gid;
                    gli.biotype = biotype;
                    gli.chr = chr;
                    gli.start = entry_start;
                    gli.end = entry_end;
                    
                    gli.strand = strand;
                    
                }

                lastgid = gid;
                sb.append(line);
                sb.append('\n');

                if (gid!=null && trid!=null)
                {
                    tr2gene.put(trid, gid);
                }
                if (gid!=null && pid!=null)
                {
                    p2gene.put(pid, gid);
                    ps_coding = true;
                }
                if (ln%100000 == 0)
                {
                    log.debug("after %d lines  read %d genes...", ln, ngenes);
                }
            }

            if (lastgid!=null)
            {
                ngenes++;
                //gene2buffer.put(lastgid, new StringBuffer(sb.toString()));
                long fp = bfr.getFilePointer();
                fp-=sb.length();
                gli.buffsize = sb.length();
                gli.fp = fp;
                gene2buffer.put(lastgid, gli.clone());
                biggest_buffer = Math.max(biggest_buffer, sb.length());
            }
    
            bfr.close();
            gtf_raf = new BufferedRandomAccessFile(gtf, "r", biggest_buffer);
            
        }
        catch(IOException ie)
        {
            throw new RuntimeException("error while reading gtf: "+gtf.getAbsolutePath(), ie);
        }
        
        long t2 = System.currentTimeMillis();
        log.info(String.format("gtf read finishing, took %2.3f sec read %d lines. %d genes to %d ids...", (t2-t1)/1000.0, ln, ngenes, gene2buffer.size()));
        create_lookup(gtf_lookup,gtf_gene2prot, gtf_gene2tr);
        compileInfos();
        
        
    }

    Integer pseudo_exon_number =0; 
    static Pattern tabP = Pattern.compile("\t");

    static final int num_fields= GTFHEADER.ATTRIBUTE.index+1;

    Gene parseGene(StringBuffer sb)
    {
       
        Gene current_gene = parseGene(organism, name,num_fields,sb, pseudo_exon_number);
        compileGene(current_gene);

        if (!no_cache)
        {
            synchronized(id2maingene)
            {
    
                if(null == id2maingene.get(current_gene.id))
                {
                    id2maingene.put(current_gene.id, current_gene);
                }
            }
        }
        
        
        return current_gene;
    }
    static Gene parseGene(String name, int num_fields, StringBuffer sb, Integer pseudo_exon_number)
    {
        return parseGene(null,name ,num_fields,sb,pseudo_exon_number);
    }
    static Gene parseGene(Organism org, String name, int num_fields, StringBuffer sb, Integer pseudo_exon_number)
    {
        
        Logger log = LogConfig.getLogger();
        BufferedReader bfr = new BufferedReader(new StringReader(sb.toString()));
        String line = null;

        Gene current_gene = null;
        Transcript current_transcript = null;
        Protein current_protein = null;

        Exon last_exon = null;
        TreeMap<Pair<Integer,Integer>, Exon> current_exons = new TreeMap<Pair<Integer,Integer>, Exon>();

        String EXON_FMT = "EXON%06d";
        
        int ln=-1;

        boolean got_startstopcodon = false;
        try
        {
            
            while (null!=(line = bfr.readLine()))
            {
                ln++;
                if (line.length()==0 || line.charAt(0)=='#')
                {
                    continue;
                }
                
                String sp[] = tabP.split(line);
    
                if (sp.length!=num_fields)
                {
                    throw new RuntimeException("invalid gtf format at line "+ln+" in file: "+name+" expected "+num_fields+" fields, got "+sp.length+"\nline: "+line);
                }
                
                String gene_id=null;
                String tr_id=null;
                String prot_id = null;
    
                String gene_name = null;
                String tr_name = null;
    
                String exon_id=null;
                int exon_number=-1;
                String chr = null;
                String biotype = null;
                String transcript_biotype = null;
                boolean strand = true;
                int start=-1;
                int end=-1;
                ENTRY_TYPE etype=null;
                Integer frame = null;
                Set<String> tags = new HashSet<String>();
                
                for (GTFHEADER gh : GTFHEADER.values())
                {
                    String v1=sp[gh.index];
                    switch(gh)
                    {
                        case CHR:
                            chr = v1;
                            break;
                        case BIOTYPE:
                            biotype=v1;
                            break;
                            
                        case ENRYTYPE:
                            if(null==(etype=ENTRY_TYPE.get(v1)))
                            {
                                throw new RuntimeException("invalid entry type found: "+v1+" at line: "+ln+" in "+name);
                            }
                            break;
                        case START:
                            start=Integer.parseInt(v1);
                            break;
                        case END:
                            end=Integer.parseInt(v1);
                            break;
                        case STRAND:
                            strand = v1.charAt(0)=='+';
                            break;
                        case FRAME:
                            try
                            {
                                frame = Integer.parseInt(v1);
                            }
                            catch(NumberFormatException nfe)
                            {
                                //not always needed, silent error
                            }
                            break;
                        case ATTRIBUTE:
                            Matcher m = attribP.matcher(sp[gh.index]);
                            int astart=0;
                            Vector<Pair<String, String> > attribs = new Vector<Pair<String, String> >();
                            while (m.find(astart))
                            {
                                String v =m.group(2);
                                attribs.add(new Pair<String,String>(m.group(1), m.group(2)));
                                
                                ATTRIB_TYPE at = ATTRIB_TYPE.get(m.group(1));
                                if (at==null)
                                {
                                    System.err.println("unexpected attrib type: "+m.group(1));
                                    continue;
    
                                }
                                switch(at)
                                {              
                                    case GENEID:
                                        gene_id=v;
                                        break;
                                    case TRANSCRIPTID:
                                        tr_id=v;
                                        break;
                                    case EXONID:
                                        exon_id=v;astart=m.end();
                                        break;
                                    case EXONNUMBER:
                                        exon_number=Integer.parseInt(v);
                                        break;
                                    case GENENAME:
                                        gene_name=v;
                                        break;
                                    case TRANSCRIPT_BIOTYPE:
                                    case TRANSCRIPT_TYPE:
                                        transcript_biotype = v;
                                        break;
                                    case TRANSCRIPTNAME:
                                        tr_name=v;
                                        break;
                                    case PROTEINID:
                                        prot_id=v;
                                        break;
                                    case GENE_TYPE:
                                    case BIOTYPE:
                                        biotype=v;
                                        break;
                                    case TAG:
                                    	tags.add(v);
                                    	break;
                                    default:
                                        break;
                                }
                                astart=m.end();
                            }
                            //System.out.printf("\t[%s]=%s\n", gh, ""+attribs);
                            break;
                        default:
                            //System.out.printf("\t[%s]=%s\n", gh, sp[gh.index]);
                            break;
                    }
                    
                
                }

                //System.out.printf("next line: %d gene: %s tr. %s prot id: %s type: %s\n", ln, gene_id, tr_id, prot_id, etype);
                /*String info = String.format("got: %s (%s) %s(%c) %d-%d exon: %d %s %s %s", 
                                          etype, biotype, chr, (strand)?'+':'-', start, end, exon_number, gene_id, tr_id, prot_id);
                 */
                if (current_gene==null)
                {
                    
                    //System.out.printf("start new gene: %s (%s) %s(%c)... \n", gene_id, gene_name, chr, (strand)?'+':'-');
                    current_gene=new Gene(gene_id,chr,strand,start,end); //have to correct after
                    current_gene.organism = org;
                    current_gene.biotype = biotype;
                    if (gene_name!=null)
                    { 
                    	if( current_gene.name == null)
                    	{
                    		current_gene.name = gene_name;
                    	}
                        current_gene.syns.add(gene_name);
                    }
    
                    current_exons.clear();
                }
                
                if (tr_id == null)
                    continue;
                
                if (etype == ENTRY_TYPE.CDS && prot_id == null)
                {
                    prot_id = "P." + tr_id;
                    //log.warn("create implicite new protein in transcript: >%s<  -> >%s< for line: >%s<", tr_id, prot_id, line);
                }

                if (current_transcript == null || !current_transcript.id.equals(tr_id))
                {
                    current_protein=null;
                    //System.out.println(info);
                    //System.out.printf("start new transcript: "+tr_id);
                    current_transcript = new Transcript(current_gene, tr_id);
                    current_transcript.biotype= (transcript_biotype == null) ? biotype : transcript_biotype;
                    current_transcript.tags = tags;
                    current_gene.transcripts.put(tr_id, current_transcript);
                }
                if (prot_id!=null && (current_protein==null || !current_protein.id.equals(prot_id)))
                {
                    //System.out.println(info);
                    //System.out.printf("start new protein: "+prot_id);
                    got_startstopcodon = false;
                    current_protein = new Protein(current_transcript, prot_id);
                    current_transcript.protein=current_protein;
                    
                    current_gene.proteins.put(prot_id, current_protein);
                }
               
                
                switch(etype)
                {
                    case EXON:
                        Pair<Integer,Integer> r = new Pair<Integer,Integer>(start, end);
                        Exon exon = current_exons.get(r);
                        if (exon==null)
                        {
                            int exnum = -1;
                            synchronized(pseudo_exon_number)
                            {
                                exnum = pseudo_exon_number++;
                            }
                            String eid=(exon_id==null)?String.format(EXON_FMT, exnum):exon_id;
                            exon = new Exon(current_gene, eid);
                            //System.out.println(info);
                            //System.out.printf("%d add new exon: %s\n", ln, exon.id);
                            current_exons.put(r,exon);
                            exon.abs_start=start;
                            exon.abs_end=end;
                            
                            current_gene.exons.put(exon.id, exon);
                        }
                        last_exon = exon;
                        current_transcript.addExon(exon);
                        break;
                    case CDS:
                        /** process cds for the case start/stop codon is not annotated*/
                        frame = (frame==null)?0:frame;
                        if (current_protein.start_exon==null)
                        {
                            //System.err.printf("%s gets start exon: %s\n", current_protein.id, last_exon.id);
                            //workaround for gb converted gtf
                            if (last_exon == null)
                            {
                                Pair<Integer,Integer> _r = new Pair<Integer,Integer>(start, end);
                                Exon _exon = current_exons.get(_r);
                                if (_exon==null)
                                {
                                    int exnum = -1;
                                    synchronized(pseudo_exon_number)
                                    {
                                        exnum = pseudo_exon_number++;
                                    }
                                    String eid=(exon_id==null)?String.format(EXON_FMT, exnum): exon_id;
                                    _exon = new Exon(current_gene, eid);
                                    //System.out.println(info);
                                    //System.out.printf("%d add new exon: %s\n", ln, exon.id);
                                    current_exons.put(_r, _exon);
                                    _exon.abs_start=start;
                                    _exon.abs_end=end;
                                    
                                    current_gene.exons.put(_exon.id, _exon);
                                }
                                last_exon = _exon;
                                current_transcript.addExon(_exon);
                                
                            }
                            current_protein.start_exon = last_exon.id;
                            current_protein.transl_start = (strand)?start+frame-last_exon.abs_start:last_exon.abs_end+frame-end;
                            
                        }

                        current_protein.end_exon=last_exon.id;

                        if (!got_startstopcodon)
                        {
                            
                            int codon_start=-1;
                            int codon_end=-1;
                            
                            if (strand)
                            {
                                codon_end=end+3;
                                codon_start=codon_end-2;
    
                            }
                            else
                            {
                                codon_start=start-3;
                                codon_end=codon_start+2;
                            }
                            current_protein.transl_end=(strand)?codon_end-last_exon.abs_start:last_exon.abs_end-codon_start;
                        }
                    
                        break;
    
                    case START_CODON:
                        got_startstopcodon = true;
                        frame = (frame==null)?0:frame;
                        //System.err.println("got start codon: "+line);
                        //System.err.println("frame: "+frame+" start: "+start+" end: "+end);
                        //System.err.printf("will update protein: %s last exon: %s (%d-%d)\n ",current_protein.id, last_exon.id, last_exon.abs_start, last_exon.abs_end);
                        current_protein.start_exon=last_exon.id;
                        current_protein.transl_start=(strand)?start+frame-last_exon.abs_start:last_exon.abs_end+frame-end;
                        
                        
                        /*System.out.println(info);
                        System.out.printf("update transl start: %d for %s last exon: %s (%d-%d)\n", 
                                          current_protein.transl_start,
                                          current_protein.id, last_exon.id, last_exon.abs_start, last_exon.abs_end);
                                          */
                        break;
                    case STOP_CODON:
                        got_startstopcodon = true;
                        current_protein.end_exon=last_exon.id;
                        current_protein.transl_end=(strand)?end-last_exon.abs_start:last_exon.abs_end-start;
                        /*System.out.println(info);
                        System.out.printf("update transl end: %d for %s last exon: %s (%d-%d)\n", 
                                          current_protein.transl_end,
                                          current_protein.id, last_exon.id, last_exon.abs_start, last_exon.abs_end);
                                          */
                        break;
                    default:
                        //CDS -don't need
                        //System.out.println(info);
                        break;
    
                }
                
            }
            bfr.close();
        }
        catch(IOException ie)
        {
            throw new RuntimeException("i/o error while parsing gene buffer: "+sb.toString(), ie);
        }
        catch(Exception e)
        {
            throw new RuntimeException("general error while parsing gene buffer:\n"+sb.toString(), e);
        }


        
        return current_gene;
    }

    
    static final String LOOKUP_SUFFIX = ".lookup";
    static final String G2PROT_SUFFIX = ".g2p";
    static final String G2TR_SUFFIX = ".g2tr";
    static final String G2NAMELOOKUP_SUFFIX = ".g2namelookup";

    public static GTFAnnotation getGTFWithLookup(File gtf)
    {
        return getGTFWithLookup(gtf, null);
    }

    public static GTFAnnotation getGTFWithLookup(File gtf, File index_dir)
    {
        return getGTFWithLookup(gtf, index_dir, false);
    }

    public static GTFAnnotation getGTFWithLookup(File gtf, File index_dir, boolean no_name_lookup)
    {
        String prefix = (index_dir == null) ? gtf.getAbsolutePath() : (new File(index_dir, gtf.getName())).getAbsolutePath();

        File l = new File(prefix + LOOKUP_SUFFIX);
        File g2p = new File(prefix + G2PROT_SUFFIX);
        File g2tr = new File(prefix + G2TR_SUFFIX);
        File g2ns = new File(prefix + G2NAMELOOKUP_SUFFIX);

        GTFAnnotation annot = new GTFAnnotation(null, gtf, l, g2p, g2tr);

        if (!no_name_lookup && !g2ns.exists())
        {
            annot.createGeneNameLookup(g2ns);
        }
        if (!no_name_lookup)
        {
            annot.setGeneNameLookup(g2ns);
        }
        

        return annot;
    }
    

    public LocationQueryResult query(String q, int around)
    {
        return new LocationQueryResult(this, q, around);
    }

    public static class LocationQueryResult
    {
        GTFAnnotation gtf;
        public Tuple3<String, Integer,Integer> reg;
        public Vector<Gene> overlapping_genes;
        public Gene gene_found;
        public Transcript transcript_found;
        public Protein protein_found;

        public String info;


        void getOverlappingInfo(BufferPrintWriter bpw)
        {
            Vector<Gene> v =  ObjectGetter.toVector(gtf.getGenes(reg.get0(), reg.get1(), reg.get2()));
            bpw.printf("%d genes overlap region\n",  v.size());
            for (Gene g : v)
            {
                if (gene_found != null && g.id.equals(gene_found.id))
                    continue;
                
                
                bpw.println(g.getLocationInfo());
            }
        }

        void getGeneOverlappingInfo(BufferPrintWriter bpw, int around)
        {
            bpw.printf("found gene: %s\n", gene_found.id);
            bpw.printf("gene: %s \n", gene_found.getLocationInfo());
            reg = Tuple3.create(gene_found.chr, gene_found.start - around, gene_found.end + 1 + around);
            getOverlappingInfo(bpw);
            info = bpw.getBuffer();
        }

        LocationQueryResult(GTFAnnotation gtf, String query, int around)
        {
            this.gtf = gtf;
            BufferPrintWriter bpw = new BufferPrintWriter();
            reg = GenomicUtils.parseChrRegion(query);
            if (reg == null)
            {
                bpw.println("no chr request: " +query);
            }
            if (reg != null)
            {
                bpw.println("got chr request: " +reg);
                getOverlappingInfo(bpw);
                info = bpw.getBuffer();
                return;
            }
                    
            String q = query;
            gene_found = gtf.getGene(q);
            if (gene_found != null)
            {
                bpw.printf("got %d proteins, %d transcripts\n", gene_found.proteins.size(), gene_found.transcripts.size());
                for (Transcript t: gene_found.transcripts.values())
                {
                    bpw.printf("\t%s type: %s got prot: %s: %s\n", t.id, t.biotype, t.protein != null, t.getRegionVector(true));
                    if (t.protein != null)
                    {
                        //t.getRegionVector(true).substract(t.protein.getCodingRegionVector(true))
                        bpw.printf("\t\tprot: %s %d : %s utrs: %s\n", t.protein.id, t.protein.hashCode(), t.protein.getCodingRegionVector(true), "tmp");
                    }
                }
                getGeneOverlappingInfo(bpw, around);
                return;

            }
                        
            String gid = gtf.getGeneIdByProteinId(q);
            if (gid != null)
            {
                bpw.printf("found gene: %s for protein id: %s\n", gid, q);
                gene_found = gtf.getGene(gid);
                getGeneOverlappingInfo(bpw, around);
                protein_found = gene_found.proteins.get(q);
                transcript_found = protein_found.parentTranscript;
                return;
            }
                        
            gid = gtf.getGeneIdByTranscriptId(q);
            if (gid != null)
            {
                bpw.printf("found gene: %s for transcript id: %s\n", gid, q);
                gene_found = gtf.getGene(gid);
                transcript_found = gene_found.transcripts.get(q);
                bpw.printf("transcript regions: %s\n",  transcript_found.getRegionVector(false));
                getGeneOverlappingInfo(bpw, around);
                return;
                
            }
            
            //log.info("gene null: %s name2geneid null: %s", (g == null), (gtf.name2geneid == null));
            if (gtf.name2geneid != null)
            {
                //log.info("check for gene name >%s<", q);
                HashSet<String> gids = gtf.name2geneid.get(q);
                gid = null;
                if (gids != null)
                {
                    gid = ObjectGetter.first(gids);
                }

                if (gids != null && gids.size() > 1)
                {
                    bpw.printf("found multiple genes for gene name: >%s< : %s\nwill take: %s\n", q, gids, gid);
                }
                
                if (gid != null)
                {
                    bpw.printf("found gene: %s for gene name: %s\n", gid, q);
                    gene_found = gtf.getGene(gid);
                    bpw.printf("got %d proteins, %d transcripts\n", gene_found.proteins.size(), gene_found.transcripts.size());
                    getGeneOverlappingInfo(bpw,around);
                    return;
                }
                
            }
            bpw.printf("no gene found for query: >%s<\n", q);
        
            info = bpw.getBuffer();
        }
    }

    public static void main (String args[]) 
    {
        SimpleOptionParser cmd = new SimpleOptionParser("i", "q", "gbuff", "getbuff", "org", "cmd", "extractRRNA", "createlookup", "overview", "uselookup", "indexdir", "intronlengths");
        cmd.setFile("i");
        cmd.setDir("indexdir");
        
        cmd.setOptional("q", "getbuff", "org", "i", "extractRRNA", "indexdir");
        cmd.setSwitches("gbuff", "cmd",  "createlookup", "overview", "uselookup", "intronlengths");

        if (!OptionParser.parseParams(args,true, false, true, cmd))
            return;
        
        if (cmd.isSet("createlookup"))
        {
            File idir = cmd.getOptionalFile("indexdir");
            getGTFWithLookup(cmd.getFile("i"), idir);
            return;
        }
        if (cmd.isSet("gbuff"))
        {
            StringBuffer sb  = FileUtils.getContentBuffer(cmd.getFile("i"));
            Gene g = parseGene("test", num_fields, sb, 0);
            g.compile();
            for (Protein p : g.proteins.values())
            {
                System.out.printf("%s exons: %d: %s\n", p.id, p.getNumExons(), p.getCodingRegionVector(true));
            }

            return;
        }

        GTFAnnotation gtf = null;
        GenomeSequenceExtractor gsex = null;

        if (cmd.isOptionSet("org"))
        {
            Organism org = GenomicConfigManager.getGenomicOrganism(cmd.getValue("org"));
            gtf = org.getAnnot();
            gsex = org.getGenome();
            
        }
        if (cmd.isOptionSet("i"))
        {
            gtf = (cmd.isSet("uselookup")) ? getGTFWithLookup(cmd.getFile("i")) : new GTFAnnotation(cmd.getFile("i"));
        }
        if (cmd.isSet("intronlengths"))
        {
            HashMap<Integer, Integer> l2c = new HashMap<>();
            int nt[]=new int[1];
            
            ObjectGetter.apply(gtf.getAllGeneIterator(), (Gene g) ->
            {
                for (Transcript t : g.transcripts.values())
                {
                    RegionVector rv = t.getRegionVector(false).getInverseRegionVector();
                    ObjectGetter.apply(rv.getRegions(), (r) -> MapBuilder.update(l2c, r.getLength()));

                    if (++nt[0] % 1_000 != 0)
                        continue;
                    
                    NumUtils.BasicNumberInfo<Integer> bni = NumUtils.getNumInfo(l2c);

                    System.out.printf("after %d transcripts: %s 90%%: %d\n", nt[0], bni, bni.quantiles.get(3));
                }
            });
            return;
        }
        if (cmd.isSet("overview"))
        {
            Vector<Gene> genes = ObjectGetter.toVector(gtf.getGenes());
            HashMap<String, HashSet<Gene>> bt2genes = ObjectGetter.buildReverseMultiMap(genes, (Gene g) -> g.biotype);
            HashMap<String, HashSet<Gene>> chr2genes = ObjectGetter.buildReverseMultiMap(genes, (Gene g) -> g.chr);

            ObjectGetter.apply(bt2genes.entrySet(), (e) -> System.out.printf("\t%s: %d\n", e.getKey(), e.getValue().size()));
            
            HashMap<Boolean, Vector<UPair<Gene>>> overlaps = new HashMap<>();

            int totalsize = NumUtils.sum(genes, (g) -> g.getMergedTranscript(true).getCoveredLength());

            System.out.printf("total expressed transcript size: %2.2f M\n", totalsize / 1_000_000.0);

            for (Map.Entry<String, HashSet<Gene>> e: chr2genes.entrySet())
            {
                Vector<Gene> v = ObjectGetter.toVector(e.getValue());
                NumUtils.sort(v, (g) -> g.start);
                for (int i = 1; i < v.size(); i++)
                {
                    Gene g = v.get(i);
                    for (int j = i - 1; j >= 0; j--)
                    {
                        Gene g2 = v.get(j);
                        if (g2.end < g.start)
                            break;
                        
            
                        MapBuilder.updateV(overlaps, g.strand == g2.strand, UPair.createU(g, g2));
                        
                    }
                }
            }
            ObjectGetter.apply(overlaps.entrySet(), (e) -> System.out.printf("\tsame strand %s overlaps: %d\n", e.getKey(), e.getValue().size()));
            return;
        }
        if (cmd.isOptionSet("extractRRNA"))
        {
            PrintWriter pw = cmd.getWriter("extractRRNA");

            for (String biotype : ObjectGetter.toArray("rRNA", "Mt_rRNA"))
            {
                Collection<String> gnames = gtf.getGenesByType(biotype);
                System.out.printf("got %d genes for %s\n", gnames.size(), biotype);


                for (String gn : gnames)
                {
                    Gene g = gtf.getGene(gn);
                    for (Transcript t : g.transcripts.values())
                    {
                        String seq = GenomicUtils.getTranscriptSeq(t, gsex);
                        
                        pw.printf(">%s\n", t.id);
                        fasta.SimpleFastaReader.writeFasta(seq, pw);
                    }
                }
                
                
            }
            
            pw.close();
            return;
            
        }

        if (cmd.isSet("cmd"))
        {
            Logger log = LogConfig.getLogger();
            BufferedReader bfr = new BufferedReader(new InputStreamReader(System.in));
            String line = null;
            
            System.out.printf(">");
            try
            {
                while (null != (line = bfr.readLine()))
                {
                       
                    System.out.printf("\n>");
                    
                    LocationQueryResult lqr = gtf.query(line, 0);
                    System.out.printf("%s\n>", lqr.info);
                    
                }
            }
            catch(IOException e)
            {

            }

            return;
            
        }
        

        if (cmd.isOptionSet("getbuff"))
        {
            System.out.println(gtf.getGeneBuffer(cmd.getValue("getbuff")).toString());
            return;
        }

        
        if (cmd.isOptionSet("q"))
        {

            LocationQueryResult lqr = gtf.query(cmd.getValue("q"), 0);
            System.out.printf("%s\n>", lqr.info);

            return;
        }
        
        for (Map.Entry<String, HashSet<String>> e: gtf.biotype2gene.entrySet())
        {
            System.out.printf("%s: %d\n", e.getKey(), e.getValue().size());
        }
        Iterator<Gene> git=gtf.getGenes(null, null, null);
        while (git.hasNext())
        {
            Gene g = git.next();
            System.out.printf("%s\t%d\t%d\n", g.id, g.transcripts.size(), g.proteins.size());
        }
    }

}

