package genome;

import java.util.*;
import lmu.utils.*;


public class Gene
{

    public Organism organism;

    public String id;
    public String biotype = null;
    public boolean strand = true;
    public String chr;
    /** coordinates are 0-based: ( take care if using Ensembl) */
    public int annotated_start;
    public int annotated_end;

    public int start;
    public int end;
    public int length;

    public String name;
    public HashSet<String> syns = new HashSet<String>();
    public HashMap<String, Exon> exons = new HashMap<String, Exon>();
    public HashMap<String, Transcript> transcripts = new HashMap<String, Transcript>();
    public HashMap<String, Protein> proteins = new HashMap<String, Protein>();

    public Organism getOrganism()
    {
        return organism;
    }
    public void finalize()
    {
        //System.out.printf("finalize: %s\n", id); System.out.flush();
        for (Transcript t : transcripts.values())
        {
            t.parentGene=null;
            if (t.protein!=null)
            {
                t.protein.parentTranscript=null;
            }

        }
        exons.clear();
        proteins.clear();
        transcripts.clear();
    }
    Vector<Region1D> exon_regions = null;

    public String toString()
    {
        if (organism!=null)
        {
            return String.format("%s (%s)", id, organism.getSymbol());
        }
        return String.format("gene: %s", id);
    }
    @Override
    public int hashCode(){
        return this.id.hashCode();
    }
    @Override
    public boolean equals(Object o)
    {
        if ( o instanceof Gene)
        {
            if ( ((Gene) o).id.equals(this.id)) return true;
        }
        return false;
    }
    
    public String getBriefInfo()
    {
        return String.format("%s(%s%c %s)", id, chr, GenomicUtils.getStrand(strand), biotype);
    }

    public String getLocationInfo()
    {
        return String.format("%s(%s%c %s %d-%d)", id, chr, GenomicUtils.getStrand(strand), biotype, start, end);
    }

    RegionVector merged_tr_global = null;
    RegionVector merged_tr_local = null;

    public synchronized RegionVector getMergedTranscript(boolean global)
    {
        RegionVector src = (global) ? merged_tr_global : merged_tr_local;
        if (src != null)
            return src;

        RegionVector rv = RegionVector.merge(lmu.utils.ObjectGetter.map(transcripts.values(), (Transcript t) -> t.getRegionVector(global)));

        ObjectGetter.decide(global, () -> merged_tr_global = rv, () -> merged_tr_local = rv);
        
        return rv;
    }

    public Vector<Region1D> getExonRegions()
    {
        if (exon_regions!=null)
            return exon_regions;


        exon_regions = new Vector<Region1D>();

        Vector<Exon> exvec = new Vector<Exon>();
        exvec.addAll(exons.values());
        Collections.sort(exvec);

        int laststart=-1;
        int lastend=0;

        Exon last_e=null;
        HashSet<Exon> current_exons = new HashSet<Exon>();

        System.out.println("gene "+ id+" exon regions: ");
        for (Exon e : exvec)
        {
            System.out.println("\t next exon: "+e);
            if (laststart<0)
            {
                laststart=e.start;
                lastend=e.end;
                current_exons.add(e);
                continue;
            }
            if (e.start<=lastend)
            {

                current_exons.add(e);
                lastend=Math.max(lastend, e.end);
                continue;
            }


            // e.start > lastend

            exon_regions.add(new Region1D(laststart, lastend, current_exons));
            current_exons=new HashSet<Exon>();
            laststart=e.start;
            lastend=e.end;


        }

        for (int i=0; i<exon_regions.size(); i++)
        {
            Region1D r = exon_regions.get(i);
            System.out.printf("\t region %d/%d: %s\n", i+1, exon_regions.size(), ""+r);
        }
        return exon_regions;
    }


    /**
     * 
    
    */
    int [] getMappingTable(double intron_factor, double exon_factor, int minlength)
    {

        int length = end-start+1;

        System.out.println("gene: "+id+" start: "+start+ " end: "+end + " length: "+length);

        boolean [] exon = new boolean[length];

        for (Exon e : exons.values())
        {
            for (int i=e.start; i<=e.end; i++)
            {
                exon[i]=true;
            }
        }
        int laststart=0;
        int mapped_pos=0;
        boolean inexon=false;

        int [] rv = new int[length];
        for (int i=0; i<length; i++)
        {
            if (exon[i]==inexon)
                continue;

            //i changes state
            double current_factor = (inexon)?exon_factor:intron_factor;

            int current_length = i-laststart;

            int mapped_length = (int)(current_factor*current_length);
            if (mapped_length<minlength)
            {
                mapped_length=minlength;
            }

            inexon=exon[i];
            laststart=i;
            if (exon[i])
            {
                inexon=true;
            }
        }
        return rv;
    }

    public void convertGlobalToLocalPosition(Exon e, int start, int end)
    {
        e.abs_start=start;
        e.abs_end=end;
        if (strand)
        {
            e.start=convertGlobalToLocalPosition(start);
            e.end=convertGlobalToLocalPosition(end);

        }
        else
        {
            e.start=convertGlobalToLocalPosition(end);
            if (end==this.end)
            {
                //e.start++;
            }
            e.end=convertGlobalToLocalPosition(start);



        }

        e.length=e.end-e.start+1;
    }
    int convertGlobalToLocalPosition(int pos)
    {
        if (strand)
            return pos-start;

        return end-pos;
    }


    public Gene(String id,  String chr, boolean strand, int start, int end)
    {
        this(id,chr,strand,start,end,null);
    }
    public Gene(String id,  String chr, boolean strand, int start, int end, String biotype)
    {
        this.id=id; this.chr=chr; this.strand=strand; this.start=start; this.end=end;
        this.length=(end-start+1);
        this.biotype=biotype;
    }



    //fixes positions
    public void compile()
    {


        Logger log = LogConfig.getLogger();

        log.debug("will compile gene: "+id);
        log.debug(ExceptionManager.logCallerStack());
        annotated_start=start;
        annotated_end=end;

        for (Exon e : exons.values())
        {
            start=Math.min(start, e.abs_start);
            end = Math.max(end, e.abs_end);
        }

        
        /*System.out.printf("compile gene: %s (%s%c %d-%d) %d transcripts %d proteins, %d exons\n",
                          g.id, g.chr, (g.strand)?'+':'-', g.start, g.end, 
                          g.transcripts.size(),g.proteins.size(), g.exons.size());
        */

        for (Exon e : exons.values())
        {
            convertGlobalToLocalPosition(e, e.abs_start, e.abs_end);
        }

        //now check for the minimal coding position
        int corr_start=start;
        int corr_end=end;
        for (Protein p : proteins.values())
        {
            p.init();
            log.debug("after initing protein %s coding exons: %d\n", p.id, p.getNumExons());
            for (CodingExon ce: p.coding_exons)
            {
                corr_start = Math.min(corr_start, ce.abs_start);
                corr_end = Math.max(corr_end, ce.abs_end);
                //log.info("coding exon end: %d %s", corr_end, ce);
            }


        }
        //log.info(String.format("gene: %s : %d-%d corrected: %d-%d\n", id, start, end, corr_start, corr_end));
        length = end-start+1;
        if (corr_start==start && corr_end==end)
        {
            return;
        }

        log.debug(String.format("correct gene: %s %s (%c) %d-%d to %d-%d\n", id, chr, (strand)?'+':'-',start,end, corr_start, corr_end));
        start=corr_start;
        end=corr_end;

        for (Exon e: exons.values())
        {
            convertGlobalToLocalPosition(e, e.abs_start, e.abs_end);
        }
        for (Protein p : proteins.values())
        {
            p.init();
        }


    }


    HashMap<UPair<Integer>, HashSet<Transcript>> introns = null;

    public synchronized HashMap<UPair<Integer>, HashSet<Transcript>> getIntrons(boolean global)
    {
        if (introns != null)
            return introns;

        HashMap<UPair<Integer>, HashSet<Transcript>> rv = new HashMap<>();

        for (Transcript t : transcripts.values())
        {
            for(Region1D r : t.getRegionVector(true).getInverseRegionVector().getRegions())
            {
                MapBuilder.update(rv, UPair.createU(r.getX1(), r.getX2()), t);
            }
        }
        return introns = rv;
    }

    public Vector<Protein> getNonSubProts(Protein lead)
    {
        HashMap<Protein, RegionVector> known = new HashMap<>();
        Vector<Protein> nnp = new Vector<>();

        if (lead == null)
            return nnp;

        known.put(lead, lead.getCodingRegionVector(false));

        Protein longest = lead;
        RegionVector lrv = null;

        while (longest!=null)
        {
            longest = null;

            for (Protein p : proteins.values())
            {
                if (known.containsKey(p))
                    continue;


                RegionVector rrv = p.getCodingRegionVector(false);


                boolean novel  = true;
                for (Map.Entry<Protein, RegionVector> e : known.entrySet())
                {
                    if (e.getValue().isSubVector(rrv))
                    {
                        novel = false;
                        break;
                    }
                }
                if (!novel)
                    continue;

                if (longest!=null && longest.getCDSLength() >= p.getCDSLength())
                    continue;

                longest = p;
                lrv = rrv;
            }

            if (longest == null)
                break;


            nnp.add(longest);
            known.put(longest, lrv);
        }

        return nnp;

    }




    public Vector<Protein> getNovelIntronIsoforms(Protein lead)
    {
        Vector<Protein> nnp = new Vector<>();

        if (lead == null)
            return nnp;

        RegionVector rv = lead.getCodingRegionVector(false);

        HashSet<Pair<Integer,Integer>> known_introns = new HashSet<>();

        //calc how many proteins introduce new introns
        known_introns.addAll(rv.getInverseRegionVector().getRegionPairs());


        boolean gotnew = known_introns.size() > 0;
        while (gotnew)
        {
            Protein np = null;
            int num_novel_intron = 0;
            for (Protein p : proteins.values())
            {
                int ni=0;
                for (Pair<Integer,Integer> i : p.getCodingRegionVector(false).getInverseRegionVector().getRegionPairs())
                {
                    if (known_introns.contains(i))
                        continue;

                    ni++;
                }
                if (ni <= num_novel_intron)
                    continue;

                np = p;
                num_novel_intron = ni;
            }
            if (! (gotnew = (np != null)))
                break;

            known_introns.addAll(np.getCodingRegionVector(false).getInverseRegionVector().getRegionPairs());
            nnp.add(np);
        }

        return nnp;
    }

    public String getName()
    {
        return name;
    }


    public String getBasicInfo()
    {
        return String.format("%s %s(%c) %d-%d type: %s", id, chr, (strand)?'+':'-', start, end, biotype);
    }

}
