package genome;

import lmu.utils.*;

import org.jdom.*;

import javax.xml.stream.*;

import algo.alignment.Alignment;

public class RegionAlignment
{
    public final static String RA_ELEM_NAME= "ra";

    public double dna_seqid;
    public double prot_seqid;
    
    public Region1D src_prot_region;
    public Region1D dst_prot_region;

    public RegionVector src;
    public RegionVector dst;

    //protein space alignment for this alignment region
    public Alignment ali = null;

    public RegionAlignment()
    {

    }

    public RegionAlignment reverse(){
    	RegionAlignment ra = new RegionAlignment();
		ra.src = dst;
		ra.dst = src;
		
		ra.dna_seqid = dna_seqid;
		ra.prot_seqid= prot_seqid;
		
		ra.src_prot_region = dst_prot_region;
		ra.dst_prot_region = src_prot_region;
		
		ra.ali = ali.reverseAli();
		
		return ra;
    }
    void initFromXML(XMLStreamReader xsr) throws XMLStreamException
    {
        dna_seqid = XMLUtils.getDoubleAttrib(xsr, "dna");
        prot_seqid = XMLUtils.getDoubleAttrib(xsr, "prot");

        while (xsr.hasNext())
        {
            int event = xsr.next();
            if (event == XMLStreamConstants.END_ELEMENT && xsr.getLocalName().equals(RA_ELEM_NAME))
            {
                break;
            }
            if (event != XMLStreamConstants.START_ELEMENT)
                continue;
            
            if (xsr.getLocalName().equals("src_p_reg"))
            {
                src_prot_region = new Region1D(xsr);
                continue;
            }
            if (xsr.getLocalName().equals("dst_p_reg"))
            {
                dst_prot_region = new Region1D(xsr);
                continue;
            }
            if (xsr.getLocalName().equals("src"))
            {
                src = new RegionVector(XMLUtils.readElemFromStream(xsr));
                continue;
            }
            if (xsr.getLocalName().equals("dst"))
            {
                dst = new RegionVector(XMLUtils.readElemFromStream(xsr));
                continue;
            }
            if (xsr.getLocalName().equals("ali"))
            {
                ali = Alignment.parseAliFromXML(XMLUtils.readElemFromStream(xsr));
                continue;
            }
        

        }

        if (dst==null || src==null)
        {
            throw new RuntimeException("invalid region alignment elem, dst/src missing!");
        }
    }


    public RegionAlignment(Element e)
    {
        try
        {
            initFromXML(XMLUtils.getReader(e));
        }
        catch(XMLStreamException xse)
        {
            throw new FRuntimeException("xml error while reading xml buffer: %s", xse, XMLUtils.prettyXML(e));
        }
    }

    public RegionAlignment(XMLStreamReader xsr) throws XMLStreamException
    {
        initFromXML(xsr);
    }

    public Element getXML()
    {
        return getXML(RA_ELEM_NAME);
    }
    public Element getXML(String elemname)
    {

        XMLUtils.Stream2ElemWriter sew = new XMLUtils.Stream2ElemWriter(true);
        try
        {
            writeXML(elemname, sew.getWriter());
        }
        catch(XMLStreamException xse)
        {
            throw new RuntimeException("xml error while writing to buffer", xse);
        }
        
        return sew.getElem();
    }

    public void writeXML(XMLStreamWriter xsw) throws XMLStreamException
    {
        writeXML(RA_ELEM_NAME, xsw);
    }
    public void writeXML(String elemname, XMLStreamWriter xsw) throws XMLStreamException
    {
        xsw.writeStartElement(elemname);
        xsw.writeAttribute("dna", ""+dna_seqid);
        xsw.writeAttribute("prot", ""+prot_seqid);
        if (ali!=null)
        {
            XMLUtils.writeElem2Stream(xsw, ali.getXML("ali"));
        }
        src.writeXML(xsw, "src");
        dst.writeXML(xsw, "dst");

        if (src_prot_region!=null)
        {
            src_prot_region.writeXML(xsw, "src_p_reg");
        }
        if (dst_prot_region!=null)
        {
            dst_prot_region.writeXML(xsw, "dst_p_reg");
        }
        
        

        xsw.writeEndElement();
    }
    @Override
    public String toString(){
    	return String.format("%s -> %s",src.getRegions() ,dst.getRegions());
    }
    public int compareTo(RegionAlignment ra)
    {
        int diff=0;

        if (0!=(diff=src.compareTo(ra.src)))
            return diff;
        
        if (0!=(diff=dst.compareTo(ra.dst)))
            return diff;

        return diff;
    }
    
    public boolean equals(RegionAlignment ra)
    {
        return 0==compareTo(ra);
    }
    public double localSeqId(String p1seq,String p2seq){
        RegionAlignment ra = this;
        
        Region1D preg1 = ra.src_prot_region;
        Region1D preg2 = ra.dst_prot_region;
        
        
        String psub1 = p1seq.substring(preg1.getX1(), preg1.getX2());
        String psub2 = p2seq.substring(preg2.getX1(), preg2.getX2());
        
        Alignment ali = ra.ali;
        int num_id = ali.getNumId(psub1, psub2);
        
        return (100.0*num_id) / Math.max(psub1.length(), psub2.length());
        
    }
    
    public int getNumId( String p1seq, String p2seq){
        RegionAlignment ra = this;
        Region1D preg1 = ra.src_prot_region;
        Region1D preg2 = ra.dst_prot_region;
        
        String psub1 = p1seq.substring(preg1.getX1(), preg1.getX2());
        String psub2 = p2seq.substring(preg2.getX1(), preg2.getX2());
        
        Alignment ali = ra.ali;
    
        int num_id = ali.getNumId(psub1, psub2);
        
        return num_id;
    }
    
    public String getFormattedProtAli(String pid1, String p1seq, String pid2, String p2seq)
    {
    	RegionAlignment ra = this;
	    Region1D preg1 = ra.src_prot_region;
	    Region1D preg2 = ra.dst_prot_region;
	    
	    Region1D r1 = ra.src.getRegions().get(0);
	    Region1D r2 = ra.dst.getRegions().get(0);
	    
	    String psub1 = p1seq.substring(preg1.getX1(), preg1.getX2());
	    String psub2 = p2seq.substring(preg2.getX1(), preg2.getX2());
	    
	    Alignment ali = ra.ali;
	
	    int num_id = ali.getNumId(psub1, psub2);
	
	    double local_seqid = (100.0*num_id) / Math.max(psub1.length(), psub2.length());
	
        BufferPrintWriter pw = new BufferPrintWriter();
        pw.printf("ra lengths: %d, %d aligned: %d seqid: %2.3f%%\n",
                          psub1.length(), psub2.length(), num_id, local_seqid);

        pw.printf("%s (%s) -> %s (%s)\n", r1, preg1, r2, preg2);
        pw.print(ali.getFormattedBy(80, pid1, psub1, pid2, psub2));

        return pw.getBuffer();
    
    }
}


