package genome;

import java.io.*;
import java.util.*;


import lmu.utils.*;


public class Transcript
 {

    static public class TranscriptIntron
    {
        Exon pre_exon=null;
        Exon post_exon = null;
        Transcript t ;

        public TranscriptIntron(Transcript t, Exon pre, Exon post)
        {
            pre_exon=pre; post_exon=post; this.t=t;

        }
        public String toString()
        {
            return String.format("intron in transcript %s between exons: %s and %s", t.id, ""+pre_exon, ""+post_exon);
        }
    }
    static public class TranscriptExon
    {
        Exon exon; 
        Transcript transcript;
    
        public TranscriptExon(Exon e, Transcript t)
        {
            exon=e; transcript = t;
        }
    
        public String toString()
        {
            return "exon: "+exon+" in transcript: "+transcript;
        }
    }

     Vector<Pair<Integer, Exon>> start2exon = null;

     synchronized Vector<Pair<Integer, Exon>> getStart2Exon()
     {
         if (start2exon != null)
             return start2exon;

         start2exon = new Vector<>();
         int start = 0;
         for (Exon e : getExons())
         {
             start2exon.add(Pair.create(start, e));
             start += e.getLength();
         }
         return start2exon;
     }

     Pair<Integer,Integer> getExonIdxAndGenePosition(int pos)
     {
         Pair<Boolean, Integer> p = VectorUtils.binarySearch(getStart2Exon(), pos, (_p, _i) -> Integer.compare(_p.getFirst(), _i));
         int exidx = (p.getFirst()) ? p.getSecond() : p.getSecond() -1;
         Exon ex = start2exon.get(exidx).getSecond();
         int rel_pos = pos - start2exon.get(exidx).getFirst();

         return Pair.create(exidx, ex.start + rel_pos);

     }

     /** null if gene_pos is not contained in the transcript, otherwise
      * relative position within spliced transcript sequence
      */
     public Integer getTranscriptPos(int gene_pos)
     {
         Pair<Boolean, Integer> p = VectorUtils.binarySearch(getStart2Exon(), gene_pos, (pe, _i) -> Integer.compare(pe.getSecond().start, _i));

         int idx = (p.getFirst())? p.getSecond() : p.getSecond() - 1;
         if (idx < 0)
             return null;
         
         
         Pair<Integer, Exon> within = start2exon.get(idx);
         
         if (gene_pos > within.getSecond().end)
             return null;
         
         int diff = gene_pos - within.getSecond().start;
         return within.getFirst() + diff;
     }

     
     public int getGenePos(int pos)
     {
         return getExonIdxAndGenePosition(pos).getSecond();
     }

     public RegionVector toGeneRegions(RegionVector transcript_regions)
     {
         RegionVector rv = new RegionVector();
         for (Region1D r : transcript_regions.getRegions())
         {
             rv.addRegions(getGeneRegions(r.getX1(), r.getX2()));
         }

         return rv;

     }

     public Vector<Region1D> getGeneRegions(int start, int end)
     {
         Pair<Integer, Integer> ex1 = getExonIdxAndGenePosition(start);
         Pair<Integer, Integer> ex2 = getExonIdxAndGenePosition(end);

         if (ex1.getFirst() - ex2.getFirst() == 0)
             return ObjectGetter.toVector(new Region1D(ex1.getSecond(), ex2.getSecond(), null));
         
         Vector<Region1D> rv = ObjectGetter.toVector(new Region1D(ex1.getSecond(), start2exon.get(ex1.getFirst()).getSecond().end +1, null));

         for (int i = ex1.getFirst() + 1; i < ex2.getFirst(); i++)
         {
             Exon ex = start2exon.get(i).getSecond();
             rv.add(new Region1D(ex.start, ex.end+1));
         }
         rv.add(new Region1D(start2exon.get(ex2.getFirst()).getSecond().start, ex2.getSecond(), null));
         
         return rv;
         
     }

     String tr_type = "mRNA";
     public Protein protein = null;
     public String id;
     public String biotype;
     public String source;
     
     
     /**
      * e.g. incomplete 3' or 5'
      */
     public Set<String> tags;
     
     public Gene parentGene = null;
     int length=-1;

     int trans_start;
     int trans_end;

     RegionVector g_rv = null;
     RegionVector l_rv = null;
     public synchronized RegionVector getRegionVector(boolean global)
     {
         RegionVector rv = (global) ? g_rv : l_rv;
         if (rv != null)
             return rv;
         
         
         RegionVector local = new RegionVector();
         for (Exon e : getExons())
         {
             local.addRegion(new Region1D(e.start, e.end+1, null));
         }
         l_rv = local;
         g_rv = GenomicUtils.toGlobal(local, parentGene);

         return (global) ? g_rv : l_rv;
     }

     public int getStart()
     {
         getLength();
         return trans_start;
     }

     public int getAbsoluteStart()
     {
         int s = getStart();
         return GenomicUtils.toGlobal(s, parentGene);
         
     }

     public int getAbsoluteEnd()
     {
         int e = getStart();
         return GenomicUtils.toGlobal(e, parentGene);
     }
     
     public int getEnd()
     {
         getLength();
         return trans_end;
     }
     public String toString()
     {
         return String.format("transcript: %s protein: %s ", id, (protein==null)?"NONE":protein.id);
     }
     
     public synchronized int getLength()
     {
         if (length>0) return length;
         
         length=0;
         if (exons.size()==0)
         {
             throw new RuntimeException("no exons found to transcript: "+id);
         }
         Collections.sort(exons);
         trans_start=exons.get(0).start;
         trans_end = exons.get(exons.size()-1).end;
         for (int i=0; i<exons.size(); i++)
         {
             Exon e = exons.get(i);
             relative_starts.add(length);
             /*System.out.printf(" exon: %d (%s) %d-%d l: %d rstart: %d (%d-%d ) gene strand: %c in transcript: %s\n", i, e.id, 
                               e.start, e.end, e.length, length, 
                               e.abs_start, e.abs_end, (parentGene.strand)?'+':'-', id);
             */
             length+=e.length;
         }
         return length;
     }

     public void getRegions(int start, int length, Vector<Region1D> regions)
     {
         if (start<0 || start+length>getLength())
             throw new RuntimeException("invalid start,length ("+start+","+length+") coordinates for the transcript "+id+" (length: "+getLength()+")");
         
         int qend = start+length;
         int matching_exon_index=0;
         
         for (int i=0; i<exons.size(); i++)
         {
             
             int rstart=relative_starts.get(i);

             int rem_length = qend-rstart;

             if (rem_length<=0)
                 break;
             
             Exon e = exons.get(i);
             if (rstart+e.length<=start)
             {
                 continue;
             }
             if (rstart<=start && rstart+e.length>start)
             {
                 int absstart=e.start+start-rstart;

                 String info = String.format(" partly overlapping exon: %d remaining length: %d in exon: %d (ex rpos: %d-%d) region was: %d-%d",
                                              matching_exon_index++, rem_length, Math.min(e.length, rem_length), rstart, rstart+e.length, start, start+length);
                 //regions.add(new Region1D(absstart, absstart+Math.min(e.length, rem_length), e, info));
                 regions.add(new Region1D(absstart, absstart+Math.min(e.length, rem_length), e));

                 matching_exon_index++;
                 continue;
             }
             
             if (start+length>rstart+e.length)
             {
                 
                 String info = String.format(" fully overlapping exon: %d remaining length: %d in exon: %d (ex rpos: %d-%d) region was: %d-%d ", 
                                             matching_exon_index++, rem_length, Math.min(e.length, rem_length), rstart, rstart+e.length,
                                             start, start+length);
                 //regions.add(new Region(e.start, e.end,e, info ));
                 regions.add(new Region1D(e.start, e.end,e ));
                 matching_exon_index++;
                 continue;
             }
             
             String info = String.format(" overlapping exon: %d remaining length: %d in exon: %d (ex rpos: %d-%d) region was: %d-%d ", 
                                             matching_exon_index++, rem_length, rem_length, rstart, rstart+rem_length,
                                             start, start+length);

             regions.add(new Region1D(e.start, e.start+rem_length, e));
             matching_exon_index++;
             
         }
         
         
     }
     
     public Transcript(Gene g, String id)
     {
         this.id=id;parentGene=g;
     }

     int t_start=0;
     int t_end=0;

     public Vector<Exon> getExons()
     {
         return exons;
     }
     Vector<Exon> exons = new Vector<Exon>();
     Vector<Integer> relative_starts = new Vector<Integer>();
    

     public void addExon(Exon e)
     {
         exons.add(e);
     }


     /*
     public static void testPositionMappings() 
     {
         genome.Organism org = config.ConfigManager.getGenomicOrganism("human");
         genome.GenomeSequenceExtractor gex = org.getGenome();

         for(String gn : org.getAnnot().getGenesByType("protein_coding"))
         {
             Gene g = org.getAnnot().getGene(gn);
             
             if (!gex.gotChr(g.chr))
                 continue;
             
             
             String gseq = GenomicUtils.getGeneSeq(g, gex);

             for (Transcript t : g.transcripts.values())
             {
                 if (t.getExons().size() < 2 || t.getLength() < 300)
                     continue;
                 
                 String tseq = GenomicUtils.getTranscriptSeq(t, gex);
                 System.out.printf("tests on %s in %s\nexons: %s\n", t.id, g.getLocationInfo(), t.getExons());

                 final int L = t.getLength();

                 for (int i = 0; i< 10; i++)
                 {
                     int p1 = (int)(Math.random() * L);
                     int p2 = (int)(Math.random() * L);
                     int x = Math.min(p1, p2);
                     int y = Math.max(p1, p2);
                     if (x == y)
                         continue;
                     
                     
                     System.out.printf("test %d-%d in %d\n", x, y, L);
                     String subt = tseq.substring(x, y);
                     Vector<Region1D> v = t.getGeneRegions(x, y);
                     System.out.printf("pos: %s\n", v);
                     RegionVector rv = new RegionVector();
                     ObjectGetter.apply(v, (r) -> rv.addRegion(r));
                     String tests = GenomicUtils.getSplicedSeq(gseq, rv, true);

                     int start = v.get(0).getX1();
                     int end = v.get(v.size() -1 ).getX2();

                     int tp1 = t.getTranscriptPos(start);
                     int tp2 = t.getTranscriptPos(end);

                     if (x != tp1 || y != tp2)
                     {
                         throw new FRuntimeException("invalid backmapping %d -> %d -> %d, %d -> %d -> %d",
                                                     x, start, tp1, y, end, tp2);
                     }
                     
                     if (tests.equals(subt))
                     {
                         System.out.printf("PASSSSED!\n");
                         continue;
                     }
                     System.out.printf("\n%s\n\n%s\n", subt, tests);
                     throw new FRuntimeException("error!");
                 }
             }
         }
     }
     
     public static void main (String args[]) 
     {
         testPositionMappings();
     }
     */
 }


