package protein.algo;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.Vector;
import java.util.regex.Matcher;

import algo.alignment.Alignment;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.NumUtils;
import lmu.utils.OptionParser;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.SimpleOptionParser;
import lmu.utils.plotting.AbstractZoomedMultiPicPlane;
import lmu.utils.plotting.RegionPicturePane;
import lmu.utils.plotting.ZoomedCoordinateConverter;
import protein.algo.external.MultipleAlignmentAlgo;
import protein.algo.external.PRRN;


public class MultipleProteinAlignment{

    Logger log = LogConfig.getLogger();
    
    public List<Pair<String,String>> idSequences ;
    private StringBuffer[] output =null;
    private List<Integer[]> seqPosMatches = new ArrayList<Integer[]>();
    private List<RegionVector> exons;
    
    public StringBuffer[] getRawAlignment(){
        return output;
    }
    

    public enum MSAFormat
    {
        CLUSTAL,FASTA,PRRN;
        
        public static MSAFormat getFormat(String mode)
        {
            for(MSAFormat m : values())
            {
                if ( m.name().toUpperCase().equals(mode.toUpperCase())) return m;
            }
            throw new RuntimeException("Unknown mode:" + mode);
        }
    }
    
    public static HashMap<String,StringBuffer> readMultipleAlignment(File f, MSAFormat mode){
        Vector<String> lines = FileUtils.getLines(f);
        HashMap<String,StringBuffer> ali = new LinkedHashMap<>();
        String id = null;
        Logger log = LogConfig.getLogger();
        
        if ( mode == null)
        {
            if (lines.get(0).startsWith("CLUSTAL"))
                mode=MSAFormat.CLUSTAL;
            else if (lines.get(0).startsWith(">"))
                mode=MSAFormat.FASTA;
            else{
                boolean simpleFormat = true;
                for(String l : lines)
                {
                    if ( l.split("\\s+").length!=2)
                        simpleFormat = false;
                }
                if ( simpleFormat)
                    mode=MSAFormat.CLUSTAL;
            }
            
        }
        if ( mode == null) new RuntimeException("Cannot determine MSA format of:" +f);
        
        
        log.info("Read %s (format: %s)",f,mode);
        //read alignment
        for(String line : lines)
        {
            if ( line.trim().length()==0) 
                continue;
            
            //if ( line.startsWith("CLUSTAL FORMAT"))
            //    continue;
            if (mode.equals(MSAFormat.CLUSTAL) )
            {
                String[] tokens = line.split("\\s+");
                if ( tokens.length != 2 || tokens[0].length() == 0) continue;
                id = tokens[0];
                if ( !ali.containsKey(id))
                {
                    ali.put(id,  new StringBuffer());
                }
                ali.get(id).append(tokens[1]);
                
                continue;
            }
            else if (mode.equals(MSAFormat.PRRN))
            {
                Matcher match = PRRN.prrAliPattern.matcher(line);
                if ( match.matches()){
                    
                    String alignment = match.group(1).trim();
                    String seqId = match.group(2).trim();
                    //System.out.println(seqId + "\t" + alignment + " " +seqToIdx.get(seqId));
                    
                    if ( !ali.containsKey(seqId))
                    {
                        ali.put(seqId,new StringBuffer());
                    }
                    ali.get(seqId).append(alignment);
                    
                    //output[seqToIdx.get(seqId)].append(alignment);
                }
                continue;
            }
            else if ( mode.equals(MSAFormat.FASTA))
            {
                if ( line.startsWith(">"))
                {
                    id = line.substring(1).split("\\s+")[0];
                    ali.put(id, new StringBuffer());
                    continue;
                }
                ali.get(id).append(line);
                
            }
            //System.out.println(id);
            
        }
        return ali;
    }
    public BufferedImage draw()
    {
        AbstractZoomedMultiPicPlane _plot = null;
        RegionPicturePane rpp = null;

        ZoomedCoordinateConverter converter = new ZoomedCoordinateConverter(seqPosMatches.size(),0.2,null);
        converter.addRegion(new Region1D(0,seqPosMatches.size()), new Object());
        converter.update();
        
        _plot = new AbstractZoomedMultiPicPlane(null, null, converter.getZoomedLength());
        rpp = _plot.createRegionPane("Multiple alignment");
        
        for(int i = 0 ; i< output.length; i++)
        {
            Integer start =0;
            RegionVector rv = new RegionVector();
            for(int j = 0 ; j< output[i].length(); j++)
            {
                boolean gap =output[i].charAt(j) == '-';
                if (  gap&& start != null)
                {
                    rv.addRegion(new Region1D(start,j,null));
                    start = null;
                    continue;
                }
                
                if ( !gap && start == null )
                    start =i;
                 
            }
            rpp.addRegion(i+"" , i, converter.getConvertedRegVec(rv));
            
        }
        rpp.update();
        
        _plot.update(true);
        
        return _plot.getCombinedImage();
        
    }
    
    public Alignment createPairwiseAlignments(int i , int k) {
        if ( i >=idSequences.size() || k>=idSequences.size() ) throw new RuntimeException(String.format("%d %d out of bounds",i,k));
        
        Vector<Pair<Integer,Integer>> pos = new Vector<Pair<Integer,Integer>>();
        
        int length = seqPosMatches.size();
        
        Integer p1Length = 0;
        Integer p2Length = 0;
        
        for(int j = 0 ; j< length; j++){
            
            Integer p1 = seqPosMatches.get(j)[i];
            Integer p2 = seqPosMatches.get(j)[k]; 
          
            if ( p1!=null && p2!=null) pos.add(new Pair<Integer,Integer>(p1,p2));
           
            if  (p1 != null) p1Length =p1;
            if  (p2 != null) p2Length =p2;
            
           
        }

        return new Alignment(p1Length+1,p2Length+1,pos);
 
        
    }
    
    public void writeAlignment(File outFile,MSAFormat format, int lineWidth)
    {
        
        if ( format.equals(MSAFormat.PRRN) )
            throw new UnsupportedOperationException();
        
        PrintWriter pw = FileUtils.getWriter(outFile);
        
        int maxIdLength = NumUtils.max(idSequences, (Pair<String,String> p)->p.getFirst().length());
        
        int length = output[0].length();
        if ( format.equals(MSAFormat.FASTA))
        {
           
            
            for(int i =  0 ; i< idSequences.size(); i++)
            {
                pw.printf(">%s\n", idSequences.get(i).first);
                for(int j = 0 ; j < length;j+=lineWidth)
                {
                    pw.println(output[i].substring(j, Math.min(length,j+lineWidth)));
                }
            }
        }
        if ( format.equals(MSAFormat.CLUSTAL))
        {
            pw.printf("CLUSTAL FORMAT for ISAR\n\n");
            pw.println();
            
            for(int j = 0 ; j < length;j+=lineWidth)
            {
                for(int i =  0 ; i< idSequences.size(); i++)
                {
                    pw.printf("%-" + maxIdLength + "s   %s\n", idSequences.get(i).first, output[i].substring(j, Math.min(length,j+lineWidth)));
                }
                pw.println();
                
            }
        }
    
        
        pw.close();
    }
    
    /*
    public void readFromFile(File file)
    {
        List<StringBuffer> o = new ArrayList<StringBuffer>();
        Iterator<String[]> lines = FileUtils.getFieldSetIterator(file, " ");
        this.idSequences = new ArrayList<Pair<String,String>>();
        while(lines.hasNext())
        {
            String[] tokens = lines.next();
            
            String id = tokens[0];
            String ali = tokens[1];
            o.add(new StringBuffer(ali));
            idSequences.add(new Pair<String,String>(id,ali.replaceAll("-", "")));
        }
        this.output = o.toArray(new StringBuffer[o.size()]);
        buildAli();
    }*/
    public MultipleProteinAlignment(List<Pair<String,String>> idSequences,MultipleAlignmentAlgo algo) {
        this(idSequences,null,algo);
    }
    public MultipleProteinAlignment(File f)
    {
        this(MultipleProteinAlignment.readMultipleAlignment(f, null));
    }
    public MultipleProteinAlignment(HashMap<String,StringBuffer> alignment)
    {
        idSequences = new ArrayList<>();
        List<StringBuffer> o = new ArrayList<StringBuffer>();
        for(Entry<String, StringBuffer> e : alignment.entrySet())
        {
            String id = e.getKey();
            String ali = e.getValue().toString();
            o.add(new StringBuffer(ali));
            idSequences.add(new Pair<String,String>(id,ali.replaceAll("-", "")));
            
        }
        this.output = o.toArray(new StringBuffer[o.size()]);
        buildAli();
    }
    private void buildAli()
    {
        int length = output[0].length();

        int[] currentProtPositions = new int[output.length];

        seqPosMatches = new ArrayList<Integer[]> ();
    
        for(int i = 0 ; i< length; i++){

            Integer[] protMatch = new Integer[output.length];
            for(int j = 0 ; j< output.length; j++){
                if ( i >= output[j].length())
                {
                    throw new RuntimeException("Strange alignment format (alignments with different lengths).");
                }
                if ( output[j].charAt(i) != '-') {
                    protMatch[j] = currentProtPositions[j];
                    currentProtPositions[j]+=1;
                }
               
            }
            /*
            System.out.println(i +  " " + protMatch[0]  );
            if (protMatch[0]!= null){
                System.out.println(output[0].charAt( protMatch[0]) + " " + output[0]);
            }
            System.out.println(i + " " + Arrays.asList(protMatch));
            */
            seqPosMatches.add(protMatch);
           
        }
        if ( seqPosMatches.size() != output[0].length()) throw new RuntimeException("!=");
        for(int i = 0 ; i< idSequences.size(); i++){
            String orig = idSequences.get(i).second;
            StringBuffer got = new StringBuffer();
            for(int j = 0 ; j< output[i].length(); j++){
                Integer pos = seqPosMatches.get(j)[i];
                if ( pos != null) got.append(orig.charAt(pos));
            }
            if ( !orig.equals(got.toString())){
                throw new RuntimeException(String.format("Orig/Got length: %d/%d",orig.length() , got.length()));
                
            }
        }
    }
    public MultipleProteinAlignment(List<Pair<String,String>> idSequences,List<RegionVector> exons,MultipleAlignmentAlgo algo) {
        this.idSequences = idSequences;
        if ( exons != null && exons.size() != idSequences.size()) throw new RuntimeException("exons != idSequences");
        this.exons = exons;
        
        HashSet<String> tmp = new HashSet<String>();
        File sequencesTmpFile = null;
        List<String> ids = new ArrayList<String>();
        try{
            sequencesTmpFile = File.createTempFile("mult.", ".tmp");
            PrintWriter pw = new PrintWriter(sequencesTmpFile);
    
            
            for(int i = 0 ; i< idSequences.size(); i++){
                Pair<String,String> idSequence = idSequences.get(i);
                if ( tmp.contains(idSequence.first)) throw new RuntimeException(String.format("%s not unique",idSequence.first));
                ids.add(idSequence.first);
                tmp.add(idSequence.first);
                
                pw.write(">" + idSequence.getFirst() + "\n");
                if ( algo.isoformStructureSupport() && exons != null && exons.get(i).size() < 100 /**because of PRRN*/ ){
                    MultipleAlignmentAlgo.printInfos(pw,exons.get(i));
                    
                }
                fasta.FastaLikeReader.writeFasta(idSequence.getSecond(), pw);
            }
            pw.close();
            

        }catch(Exception e){
            throw new RuntimeException("Cannot write tmp files",e);
        }
        long l1 = System.currentTimeMillis();
        try{
            output = algo.align(sequencesTmpFile, ids);
        }catch(Exception e){
            throw new RuntimeException("Cannot align",e);
        }
        
        if (sequencesTmpFile.exists()){
            sequencesTmpFile.delete();
        }
        
        long time = System.currentTimeMillis()-l1;
        log.info("Alignment took: %d ms",time);
        
        buildAli();
    }
    
    public static void main(String[] args) throws Exception
    {
        SimpleOptionParser cmdLine = new SimpleOptionParser("in");
        cmdLine.setFile("in");
        
        if (!OptionParser.parseParams(args,true, false, true,  cmdLine)){
            return;
        }
        
        MultipleProteinAlignment ali = new MultipleProteinAlignment(cmdLine.getFile("in"));
        ali.writeAlignment(new File("test.aln"), MSAFormat.CLUSTAL,70);
     
    }
}
