package protein.algo;

import java.io.*;
import java.util.*;
import protein.aminoacid.*;
import algo.alignment.*;
import algo.alignment.Gotoh.AlignMode;
import algo.alignment.Gotoh.SimCalculator;
import lmu.utils.*;

public class ProteinAlign 
{

    static ProteinAligner aligner = new ProteinAligner();

    public static Alignment align(String seq1, String seq2, AlignMode mode, String matrix, double gapOpen, double gapExtend)
    {
        aligner.init(matrix,mode,gapOpen,gapExtend);
        return aligner.align(seq1,seq2);
        
    }
    public static double score(String seq1, String seq2, String mode, String matrix, double gapOpen, double gapExtend)
    {
        aligner.init(matrix,mode,gapOpen,gapExtend);
        return aligner.score(seq1,seq2);
    }
    public static double score(String seq1, String seq2, AlignMode mode, String matrix, double gapOpen, double gapExtend)
    {
        aligner.init(matrix,mode,gapOpen,gapExtend);
        return aligner.score(seq1,seq2);
        
    }
    
    static HashMap<String, String> readSeqlib(File f) throws IOException
    {
        HashMap<String, String> rv =new HashMap<String, String>();
        BufferedReader bfr = new BufferedReader(new FileReader(f));

        String line;
        while (null!=(line=bfr.readLine()))
        {
            String[] sp =line.split(":");
            if (sp.length!=2)
            {
                System.err.println("unexpecteD: "+line);
                continue;
            }
            rv.put(sp[0],sp[1]);
        }
        bfr.close();
        return rv;
    }

    static class Query2TargetIterator implements Iterator<Pair<String,String>>
    {
        Vector<String> queries = new Vector<String>();
        Vector<String> targets = new Vector<String>();
        int i=0; 
        int j=0;

        public Query2TargetIterator(Collection<String> queries, Collection<String> targets) 
        {
            this.queries.addAll(queries);
            this.targets.addAll(targets);

        }
        public boolean hasNext()
        {
            return (i<targets.size());
        }
        public Pair<String,String> next()
        {
            
            Pair<String,String> p = new Pair<String,String>(queries.get(j), targets.get(i));
            if (++j>=queries.size() && i<targets.size())
            {
                j=0;
                i++;
            }
            return p;
            
        }

        public void remove()
        {
            throw new RuntimeException("not implemented");
        }

    }
    static class SeqPairIterator implements Iterator<Pair<String,String>>
    {
        BufferedReader bfr = null;

        File f = null;
        int ln=0;
        Pair<String,String> nextE = null;
        public SeqPairIterator(File f) throws IOException
        {
            bfr = new BufferedReader(new FileReader(f));
            readNext();
        }
        
        public boolean hasNext()
        {
            return nextE!=null;
        }
        public Pair<String,String> next()
        {
            Pair<String,String> rv = nextE;
            readNext();
            return rv;
        }

        public void remove()
        {
            throw new RuntimeException("not implemented");
        }
        public void readNext()
        {
            nextE=null;
            try
            {
                
                String line;
                
                if (bfr==null || null== (line = bfr.readLine()))
                {
                    if(bfr!=null)
                    {
                        bfr.close();
                        bfr=null;
                    }
                    
                    return;
                }
                ln++;
                String[] sp = line.split("\\s+");
                nextE = new Pair<String,String>(sp[0], sp[1]);
            }
            catch(IOException e)
            {
                throw new RuntimeException("error:"+e.getMessage()+" at line: "+ln+" in file "+f.getAbsolutePath(), e);
            }
        }

            
    }
    static class SeqAlign implements Comparable<SeqAlign>
    {
        Alignment a;
        String id1;
        String id2;
        public SeqAlign(String id1, String id2, Alignment a)
        {
            this.a=a; this.id1=id1; this.id2=id2;
        }
        public int compareTo(SeqAlign sa)
        {
            return sa.a.compareTo(a);
        }

    }

    static void report(Vector<SeqAlign> alis, HashMap<String, String> seqlib, boolean needali)
    {
        Collections.sort(alis);
        
        for (SeqAlign sa: alis)
        {
            
            if (!needali)
            {
                System.out.printf(">%s %s %2.3f\n", sa.id1, sa.id2, sa.a.score);
                continue;
            }
            String target=seqlib.get(sa.id2);
            String query = seqlib.get(sa.id1);
            int nid=sa.a.getNumId(query, target);
            System.out.printf(">%s %s %2.3f idendity: %d/%d\n", sa.id1, sa.id2, sa.a.score, nid, target.length());
            System.out.println(sa.a.getFormatted(sa.id1,query,sa.id2, target));
        }
    }
    public static void main (String args[]) throws Exception
    {
        Locale.setDefault(Locale.UK);
        SimpleOptionParser sp = new SimpleOptionParser("mode", "matrix", "go", "ge", "seqlib", "pairs", "printali", "target");
        sp.setFile("seqlib", "pairs", "target");
        sp.setOptional("target", "pairs");
        sp.setDouble("go", "ge");
        sp.setDefault("mode", "freeshift");
        sp.setDefault("matrix", "dayhoff");
        sp.setDefault("go", "-11");
        sp.setDefault("ge", "-1");
        sp.setSwitches("printali");

        if(!OptionParser.parseParams(args,true,false,true, sp))
            return;


        HashMap<String, String> seqlib = readSeqlib(sp.getFile("seqlib"));
        Iterator<Pair<String,String>> pairIt = null;
        if (sp.isOptionSet("pairs"))
        {
            pairIt = new SeqPairIterator(sp.getFile("pairs"));
        }
        
        if (sp.isOptionSet("target"))
        {
            HashMap<String, String> tseqlib = readSeqlib(sp.getFile("target"));

            pairIt = new Query2TargetIterator(seqlib.keySet(), tseqlib.keySet());

            seqlib.putAll(tseqlib);
        }


        boolean needali = sp.isSet("printali");
        AlignMode am = AlignMode.get(sp.getValue("mode"));
        double go = sp.getDouble("go");
        double ge = sp.getDouble("ge");
        String matrix = sp.getValue("matrix");
        ProteinAligner aligner = new ProteinAligner(am,matrix,go,ge);

        Vector<SeqAlign> alis = new Vector<SeqAlign>();
        
        SeqAlign last = null;
        final boolean sortalis = sp.isOptionSet("target");
        
        while (pairIt.hasNext())
        {
            Pair<String,String> ids = pairIt.next();
            String s1 = seqlib.get(ids.getFirst());
            String s2 = seqlib.get(ids.getSecond());

            if (sortalis)
            {
                algo.alignment.Alignment a = aligner.align(s1, s2);
                SeqAlign sa = new SeqAlign(ids.getFirst(), ids.getSecond(), a);

                if (last!=null && last.id2.equals(ids.getSecond()))
                {
                    alis.add(last=sa);
                    continue;
                }
                report(alis,seqlib ,needali);
                alis.clear();
                alis.add(last=sa);
                continue;
            }
            if (!needali)
            {
                double score = aligner.score(s1, s2);
                System.out.printf("%s %s %2.3f\n", ids.getFirst(), ids.getSecond(), score);
            }
            else
            {
                algo.alignment.Alignment a = aligner.align(s1, s2);
                System.out.printf("%s %s %2.3f\n", ids.getFirst(), ids.getSecond(), a.score);
                //System.out.printf(">%s %s %2.3f\n%s\n%s\n", ids.getFirst(), ids.getSecond(), a.score);
            }
        }
        if (sortalis)
        {
            report(alis,seqlib ,needali);
        }
    }
}
