package protein.algo;

import java.io.*;
import java.util.*;

import fasta.SimpleFastaReader;

import protein.aminoacid.*;
import algo.alignment.*;
import algo.alignment.Gotoh.AlignMode;
import algo.alignment.Gotoh.SimCalculator;
import lmu.utils.*;

public class ProteinAligner
{
    Gotoh g = new Gotoh();
    AminoAcidExchangeMatrix matrix=null;

    static class AAExchangeCalc implements Gotoh.SimCalculator
    {

        AminoAcidExchangeMatrix matrix;
        String s1; String s2;
        public AAExchangeCalc(String s1, String s2, AminoAcidExchangeMatrix matrix)
        {
            this.s1=s1;
            this.s2=s2;
            this.matrix=matrix;
        }
        public long calcSim(int i, int j)
        {
            return matrix.getL(s1.charAt(i), s2.charAt(j));
        }
        public long[] calcSim(int i, long[] target)
        {

            //System.out.printf("will get sim row for %d: %c target: %d \n", i, s1.charAt(i), s2.length());
            long[] r = matrix.getLRow(s1.charAt(i));
            for (int j=0; j<s2.length(); j++)
            {
                int index = s2.charAt(j)-'A';
                if ( index < 0 || index >= r.length)
                {
                    throw new RuntimeException(String.format("Unknown character: %c",s2.charAt(j)));
                }
                
                target[j]=r[s2.charAt(j)-'A'];
            }
            return target;
        }
        public long getFactor()
        {
            return matrix.getFactor();
        }
    }

    public ProteinAligner()
    {
        this("freeshift", "dayhoff", -12, -1);
    }
    public ProteinAligner(String mode, String matrix, double gap_open, double gap_extend)
    {
        this(AlignMode.get(mode), matrix, gap_open, gap_extend);

    }

    public ProteinAligner(AlignMode mode, String matrix, double gap_open, double gap_extend)
    {
        init(matrix,mode,gap_open, gap_extend);
    
    }

    
    public void init(String matrix, String mode, double gapOpen, double gapExtend)
    {
        init(matrix,AlignMode.get(mode),gapOpen,gapExtend);
    }

    public void init(String matrix, AlignMode mode, double gapOpen, double gapExtend)
    {
        this.matrix = AminoAcidExchangeMatrix.getMatrix(matrix);
        g.setAlignMode(mode);
        g.setGapOpen(gapOpen);
        g.setGapExtend(gapExtend);
    }


    
    public double score(String seq1, String seq2)
    {
        AAExchangeCalc sc = new AAExchangeCalc(seq1, seq2, matrix);
        return g.score(seq1.length(), seq2.length(), sc);
    }

    public Alignment align(String seq1, String seq2)
    {
        AAExchangeCalc sc = new AAExchangeCalc(seq1, seq2, matrix);
        return g.align(seq1.length(), seq2.length(), sc);
    }
    
    public static void main(String[] args) throws Exception{
        Locale.setDefault(Locale.UK);
        SimpleOptionParser sp = new SimpleOptionParser("mode", "matrix", "go", "ge", "seq1","seq2","seq_fasta_1","seq_fasta_2");
        sp.setDouble("go", "ge");
        sp.setDefault("mode", "freeshift");
        sp.setDefault("matrix", "dayhoff");
        sp.setDefault("go", "-11");
        sp.setDefault("ge", "-1");

        sp.setFile("seq_fasta_1");
        sp.setFile("seq_fasta_2");
        sp.setOptional("seq1","seq2","seq_fasta_1","seq_fasta_2");
        if(!OptionParser.parseParams(args,true,false,true, sp))
            return;
        
        String s1 = null;
        String s2 =null; 
        if ( sp.isOptionSet("seq1")){
        	s1 = sp.getValue("seq1");
        }
        if ( sp.isOptionSet("seq2")){
        	s2 = sp.getValue("seq2");;
        }
        if ( sp.isOptionSet("seq_fasta_1")){
        	SimpleFastaReader reader = new SimpleFastaReader(sp.getFile("seq_fasta_1"));
        	if ( reader.hasNext()) s1 = reader.next().getSecond();
        	reader.close();
        }
        if ( sp.isOptionSet("seq_fasta_2")){
        	SimpleFastaReader reader = new SimpleFastaReader(sp.getFile("seq_fasta_2"));
        	if ( reader.hasNext())s2 = reader.next().getSecond();
        	reader.close();
        }
        if ( s1 == null || s2 == null){
        	System.err.println(sp.getHelp());
        	throw new RuntimeException("Neither sequence strings, nor sequence files are given ");
        	
        }
        
        System.out.println(s1);
        System.out.println(s2);
        AlignMode am = AlignMode.get(sp.getValue("mode"));
        double go = sp.getDouble("go");
        double ge = sp.getDouble("ge");
        String matrix = sp.getValue("matrix"); 
        ProteinAligner aligner = new ProteinAligner(am,matrix,go,ge);
        
        
        
        Alignment ali = aligner.align(s1,s2);
        
        int nid=ali.getNumId(s1, s2);
        System.out.printf(">%s %s %2.3f idendity: %d/%d\n", "Seq1","Seq", ali.score, nid, s2.length());
        System.out.println(ali.getFormatted("Seq1",s1,"Seq2", s2));
        System.out.printf(">%s %s %2.3f idendity: %d/%d\n", "Seq1","Seq", ali.score, nid, s2.length());
        System.out.println(ali.getAlignmentString(s1, s2).getFirst());
        System.out.println(ali.getAlignmentString(s1, s2).getSecond());
        
        
    }
}

