#!/usr/local/bin/perl

use Bio::EnsEMBL::Registry;
use Getopt::Long;
use Bio::SeqIO;
use Getopt::Long; 
use Text::Wrap;
use Bio::DB::Taxonomy;
use Error qw(:try);

# initialize some defaults
my $species = 'homo_sapiens';
my $source  = 'core'; # core or vega
# allow identifier being passed as the first argument in the command line or by an option -n or -gene_symbol
my $identifier;
GetOptions(
  "g|gene_symbol=s" => \$identifier,
  "species=s"       => \$species
);
 
if (!$identifier)
{
  print "\nThe following non-optional parameter(s) are missing:\n\n"; 
  print "\t\t-s ENSEMBL species name    e.g. homo_sapiens, saccharomyces_cerevisiae (default: homo_sapiens) \n";
  print "\t\t-g ENSEMBL gene id         e.g. ENSG00000138069, YPT1\n";
  print "\n";  
  exit;
}

my $db = Bio::DB::Taxonomy->new(-source => 'entrez');
my $taxonid=-1;

# use NCBI Entrez over HTTP
try {
  $taxonid = $db->get_taxonid($species) ;
}catch Bio::Root::Exception with{
	print STDERR "Could not query tax id from NCBI.\n";
}
;

sub reverse_complement {
  my $dna = shift;

  # reverse the DNA sequence
  my $revcomp = reverse($dna);

  #complement the reversed DNA sequence
  $revcomp =~ tr/ACGTacgt/TGCAtgca/;
  return $revcomp;
}

my $out_seq = Bio::SeqIO->new(
  -fg => \*STDOUT,
  -format => 'fasta',
);
 
# The current way for accesing ensemble is using the registry
# it matches your API with its corresponding ensembl database version
# Also takes care of the mysql port (now is in a non standard port 5306)
my $reg = 'Bio::EnsEMBL::Registry';
 
$reg->load_registry_from_db(-host => 'ensembldb.ensembl.org', -user=>'anonymous');

my $gene_adaptor   = $reg->get_adaptor($species, $source, 'Gene' );

$Text::Wrap::columns = 80;

my $print_DNA = 1;
my $counter=0;

foreach my $gene (@{$gene_adaptor->fetch_all_by_external_name($identifier)}) {

  $counter++;
  print ">", $gene->display_id(), " ",$gene->external_name() , " ",$taxonid, " ", $gene->chr_name, ":",$gene->start, "-", $gene->end, ":", $gene->strand, "\n";
     
  # the seq method in gene returns the nucleotide sequence
  # [warning] in transcript and exon objects, the seq method returns a biperl Bio::Seq object
  #print "gene sequence for " . $identifier.":\n". $gene->seq() . "\n";
  print wrap('','',$gene->seq()), "\n";  
  my $seq = $gene->seq();
  if ( $gene->strand()== -1){
    reverse_complement($seq);
  }
    
  foreach my $trans (@{$gene->get_all_Transcripts}) {
    #check for coding
    if ( $trans->translation() ) {
      my $pep = $trans->translate();
    } else {
      next;
    }
    my $dna="";
    my $exons=$trans->get_all_Exons( );
    if ( $gene->strand() == -1) {
      reverse($exons);
    }
     
    print ">", $trans->stable_id() . "\n";

    foreach my $exon ( @{ $exons } ) {

      #skip non-coding exons
      if ( $exon->coding_region_start($trans) == "" ){
	next;
      }
      #compute local positions
      my $start = $exon->coding_region_start($trans)-$gene->start;
      my $end =$exon->coding_region_end($trans) -$gene->start+1;
	  
      if ( $gene->strand() == -1){
	$start = $gene->end-$exon->coding_region_end($trans);
	$end = $gene->end-$exon->coding_region_start($trans)+1;
      }
      print ";", $start,  " ", $end, "\n";

      #extract exon sequence
      my $exonseq =  substr($seq,$start,$end-$start);
      $dna="$dna$exonseq";

    }
	
    if ( length($dna) > 0 ) {
      if ( $print_DNA==1) {
	print wrap('','',$dna),"\n";
      }else{
	my $dnaO=Bio::Seq->new(-seq => $dna);
	my $as=$dnaO->translate()->seq();
	$as=~s/\*/X/g; 
	print wrap('','',$as), "\n";
      }
    }
  }
  #only consider first gene
  last;
}
if ( $counter == 0)
{
  print STDERR "Unknown gene id: ", $identifier, "\n";
}
